/*
 * Copyright (c) 2005-2014 Brocade Communications Systems, Inc.
 * Copyright (c) 2014 QLogic Corporation.
 * All rights reserved
 * www.qlogic.com
 *
 * See LICENSE.bfa for copyright and licensing details.
 */
/**
 * @file bfad.c Linux driver PCI interface module.
 */

#include <linux/module.h>
#include "bfad_drv.h"
#include "bfad_ioctl.h"
#include "bfad_im.h"
#include "bfad_tm.h"
#include "bfad_ipfc.h"
#include "bfa/bfa_ipfc.h"
#include <fcs/bfa_fcs.h>
#include "bfad_aen_cb.h"
#include <fcs/bfa_fcbuild.h>

BFA_TRC_FILE(LDRV, BFAD);

static bfa_status_t bfad_fc4_port_new(struct bfad_s *bfad,
				      struct bfad_port_s *port, int roles);
static void bfad_fc4_port_delete(struct bfad_s *bfad,
				 struct bfad_port_s *port, int roles);
struct bfad_ipfc_ops ipfc_ops;
struct bfad_tm_ops_s tm_ops;
atomic_t tm_ops_flag = ATOMIC_INIT(0);

struct semaphore bfad_sem;
int bfad_scan_done;

LIST_HEAD(bfad_list);

static	int	bfad_inst;
int		supported_fc4s;

char		*host_name;
char		*os_name;
char		*os_patch;
int		num_rports;
int		num_ioims;
int		num_tios;
int		num_tms;
int		num_fcxps;
int		num_ufbufs;
int		reqq_size;
int		rspq_size;
int		num_sgpgs;
int		rport_del_timeout = BFA_FCS_RPORT_DEF_DEL_TIMEOUT;
int		bfa_lun_queue_depth = BFAD_LUN_QUEUE_DEPTH;
int		bfa_io_max_sge = BFAD_IO_MAX_SGE;
int		log_level = BFA_LOG_WARNING;
int		ioc_auto_recover = BFA_TRUE;
int		linkup_delay = -1;
int		fdmi_enable = BFA_TRUE;
int		pcie_max_read_reqsz;
uint8_t		port_role = BFA_LPORT_ROLE_FCP_IM;
int		max_xfer_size = BFAD_MAX_SECTORS >> 1;
int		max_rport_logins = BFA_FCS_MAX_RPORT_LOGINS;

/*
 * Stores the module parm num_sgpgs value;
 * used to reset for bfad next instance.
 */
int		num_sgpgs_parm = 0;

static void
bfad_sm_uninit(struct bfad_s *bfad, enum bfad_sm_event event);
static void
bfad_sm_created(struct bfad_s *bfad, enum bfad_sm_event event);
static void
bfad_sm_initializing(struct bfad_s *bfad, enum bfad_sm_event event);
static void
bfad_sm_operational(struct bfad_s *bfad, enum bfad_sm_event event);
static void
bfad_sm_stopping(struct bfad_s *bfad, enum bfad_sm_event event);
static void
bfad_sm_failed(struct bfad_s *bfad, enum bfad_sm_event event);
static void
bfad_sm_fcs_exit(struct bfad_s *bfad, enum bfad_sm_event event);

/**
 * Beginning state for the driver instance, awaiting the pci_probe event
 */
static void
bfad_sm_uninit(struct bfad_s *bfad, enum bfad_sm_event event)
{
	bfa_trc(bfad, event);

	switch (event) {
	case BFAD_E_CREATE:
		bfa_sm_set_state(bfad, bfad_sm_created);
		bfad_os_kthread_create(bfad);
		break;

	case BFAD_E_STOP:
		/* Ignore stop; already in uninit */
		break;

	default:
		bfa_sm_fault(bfad, event);
	}
}

/**
 * Driver Instance is created, awaiting event INIT to initialize the bfad
 */
static void
bfad_sm_created(struct bfad_s *bfad, enum bfad_sm_event event)
{
	unsigned long flags;
	bfa_status_t ret;

	bfa_trc(bfad, event);

	switch (event) {
	case BFAD_E_INIT:
		bfa_sm_set_state(bfad, bfad_sm_initializing);

		init_completion(&bfad->comp);

		/* Enable Interrupt and wait bfa_init completion */
		if (bfad_setup_intr(bfad)) {
			printk(KERN_WARNING "bfa %s: bfad_setup_intr failed\n",
					bfad->pci_name);
			bfa_sm_send_event(bfad, BFAD_E_INIT_FAILED);
			break;
		}

		spin_lock_irqsave(&bfad->bfad_lock, flags);
		bfa_init(&bfad->bfa);
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);

#ifdef CONFIG_PCI_MSI
		/* Set up interrupt handler for each vectors */
		if ((bfad->bfad_flags & BFAD_MSIX_ON) &&
			bfad_install_msix_handler(bfad)) {
				printk(KERN_WARNING
					"bfa %s: install_msix failed.\n",
					bfad->pci_name);
		}
#endif
		bfad_init_timer(bfad);

		wait_for_completion(&bfad->comp);

		if ((bfad->bfad_flags & BFAD_HAL_INIT_DONE)) {
			bfa_sm_send_event(bfad, BFAD_E_INIT_SUCCESS);
		} else {
			printk(KERN_WARNING "bfa %s: bfa init failed.\n",
				bfad->pci_name);
			spin_lock_irqsave(&bfad->bfad_lock, flags);
			bfa_fcs_init(&bfad->bfa_fcs);
			spin_unlock_irqrestore(&bfad->bfad_lock, flags);

			ret = bfad_cfg_pport(bfad, port_role);
			if (ret != BFA_STATUS_OK) {
				init_completion(&bfad->comp);

				spin_lock_irqsave(&bfad->bfad_lock, flags);
				bfad->pport.flags |= BFAD_PORT_DELETE;
				bfa_fcs_exit(&bfad->bfa_fcs);
				spin_unlock_irqrestore(&bfad->bfad_lock, flags);

				wait_for_completion(&bfad->comp);

				bfa_sm_send_event(bfad, BFAD_E_INIT_FAILED);
				break;
			}

			bfad->bfad_flags |= BFAD_HAL_INIT_FAIL;
			bfa_sm_send_event(bfad, BFAD_E_HAL_INIT_FAILED);
		}

		break;

	case BFAD_E_KTHREAD_CREATE_FAILED:
		bfa_sm_set_state(bfad, bfad_sm_uninit);
		break;

	default:
		bfa_sm_fault(bfad, event);
	}
}

static void
bfad_sm_initializing(struct bfad_s *bfad, enum bfad_sm_event event)
{
	int	retval;
	unsigned long	flags;

	bfa_trc(bfad, event);

	switch (event) {
	case BFAD_E_INIT_SUCCESS:
		bfad_os_kthread_stop(bfad);
		spin_lock_irqsave(&bfad->bfad_lock, flags);
		bfad->bfad_tsk = NULL;
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);

		retval = bfad_start_ops(bfad);
		if (retval != BFA_STATUS_OK)
			bfa_sm_set_state(bfad, bfad_sm_failed);

		bfa_sm_set_state(bfad, bfad_sm_operational);
		break;

	case BFAD_E_INIT_FAILED:
		bfa_sm_set_state(bfad, bfad_sm_uninit);
		bfad_os_kthread_stop(bfad);
		spin_lock_irqsave(&bfad->bfad_lock, flags);
		bfad->bfad_tsk = NULL;
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		break;

	case BFAD_E_HAL_INIT_FAILED:
		bfa_sm_set_state(bfad, bfad_sm_failed);
		break;

	default:
		bfa_sm_fault(bfad, event);
	}
}

static void
bfad_sm_failed(struct bfad_s *bfad, enum bfad_sm_event event)
{
	int	retval;

	bfa_trc(bfad, event);

	switch (event) {
	case BFAD_E_INIT_SUCCESS:
		retval = bfad_start_ops(bfad);
		if (retval != BFA_STATUS_OK)
			break;
		bfa_sm_set_state(bfad, bfad_sm_operational);
		break;

	case BFAD_E_STOP:
		bfa_sm_set_state(bfad, bfad_sm_fcs_exit);
		bfa_sm_send_event(bfad, BFAD_E_FCS_EXIT_COMP);
		break;

	case BFAD_E_EXIT_COMP:
		bfa_sm_set_state(bfad, bfad_sm_uninit);
		bfad_remove_intr(bfad);
		del_timer_sync(&bfad->hal_tmo);
		break;

	default:
		bfa_sm_fault(bfad, event);
	}
}

static void
bfad_sm_operational(struct bfad_s *bfad, enum bfad_sm_event event)
{
	bfa_trc(bfad, event);

	switch (event) {
	case BFAD_E_STOP:
		bfa_sm_set_state(bfad, bfad_sm_fcs_exit);
		bfad_fcs_stop(bfad);
		break;

	default:
		bfa_sm_fault(bfad, event);
	}
}

static void
bfad_sm_fcs_exit(struct bfad_s *bfad, enum bfad_sm_event event)
{
	bfa_trc(bfad, event);

	switch (event) {
	case BFAD_E_FCS_EXIT_COMP:
		bfa_sm_set_state(bfad, bfad_sm_stopping);
		bfad_stop(bfad);
		break;

	default:
		bfa_sm_fault(bfad, event);
	}
}

static void
bfad_sm_stopping(struct bfad_s *bfad, enum bfad_sm_event event)
{
	bfa_trc(bfad, event);

	switch (event) {
	case BFAD_E_EXIT_COMP:
		bfa_sm_set_state(bfad, bfad_sm_uninit);
		bfad_remove_intr(bfad);
		del_timer_sync(&bfad->hal_tmo);
		bfad_fc4_probe_undo(bfad);
		bfad_os_iodm_disable_events(bfad->pport.im_port->shost);
		bfad_uncfg_pport(bfad);
		break;

	default:
		bfa_sm_fault(bfad, event);
		break;
	}
}

bfa_boolean_t
bfad_is_ready(void)
{
	struct bfad_s *bfad = NULL;

	down(&bfad_sem);
	if (!bfad_scan_done) {
		up(&bfad_sem);
		return BFA_FALSE;
	}

	list_for_each_entry(bfad, &bfad_list, list_entry) {
		if (!(bfad->bfad_flags & BFAD_DRV_INIT_DONE)) {
			up(&bfad_sem);
			return BFA_FALSE;
		}
	}
	up(&bfad_sem);

	return BFA_TRUE;
}

void
bfad_flags_set(struct bfad_s *bfad, uint32_t flags)
{
	if (bfad)
		bfad->bfad_flags |= flags;
}

void
bfad_flags_clear(struct bfad_s *bfad, uint32_t flags)
{
	if (bfad)
		bfad->bfad_flags &= ~flags;
}

/**
 * @dg BFA callbacks
 * @{
 */
void
bfad_hcb_comp(void *arg, bfa_status_t status)
{
	struct bfad_hal_comp *fcomp = (struct bfad_hal_comp *)arg;

	fcomp->status = status;
	complete(&fcomp->comp);
}

/**
 * @brief
 * bfa_init callback
 */
void
bfa_cb_init(void *drv, bfa_status_t init_status)
{
	struct bfad_s	      *bfad = drv;

	if (init_status == BFA_STATUS_OK) {
		bfad->bfad_flags |= BFAD_HAL_INIT_DONE;

		/*
		 * If BFAD_HAL_INIT_FAIL flag is set:
		 * Wake up the kernel thread to start
		 * the bfad operations after HAL init done
		 */
		if ((bfad->bfad_flags & BFAD_HAL_INIT_FAIL)) {
			bfad->bfad_flags &= ~BFAD_HAL_INIT_FAIL;
			bfad_os_kthread_wakeup(bfad);
		}
	}

	complete(&bfad->comp);
}

void *
bfa_fcb_fcs(struct bfad_s *bfad)
{
	return (void*)&bfad->bfa_fcs;
}

/**
 * @brief
 * bfa_stop callback
 */
void
bfa_cb_stop(void *drv, bfa_status_t stop_status)
{
	struct bfad_s	      *bfad = drv;

	/* Signal the BFAD stop waiting thread */
	complete(&bfad->comp);
}

void
bfa_cb_get_bootwwns(void *drv, uint8_t *nwwns, wwn_t *wwns,
					bfa_boolean_t *bflash, uint8_t *bus,
					uint8_t *target)
{
	struct bfad_s		*bfad = drv;
	bfa_iocfc_get_bootwwns(&bfad->bfa, nwwns, wwns);
}

/**
 * @brief
 * bfa_ioc_enable() callback.
 */
void
bfa_cb_ioc_enable(void *drv)
{
	struct bfad_s *bfad = drv;

	complete(&bfad->enable_comp);
}

/**
 * @brief
 * bfa_ioc_diable() callback.
 */
void
bfa_cb_ioc_disable(void *drv)
{
	struct bfad_s	      *bfad = drv;

	complete(&bfad->disable_comp);
}

void
bfa_fcb_exit(void *drv)
{
	struct bfad_s	      *bfad = drv;

	complete(&bfad->comp);
}

/**
 * @}
 */

/**
 * @dg BFA_FCS callbacks
 * @{
 */
static struct bfad_port_s *
bfad_get_drv_port(struct bfad_s *bfad, struct bfad_vf_s *vf_drv,
		  struct bfad_vport_s *vp_drv)
{
	return ((vp_drv) ? (&(vp_drv)->drv_port)
		: ((vf_drv) ? (&(vf_drv)->base_port) : (&(bfad)->pport)));
}

struct bfad_port_s *
bfa_fcb_lport_new(struct bfad_s *bfad, struct bfa_fcs_lport_s *port,
		 enum bfa_lport_role roles, struct bfad_vf_s *vf_drv,
		 struct bfad_vport_s *vp_drv)
{
	bfa_status_t	rc;
	struct bfad_port_s    *port_drv;

	if (!vp_drv && !vf_drv) {
		port_drv = &bfad->pport;
		port_drv->pvb_type = BFAD_PORT_PHYS_BASE;
	} else if (!vp_drv && vf_drv) {
		port_drv = &vf_drv->base_port;
		port_drv->pvb_type = BFAD_PORT_VF_BASE;
	} else if (vp_drv && !vf_drv) {
		port_drv = &vp_drv->drv_port;
		port_drv->pvb_type = BFAD_PORT_PHYS_VPORT;
	} else {
		port_drv = &vp_drv->drv_port;
		port_drv->pvb_type = BFAD_PORT_VF_VPORT;
	}

	port_drv->fcs_port = port;
	port_drv->roles = roles;
	rc = bfad_fc4_port_new(bfad, port_drv, roles);
	if (rc != BFA_STATUS_OK) {
		bfad_fc4_port_delete(bfad, port_drv, roles);
		port_drv = NULL;
	}

	return port_drv;
}

void
bfa_fcb_lport_delete(struct bfad_s *bfad, enum bfa_lport_role roles,
		    struct bfad_vf_s *vf_drv, struct bfad_vport_s *vp_drv)
{
	struct bfad_port_s    *port_drv;

	if (vp_drv) {
		bfa_trc(bfad, roles);
		/*
		 * For kernels > 2.6.31 in IM mode - we queue the vport delete
		 * work to the IM work_q from the FCS vport delete callback fn
		 * The reason being - If we queue the work to the port_delete
		 * handler here; there can be a case where we invoke the
		 * fc_vport_terminate() on the vport from the port_del handler
		 * while we are waiting for the LOGO for that particular vport.
		 * Since the FCS vport del callbk is called only after the FCS
		 * vport sm is cleaned up we can avoid this race.
		 */
#if (LINUX_VERSION_CODE <= KERNEL_VERSION(2, 6, 31))
		if ((!vp_drv->comp_del) || (roles & BFA_LPORT_ROLE_FCP_TM)) {
#else
		if (roles & BFA_LPORT_ROLE_FCP_TM) {
#endif
			bfad_os_iodm_disable_events(
				vp_drv->drv_port.im_port->shost);
			port_drv = bfad_get_drv_port(bfad, vf_drv, vp_drv);
			bfad_fc4_port_delete(bfad, port_drv, roles);
		}
	}
}

void
bfa_fcb_lport_online(struct bfad_s *bfad, enum bfa_lport_role roles,
		    struct bfad_vf_s *vf_drv, struct bfad_vport_s *vp_drv)
{
	struct bfad_port_s *port_drv = bfad_get_drv_port(bfad, vf_drv, vp_drv);

	if (roles & BFA_LPORT_ROLE_FCP_IM)
		bfad_im_port_online(bfad, port_drv);

	if ((roles & BFA_LPORT_ROLE_FCP_TM) &&
			atomic_read(&tm_ops_flag))
		tm_ops.lport_online(bfad, port_drv);

	if (roles & BFA_LPORT_ROLE_FCP_IPFC)
		if (atomic_read(&bfad->ipfc_ops_flag))
			ipfc_ops.port_online(bfad, port_drv);

	bfad_flags_set(bfad, BFAD_PORT_ONLINE);
}

void
bfa_fcb_lport_offline(struct bfad_s *bfad, enum bfa_lport_role roles,
		     struct bfad_vf_s *vf_drv, struct bfad_vport_s *vp_drv)
{
	struct bfad_port_s *port_drv = bfad_get_drv_port(bfad, vf_drv, vp_drv);

	if (roles & BFA_LPORT_ROLE_FCP_IM)
		bfad_im_port_offline(bfad, port_drv);

	if ((roles & BFA_LPORT_ROLE_FCP_TM) &&
			atomic_read(&tm_ops_flag))
		tm_ops.lport_offline(bfad, port_drv);

	if (roles & BFA_LPORT_ROLE_FCP_IPFC)
		if (atomic_read(&bfad->ipfc_ops_flag))
			ipfc_ops.port_offline(bfad, port_drv);
}

void
bfa_fcb_vf_stop(struct bfad_vf_s *vf_drv)
{
}

void
bfa_fcb_vport_delete(struct bfad_vport_s *vport_drv)
{
	list_del_init(&vport_drv->list_entry);
	if (vport_drv->comp_del) {
		bfad_os_iodm_disable_events(vport_drv->drv_port.im_port->shost);
		complete(vport_drv->comp_del);
		return;
	}

	/*
	 * For kernels > 2.6.31 in IM mode - we queue the vport delete
	 * work to the IM work_q from this FCS vport delete callback fn
	 * The memory for the bfad_vport_s is freed from the FC function
	 * template vport_delete entry point.
	 */
#if (LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 31))
	bfad_os_iodm_disable_events(vport_drv->drv_port.im_port->shost);
	if (vport_drv && vport_drv->drv_port.roles == BFA_LPORT_ROLE_FCP_IM)
		bfad_fc4_port_delete(vport_drv->drv_port.bfad,
			&vport_drv->drv_port, BFA_LPORT_ROLE_FCP_IM);
#else
	kfree(vport_drv);
#endif
}

void
bfa_fcb_vport_login(struct bfad_vport_s *vport_drv, bfa_status_t status)
{
}

bfa_lport_role_t
bfa_fcb_lport_get_role(struct bfad_s *bfad)
{
	return port_role;
}

bfa_trs_cfg_t *
bfad_trs_get_cfg(struct bfad_s *bfad)
{
	return &(bfad->trs_cfg);
}

bfa_status_t
bfa_cb_trs_enqueue_entry(struct bfad_s *bfad, wwn_t lpwwn, wwn_t rpwwn)
{
	bfa_trs_cfg_t *trs_cfg = bfad_trs_get_cfg(bfad);
	bfa_trs_mask_t *trs_mask = NULL;

	/* If TRS list is full, return immediately */
	if (trs_cfg->count == BFA_MAX_TRS_MASK_CFG)
		return BFA_STATUS_MAX_ENTRY_REACHED;

	trs_mask = kzalloc(sizeof(bfa_trs_mask_t), GFP_ATOMIC);
	if (!trs_mask) {
		bfa_trc(bfad, 0);
		return BFA_STATUS_ENOMEM;
	}

	trs_mask->trs_list.lpwwn = lpwwn;
	trs_mask->trs_list.rpwwn = rpwwn;
	bfa_q_enq(&trs_cfg->trs_q, trs_mask);
	++(trs_cfg->count);
	return BFA_STATUS_OK;
}

void
bfa_cb_trs_dequeue_entry(struct bfad_s *bfad, wwn_t lpwwn, wwn_t rpwwn)
{
	bfa_trs_cfg_t *trs_cfg = bfad_trs_get_cfg(bfad);
	bfa_trs_mask_t *trs_mask = NULL;
	bfa_q_t *qe, *qen;

	/* Check for the requested wwns in the TRS queue */
	bfa_q_iter_safe(&trs_cfg->trs_q, qe, qen) {
		trs_mask = (bfa_trs_mask_t *)qe;
		if (trs_mask->trs_list.lpwwn == lpwwn &&
				trs_mask->trs_list.rpwwn == rpwwn) {
			bfa_q_qe_deq(trs_mask);
			kfree(trs_mask);
			--(trs_cfg->count);
			return;
		}
	}
}

void
bfa_cb_trs_clear_list(struct bfad_s *bfad)
{
	bfa_trs_cfg_t *trs_cfg = bfad_trs_get_cfg(bfad);
	bfa_trs_mask_t *trs_mask = NULL;
	bfa_q_t *qe, *qen;

	/* Dequeue the elements in TRS queue and free the memory */
	bfa_q_iter_safe(&trs_cfg->trs_q, qe, qen) {
		trs_mask = (bfa_trs_mask_t *)qe;
		bfa_q_qe_deq(trs_mask);
		kfree(trs_mask);
	}

	trs_cfg->count = 0;
}

/**
 * @brief
 * FCS RPORT alloc callback, after successful PLOGI by FCS
 */
bfa_status_t
bfa_fcb_rport_alloc(struct bfad_s *bfad, struct bfa_fcs_rport_s **rport,
		    struct bfad_rport_s **rport_drv)
{
	bfa_status_t	rc = BFA_STATUS_OK;

	*rport_drv = kzalloc(sizeof(struct bfad_rport_s), GFP_ATOMIC);
	if (*rport_drv == NULL) {
		rc = BFA_STATUS_ENOMEM;
		goto ext;
	}

	*rport = &(*rport_drv)->fcs_rport;

ext:
	return rc;
}

/**
 * @brief
 * FCS RPORT free callback.
 */
void
bfa_fcb_rport_free(struct bfad_s *bfad, struct bfad_rport_s **rport_drv)
{
	kfree(*rport_drv);
}

/**
 * @brief
 * FCS PBC VPORT Create
 */
void
bfa_fcb_pbc_vport_create(struct bfad_s *bfad, struct bfi_pbc_vport_s pbc_vport)
{

	struct bfa_lport_cfg_s port_cfg = {0};
	struct bfad_vport_s   *vport;
	int rc;

	vport = kzalloc(sizeof(struct bfad_vport_s), GFP_KERNEL);
	if (!vport) {
		bfa_trc(bfad, 0);
		return;
	}

	vport->drv_port.bfad = bfad;
	port_cfg.roles = BFA_LPORT_ROLE_FCP_IM;
	port_cfg.pwwn = pbc_vport.vp_pwwn;
	port_cfg.nwwn = pbc_vport.vp_nwwn;
	port_cfg.preboot_vp  = BFA_TRUE;

	rc = bfa_fcs_pbc_vport_create(&vport->fcs_vport, &bfad->bfa_fcs, 0,
				  &port_cfg, vport);

	if (rc != BFA_STATUS_OK) {
		bfa_trc(bfad, 0);
		return;
	}

	list_add_tail(&vport->list_entry, &bfad->pbc_vport_list);

}

/**
 * @}
 */

void
bfad_hal_mem_release(struct bfad_s *bfad)
{
	struct bfa_meminfo_s *hal_meminfo = &bfad->meminfo;
	struct bfa_mem_dma_s *dma_info, *dma_elem;
	struct bfa_mem_kva_s *kva_info, *kva_elem;
	struct bfa_q_s	*dm_qe, *km_qe;

	dma_info = &hal_meminfo->dma_info;
	kva_info = &hal_meminfo->kva_info;

	/* Iterate through the KVA meminfo queue */
	bfa_q_iter(&kva_info->qe, km_qe) {
		kva_elem = (struct bfa_mem_kva_s *) km_qe;
		if (kva_elem->kva)
			vfree(kva_elem->kva);
	}

	/* Iterate through the DMA meminfo queue */
	bfa_q_iter(&dma_info->qe, dm_qe) {
		dma_elem = (struct bfa_mem_dma_s *) dm_qe;
		if (dma_elem->kva && dma_elem->dma)
			bfad_os_dma_free(bfad, dma_elem);
	}

	memset(hal_meminfo, 0, sizeof(struct bfa_meminfo_s));
}

void
bfad_update_hal_cfg(struct bfa_iocfc_cfg_s *bfa_cfg)
{
	if (num_rports > 0)
		bfa_cfg->fwcfg.num_rports = num_rports;
	if (num_ioims > 0)
		bfa_cfg->fwcfg.num_ioim_reqs = num_ioims;
	if (num_tios > 0)
		bfa_cfg->fwcfg.num_fwtio_reqs = num_tios;
	if (num_tms > 0)
		bfa_cfg->fwcfg.num_tskim_reqs = num_tms;
	if (num_fcxps > 0 && num_fcxps <= BFA_FCXP_MAX)
		bfa_cfg->fwcfg.num_fcxp_reqs = num_fcxps;
	if (num_ufbufs > 0 && num_ufbufs <= BFA_UF_MAX)
		bfa_cfg->fwcfg.num_uf_bufs = num_ufbufs;
	if (reqq_size > 0)
		bfa_cfg->drvcfg.num_reqq_elems = reqq_size;
	if (rspq_size > 0)
		bfa_cfg->drvcfg.num_rspq_elems = rspq_size;
	if (num_sgpgs > 0 && num_sgpgs <= BFA_SGPG_MAX)
		bfa_cfg->drvcfg.num_sgpgs = num_sgpgs;

	if ((bfa_cfg->fwcfg.num_ioim_reqs +
			bfa_cfg->fwcfg.num_fwtio_reqs) > BFA_IO_MAX) {
		if (num_ioims && num_tios) {
			bfa_cfg->fwcfg.num_ioim_reqs = BFA_IO_MAX/2;
			bfa_cfg->fwcfg.num_fwtio_reqs = BFA_IO_MAX/2;
		} else if (num_tios) {
			bfa_cfg->fwcfg.num_fwtio_reqs = BFA_FWTIO_MAX;
		} else if (num_ioims) {
			bfa_cfg->fwcfg.num_ioim_reqs = BFA_IOIM_MAX;
		} else {
			/*
			 * Module params num_ioims, num_tios are not available
			 * and default config for num_ioims and num_fwtios is
			 * invalid. These values are adjusted by BFA_FCP_MOD.
			 */
		}
	}

	/*
	 * populate the hal values back to the driver for sysfs use.
	 * otherwise, the default values will be shown as 0 in sysfs
	 */
	num_rports = bfa_cfg->fwcfg.num_rports;
	num_ioims = bfa_cfg->fwcfg.num_ioim_reqs;
	num_tios = bfa_cfg->fwcfg.num_fwtio_reqs;
	num_tms = bfa_cfg->fwcfg.num_tskim_reqs;
	num_fcxps = bfa_cfg->fwcfg.num_fcxp_reqs;
	num_ufbufs = bfa_cfg->fwcfg.num_uf_bufs;
	reqq_size = bfa_cfg->drvcfg.num_reqq_elems;
	rspq_size = bfa_cfg->drvcfg.num_rspq_elems;
	num_sgpgs = bfa_cfg->drvcfg.num_sgpgs;
}

bfa_status_t
bfad_hal_mem_alloc(struct bfad_s *bfad)
{
	struct bfa_meminfo_s *hal_meminfo = &bfad->meminfo;
	struct bfa_mem_dma_s *dma_info, *dma_elem;
	struct bfa_mem_kva_s *kva_info, *kva_elem;
	struct bfa_q_s	*dm_qe, *km_qe;
	bfa_status_t	rc = BFA_STATUS_OK;
	dma_addr_t phys_addr;

	bfa_cfg_get_default(&bfad->ioc_cfg);
	bfad_update_hal_cfg(&bfad->ioc_cfg);
	bfad->cfg_data.ioc_queue_depth = bfad->ioc_cfg.fwcfg.num_ioim_reqs;
	bfa_cfg_get_meminfo(&bfad->ioc_cfg, hal_meminfo, &bfad->bfa);

	dma_info = &hal_meminfo->dma_info;
	kva_info = &hal_meminfo->kva_info;

	bfa_q_iter(&kva_info->qe, km_qe) {
		kva_elem = (struct bfa_mem_kva_s *) km_qe;
		kva_elem->kva = NULL;
	}

	bfa_q_iter(&dma_info->qe, dm_qe) {
		dma_elem = (struct bfa_mem_dma_s *) dm_qe;
		dma_elem->kva = NULL;
		dma_elem->dma = 0;
	}

	/* Iterate through the KVA meminfo queue */
	bfa_q_iter(&kva_info->qe, km_qe) {
		kva_elem = (struct bfa_mem_kva_s *) km_qe;
		kva_elem->kva = vmalloc(kva_elem->mem_len);
		if (kva_elem->kva == NULL) {
			bfad_hal_mem_release(bfad);
			rc = BFA_STATUS_ENOMEM;
			goto ext;
		}
		memset(kva_elem->kva, 0, kva_elem->mem_len);
	}

	/* Iterate through the DMA meminfo queue */
	bfa_q_iter(&dma_info->qe, dm_qe) {
		dma_elem = (struct bfa_mem_dma_s *) dm_qe;
		dma_elem->kva = bfad_os_dma_alloc(bfad, dma_elem, &phys_addr);
		if (dma_elem->kva == NULL) {
			bfad_hal_mem_release(bfad);
			rc = BFA_STATUS_ENOMEM;
			goto ext;
		}
		dma_elem->dma = phys_addr;
		memset(dma_elem->kva, 0, dma_elem->mem_len);
	}
ext:
	return rc;
}

struct bfad_s *
bfad_find_bfad_by_inst_no(int inst_no)
{
	struct bfad_s	      *bfad = NULL, *rc = NULL;

	down(&bfad_sem);
	list_for_each_entry(bfad, &bfad_list, list_entry) {
		if (bfad->inst_no == inst_no) {
			rc = bfad;
			break;
		}
	}
	up(&bfad_sem);
	return rc;
}

bfa_status_t
bfad_fc4_probe(struct bfad_s *bfad)
{
	int		rc;

	rc = bfad_im_probe(bfad);
	if (rc != BFA_STATUS_OK)
		goto ext;

	if (atomic_read(&bfad->ipfc_ops_flag))
		ipfc_ops.probe(bfad);

	bfad->bfad_flags |= BFAD_FC4_PROBE_DONE;

ext:
	return rc;
}

void
bfad_fc4_probe_undo(struct bfad_s *bfad)
{

	bfad_im_probe_undo(bfad);
	if (atomic_read(&bfad->ipfc_ops_flag))
		ipfc_ops.probe_undo(bfad);

	bfad->bfad_flags &= ~BFAD_FC4_PROBE_DONE;
}

static void
bfad_fc4_probe_post(struct bfad_s *bfad)
{
	if (bfad->im)
		bfad_im_probe_post(bfad->im);
	if (atomic_read(&bfad->ipfc_ops_flag))
		ipfc_ops.probe_post(bfad);
}

static bfa_status_t
bfad_fc4_port_new(struct bfad_s *bfad, struct bfad_port_s *port, int roles)
{
	int		rc = BFA_STATUS_OK;

	if (roles & BFA_LPORT_ROLE_FCP_IM)
		rc = bfad_im_port_new(bfad, port);
	if (rc != BFA_STATUS_OK)
		goto ext;
	if ((roles & BFA_LPORT_ROLE_FCP_TM) &&
			atomic_read(&tm_ops_flag)) {
		rc = tm_ops.lport_new(bfad, port);
		if (rc != BFA_STATUS_OK)
			goto ext;
	}

	if (roles & BFA_LPORT_ROLE_FCP_IPFC)
		if (atomic_read(&bfad->ipfc_ops_flag))
		ipfc_ops.probe(bfad);
ext:
	return rc;
}

static void
bfad_fc4_port_delete(struct bfad_s *bfad, struct bfad_port_s *port, int roles)
{

	if (roles & BFA_LPORT_ROLE_FCP_IM)
		bfad_im_port_delete(bfad, port);

	if ((roles & BFA_LPORT_ROLE_FCP_TM) &&
			atomic_read(&tm_ops_flag))
		tm_ops.lport_delete(bfad, port);

	if (roles & BFA_LPORT_ROLE_FCP_IPFC)
	   if (atomic_read(&bfad->ipfc_ops_flag))
		ipfc_ops.probe_undo(bfad);
}

/**
 * @brief
 * Create a vport under a vf.
 */
bfa_status_t
bfad_vport_create(struct bfad_s *bfad, uint16_t vf_id,
		  struct bfa_lport_cfg_s *port_cfg, struct device *dev)
{
	struct bfad_vport_s   *vport;
	int		rc = BFA_STATUS_OK;
	unsigned long	flags;
	struct completion fcomp;

	vport = kzalloc(sizeof(struct bfad_vport_s), GFP_KERNEL);
	if (!vport) {
		rc = BFA_STATUS_ENOMEM;
		goto ext;
	}

	vport->drv_port.bfad = bfad;
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	rc = bfa_fcs_vport_create(&vport->fcs_vport, &bfad->bfa_fcs, vf_id,
				  port_cfg, vport);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	if (rc != BFA_STATUS_OK)
		goto ext_free_vport;

	if (port_cfg->roles & BFA_LPORT_ROLE_FCP_IM) {
		rc = bfad_im_scsi_host_alloc(bfad, vport->drv_port.im_port,
							dev);
		if (rc != BFA_STATUS_OK)
			goto ext_free_fcs_vport;
	}

	bfad_os_scsi_ioqueue_register(&vport->drv_port);

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	if (!(port_cfg->roles & BFA_LPORT_ROLE_FCP_TM) ||
			(atomic_read(&tm_ops_flag))) {
		bfa_fcs_vport_start(&vport->fcs_vport);
	}
	list_add_tail(&vport->list_entry, &bfad->vport_list);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	bfad_os_iodm_enable_events(vport->drv_port.im_port->shost);

	return BFA_STATUS_OK;

ext_free_fcs_vport:
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	vport->comp_del = &fcomp;
	init_completion(vport->comp_del);
	bfa_fcs_vport_delete(&vport->fcs_vport);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	wait_for_completion(vport->comp_del);
ext_free_vport:
	kfree(vport);
ext:
	return rc;
}

/**
 * @brief
 * Create a vf and its base vport implicitely.
 */
bfa_status_t
bfad_vf_create(struct bfad_s *bfad, uint16_t vf_id,
	       struct bfa_lport_cfg_s *port_cfg)
{
	struct bfad_vf_s      *vf;
	int		rc = BFA_STATUS_OK;

	vf = kzalloc(sizeof(struct bfad_vf_s), GFP_KERNEL);
	if (!vf) {
		rc = BFA_STATUS_FAILED;
		goto ext;
	}

	rc = bfa_fcs_vf_create(&vf->fcs_vf, &bfad->bfa_fcs, vf_id, port_cfg,
			       vf);
	if (rc != BFA_STATUS_OK)
		kfree(vf);
ext:
	return rc;
}

void
bfad_bfa_tmo(unsigned long data)
{
	struct bfad_s	      *bfad = (struct bfad_s *) data;
	unsigned long	flags;
	struct bfa_q_s	       doneq;

	spin_lock_irqsave(&bfad->bfad_lock, flags);

	bfa_timer_tick(&bfad->bfa);

	bfa_comp_deq(&bfad->bfa, &doneq);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	if (!bfa_q_is_empty(&doneq)) {
		bfa_comp_process(&bfad->bfa, &doneq);
		spin_lock_irqsave(&bfad->bfad_lock, flags);
		bfa_comp_free(&bfad->bfa, &doneq);
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	}

	mod_timer(&bfad->hal_tmo,
		  jiffies + msecs_to_jiffies(BFA_TIMER_FREQ));
}

void
bfad_init_timer(struct bfad_s *bfad)
{
	init_timer(&bfad->hal_tmo);
	bfad->hal_tmo.function = bfad_bfa_tmo;
	bfad->hal_tmo.data = (unsigned long)bfad;

	mod_timer(&bfad->hal_tmo,
		  jiffies + msecs_to_jiffies(BFA_TIMER_FREQ));
}

int
bfad_pci_init(struct pci_dev *pdev, struct bfad_s *bfad)
{
	unsigned long	bar_len;
	int		rc = -ENODEV;

	if (pci_enable_device(pdev)) {
		BFA_PRINTF(BFA_ERR, "pci_enable_device fail %p\n", pdev);
		goto out;
	}

	if (pci_request_regions(pdev, BFAD_DRIVER_NAME))
		goto out_disable_device;

	pci_set_master(pdev);

	bfad_os_pci_set_mwi(pdev);

	if ((pci_set_dma_mask(pdev, DMA_BIT_MASK(64)) != 0) ||
	    (pci_set_consistent_dma_mask(pdev, DMA_BIT_MASK(64)) != 0))
		if ((pci_set_dma_mask(pdev, DMA_BIT_MASK(32)) != 0) ||
		   (pci_set_consistent_dma_mask(pdev, DMA_BIT_MASK(32)) != 0)) {
			BFA_PRINTF(BFA_ERR, "pci_set_dma_mask fail %p\n", pdev);
			goto out_release_region;
		}

#ifdef SUPPORT_PCI_AER
	/*
	 * Enable PCIE Advanced Error Recovery (AER) if the kernel version
	 * supports.
	 */
	BFAD_ENABLE_PCIE_AER(pdev);
#endif

	bfad->pci_bar0_map = pci_resource_start(pdev, 0);
	bar_len = pci_resource_len(pdev, 0);
	bfad->pci_bar0_kva = ioremap(bfad->pci_bar0_map, bar_len);

	bfad->pci_bar2_map = pci_resource_start(pdev, 2);
	bar_len = pci_resource_len(pdev, 2);
	bfad->pci_bar2_kva = ioremap(bfad->pci_bar2_map, bar_len);

	if (bfad->pci_bar0_kva == NULL) {
		BFA_PRINTF(BFA_ERR, "Fail to map bar0\n");
		goto out_release_region;
	}

	bfad->hal_pcidev.pci_slot = PCI_SLOT(pdev->devfn);
	bfad->hal_pcidev.pci_func = PCI_FUNC(pdev->devfn);
	bfad->hal_pcidev.pci_bar_kva = bfad->pci_bar0_kva;
	bfad->hal_pcidev.device_id = pdev->device;
	bfad->hal_pcidev.ssid = pdev->subsystem_device;
	bfad->pci_name = pci_name(pdev);

	bfad->pci_attr.vendor_id = pdev->vendor;
	bfad->pci_attr.device_id = pdev->device;
	bfad->pci_attr.ssid = pdev->subsystem_device;
	bfad->pci_attr.ssvid = pdev->subsystem_vendor;
	bfad->pci_attr.pcifn = PCI_FUNC(pdev->devfn);

	bfad->pcidev = pdev;

	/* Adjust PCIe Maximum Read Request Size */
	if (pcie_max_read_reqsz > 0) {
		int pcie_cap_reg;
		uint16_t pcie_dev_ctl;
		uint16_t mask = 0xffff;

		switch (pcie_max_read_reqsz) {
		case 128:
			mask = 0x0;
			break;
		case 256:
			mask = 0x1000;
			break;
		case 512:
			mask = 0x2000;
			break;
		case 1024:
			mask = 0x3000;
			break;
		case 2048:
			mask = 0x4000;
			break;
		case 4096:
			mask = 0x5000;
			break;
		default:
			break;
		}

		pcie_cap_reg = pci_find_capability(pdev, PCI_CAP_ID_EXP);
		if (mask != 0xffff && pcie_cap_reg) {
			pcie_cap_reg += 0x08;
			pci_read_config_word(pdev, pcie_cap_reg, &pcie_dev_ctl);
			if ((pcie_dev_ctl & 0x7000) != mask) {
				printk(KERN_WARNING "bfa %s: "
				"pcie_max_read_request_size is %d, "
				"reset to %d\n", bfad->pci_name,
				(1 << ((pcie_dev_ctl & 0x7000) >> 12)) << 7,
				pcie_max_read_reqsz);

				pcie_dev_ctl &= ~0x7000;
				pci_write_config_word(pdev, pcie_cap_reg,
						pcie_dev_ctl | mask);
			}
		}
	}

	bfad_os_pci_save_state(pdev);

	return 0;

out_release_region:
	pci_release_regions(pdev);
out_disable_device:
	pci_disable_device(pdev);
out:
	return rc;
}

void
bfad_pci_uninit(struct pci_dev *pdev, struct bfad_s *bfad)
{
	bfad_os_iounmap(pdev, bfad);
	pci_release_regions(pdev);
#ifdef SUPPORT_PCI_AER
	/*
	 * Disable PCIE Advanced Error Recovery (AER).
	 */
	BFAD_DISABLE_PCIE_AER(pdev);
#endif
	pci_disable_device(pdev);
	pci_set_drvdata(pdev, NULL);
}

void
bfad_fcs_port_cfg(struct bfad_s *bfad)
{
	struct bfa_lport_cfg_s  port_cfg;
	struct bfa_port_attr_s attr;
	char		symname[BFA_SYMNAME_MAXLEN];

	sprintf(symname, "%s-%d", BFAD_DRIVER_NAME, bfad->inst_no);
	memcpy(port_cfg.sym_name.symname, symname, strlen(symname));
	bfa_fcport_get_attr(&bfad->bfa, &attr);
	port_cfg.nwwn = attr.nwwn;
	port_cfg.pwwn = attr.pwwn;

	bfa_fcs_cfg_base_port(&bfad->bfa_fcs, &port_cfg);
}

bfa_status_t
bfad_drv_init(struct bfad_s *bfad)
{
	bfa_status_t	rc;
	unsigned long	flags;

	bfad->cfg_data.rport_del_timeout = rport_del_timeout;
	bfad->cfg_data.lun_queue_depth = bfa_lun_queue_depth;
	bfad->cfg_data.io_max_sge = bfa_io_max_sge;
	bfad->cfg_data.binding_method = FCP_PWWN_BINDING;

	rc = bfad_hal_mem_alloc(bfad);
	if (rc != BFA_STATUS_OK) {
		printk(KERN_WARNING "bfa %s: bfad_hal_mem_alloc failure\n",
		       bfad->pci_name);
		printk(KERN_WARNING
			"Not enough memory to attach all QLogic BR-series HBA "
			"ports, %s", "System may need more memory.\n");
		return BFA_STATUS_FAILED;
	}

	bfa_init_log(&bfad->bfa, bfad->logm);
	bfa_init_trc(&bfad->bfa, bfad->trcmod);
	bfa_init_aen(&bfad->bfa, bfad->aen);
	memset(bfad->file_map, 0, sizeof(bfad->file_map));
	bfa_init_plog(&bfad->bfa, &bfad->plog_buf);
	bfa_plog_init(&bfad->plog_buf);
	bfa_plog_str(&bfad->plog_buf, BFA_PL_MID_DRVR, BFA_PL_EID_DRIVER_START,
		     0, "Driver Attach");

	bfa_attach(&bfad->bfa, bfad, &bfad->ioc_cfg, &bfad->meminfo,
		   &bfad->hal_pcidev);

	/* FCS INIT */
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	bfa_fcs_log_init(&bfad->bfa_fcs, bfad->logm);
	bfa_fcs_trc_init(&bfad->bfa_fcs, bfad->trcmod);
	bfa_fcs_aen_init(&bfad->bfa_fcs, bfad->aen);
	bfa_fcs_attach(&bfad->bfa_fcs, &bfad->bfa, bfad, BFA_FALSE);
	bfa_fcs_set_fdmi_param(&bfad->bfa_fcs, fdmi_enable);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	bfad->bfad_flags |= BFAD_DRV_INIT_DONE;

	return BFA_STATUS_OK;
}

void
bfad_drv_start(struct bfad_s *bfad)
{
	unsigned long	flags;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	bfa_start(&bfad->bfa);
	if (!(bfad->pport.roles & BFA_LPORT_ROLE_FCP_TM)) {
		bfa_fcs_start(&bfad->bfa_fcs);
	}
	bfad->bfad_flags |= BFAD_HAL_START_DONE;
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	bfad_fc4_probe_post(bfad);
}

void
bfad_fcs_stop(struct bfad_s *bfad)
{
	unsigned long	flags;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	init_completion(&bfad->comp);
	bfad->pport.flags |= BFAD_PORT_DELETE;
	bfa_fcs_exit(&bfad->bfa_fcs);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	wait_for_completion(&bfad->comp);

	bfa_sm_send_event(bfad, BFAD_E_FCS_EXIT_COMP);
}

void
bfad_stop(struct bfad_s *bfad)
{
	unsigned long	flags;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	init_completion(&bfad->comp);
	bfa_stop(&bfad->bfa);
	bfad->bfad_flags &= ~BFAD_HAL_START_DONE;
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	wait_for_completion(&bfad->comp);

	bfa_sm_send_event(bfad, BFAD_E_EXIT_COMP);
}

bfa_status_t
bfad_cfg_pport(struct bfad_s *bfad, enum bfa_lport_role role)
{
	int		rc = BFA_STATUS_OK;

	/* Allocate scsi_host for the physical port */
	if ((supported_fc4s & BFA_LPORT_ROLE_FCP_IM) &&
	    (role & BFA_LPORT_ROLE_FCP_IM)) {
		if (bfad->pport.im_port == NULL) {
			rc = BFA_STATUS_FAILED;
			goto out;
		}
		rc = bfad_im_scsi_host_alloc(bfad, bfad->pport.im_port,
						&bfad->pcidev->dev);
		if (rc != BFA_STATUS_OK)
			goto out;

		bfad->pport.roles |= BFA_LPORT_ROLE_FCP_IM;
	}

	if ((supported_fc4s & BFA_LPORT_ROLE_FCP_TM) &&
			(role & BFA_LPORT_ROLE_FCP_TM)) {
		bfad->pport.roles |= BFA_LPORT_ROLE_FCP_TM;
	}

	bfad->bfad_flags |= BFAD_CFG_PPORT_DONE;

out:
	return rc;
}

void
bfad_uncfg_pport(struct bfad_s *bfad)
{
	if (bfad->pport.roles & BFA_LPORT_ROLE_FCP_IPFC) {
		if (atomic_read(&bfad->ipfc_ops_flag))
			ipfc_ops.probe_undo(bfad);
		bfad->pport.roles &= ~BFA_LPORT_ROLE_FCP_IPFC;
	}

	if ((supported_fc4s & BFA_LPORT_ROLE_FCP_IM) &&
	    (bfad->pport.roles & BFA_LPORT_ROLE_FCP_IM)) {
		bfad_im_scsi_host_free(bfad, bfad->pport.im_port);
		bfad_im_port_clean(bfad->pport.im_port);
		kfree(bfad->pport.im_port);
		bfad->pport.roles &= ~BFA_LPORT_ROLE_FCP_IM;
	}

	if (bfad->pport.roles & BFA_LPORT_ROLE_FCP_TM) {
		bfad->pport.roles &= ~BFA_LPORT_ROLE_FCP_TM;
	}

	bfad->bfad_flags &= ~BFAD_CFG_PPORT_DONE;
}

void
bfad_drv_log_level_set(struct bfad_s *bfad)
{
	if (log_level > BFA_LOG_INVALID && log_level <= BFA_LOG_LEVEL_MAX)
		bfa_log_set_level_all(&bfad->log_data, log_level);
}

int
bfad_get_num_pbc_vports(struct bfad_s *bfad)
{
	struct bfad_vport_s *vport;
	int count = 0;

	list_for_each_entry(vport, &bfad->pbc_vport_list, list_entry)
		count++;

	return count;
}

bfa_status_t
bfad_start_ops(struct bfad_s *bfad) {

	int	retval;
	struct bfa_fcs_driver_info_s driver_info;
	struct bfad_vport_s *vport;
#if (LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 31))
	struct bfad_vport_s *vport_new;
#endif
	unsigned long flags;

	/* Limit min/max. xfer size to [64k-32MB] */
	if (max_xfer_size < BFAD_MIN_SECTORS >> 1)
		max_xfer_size = BFAD_MIN_SECTORS >> 1;
	if (max_xfer_size > BFAD_MAX_SECTORS >> 1)
		max_xfer_size = BFAD_MAX_SECTORS >> 1;

	/* Fill the driver_info info to fcs */
	memset(&driver_info, 0, sizeof(driver_info));
	strncpy(driver_info.version, BFAD_DRIVER_VERSION,
		sizeof(driver_info.version) - 1);
	if (host_name)
		strncpy(driver_info.host_machine_name, host_name,
			sizeof(driver_info.host_machine_name) - 1);
	if (os_name)
		strncpy(driver_info.host_os_name, os_name,
			sizeof(driver_info.host_os_name) - 1);
	if (os_patch)
		strncpy(driver_info.host_os_patch, os_patch,
			sizeof(driver_info.host_os_patch) - 1);

	strncpy(driver_info.os_device_name, bfad->pci_name,
		sizeof(driver_info.os_device_name) - 1);

	/* FCS driver info init */
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	bfa_fcs_driver_info_init(&bfad->bfa_fcs, &driver_info);

	if (bfad->bfad_flags & BFAD_CFG_PPORT_DONE)
		bfa_fcs_update_cfg(&bfad->bfa_fcs);
	else
		bfa_fcs_init(&bfad->bfa_fcs);

	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	/* PPORT FCS config */
	bfad_fcs_port_cfg(bfad);

	if (!(bfad->bfad_flags & BFAD_CFG_PPORT_DONE)) {
		retval = bfad_cfg_pport(bfad, port_role);
		if (retval != BFA_STATUS_OK)
			return BFA_STATUS_FAILED;
	}

	bfad_os_scsi_ioqueue_register(&bfad->pport);

	/* Setup fc host fixed attribute if the lk supports */
	bfad_os_fc_host_init(bfad->pport.im_port);

	/* BFAD level FC4 (IM/TM/IPFC) specific resource allocation */
	retval = bfad_fc4_probe(bfad);
	if (retval != BFA_STATUS_OK) {
		printk(KERN_WARNING "bfad_fc4_probe failed\n");
		return BFA_STATUS_FAILED;
	}

	bfad_os_iodm_enable_events(bfad->pport.im_port->shost);

	bfad_drv_start(bfad);
#if (LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 31))
	list_for_each_entry_safe(vport, vport_new, &bfad->pbc_vport_list,
					list_entry) {
		struct fc_vport_identifiers vid;
		struct fc_vport *fc_vport;
		char pwwn_buf[BFA_STRING_32];

		memset(&vid, 0, sizeof(vid));
		vid.roles = FC_PORT_ROLE_FCP_INITIATOR;
		vid.vport_type = FC_PORTTYPE_NPIV;
		vid.disable = false;
		vid.node_name = wwn_to_u64((u8 *)
				(&((vport->fcs_vport).lport.port_cfg.nwwn)));
		vid.port_name = wwn_to_u64((u8 *)
				(&((vport->fcs_vport).lport.port_cfg.pwwn)));
		fc_vport = fc_vport_create(bfad->pport.im_port->shost, 0, &vid);
		if (!fc_vport) {
			wwn2str(pwwn_buf, sizeof(pwwn_buf), vid.port_name);
			printk(KERN_WARNING "bfa %s: failed to create pbc vport"
				" %s\n", bfad->pci_name, pwwn_buf);
		}
		list_del(&vport->list_entry);
		kfree(vport);
	}
#else
	/* Complete pbc vport create */
	list_for_each_entry(vport, &bfad->pbc_vport_list, list_entry) {
		retval = bfad_im_scsi_host_alloc(bfad, vport->drv_port.im_port,
						 0);
		if (retval == BFA_STATUS_OK) {
			bfa_fcs_vport_start(&vport->fcs_vport);
		} else {
			printk("bfa %s: pbc_vport %p failed to allocate"
				" shost\n", bfad->pci_name, vport);
		}
	}
#endif
	/*
	 * If linkup_delay is set to -1 default; try to retrive the
	 * value using the bfad_os_get_linkup_delay(); else use the
	 * passed in module param value as the linkup_delay.
	 */
	if (linkup_delay < 0) {
		linkup_delay = bfad_os_get_linkup_delay(bfad);
		bfad_os_rport_online_wait(bfad);
		linkup_delay = -1;
	} else {
		bfad_os_rport_online_wait(bfad);
	}

	bfad_os_scsi_scan(bfad->pport.im_port);
	bfa_log(bfad->logm, BFA_LOG_LINUX_DEVICE_CLAIMED, bfad->pci_name);
	return BFA_STATUS_OK;
}

int
bfad_worker(void *ptr)
{
	struct bfad_s *bfad;
	unsigned long   flags;

	bfad = (struct bfad_s *)ptr;

	while (!bfad_os_kthread_should_stop()) {

		/* Send event BFAD_E_INIT_SUCCESS */
		bfa_sm_send_event(bfad, BFAD_E_INIT_SUCCESS);

		spin_lock_irqsave(&bfad->bfad_lock, flags);
		bfad->bfad_tsk = NULL;
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);

		break;
	}

	return 0;
}

/**
 * @dg bfa_isr BFA driver interrupt functions
 * @{
 */
irqreturn_t bfad_intx(int irq, void *dev_id);
int msix_disable_cb = 0, msix_disable_ct = 0;

const char *msix_name_ct[] = {
	"ctrl",
	"cpe0", "cpe1", "cpe2", "cpe3",
	"rme0", "rme1", "rme2", "rme3" };

const char *msix_name_cb[] = {
	"cpe0", "cpe1", "cpe2", "cpe3",
	"rme0", "rme1", "rme2", "rme3",
	"eemc", "elpu0", "elpu1", "epss", "mlpu" };

/**
 * @brief
 * Line based interrupt handler.
 */
irqreturn_t
bfad_intx(int irq, void *dev_id)
{
	struct bfad_s	*bfad = dev_id;
	struct bfa_q_s	doneq;
	unsigned long	flags;
	bfa_boolean_t rc;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	rc = bfa_intx(&bfad->bfa);
	if (!rc) {
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		return IRQ_NONE;
	}

	bfa_comp_deq(&bfad->bfa, &doneq);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	if (!bfa_q_is_empty(&doneq)) {
		bfa_comp_process(&bfad->bfa, &doneq);

		spin_lock_irqsave(&bfad->bfad_lock, flags);
		bfa_comp_free(&bfad->bfa, &doneq);
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		bfa_trc_fp(bfad, irq);
	}

	return IRQ_HANDLED;

}

#ifdef CONFIG_PCI_MSI

static irqreturn_t
bfad_msix(int irq, void *dev_id)
{
	struct bfad_msix_s *vec = dev_id;
	struct bfad_s *bfad = vec->bfad;
	struct bfa_q_s doneq;
	unsigned long   flags;

	spin_lock_irqsave(&bfad->bfad_lock, flags);

	bfa_msix(&bfad->bfa, vec->msix.entry);
	bfa_comp_deq(&bfad->bfa, &doneq);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	if (!bfa_q_is_empty(&doneq)) {
		bfa_comp_process(&bfad->bfa, &doneq);

		spin_lock_irqsave(&bfad->bfad_lock, flags);
		bfa_comp_free(&bfad->bfa, &doneq);
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	}

	return IRQ_HANDLED;
}

/**
 * @brief
 * Initialize the MSIX entry table.
 */
static void
bfad_init_msix_entry(struct bfad_s *bfad, struct msix_entry *msix_entries,
			 int mask, int max_bit)
{
	int	i;
	int	match = 0x00000001;

	for (i = 0, bfad->nvec = 0; i < MAX_MSIX_ENTRY; i++) {
		if (mask & match) {
			bfad->msix_tab[bfad->nvec].msix.entry = i;
			bfad->msix_tab[bfad->nvec].bfad = bfad;
			msix_entries[bfad->nvec].entry = i;
			bfad->nvec ++;
		}

		match <<= 1;
	}

}

int
bfad_install_msix_handler(struct bfad_s *bfad)
{
	int i, error = 0;

	for (i = 0; i < bfad->nvec; i++) {
		sprintf(bfad->msix_tab[i].name, "bfa-%s-%s",
				bfad->pci_name,
				((bfa_asic_id_cb(bfad->hal_pcidev.device_id)) ?
				msix_name_cb[i] : msix_name_ct[i]));

		error = request_irq(bfad->msix_tab[i].msix.vector,
				    (bfad_isr_t) bfad_msix, 0,
				    bfad->msix_tab[i].name, &bfad->msix_tab[i]);
		bfa_trc(bfad, i);
		bfa_trc(bfad, bfad->msix_tab[i].msix.vector);
		if (error) {
			int	j;

			for (j = 0; j < i; j++)
				free_irq(bfad->msix_tab[j].msix.vector,
						&bfad->msix_tab[j]);

			bfad->bfad_flags &= ~BFAD_MSIX_ON;
			pci_disable_msix(bfad->pcidev);

			return 1;
		}
	}

	return 0;
}

/**
 * @brief
 * Setup MSIX based interrupt.
 */
int
bfad_setup_intr(struct bfad_s *bfad)
{
	int error = 0;
	uint32_t mask = 0, i, num_bit = 0, max_bit = 0;
	struct msix_entry msix_entries[MAX_MSIX_ENTRY];
	struct pci_dev *pdev = bfad->pcidev;
	uint16_t reg;

	/* Call BFA to get the msix map for this PCI function.  */
	bfa_msix_getvecs(&bfad->bfa, &mask, &num_bit, &max_bit);

	/* Set up the msix entry table */
	bfad_init_msix_entry(bfad, msix_entries, mask, max_bit);

	if ((bfa_asic_id_ctc(pdev->device) && !msix_disable_ct) ||
	    (bfa_asic_id_cb(pdev->device) && !msix_disable_cb)) {

		error = pci_enable_msix(bfad->pcidev, msix_entries, bfad->nvec);
		if (error) {
			/* In CT1 & CT2, try to allocate just one vector */
			if (bfa_asic_id_ctc(pdev->device)) {
			    printk(KERN_WARNING "bfa %s: trying one msix "
			    "vector failed to allocate %d[%d]\n",
					bfad->pci_name, bfad->nvec, error);

			    bfad->nvec = 1;
			    error = pci_enable_msix(bfad->pcidev,
						msix_entries, bfad->nvec);
			}

			/*
			 * Only error number of vector is available.
			 * We don't have a mechanism to map multiple
			 * interrupts into one vector, so even if we
			 * can try to request less vectors, we don't
			 * know how to associate interrupt events to
			 *  vectors. Linux doesn't dupicate vectors
			 * in the MSIX table for this case.
			 */

			if (error) {
				printk(KERN_WARNING "bfa %s: "
				"pci_enable_msix failed (%d),"
				" use line based.\n", bfad->pci_name, error);

				goto line_based;
			}
		}

		/* Disable INTX in MSI-X mode */
		pci_read_config_word(pdev, PCI_COMMAND, &reg);

		if (!(reg & PCI_COMMAND_INTX_DISABLE))
			pci_write_config_word(pdev, PCI_COMMAND,
				reg | PCI_COMMAND_INTX_DISABLE);

		/* Save the vectors */
		for (i = 0; i < bfad->nvec; i++) {
			bfa_trc(bfad, msix_entries[i].vector);
			bfad->msix_tab[i].msix.vector = msix_entries[i].vector;
		}

		bfa_msix_init(&bfad->bfa, bfad->nvec);

		bfad->bfad_flags |= BFAD_MSIX_ON;

		return error;
	}

line_based:
	error = 0;
	if (request_irq
	    (bfad->pcidev->irq, (bfad_isr_t) bfad_intx, BFAD_IRQ_FLAGS,
	     BFAD_DRIVER_NAME, bfad) != 0) {
		/* Enable interrupt handler failed */
		return 1;
	}
	bfad->bfad_flags |= BFAD_INTX_ON;

	return error;
}

void
bfad_remove_intr(struct bfad_s *bfad)
{
	int	i;

	if (bfad->bfad_flags & BFAD_MSIX_ON) {
		for (i = 0; i < bfad->nvec; i++)
			free_irq(bfad->msix_tab[i].msix.vector,
					&bfad->msix_tab[i]);

		pci_disable_msix(bfad->pcidev);
		bfad->bfad_flags &= ~BFAD_MSIX_ON;
	} else if (bfad->bfad_flags & BFAD_INTX_ON) {
		free_irq(bfad->pcidev->irq, bfad);
	}
}

#else /* CONFIG_PCI_MSI */
/**
 * @brief
 * Setup line-based interrupt.
 */
int
bfad_setup_intr(struct bfad_s *bfad)
{
	if (request_irq
	    (bfad->pcidev->irq, (bfad_isr_t) bfad_intx, SA_SHIRQ,
	     BFAD_DRIVER_NAME, bfad) != 0) {
		/* Enable interrupt handler failed */
		return 1;
	}
	bfad->bfad_flags |= BFAD_INTX_ON;

	return 0;
}

void
bfad_remove_intr(struct bfad_s *bfad)
{
	bfa_trc(bfad, bfad->pcidev->irq);
	if (bfad->bfad_flags & BFAD_INTX_ON)
		free_irq(bfad->pcidev->irq, bfad);
}

#endif /* CONFIG_PCI_MSI */

/**
 * @}
 */
/*
 * @dg PCI_entry PCI driver entries *@{
 */

/**
 * @brief
 * PCI probe entry.
 */
int
bfad_pci_probe(struct pci_dev *pdev, const struct pci_device_id *pid)
{
	struct bfad_s	*bfad;
	int		error = -ENODEV, retval;

	/* For single port cards - only claim function 0 */
	if ((pdev->device == BFA_PCI_DEVICE_ID_FC_8G1P) &&
		(PCI_FUNC(pdev->devfn) != 0))
		return -ENODEV;

	BFA_TRACE(BFA_INFO, "bfad_pci_probe entry");

	bfad = kzalloc(sizeof(struct bfad_s), GFP_KERNEL);
	if (!bfad) {
		error = -ENOMEM;
		goto out;
	}

	bfad->trcmod = kzalloc(sizeof(struct bfa_trc_mod_s), GFP_KERNEL);
	if (!bfad->trcmod) {
		printk(KERN_WARNING "Error alloc trace buffer!\n");
		error = -ENOMEM;
		goto out_alloc_trace_failure;
	}

	/* LOG/TRACE INIT */
	bfa_trc_init(bfad->trcmod);
	bfa_trc(bfad, bfad_inst);

	bfad->logm = &bfad->log_data;
	bfa_log_init(bfad->logm, pci_name(pdev), bfa_os_printf);

	bfad_drv_log_level_set(bfad);

	bfad->aen = &bfad->aen_buf;

	if (!(bfad_load_fwimg(pdev))) {
		kfree(bfad->trcmod);
		goto out_alloc_trace_failure;
	}

	retval = bfad_pci_init(pdev, bfad);
	if (retval) {
		printk(KERN_WARNING "bfad_pci_init failure!\n");
		error = retval;
		goto out_pci_init_failure;
	}

	down(&bfad_sem);
	bfad->inst_no = bfad_inst++;
	list_add_tail(&bfad->list_entry, &bfad_list);
	up(&bfad_sem);

	/* Initializing the state machine: State set to uninit */
	bfa_sm_set_state(bfad, bfad_sm_uninit);

	spin_lock_init(&bfad->bfad_lock);
	pci_set_drvdata(pdev, bfad);

	bfad->ref_count = 0;
	bfad->pport.bfad = bfad;
	bfa_aen_init(bfad->aen, bfad->trcmod, bfad,
		     bfa_os_get_instance_id(bfad),
		     bfad_aen_cb_notify, bfa_os_gettimeofday);
	bfa_q_init(&bfad->trs_cfg.trs_q);
	INIT_LIST_HEAD(&bfad->pbc_vport_list);
	INIT_LIST_HEAD(&bfad->vport_list);

	/* Setup the debugfs node for this bfad */
#if (LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 31))
	bfad_debugfs_init(&bfad->pport);
#endif

	retval = bfad_drv_init(bfad);
	if (retval != BFA_STATUS_OK)
		goto out_drv_init_failure;

	bfa_sm_send_event(bfad, BFAD_E_CREATE);

	if (bfa_sm_cmp_state(bfad, bfad_sm_uninit))
		goto out_bfad_sm_failure;

	return 0;

out_bfad_sm_failure:
	bfad_hal_mem_release(bfad);
out_drv_init_failure:
	/* Remove the debugfs node for this bfad */
#if (LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 31))
	kfree(bfad->regdata);
	bfad_debugfs_exit(&bfad->pport);
#endif
	down(&bfad_sem);
	bfad_inst--;
	list_del(&bfad->list_entry);
	up(&bfad_sem);
	bfad_pci_uninit(pdev, bfad);
out_pci_init_failure:
	kfree(bfad->trcmod);
out_alloc_trace_failure:
	kfree(bfad);
out:
	return error;
}

/**
 * @brief
 * PCI remove entry.
 */
void
bfad_pci_remove(struct pci_dev *pdev)
{
	struct bfad_s	      *bfad = pci_get_drvdata(pdev);
	unsigned long	flags;

	bfa_trc(bfad, bfad->inst_no);

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	if (bfad->bfad_tsk != NULL) {
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		bfad_os_kthread_stop(bfad);
	} else {
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	}

	/* Send Event BFAD_E_STOP */
	bfa_sm_send_event(bfad, BFAD_E_STOP);

	/* Driver detach and dealloc mem */
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	bfa_detach(&bfad->bfa);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	bfad_hal_mem_release(bfad);

	/* Cleaning the BFAD instance */
	down(&bfad_sem);
	bfad_inst--;
	list_del(&bfad->list_entry);
	up(&bfad_sem);
	bfad_pci_uninit(pdev, bfad);

	/* Remove the debugfs node for this bfad */
#if (LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 31))
	kfree(bfad->regdata);
	bfad_debugfs_exit(&bfad->pport);
#endif
	kfree(bfad->trcmod);
	kfree(bfad);
}

struct pci_device_id bfad_id_table[] = {
	{
		.vendor = BFA_PCI_VENDOR_ID_BROCADE,
		.device = BFA_PCI_DEVICE_ID_FC_8G2P,
		.subvendor = PCI_ANY_ID,
		.subdevice = PCI_ANY_ID,
	},
	{
		.vendor = BFA_PCI_VENDOR_ID_BROCADE,
		.device = BFA_PCI_DEVICE_ID_FC_8G1P,
		.subvendor = PCI_ANY_ID,
		.subdevice = PCI_ANY_ID,
	},
	{
		.vendor = BFA_PCI_VENDOR_ID_BROCADE,
		.device = BFA_PCI_DEVICE_ID_CT,
		.subvendor = PCI_ANY_ID,
		.subdevice = PCI_ANY_ID,
		.class = (PCI_CLASS_SERIAL_FIBER << 8),
		.class_mask = ~0,
	},
	{
		.vendor = BFA_PCI_VENDOR_ID_BROCADE,
		.device = BFA_PCI_DEVICE_ID_CT_FC,
		.subvendor = PCI_ANY_ID,
		.subdevice = PCI_ANY_ID,
		.class = (PCI_CLASS_SERIAL_FIBER << 8),
		.class_mask = ~0,
	},
	{
		.vendor = BFA_PCI_VENDOR_ID_BROCADE,
		.device = BFA_PCI_DEVICE_ID_CT2,
		.subvendor = PCI_ANY_ID,
		.subdevice = PCI_ANY_ID,
		.class = (PCI_CLASS_SERIAL_FIBER << 8),
		.class_mask = ~0,
	},
	{
		.vendor = BFA_PCI_VENDOR_ID_BROCADE,
		.device = BFA_PCI_DEVICE_ID_CT2_QUAD,
		.subvendor = PCI_ANY_ID,
		.subdevice = PCI_ANY_ID,
		.class = (PCI_CLASS_SERIAL_FIBER << 8),
		.class_mask = ~0,
	},

	{0, 0},
};

MODULE_DEVICE_TABLE(pci, bfad_id_table);

/**
 * @}
 * @dg Linux driver module functions
 * @{
 */

bfa_status_t
bfad_fc4_module_init(void)
{
	int		rc;

	rc = bfad_im_module_init();
	if (rc != BFA_STATUS_OK)
		goto ext;

ext:
	return rc;
}

void
bfad_fc4_module_exit(void)
{
	bfad_im_module_exit();
}

/**
 * @brief
 * Driver module init.
 */
static int __init
bfad_init(void)
{
	int		error = 0;

	printk(KERN_INFO "QLogic BR-series BFA FC/FCOE SCSI driver - version: "
		"%s\n", bfa_version);

	if (num_sgpgs > 0)
		num_sgpgs_parm = num_sgpgs;

	SET_MODULE_VERSION(BFAD_DRIVER_VERSION);
	bfad_os_idr_init(&bfad_im_port_index);
	error = bfad_fc4_module_init();
	if (error) {
		error = -ENOMEM;
		printk(KERN_WARNING "bfad_fc4_module_init failure\n");
		goto ext;
	}

	sema_init(&bfad_sem, 1);	// Must be before bfad_os_ioctl_init()
	error = bfad_os_ioctl_init();
	if (error) {
		printk(KERN_WARNING "bfad_os_ioctl_init failure\n");
		goto ext;
	}

	if (strcmp(FCPI_NAME, " fcpim") == 0)
		supported_fc4s |= BFA_LPORT_ROLE_FCP_IM;
	if (strcmp(FCPT_NAME, " fcptm") == 0)
		supported_fc4s |= BFA_LPORT_ROLE_FCP_TM;
	if (strcmp(IPFC_NAME, " ipfc") == 0)
		supported_fc4s |= BFA_LPORT_ROLE_FCP_IPFC;

	bfa_ioc_auto_recover(ioc_auto_recover);
	bfa_fcs_rport_set_del_timeout(rport_del_timeout);
	bfa_fcs_rport_set_max_logins(max_rport_logins);
	error = bfad_os_pci_register_driver(bfad_pci_driver_p);

	down(&bfad_sem);
	bfad_scan_done = 1;
	up(&bfad_sem);

	if (error) {
		printk(KERN_WARNING "bfad_os_pci_register_driver failure\n");
		goto ext1;
	}


	return 0;

ext1:
	bfad_os_ioctl_exit();
ext:
	bfad_fc4_module_exit();
	bfad_os_idr_destroy(&bfad_im_port_index);
	return error;
}

/**
 * @brief
 * Driver module exit.
 */
static void __exit
bfad_exit(void)
{
	bfad_os_pci_unregister_driver(bfad_pci_driver_p);
	bfad_os_ioctl_exit();
	bfad_fc4_module_exit();
	bfad_os_idr_destroy(&bfad_im_port_index);
}

void
bfad_register(struct bfad_ipfc_ops *ops)
{
	struct bfad_s *bfad;
	bfa_fcs_lport_t  *fcs_port;

	ipfc_ops = *ops;

	down(&bfad_sem);

	list_for_each_entry(bfad, &bfad_list, list_entry) {
		atomic_set(&bfad->ipfc_ops_flag, 1);
		ipfc_ops.probe(bfad);
		fcs_port = bfa_fcs_lookup_port(&bfad->bfa_fcs, 0, 0);
		if (fcs_port)
			bfa_fcs_lport_enable_ipfc_roles(fcs_port);
	}

	up(&bfad_sem);
}

void
bfad_unregister(void)
{
	struct bfad_s *bfad;
	bfa_fcs_lport_t *fcs_port;

	down(&bfad_sem);

	list_for_each_entry(bfad, &bfad_list, list_entry) {
		if (atomic_read(&bfad->ipfc_ops_flag)) {
			fcs_port = bfa_fcs_lookup_port(&bfad->bfa_fcs, 0, 0);
			if (fcs_port)
				bfa_fcs_lport_disable_ipfc_roles(fcs_port);
			ipfc_ops.probe_undo(bfad);
		}
		atomic_set(&bfad->ipfc_ops_flag, 0);
	}

	up(&bfad_sem);

	memset(&ipfc_ops, 0x00, sizeof (struct bfad_ipfc_ops));
}

void
bfad_stopall(void)
{
}

void
bfa_fcb_process_farp(void *lport, void *buf)
{
	struct bfad_s *bfad =
			((struct bfa_fcs_lport_s *)lport)->bfad_port->bfad;

	if (atomic_read(&bfad->ipfc_ops_flag))
		ipfc_ops.process_farp(lport, buf);
}

uint16_t
bfa_cb_ipfc_pktlen(bfad_pkt_t *dpkt, void *bfad)
{

	if (atomic_read(&((struct bfad_s *)bfad)->ipfc_ops_flag))
		return ipfc_ops.pktlen(dpkt);
	else
		return 0;
}

void
bfa_cb_ipfc_sge(bfad_pkt_t *dpkt, int sge_index, bfa_sge_t *sge, void *bfad)
{
	if (atomic_read(&((struct bfad_s *)bfad)->ipfc_ops_flag))
		ipfc_ops.sge(dpkt, sge_index, sge);
}


int
bfa_cb_ipfc_nsgles(bfad_pkt_t *dpkt, void *bfad)
{
	if (atomic_read(&((struct bfad_s *)bfad)->ipfc_ops_flag))
		return ipfc_ops.nsgles(dpkt);

	return -1;
}

void bfa_cb_ipfc_pktrecv(void *dipfc, bfad_pkt_t *dpkt, bfa_ipfc_status_t
		status, uint32_t pktlen, uint32_t residue)
{
	if (atomic_read(&((struct bfad_s *)dipfc)->ipfc_ops_flag))
		ipfc_ops.pkt_recv(dipfc, dpkt, status, pktlen, residue);
}

void
bfad_tm_register(struct bfad_tm_ops_s *bfad_tm_ops)
{
	tm_ops = *bfad_tm_ops;
	atomic_set(&tm_ops_flag, 1);
}

void
bfad_tm_unregister(void)
{
	atomic_set(&tm_ops_flag, 0);
	memset(&tm_ops, 0, sizeof(struct bfad_tm_ops_s));
}


void
bfad_get_port_instances(struct list_head **bfad_port_list)
{
	down(&bfad_sem);
	*bfad_port_list = &bfad_list;
	up(&bfad_sem);
}

void
bfad_acquire_lock(struct bfad_s *bfad, unsigned long *flags)
{
	spin_lock_irqsave(&bfad->bfad_lock, *flags);
}

void
bfad_release_lock(struct bfad_s *bfad, unsigned long flags)
{
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
}

void
bfad_acquire_semaphore(void)
{
	down(&bfad_sem);
}

void
bfad_release_semaphore(void)
{
	up(&bfad_sem);
}

bfa_status_t
bfad_target_mode_ioctl(struct bfad_s *bfad, void *arg)
{
	if (atomic_read(&tm_ops_flag)) {
		tm_ops.bfat_ioctl(bfad, arg);
		return BFA_TRUE;
	}

	return BFA_FALSE;
}

int
bfad_cp_to_user(int size, unsigned long arg, uint8_t *iocmd)
{
	if (copy_to_user((void __user *)arg, (uint8_t *)iocmd, size)) {
		kfree(iocmd);
		return EIO;
	}
	kfree(iocmd);
	return 0;
}

void *
bfad_cp_from_user(int size, unsigned long arg)
{
	void *iocmd;

	iocmd = (uint8_t *)kmalloc(size, GFP_KERNEL);
	if (iocmd)
		memset(iocmd, 0, size);
	else
		return NULL;
	if (copy_from_user((uint8_t *)iocmd, (void __user *)arg, size)) {
		kfree(iocmd);
		return NULL;
	}
	return iocmd;
}

#ifndef __VMKLNX__
EXPORT_SYMBOL(bfad_acquire_semaphore);
EXPORT_SYMBOL(bfad_release_semaphore);
EXPORT_SYMBOL(bfad_acquire_lock);
EXPORT_SYMBOL(bfad_release_lock);
EXPORT_SYMBOL(bfad_get_port_instances);
EXPORT_SYMBOL(bfad_tm_register);
EXPORT_SYMBOL(bfad_tm_unregister);
EXPORT_SYMBOL(bfa_fcs_tm_register);
EXPORT_SYMBOL(bfa_fcs_tm_unregister);
EXPORT_SYMBOL(bfa_tm_register);
EXPORT_SYMBOL(bfa_tm_unregister);
EXPORT_SYMBOL(bfa_fcs_start);
EXPORT_SYMBOL(bfa_fcs_stop);
EXPORT_SYMBOL(bfa_fcs_vport_start);
EXPORT_SYMBOL(bfa_fcs_vport_stop);

EXPORT_SYMBOL(bfa_itn_create);
EXPORT_SYMBOL(bfa_fcp_alloc_iotag);
EXPORT_SYMBOL(bfa_fcp_release_iotag);
EXPORT_SYMBOL(bfa_sm_panic);
EXPORT_SYMBOL(bfa_q_is_on_q_func);
EXPORT_SYMBOL(bfa_sgpg_mfree);
EXPORT_SYMBOL(bfa_sgpg_malloc);
EXPORT_SYMBOL(bfa_sgpg_winit);
EXPORT_SYMBOL(bfa_sgpg_wcancel);
EXPORT_SYMBOL(bfa_sgpg_wait);
EXPORT_SYMBOL(bfa_fcs_rport_tin_ack);
EXPORT_SYMBOL(bfa_fcs_rport_lookup);
EXPORT_SYMBOL(bfa_fcs_lookup_port);
EXPORT_SYMBOL(bfa_fcxp_walloc_cancel);
EXPORT_SYMBOL(bfa_fcxp_alloc_wait);
EXPORT_SYMBOL(bfa_fcxp_discard);
EXPORT_SYMBOL(bfa_timer_begin);
EXPORT_SYMBOL(bfa_timer_stop);
EXPORT_SYMBOL(fc_ba_acc_build);
EXPORT_SYMBOL(fc_rrq_build);
EXPORT_SYMBOL(fc_ls_rjt_build);
EXPORT_SYMBOL(fc_srr_acc_build);
EXPORT_SYMBOL(fc_rrq_rsp_build);
EXPORT_SYMBOL(fc_prlo_acc_build);
EXPORT_SYMBOL(fc_rec_acc_build);
EXPORT_SYMBOL(fc_prli_acc_build);
EXPORT_SYMBOL(fcp_rsp_fchdr_build);

EXPORT_SYMBOL(bfa_get_attr);
EXPORT_SYMBOL(bfa_ipfc_flushpkt);
EXPORT_SYMBOL(bfa_fcs_get_base_port);
EXPORT_SYMBOL(bfa_fcport_get_attr);
EXPORT_SYMBOL(bfa_fcxp_alloc);
EXPORT_SYMBOL(bfa_fcxp_send);
EXPORT_SYMBOL(bfa_isr_bind);
EXPORT_SYMBOL(bfa_lps_get_tag);
EXPORT_SYMBOL(bfa_ipfc_meminfo);
EXPORT_SYMBOL(bfa_ipfc_attach);
EXPORT_SYMBOL(bfa_ipfc_detach);
EXPORT_SYMBOL(bfa_ipfc_postpkt);
EXPORT_SYMBOL(bfad_register);
EXPORT_SYMBOL(bfad_unregister);
EXPORT_SYMBOL(bfad_stopall);
EXPORT_SYMBOL(bfa_fcxp_get_reqbuf);
EXPORT_SYMBOL(bfa_log_init);
EXPORT_SYMBOL(bfa_os_printf);
EXPORT_SYMBOL(bfa_panic);
EXPORT_SYMBOL(bfa_log);
EXPORT_SYMBOL(bfad_find_bfad_by_inst_no);
EXPORT_SYMBOL(bfad_cp_from_user);
EXPORT_SYMBOL(bfad_ioctl_ioc_get_inst);
EXPORT_SYMBOL(bfad_is_ready);
EXPORT_SYMBOL(bfad_ioctl_ioc_get_version);
EXPORT_SYMBOL(bfad_cp_to_user);
#endif

#define BFAD_PROTO_NAME FCPI_NAME FCPT_NAME IPFC_NAME

module_init(bfad_init);
module_exit(bfad_exit);
MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("QLogic BR-series Fibre Channel HBA Driver" BFAD_PROTO_NAME);
MODULE_AUTHOR("QLogic Corporation");
MODULE_VERSION(BFAD_DRIVER_VERSION);

/*
 * @}
 */
