/*
 * Copyright (c) 2005-2014 Brocade Communications Systems, Inc.
 * Copyright (c) 2014 QLogic Corporation.
 * All rights reserved
 * www.qlogic.com
 *
 * See LICENSE.bfa for copyright and licensing details.
 */

#include <linux/poll.h>		/* poll_wait() */
#include <linux/spinlock.h>	/* spin_lock_irq() */
#include <asm/uaccess.h>	/* copy_from_user(), copy_to_user() */
#include <linux/fs.h>		/* struct file_operations */
#include <linux/string.h>	/* memset() */

#include "bfad_aen.h"
#include "bfad_drv.h"
#include "bfad_ioctl_cna.h"
#include "bfad_ioctl.h"
#include "bfad_im.h"

#ifdef DEFINE_SPINLOCK
static DEFINE_SPINLOCK(bfad_aen_spinlock);
#else
static spinlock_t bfad_aen_spinlock = SPIN_LOCK_UNLOCKED;
#endif
DECLARE_WAIT_QUEUE_HEAD(bfad_aen_readq);
struct fasync_struct *bfad_aen_asyncq;
int fasync_enabled;

/**
 * @dg bfad2aen_api Base Linux ==> AEN module interface to support file
 * 			operations
 * @{
 */

/**
 * @brief IOCTL function to copy AEN module data onto user space.
 * @param[in,out] arg   - Ptr to user bfa_ioctl_aen_t where data will be
 * 			copied from/to
 * @retval 0            Success
 * @retval -EINVAL      Invalid app_id passed in user command
 * @retval -EFAULT      Error copying data onto user buffer
 */
int
bfad_ioctl_aen_get(struct file *file, unsigned long arg)
{
	bfa_ioctl_aen_t ioc_aen;
	unsigned int    req_cnt, cnt = 0;
	int      	size = sizeof(bfa_ioctl_aen_t);
	int		rc = 0;
	bfa_status_t	status = BFA_STATUS_OK;
	unsigned long	flags;
	struct bfad_s	*bfad;
	bfa_aen_entry_t __user *usr_aen_list =
			(bfa_aen_entry_t __user *)
			(((bfa_ioctl_aen_t __user *)arg)->aen_list);

	if (copy_from_user((void *)&ioc_aen, (void __user *)arg, size))
		return -EFAULT;
	bfad = bfad_find_bfad_by_inst_no(ioc_aen.bfad_num);
	bfa_assert(bfad);

	if ((ioc_aen.app_id < bfa_aen_app_bcu) ||
		(ioc_aen.app_id >= BFA_AEN_MAX_APP))
		return -EINVAL;

	/* Get events from AEN module */
	req_cnt = ioc_aen.aen_count;
	while (req_cnt--) {
		spin_lock_irqsave(&bfad_aen_spinlock, flags);
		status = bfa_aen_fetch(bfad->aen, ioc_aen.aen_list, 1,
					ioc_aen.app_id, &rc);
		spin_unlock_irqrestore(&bfad_aen_spinlock, flags);
		if ((status != BFA_STATUS_OK) || !rc)
			break;
		if (copy_to_user((void __user *)&usr_aen_list[cnt],
					(void *)ioc_aen.aen_list,
					sizeof(bfa_aen_entry_t)))
			return -EFAULT;
		cnt++;
	}

	ioc_aen.aen_count = cnt;

	if (status != BFA_STATUS_OK)
		ioc_aen.status = status;
	else
		ioc_aen.status = (cnt > 0) ? BFA_STATUS_OK : BFA_STATUS_FAILED;

	if (copy_to_user((void __user *)arg, (void *)&ioc_aen,
				size - sizeof(bfa_aen_entry_t)))
		return -EFAULT;

	return 0;
}

int
bfad_ioctl_aen_update(struct file *file, unsigned long arg)
{
	bfa_ioctl_aen_t ioc_aen;
	unsigned long	flags;
	struct bfad_s	*bfad;

	if (copy_from_user((void *)&ioc_aen, (void __user *)arg,
			sizeof(bfa_ioctl_aen_t)))
		return -EFAULT;

	if ((ioc_aen.app_id < bfa_aen_app_bcu) ||
		(ioc_aen.app_id >= BFA_AEN_MAX_APP))
		return -EINVAL;

	// Update all instances to have same file map in Linux
	down(&bfad_sem);
	spin_lock_irqsave(&bfad_aen_spinlock, flags);
	list_for_each_entry(bfad, &bfad_list, list_entry) {
		bfad->file_map[ioc_aen.app_id] = file;
		/* Update file handle corresponding to this app */
	}
	spin_unlock_irqrestore(&bfad_aen_spinlock, flags);
	up(&bfad_sem);

	return 0;
}

static bfa_aen_app_t
bfad_aen_get_appid(struct file *file)
{
	int		i;
	bfa_aen_app_t	app_id = BFA_AEN_MAX_APP;
	struct bfad_s	*bfad;

	// Just find the app_id in any bfad instance
	// Assuming that all other instances will have same file map in Linux
	down(&bfad_sem);
	list_for_each_entry(bfad, &bfad_list, list_entry) {
		for (i = 0; i < BFA_AEN_MAX_APP; i++) {
			if (bfad->file_map[i] == file) {
				app_id = i;
				break;
			}
		}
	}
	up(&bfad_sem);

	return app_id;
}

/**
 * @brief Poll function for AEN module file operation through Base Linux driver.
 * @note  To support poll/select interface for non-blocking read/write file
 * 		operations
 * @param[in] file  - File ptr
 * @param[in] wait  - Ptr to fd for which data availability will be checked
 * @return Mask to inform the current data availability status
 */
unsigned int
bfad_aen_poll(struct file *file, struct poll_table_struct *wait)
{
	struct bfad_s   *bfad;
	unsigned int    mask = 0;
	unsigned long   flags;
	bfa_aen_app_t	app_id;

	app_id = bfad_aen_get_appid(file);
	if (app_id == BFA_AEN_MAX_APP)
		return POLLNVAL;

	poll_wait(file, &bfad_aen_readq, wait);

	/* See if there are any events posted */
	down(&bfad_sem);
	spin_lock_irqsave(&bfad_aen_spinlock, flags);
	list_for_each_entry(bfad, &bfad_list, list_entry) {
		bfa_assert(bfad->file_map[app_id] == file);
		if (bfa_aen_fetch_count(bfad->aen, app_id) > 0) {
			mask |= POLLIN | POLLRDNORM;	/* Readable */
			break;
		}
	}
	spin_unlock_irqrestore(&bfad_aen_spinlock, flags);
	up(&bfad_sem);

	return (mask);
}

int
bfad_aen_fasync(int fd, struct file *file, int on)
{
	fasync_enabled = on;
	return fasync_helper(fd, file, on, &bfad_aen_asyncq);
}

#if  LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 31)
void
bfad_aen_im_notify(struct bfad_s *bfad)
{
	struct bfad_im_s *im = bfad->im;

	if (bfad->bfad_flags & BFAD_FC4_PROBE_DONE)
		bfad_os_queue_work(bfad->im->drv_workq,
				&im->aen_im_notify_work);
}

BFAD_WORK_HANDLER(bfad_aen_im_notify_handler)
{	 /* work struct is the function arg */
	struct bfad_im_s *im =
		container_of(work, struct bfad_im_s, aen_im_notify_work);
	struct bfa_aen_entry_s aen_entry;
	int     rc = 0;
	unsigned long   flags;
	bfa_status_t status;
	struct bfad_s *bfad = im->bfad;
	struct bfad_port_s pport = bfad->pport;
	struct bfad_im_port_s *im_port = pport.im_port;
	struct Scsi_Host *shost;

	while (1) {
		spin_lock_irqsave(&bfad_aen_spinlock, flags);
		status = bfa_aen_fetch(bfad->aen,
				&aen_entry, 1, 0, &rc);
		spin_unlock_irqrestore(&bfad_aen_spinlock, flags);
		if ((status != BFA_STATUS_OK) || !rc)
			break;

		shost = im_port->shost;
		fc_host_post_vendor_event(shost, fc_get_event_number(),
				sizeof(struct bfa_aen_entry_s),
				(char *)&aen_entry, BFAD_NL_VENDOR_ID);
	}
}
#endif

/**
 * @}
 */
