/*
 * Copyright (c) 2005-2014 Brocade Communications Systems, Inc.
 * Copyright (c) 2014 QLogic Corporation.
 * All rights reserved
 * www.qlogic.com
 *
 * See LICENSE.bfa for copyright and licensing details.
 */

/**
 * @file bfa_attr.c Linux driver configuration interface module.
 */

#include "bfad_drv.h"
#include "bfad_im.h"
#include "bfad_attr.h"
#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 31)
#include "bfad_bsg.h"
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 9)
struct Scsi_Host *
bfad_os_starget_to_shost(struct scsi_target *starget)
{
	return dev_to_shost(starget->dev.parent);
}
#endif


/*
* TBD: Need to find out the kernel version starting to support
 * FC transport.
 */
#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 10)
/**
 * @brief
 * FC transport template entry, get SCSI host port type.
 */
void
bfad_im_get_host_port_type(struct Scsi_Host *shost)
{
	struct bfad_im_port_s *im_port =
			(struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_s *bfad = im_port->bfad;
	struct bfa_lport_attr_s port_attr;

	if (bfad_os_validate_bfad(bfad, shost) != BFA_STATUS_OK)
		return;

	if (bfad_os_set_port_npiv_type(im_port, shost))
		return;

	bfa_fcs_lport_get_attr(&bfad->bfa_fcs.fabric.bport, &port_attr);

	switch (port_attr.port_type) {
	case BFA_PORT_TYPE_NPORT:
		fc_host_port_type(shost) = FC_PORTTYPE_NPORT;
		break;
	case BFA_PORT_TYPE_NLPORT:
		fc_host_port_type(shost) = FC_PORTTYPE_NLPORT;
		break;
	case BFA_PORT_TYPE_P2P:
		fc_host_port_type(shost) = FC_PORTTYPE_PTP;
		break;
	case BFA_PORT_TYPE_LPORT:
		fc_host_port_type(shost) = FC_PORTTYPE_LPORT;
		break;
	default:
		fc_host_port_type(shost) = FC_PORTTYPE_UNKNOWN;
		break;
	}
}

/**
 * @brief
 * FC transport template entry, get SCSI host port state.
 */
void
bfad_im_get_host_port_state(struct Scsi_Host *shost)
{
	struct bfad_im_port_s *im_port =
			(struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_s *bfad = im_port->bfad;
	struct bfa_port_attr_s attr;

	if (bfad_os_validate_bfad(bfad, shost) != BFA_STATUS_OK)
		return;

	bfa_fcport_get_attr(&bfad->bfa, &attr);

	switch (attr.port_state) {
	case BFA_PORT_ST_LINKDOWN:
		fc_host_port_state(shost) = FC_PORTSTATE_LINKDOWN;
		break;
	case BFA_PORT_ST_LINKUP:
		fc_host_port_state(shost) = FC_PORTSTATE_ONLINE;
		break;
	case BFA_PORT_ST_DISABLED:
	case BFA_PORT_ST_STOPPED:
	case BFA_PORT_ST_IOCDOWN:
	case BFA_PORT_ST_IOCDIS:
		fc_host_port_state(shost) = FC_PORTSTATE_OFFLINE;
		break;
	case BFA_PORT_ST_UNINIT:
	case BFA_PORT_ST_ENABLING_QWAIT:
	case BFA_PORT_ST_ENABLING:
	case BFA_PORT_ST_DISABLING_QWAIT:
	case BFA_PORT_ST_DISABLING:
	default:
		fc_host_port_state(shost) = FC_PORTSTATE_UNKNOWN;
		break;
	}
}

/**
 * @brief
 * FC transport template entry, get SCSI host active fc4s.
 */
void
bfad_im_get_host_active_fc4s(struct Scsi_Host *shost)
{
	struct bfad_im_port_s *im_port =
			(struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_port_s    *port = im_port->port;

	memset(fc_host_active_fc4s(shost), 0,
	       sizeof(fc_host_active_fc4s(shost)));

	if (port->
	    supported_fc4s & (BFA_LPORT_ROLE_FCP_IM | BFA_LPORT_ROLE_FCP_TM))
		fc_host_active_fc4s(shost)[2] = 1;

	if (port->supported_fc4s & BFA_LPORT_ROLE_FCP_IPFC)
		fc_host_active_fc4s(shost)[3] = 0x20;

	fc_host_active_fc4s(shost)[7] = 1;
}

/**
 * @brief
 * FC transport template entry, get SCSI host link speed.
 */
void
bfad_im_get_host_speed(struct Scsi_Host *shost)
{
	struct bfad_im_port_s *im_port =
			(struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_s *bfad = im_port->bfad;
	struct bfa_port_attr_s attr;

	if (bfad_os_validate_bfad(bfad, shost) != BFA_STATUS_OK)
		return;

	bfa_fcport_get_attr(&bfad->bfa, &attr);
	switch (attr.speed) {
	case BFA_PORT_SPEED_10GBPS:
		fc_host_speed(shost) = FC_PORTSPEED_10GBIT;
		break;
	case BFA_PORT_SPEED_16GBPS:
		fc_host_speed(shost) = FC_PORTSPEED_16GBIT;
		break;
	case BFA_PORT_SPEED_8GBPS:
		fc_host_speed(shost) = FC_PORTSPEED_8GBIT;
		break;
	case BFA_PORT_SPEED_4GBPS:
		fc_host_speed(shost) = FC_PORTSPEED_4GBIT;
		break;
	case BFA_PORT_SPEED_2GBPS:
		fc_host_speed(shost) = FC_PORTSPEED_2GBIT;
		break;
	case BFA_PORT_SPEED_1GBPS:
		fc_host_speed(shost) = FC_PORTSPEED_1GBIT;
		break;
	default:
		fc_host_speed(shost) = FC_PORTSPEED_UNKNOWN;
		break;
	}

}

/**
 * @brief
 * FC transport template entry, get SCSI host port type.
 */
void
bfad_im_get_host_fabric_name(struct Scsi_Host *shost)
{
	struct bfad_im_port_s *im_port =
			(struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_port_s    *port = im_port->port;
	wwn_t fabric_nwwn = 0;

	/*
	 * TBD: Set to fabric nwwn if in switch enviroment. Set to local port
	 * nwwn if private loop or P2P.
	 */
	fabric_nwwn = bfa_fcs_lport_get_fabric_name(port->fcs_port);

	fc_host_fabric_name(shost) = bfa_os_htonll(fabric_nwwn);

}

/**
 * @brief
 * FC transport template entry, get BFAD statistics.
 */
struct fc_host_statistics *
bfad_im_get_stats(struct Scsi_Host *shost)
{
	struct bfad_im_port_s *im_port =
			(struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_s *bfad = im_port->bfad;
	struct bfad_hal_comp fcomp;
	union bfa_port_stats_u *fcstats;
	struct fc_host_statistics *hstats;
	bfa_status_t    rc;
	unsigned long   flags;

	fcstats = kzalloc(sizeof(union bfa_port_stats_u), GFP_KERNEL);
	if (fcstats == NULL)
		return NULL;

	hstats = &bfad->link_stats;
	init_completion(&fcomp.comp);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	memset(hstats, 0, sizeof(struct fc_host_statistics));
	rc = bfa_port_get_stats(BFA_FCPORT(&bfad->bfa),
				fcstats, bfad_hcb_comp, &fcomp);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	if (rc != BFA_STATUS_OK)
		return NULL;

	wait_for_completion(&fcomp.comp);

	/* Fill the fc_host_statistics structure */
	hstats->seconds_since_last_reset = fcstats->fc.secs_reset;
	hstats->tx_frames = fcstats->fc.tx_frames;
	hstats->tx_words  = fcstats->fc.tx_words;
	hstats->rx_frames = fcstats->fc.rx_frames;
	hstats->rx_words  = fcstats->fc.rx_words;
	hstats->lip_count = fcstats->fc.lip_count;
	hstats->nos_count = fcstats->fc.nos_count;
	hstats->error_frames = fcstats->fc.error_frames;
	hstats->dumped_frames = fcstats->fc.dropped_frames;
	hstats->link_failure_count = fcstats->fc.link_failures;
	hstats->loss_of_sync_count = fcstats->fc.loss_of_syncs;
	hstats->loss_of_signal_count = fcstats->fc.loss_of_signals;
	hstats->prim_seq_protocol_err_count = fcstats->fc.primseq_errs;
	hstats->invalid_crc_count = fcstats->fc.invalid_crcs;

	kfree(fcstats);

	return hstats;
}

/**
 * @brief
 * FC transport template entry, reset BFAD statistics.
 */
void
bfad_im_reset_stats(struct Scsi_Host *shost)
{
	struct bfad_im_port_s *im_port =
			(struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_s *bfad = im_port->bfad;
	struct bfad_hal_comp fcomp;
	unsigned long   flags;
	bfa_status_t    rc;

	init_completion(&fcomp.comp);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	rc = bfa_port_clear_stats(BFA_FCPORT(&bfad->bfa), bfad_hcb_comp,
			&fcomp);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	if (rc != BFA_STATUS_OK)
		return;

	wait_for_completion(&fcomp.comp);
}
/**
 * @brief
 * FC transport template entry, set rport loss timeout.
 * Update dev_loss_tmo based on the value pushed down by the stack
 * In case it is lesser than path_tov of driver, set it to path_tov + 1
 * to ensure that the driver times out before the application
 */
void
bfad_im_set_rport_loss_tmo(struct fc_rport *rport, uint32_t timeout)
{
	struct bfad_itnim_data_s *itnim_data = rport->dd_data;
	struct bfad_itnim_s   *itnim = itnim_data->itnim;
	struct bfad_s *bfad = itnim->im->bfad;
	uint16_t path_tov = bfa_fcpim_path_tov_get(&bfad->bfa);

	rport->dev_loss_tmo = timeout;

	if (timeout < path_tov)
		rport->dev_loss_tmo = path_tov + 1;
}
#endif

#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 31)
static int
bfad_im_vport_create(struct fc_vport *fc_vport, bool disable)
{
	char *vname = fc_vport->symbolic_name;
	struct Scsi_Host *shost = fc_vport->shost;
	struct bfad_im_port_s *im_port =
		(struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_s *bfad = im_port->bfad;
	struct bfa_lport_cfg_s port_cfg;
	struct bfad_vport_s *vp;
	int status = 0, rc;
	unsigned long flags;

	memset(&port_cfg, 0, sizeof(port_cfg));
	u64_to_wwn(fc_vport->node_name, (u8 *)&port_cfg.nwwn);
	u64_to_wwn(fc_vport->port_name, (u8 *)&port_cfg.pwwn);
	if (strlen(vname) > 0)
		strcpy((char *)&port_cfg.sym_name, vname);
	port_cfg.roles = BFA_LPORT_ROLE_FCP_IM;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	list_for_each_entry(vp, &bfad->pbc_vport_list, list_entry) {
		if (port_cfg.pwwn ==
			vp->fcs_vport.lport.port_cfg.pwwn) {
			port_cfg.preboot_vp =
				vp->fcs_vport.lport.port_cfg.preboot_vp;
			break;
		}
	}
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	rc = bfad_vport_create(bfad, 0, &port_cfg, &fc_vport->dev);
	if (rc == BFA_STATUS_OK) {
		struct bfad_vport_s *vport;
		struct bfa_fcs_vport_s *fcs_vport;
		struct Scsi_Host *vshost;

		spin_lock_irqsave(&bfad->bfad_lock, flags);
		fcs_vport = bfa_fcs_vport_lookup(&bfad->bfa_fcs, 0,
				port_cfg.pwwn);
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		if (fcs_vport == NULL)
			return VPCERR_BAD_WWN;

		fc_vport_set_state(fc_vport, FC_VPORT_ACTIVE);
		if (disable) {
			spin_lock_irqsave(&bfad->bfad_lock, flags);
			bfa_fcs_vport_stop(fcs_vport);
			spin_unlock_irqrestore(&bfad->bfad_lock, flags);
			fc_vport_set_state(fc_vport, FC_VPORT_DISABLED);
		}

		vport = fcs_vport->vport_drv;
		vshost = vport->drv_port.im_port->shost;
		fc_host_node_name(vshost) = wwn_to_u64((u8 *)&port_cfg.nwwn);
		fc_host_port_name(vshost) = wwn_to_u64((u8 *)&port_cfg.pwwn);
		fc_host_supported_classes(vshost) = FC_COS_CLASS3;

		memset(fc_host_supported_fc4s(vshost), 0,
		       sizeof(fc_host_supported_fc4s(vshost)));
		if (supported_fc4s & (BFA_LPORT_ROLE_FCP_IM |
				      BFA_LPORT_ROLE_FCP_TM))
			/* For FCP type 0x08 */
			fc_host_supported_fc4s(vshost)[2] = 1;
		if (supported_fc4s | BFA_LPORT_ROLE_FCP_IPFC)
			/* For LLC/SNAP type 0x05 */
			fc_host_supported_fc4s(vshost)[3] = 0x20;

		/* For fibre channel services type 0x20 */
		fc_host_supported_fc4s(vshost)[7] = 1;

		fc_host_supported_speeds(vshost) =
				bfad_im_supported_speeds(&bfad->bfa);
		fc_host_maxframe_size(vshost) =
				bfa_fcport_get_maxfrsize(&bfad->bfa);
		fc_vport->dd_data = vport->drv_port.im_port;
		vport->drv_port.im_port->fc_vport = fc_vport;
	} else if (rc == BFA_STATUS_INVALID_WWN)
		return VPCERR_BAD_WWN;
	else if (rc == BFA_STATUS_VPORT_EXISTS)
		return VPCERR_BAD_WWN;
	else if (rc == BFA_STATUS_VPORT_MAX)
		return VPCERR_NO_FABRIC_SUPP;
	else if (rc == BFA_STATUS_VPORT_WWN_BP)
		return VPCERR_BAD_WWN;
	else
		return FC_VPORT_FAILED;

	return status;
}

static int
bfad_im_vport_delete(struct fc_vport *fc_vport)
{
	struct bfad_vport_s *vport;
	struct bfad_im_port_s *im_port =
				(struct bfad_im_port_s *)fc_vport->dd_data;
	struct bfad_s *bfad = im_port->bfad;
	struct bfa_fcs_vport_s *fcs_vport;
	struct Scsi_Host *vshost;
	wwn_t   pwwn;
	int rc;
	unsigned long flags;
	struct completion fcomp;

	vport = container_of(im_port->port, struct bfad_vport_s, drv_port);

	if (im_port->flags & BFAD_PORT_DELETE) {
		bfad_os_scsi_host_free(bfad, im_port);
		kfree(vport);
		return 0;
	}

	vshost = vport->drv_port.im_port->shost;
	u64_to_wwn(fc_host_port_name(vshost), (u8 *)&pwwn);

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	fcs_vport = bfa_fcs_vport_lookup(&bfad->bfa_fcs, 0, pwwn);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	if (fcs_vport == NULL)
		return VPCERR_BAD_WWN;

	vport->drv_port.flags |= BFAD_PORT_DELETE;

	vport->comp_del = &fcomp;
	init_completion(vport->comp_del);

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	rc = bfa_fcs_vport_delete(&vport->fcs_vport);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	if (rc == BFA_STATUS_PBC) {
		vport->drv_port.flags &= ~BFAD_PORT_DELETE;
		vport->comp_del = NULL;
		return -1;
	}
	wait_for_completion(vport->comp_del);

	bfad_os_scsi_host_free(bfad, im_port);
	list_del(&vport->list_entry);
	kfree(vport);

	return 0;
}

static int
bfad_im_vport_disable(struct fc_vport *fc_vport, bool disable)
{
	struct bfad_im_port_s *im_port =
				(struct bfad_im_port_s *)fc_vport->dd_data;
	struct bfad_vport_s *vport = container_of(im_port->port,
					struct bfad_vport_s, drv_port);
	struct bfad_s *bfad;
	struct bfa_fcs_vport_s *fcs_vport;
	struct Scsi_Host *vshost;
	wwn_t   pwwn;
	unsigned long flags;

	bfad = vport->drv_port.bfad;
	vshost = im_port->shost;
	u64_to_wwn(fc_host_port_name(vshost), (u8 *)&pwwn);

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	fcs_vport = bfa_fcs_vport_lookup(&bfad->bfa_fcs, 0, pwwn);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	if (fcs_vport == NULL)
		return VPCERR_BAD_WWN;

	if (disable) {
		bfa_fcs_vport_stop(fcs_vport);
		fc_vport_set_state(fc_vport, FC_VPORT_DISABLED);
	} else {
		bfa_fcs_vport_start(fcs_vport);
		fc_vport_set_state(fc_vport, FC_VPORT_ACTIVE);
	}

	return 0;
}

void
bfad_im_vport_set_symbolic_name(struct fc_vport *fc_vport)
{
	struct bfad_im_port_s *im_port =
				(struct bfad_im_port_s *)fc_vport->dd_data;
	struct bfad_vport_s *vport = container_of(
				im_port->port, struct bfad_vport_s, drv_port);
	struct bfa_fcs_vport_s *fcs_vport;
	struct bfad_s *bfad;
	struct Scsi_Host *vshost;
	wwn_t   pwwn;
	char *sym_name = fc_vport->symbolic_name;
	unsigned long flags;

	bfad = vport->drv_port.bfad;
	vshost = vport->drv_port.im_port->shost;
	u64_to_wwn(fc_host_port_name(vshost), (u8 *)&pwwn);

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	fcs_vport = bfa_fcs_vport_lookup(&bfad->bfa_fcs, 0, pwwn);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	if (fcs_vport == NULL)
		return;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	if (strlen(sym_name) > 0)
		bfa_fcs_lport_set_symname(
			&fcs_vport->lport, sym_name);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
}

#else
static int
bfad_im_parse_wwn(const char *buf, uint8_t *wwn)
{
	unsigned int i, j;
	memset(wwn, 0, 8);

	/* Validate and store the new name */
	for (i = 0, j = 0; i < 16; i++) {
		if ((*buf >= '0') && (*buf <= '9'))
			j = ((j << 4) | (*buf++ -'0'));
		else if ((*buf >= 'a') && (*buf <= 'f'))
			j = ((j << 4) | ((*buf++ -'a') + 10));
		else if ((*buf >= 'A') && (*buf <= 'F'))
			j = ((j << 4) | ((*buf++ -'A') + 10));
		else
			return -EINVAL;
		if (i % 2) {
			wwn[i/2] = j & 0xff;
			j = 0;
		}
	}

	return 0;
}

static ssize_t
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 26)
bfad_im_vport_create(struct device *cdev, struct device_attribute *attr,
					 const char *buf, size_t count)
#else
bfad_im_vport_create(struct class_device *cdev, const char *buf, size_t count)
#endif
{
	struct Scsi_Host *shost = class_to_shost(cdev);
	struct bfad_im_port_s *im_port =
			(struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_s *bfad = im_port->bfad;
	bfa_lport_cfg_t port_cfg;
	uint8_t wwn[8];
	int status;

	memset(&port_cfg, 0, sizeof(port_cfg));
	status = bfad_im_parse_wwn(&buf[0], wwn);
	if (status)
		return status;
	memcpy(&port_cfg.pwwn, wwn, sizeof(wwn));

	status = bfad_im_parse_wwn(&buf[17], wwn);
	if (status)
		return status;
	memcpy(&port_cfg.nwwn, wwn, sizeof(wwn));

	port_cfg.roles = BFA_LPORT_ROLE_FCP_IM;
	status = bfad_vport_create(bfad, 0, &port_cfg, 0);
	if (status != BFA_STATUS_OK)
		return -EIO;

	return count;
}

static ssize_t
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 26)
bfad_im_vport_delete(struct device *cdev, struct device_attribute *attr,
					 const char *buf, size_t count)
#else
bfad_im_vport_delete(struct class_device *cdev, const char *buf, size_t count)
#endif
{
	struct Scsi_Host  *shost = class_to_shost(cdev);
	struct bfad_im_port_s *im_port =
			(struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_s *bfad = im_port->bfad;
	uint8_t wwn[8];
	int status;
	unsigned long flags;
	bfa_fcs_vport_t *fcs_vport;
	struct bfad_vport_s *bfad_vport;
	struct completion fcomp;

	status = bfad_im_parse_wwn(&buf[0], wwn);
	if (status)
		return status;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	fcs_vport = bfa_fcs_vport_lookup(&bfad->bfa_fcs, 0, *(wwn_t *)wwn);
	if (fcs_vport == NULL) {
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		return -EINVAL;
	}

	bfad_vport = fcs_vport->vport_drv;
	bfad_vport->drv_port.flags |= BFAD_PORT_DELETE;
	bfad_vport->comp_del = &fcomp;
	init_completion(bfad_vport->comp_del);

	status = bfa_fcs_vport_delete(fcs_vport);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	wait_for_completion(bfad_vport->comp_del);

	if (status != BFA_STATUS_OK) {
		return -EIO;
	}

	bfad_os_scsi_host_free(bfad, bfad_vport->drv_port.im_port);
	list_del(&bfad_vport->list_entry);
	kfree(bfad_vport);
	return count;
}

#endif

struct fc_function_template bfad_im_fc_function_template = {

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 9)
	/*
	 * Target dynamic attributes
	 */
	.get_starget_port_id = bfad_im_get_starget_port_id,
	.show_starget_port_id = 1,
	.get_starget_node_name = bfad_im_get_starget_node_name,
	.show_starget_node_name = 1,
	.get_starget_port_name = bfad_im_get_starget_port_name,
	.show_starget_port_name = 1,

	/*
	 * Host dynamic attribute
	 */
	.get_host_port_id = bfad_im_get_host_port_id,
	.show_host_port_id = 1,

	.issue_fc_host_lip = bfad_im_issue_fc_host_lip,
#endif

	/*
	 * TBD: Need to find out the kernel version starting to support
	 * FC transport
	 */
#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 10)
	/*
	 * Host fixed attributes
	 */
	.show_host_node_name = 1,
	.show_host_port_name = 1,
	.show_host_supported_classes = 1,
	.show_host_supported_fc4s = 1,
	.show_host_supported_speeds = 1,
	.show_host_maxframe_size = 1,

	/*
	 * More host dynamic attributes
	 */
	.show_host_port_type = 1,
	.get_host_port_type = bfad_im_get_host_port_type,
	.show_host_port_state = 1,
	.get_host_port_state = bfad_im_get_host_port_state,
	.show_host_active_fc4s = 1,
	.get_host_active_fc4s = bfad_im_get_host_active_fc4s,
	.show_host_speed = 1,
	.get_host_speed = bfad_im_get_host_speed,
	.show_host_fabric_name = 1,
	.get_host_fabric_name = bfad_im_get_host_fabric_name,

	/*
	 * This is host fixed attribute in lk 2.6.16 but changed to dynamic
	 * in lk 2.6.?
	 */
	.show_host_symbolic_name = 1,

	/*
	 * Statistics
	 */
	.get_fc_host_stats = bfad_im_get_stats,
	.reset_fc_host_stats = bfad_im_reset_stats,

	/*
	 * Allocation length for host specific data
	 */
	.dd_fcrport_size = sizeof(struct bfad_itnim_data_s *),

	/*
	 * Remote port fixed attributes
	 */
	.show_rport_maxframe_size = 1,
	.show_rport_supported_classes = 1,
	.show_rport_dev_loss_tmo = 1,
	.set_rport_dev_loss_tmo = bfad_im_set_rport_loss_tmo,
#endif
#if defined(__VMKLNX__)
	.vport_create = bfad_im_vmw_vport_create,
	.vport_delete = bfad_im_vmw_vport_delete,
#endif
#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 31)
	.bsg_request = bfad_im_bsg_request,
	.bsg_timeout = bfad_im_bsg_timeout,
	.vport_create = bfad_im_vport_create,
	.vport_delete = bfad_im_vport_delete,
	.vport_disable = bfad_im_vport_disable,
	.set_vport_symbolic_name = bfad_im_vport_set_symbolic_name,
#endif
};

#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 31)
struct fc_function_template bfad_im_vport_fc_function_template = {

	/* Target dynamic attributes */
	.get_starget_port_id = bfad_im_get_starget_port_id,
	.show_starget_port_id = 1,
	.get_starget_node_name = bfad_im_get_starget_node_name,
	.show_starget_node_name = 1,
	.get_starget_port_name = bfad_im_get_starget_port_name,
	.show_starget_port_name = 1,

	/* Host dynamic attribute */
	.get_host_port_id = bfad_im_get_host_port_id,
	.show_host_port_id = 1,

	/* Host fixed attributes */
	.show_host_node_name = 1,
	.show_host_port_name = 1,
	.show_host_supported_classes = 1,
	.show_host_supported_fc4s = 1,
	.show_host_supported_speeds = 1,
	.show_host_maxframe_size = 1,

	/* More host dynamic attributes */
	.show_host_port_type = 1,
	.get_host_port_type = bfad_im_get_host_port_type,
	.show_host_port_state = 1,
	.get_host_port_state = bfad_im_get_host_port_state,
	.show_host_active_fc4s = 1,
	.get_host_active_fc4s = bfad_im_get_host_active_fc4s,
	.show_host_speed = 1,
	.get_host_speed = bfad_im_get_host_speed,
	.show_host_fabric_name = 1,
	.get_host_fabric_name = bfad_im_get_host_fabric_name,

	.show_host_symbolic_name = 1,

	/* Statistics */
	.get_fc_host_stats = bfad_im_get_stats,
	.reset_fc_host_stats = bfad_im_reset_stats,

	/* Allocation length for host specific data */
	.dd_fcrport_size = sizeof(struct bfad_itnim_data_s *),

	/* Remote port fixed attributes */
	.show_rport_maxframe_size = 1,
	.show_rport_supported_classes = 1,
	.show_rport_dev_loss_tmo = 1,
	.set_rport_dev_loss_tmo = bfad_im_set_rport_loss_tmo,

};
#endif

/**
 * @}
 * @dg Scsi_Host_attrs SCSI host attributes
 * @{
 */
static ssize_t
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 26)
bfad_im_serial_num_show(struct device *cdev,
			struct device_attribute *attr, char *buf)
#else
bfad_im_serial_num_show(struct class_device *cdev, char *buf)
#endif
{
	struct Scsi_Host *shost = class_to_shost(cdev);
	struct bfad_im_port_s *im_port =
			(struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_s *bfad = im_port->bfad;
	char serial_num[BFA_ADAPTER_SERIAL_NUM_LEN];

	bfa_get_adapter_serial_num(&bfad->bfa, serial_num);
	return snprintf(buf, PAGE_SIZE, "%s\n", serial_num);
}

static ssize_t
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 26)
bfad_im_model_show(struct device *cdev,
			struct device_attribute *attr, char *buf)
#else
bfad_im_model_show(struct class_device *cdev, char *buf)
#endif
{
	struct Scsi_Host *shost = class_to_shost(cdev);
	struct bfad_im_port_s *im_port =
			(struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_s *bfad = im_port->bfad;
	char model[BFA_ADAPTER_MODEL_NAME_LEN];

	bfa_get_adapter_model(&bfad->bfa, model);
	return snprintf(buf, PAGE_SIZE, "%s\n", model);
}

static ssize_t
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 26)
bfad_im_model_desc_show(struct device *cdev,
			struct device_attribute *attr, char *buf)
#else
bfad_im_model_desc_show(struct class_device *cdev, char *buf)
#endif
{
	struct Scsi_Host *shost = class_to_shost(cdev);
	struct bfad_im_port_s *im_port =
			(struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_s *bfad = im_port->bfad;
	struct bfa_s *bfa = &bfad->bfa;
	struct bfa_ioc_s *ioc = &bfa->ioc;
	char model[BFA_ADAPTER_MODEL_NAME_LEN];
	char model_descr[BFA_ADAPTER_MODEL_DESCR_LEN];
	int nports = 0;

	bfa_get_adapter_model(bfa, model);
	nports = bfa_get_nports(bfa);
	if (!strcmp(model, "QLogic-425")) {
		snprintf(model_descr, BFA_ADAPTER_MODEL_DESCR_LEN,
			"QLogic BR-series 4Gbps PCIe dual port FC HBA");
	} else if (!strcmp(model, "QLogic-825")) {
		snprintf(model_descr, BFA_ADAPTER_MODEL_DESCR_LEN,
			"QLogic BR-series 8Gbps PCIe dual port FC HBA");
	} else if (!strcmp(model, "QLogic-42B")) {
		snprintf(model_descr, BFA_ADAPTER_MODEL_DESCR_LEN,
			"QLogic BR-series 4Gbps PCIe dual port FC HBA for HP");
	} else if (!strcmp(model, "QLogic-82B")) {
		snprintf(model_descr, BFA_ADAPTER_MODEL_DESCR_LEN,
			"QLogic BR-series 8Gbps PCIe dual port FC HBA for HP");
	} else if (!strcmp(model, "QLogic-1010")) {
		snprintf(model_descr, BFA_ADAPTER_MODEL_DESCR_LEN,
			"QLogic BR-series 10Gbps single port CNA");
	} else if (!strcmp(model, "QLogic-1020")) {
		snprintf(model_descr, BFA_ADAPTER_MODEL_DESCR_LEN,
			"QLogic BR-series 10Gbps dual port CNA");
	} else if (!strcmp(model, "QLogic-1007")) {
		snprintf(model_descr, BFA_ADAPTER_MODEL_DESCR_LEN,
			"QLogic BR-series 10Gbps CNA for IBM Blade Center");
	} else if (!strcmp(model, "QLogic-415")) {
		snprintf(model_descr, BFA_ADAPTER_MODEL_DESCR_LEN,
			"QLogic BR-series 4Gbps PCIe single port FC HBA");
	} else if (!strcmp(model, "QLogic-815")) {
		snprintf(model_descr, BFA_ADAPTER_MODEL_DESCR_LEN,
			"QLogic BR-series 8Gbps PCIe single port FC HBA");
	} else if (!strcmp(model, "QLogic-41B")) {
		snprintf(model_descr, BFA_ADAPTER_MODEL_DESCR_LEN,
			"QLogic BR-series 4Gbps PCIe single port FC HBA for HP");
	} else if (!strcmp(model, "QLogic-81B")) {
		snprintf(model_descr, BFA_ADAPTER_MODEL_DESCR_LEN,
			"QLogic BR-series 8Gbps PCIe single port FC HBA for HP");
	} else if (!strcmp(model, "QLogic-804")) {
		snprintf(model_descr, BFA_ADAPTER_MODEL_DESCR_LEN,
			"QLogic BR-series 8Gbps FC HBA for HP Bladesystem C-class");
	} else if (!strcmp(model, "QLogic-902") ||
		   !strcmp(model, "QLogic-1741")) {
		snprintf(model_descr, BFA_ADAPTER_MODEL_DESCR_LEN,
			"QLogic BR-series 10Gbps CNA for Dell M-Series Blade Servers");
	} else if (!strcmp(model, "QLogic-1560")) {
		if (nports == 1)
			snprintf(model_descr, BFA_ADAPTER_MODEL_DESCR_LEN,
				"QLogic BR-series 16Gbps PCIe single port FC HBA");
		else
			snprintf(model_descr, BFA_ADAPTER_MODEL_DESCR_LEN,
				"QLogic BR-series 16Gbps PCIe dual port FC HBA");
	} else if (!strcmp(model, "QLogic-1710")) {
		if (nports == 1)
			snprintf(model_descr, BFA_ADAPTER_MODEL_DESCR_LEN,
				"QLogic BR-series 10Gbps single port CNA");
		else
			snprintf(model_descr, BFA_ADAPTER_MODEL_DESCR_LEN,
				"QLogic BR-series 10Gbps dual port CNA");
	} else if (strstr(model, "QLogic-1860")) {
		if (nports == 1 && bfa_ioc_is_cna(ioc))
			snprintf(model_descr, BFA_ADAPTER_MODEL_DESCR_LEN,
				"QLogic BR-series 10Gbps single port CNA");
		else if (nports == 1 && !bfa_ioc_is_cna(ioc))
			snprintf(model_descr, BFA_ADAPTER_MODEL_DESCR_LEN,
				"QLogic BR-series 16Gbps PCIe single port FC HBA");
		else if (nports == 2 && bfa_ioc_is_cna(ioc))
			snprintf(model_descr, BFA_ADAPTER_MODEL_DESCR_LEN,
				"QLogic BR-series 10Gbps dual port CNA");
		else if (nports == 2 && !bfa_ioc_is_cna(ioc))
			snprintf(model_descr, BFA_ADAPTER_MODEL_DESCR_LEN,
				"QLogic BR-series 16Gbps PCIe dual port FC HBA");
	} else if (!strcmp(model, "QLogic-1867")) {
		if (nports == 1 && !bfa_ioc_is_cna(ioc))
			snprintf(model_descr, BFA_ADAPTER_MODEL_DESCR_LEN,
			     "QLogic BR-series 16Gbps PCIe single port FC HBA for IBM");
		else if (nports == 2 && !bfa_ioc_is_cna(ioc))
			snprintf(model_descr, BFA_ADAPTER_MODEL_DESCR_LEN,
			     "QLogic BR-series 16Gbps PCIe dual port FC HBA for IBM");
	} else {
		snprintf(model_descr, BFA_ADAPTER_MODEL_DESCR_LEN,
			"Invalid Model");
	}

	return snprintf(buf, PAGE_SIZE, "%s\n", model_descr);
}

static ssize_t
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 26)
bfad_im_node_name_show(struct device *cdev,
			struct device_attribute *attr, char *buf)
#else
bfad_im_node_name_show(struct class_device *cdev, char *buf)
#endif
{
	struct Scsi_Host *shost = class_to_shost(cdev);
	struct bfad_im_port_s *im_port =
			(struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_port_s    *port = im_port->port;
	uint64_t nwwn;

	nwwn = bfa_fcs_lport_get_nwwn(port->fcs_port);
	return snprintf(buf, PAGE_SIZE, "0x%llx\n", bfa_os_htonll(nwwn));
}

static ssize_t
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 26)
bfad_im_symbolic_name_show(struct device *cdev,
			struct device_attribute *attr, char *buf)
#else
bfad_im_symbolic_name_show(struct class_device *cdev, char *buf)
#endif
{
	struct Scsi_Host *shost = class_to_shost(cdev);
	struct bfad_im_port_s *im_port =
			(struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_s *bfad = im_port->bfad;
	struct bfa_lport_attr_s port_attr;
	char symname[BFA_SYMNAME_MAXLEN];

	bfa_fcs_lport_get_attr(&bfad->bfa_fcs.fabric.bport, &port_attr);
	strncpy(symname,
		port_attr.port_cfg.sym_name.symname, BFA_SYMNAME_MAXLEN);
	return snprintf(buf, PAGE_SIZE, "%s\n", symname);
}

static ssize_t
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 26)
bfad_im_hw_version_show(struct device *cdev,
			struct device_attribute *attr, char *buf)
#else
bfad_im_hw_version_show(struct class_device *cdev, char *buf)
#endif
{
	struct Scsi_Host *shost = class_to_shost(cdev);
	struct bfad_im_port_s *im_port =
			(struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_s *bfad = im_port->bfad;
	char hw_ver[BFA_VERSION_LEN];

	bfa_get_pci_chip_rev(&bfad->bfa, hw_ver);
	return snprintf(buf, PAGE_SIZE, "%s\n", hw_ver);
}

static ssize_t
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 26)
bfad_im_drv_version_show(struct device *cdev,
			struct device_attribute *attr, char *buf)
#else
bfad_im_drv_version_show(struct class_device *cdev, char *buf)
#endif
{
	return snprintf(buf, PAGE_SIZE, "%s\n", BFAD_DRIVER_VERSION);
}

static ssize_t
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 26)
bfad_im_optionrom_version_show(struct device *cdev,
			struct device_attribute *attr, char *buf)
#else
bfad_im_optionrom_version_show(struct class_device *cdev, char *buf)
#endif
{
	struct Scsi_Host *shost = class_to_shost(cdev);
	struct bfad_im_port_s *im_port =
			(struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_s *bfad = im_port->bfad;
	char optrom_ver[BFA_VERSION_LEN];

	bfa_get_adapter_optrom_ver(&bfad->bfa, optrom_ver);
	return snprintf(buf, PAGE_SIZE, "%s\n", optrom_ver);
}

static ssize_t
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 26)
bfad_im_fw_version_show(struct device *cdev,
			struct device_attribute *attr, char *buf)
#else
bfad_im_fw_version_show(struct class_device *cdev, char *buf)
#endif
{
	struct Scsi_Host *shost = class_to_shost(cdev);
	struct bfad_im_port_s *im_port =
			(struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_s *bfad = im_port->bfad;
	char fw_ver[BFA_VERSION_LEN];

	bfa_get_adapter_fw_ver(&bfad->bfa, fw_ver);
	return snprintf(buf, PAGE_SIZE, "%s\n", fw_ver);
}

static ssize_t
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 26)
bfad_im_num_of_ports_show(struct device *cdev,
			struct device_attribute *attr, char *buf)
#else
bfad_im_num_of_ports_show(struct class_device *cdev, char *buf)
#endif
{
	struct Scsi_Host *shost = class_to_shost(cdev);
	struct bfad_im_port_s *im_port =
			(struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_s *bfad = im_port->bfad;

	return snprintf(buf, PAGE_SIZE, "%d\n",
			bfa_get_nports(&bfad->bfa));
}

static ssize_t
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 26)
bfad_im_drv_name_show(struct device *cdev,
		struct device_attribute *attr, char *buf)
#else
bfad_im_drv_name_show(struct class_device *cdev, char *buf)
#endif
{
	return snprintf(buf, PAGE_SIZE, "%s\n", BFAD_DRIVER_NAME);
}

static ssize_t
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 26)
bfad_im_num_of_discovered_ports_show(struct device *cdev,
			struct device_attribute *attr, char *buf)
#else
bfad_im_num_of_discovered_ports_show(struct class_device *cdev, char *buf)
#endif
{
	struct Scsi_Host *shost = class_to_shost(cdev);
	struct bfad_im_port_s *im_port =
			(struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_port_s    *port = im_port->port;
	struct bfad_s *bfad = im_port->bfad;
	uint32_t nrports = 2048;
	wwn_t *rports = NULL;
	unsigned long flags;

	rports = kzalloc(sizeof(wwn_t) * nrports, GFP_ATOMIC);
	if (rports == NULL)
		return snprintf(buf, PAGE_SIZE, "Failed\n");

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	bfa_fcs_lport_get_rports(port->fcs_port, rports, &nrports);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	kfree(rports);

	return snprintf(buf, PAGE_SIZE, "%d\n", nrports);
}

static ssize_t
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 26)
bfad_im_irq_show(struct device *cdev, struct device_attribute *attr, char *buf)
#else
bfad_im_irq_show(struct class_device *cdev, char *buf)
#endif
{
	struct Scsi_Host *shost = class_to_shost(cdev);
	struct bfad_im_port_s *im_port =
			(struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_s *bfad = im_port->bfad;

#ifdef CONFIG_PCI_MSI
	ssize_t tot_size = 0, rtn = 0;
	int i;

	if (!(bfad->bfad_flags & BFAD_MSIX_ON))
		return snprintf(buf, PAGE_SIZE, "%d\n", bfad->pcidev->irq);
	else {
		buf[0] = '\0';
		for (i = 0; i < bfad->nvec; i++) {
			rtn = snprintf(buf, PAGE_SIZE - tot_size, "%d ",
						bfad->msix_tab[i].msix.vector);

			if (rtn <= 0)
				goto out;

			buf += rtn;
			tot_size += rtn;
		}

		rtn = snprintf(buf, PAGE_SIZE - tot_size, "\n");
	}
out:
	if (rtn < 0)
		return rtn;
	else
		return tot_size + rtn;
#else
	return snprintf(buf, PAGE_SIZE, "%d\n", bfad->pcidev->irq);
#endif


}



#if (LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 10))
static ssize_t
bfad_im_port_name_show(struct class_device *cdev, char *buf)
{
	struct Scsi_Host *shost = class_to_shost(cdev);
	struct bfad_im_port_s *im_port =
			(struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_port_s    *port = im_port->port;
	uint64_t pwwn;

	pwwn = bfa_fcs_lport_get_pwwn(port->fcs_port);
	return snprintf(buf, PAGE_SIZE, "0x%llx\n", bfa_os_htonll(pwwn));
}

static ssize_t
bfad_im_port_id_show(struct class_device *cdev, char *buf)
{
	struct Scsi_Host *shost = class_to_shost(cdev);
	struct bfad_im_port_s *im_port =
			(struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_port_s    *port = im_port->port;
	uint32_t fc_id;

	fc_id = bfa_fcs_lport_get_fcid(port->fcs_port);
	return snprintf(buf, PAGE_SIZE, "0x%x\n", bfa_os_hton3b(fc_id));
}

static ssize_t
bfad_im_port_type_show(struct class_device *cdev, char *buf)
{
	struct Scsi_Host *shost = class_to_shost(cdev);
	struct bfad_im_port_s *im_port =
			(struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_s *bfad = im_port->bfad;
	struct bfa_port_attr_s attr;
	char porttype[16];

	bfa_fcport_get_attr(&bfad->bfa, &attr);

	switch (attr.port_type) {
	case BFA_PORT_TYPE_NPORT:
		strcpy(porttype, "NPORT");
		break;
	case BFA_PORT_TYPE_NLPORT:
		strcpy(porttype, "NLPORT");
		break;
	case BFA_PORT_TYPE_P2P:
		strcpy(porttype, "P2P");
		break;
	case BFA_PORT_TYPE_LPORT:
		strcpy(porttype, "LPORT");
		break;
	default:
		strcpy(porttype, "UNKNOWN");
		break;
	}

	return snprintf(buf, PAGE_SIZE, "%s\n", porttype);
}

static ssize_t
bfad_im_port_state_show(struct class_device *cdev, char *buf)
{
	struct Scsi_Host *shost = class_to_shost(cdev);
	struct bfad_im_port_s *im_port =
			(struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_s *bfad = im_port->bfad;
	struct bfa_port_attr_s attr;
	char portstate[16];

	bfa_fcport_get_attr(&bfad->bfa, &attr);

	switch (attr.port_state) {
	case BFA_PORT_ST_LINKDOWN:
		strcpy(portstate, "Link Down");
		break;
	case BFA_PORT_ST_LINKUP:
		strcpy(portstate, "Link Up");
		break;
	case BFA_PORT_ST_UNINIT:
	case BFA_PORT_ST_ENABLING_QWAIT:
	case BFA_PORT_ST_ENABLING:
	case BFA_PORT_ST_DISABLING_QWAIT:
	case BFA_PORT_ST_DISABLING:
	case BFA_PORT_ST_DISABLED:
	case BFA_PORT_ST_STOPPED:
	case BFA_PORT_ST_IOCDOWN:
	default:
		strcpy(portstate, "Unknown");
		break;
	}

	return snprintf(buf, PAGE_SIZE, "%s\n", portstate);
}

static ssize_t
bfad_im_supported_classes_show(struct class_device *cdev, char *buf)
{
	struct Scsi_Host *shost = class_to_shost(cdev);
	struct bfad_im_port_s *im_port =
			(struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_s *bfad = im_port->bfad;
	struct bfa_port_attr_s attr;
	uint32_t classes;

	bfa_fcport_get_attr(&bfad->bfa, &attr);
	classes = attr.cos_supported;
	return snprintf(buf, PAGE_SIZE, "%d\n", classes);
}

static ssize_t
bfad_im_supported_fc4s_show(struct class_device *cdev, char *buf)
{
	u8 fc4s[32];
	int i, len = 0, total = 0;

	memset(fc4s, 0, sizeof(fc4s));
	if (supported_fc4s & (BFA_LPORT_ROLE_FCP_IM | BFA_LPORT_ROLE_FCP_TM))
		/*
		 * For FCP type 0x08
		 */
		fc4s[2] = 1;
	if (supported_fc4s | BFA_LPORT_ROLE_FCP_IPFC)
		/*
		 * For LLC/SNAP type 0x05
		 */
		fc4s[3] = 0x20;
	/*
	 * For fibre channel services type 0x20
	 */
	fc4s[7] = 1;

	for (i = 0; i < 32; i++) {
		len = sprintf(buf + total, "0x%02x", fc4s[i]);
		total += len;
		if (i != 31) {
			len = sprintf(buf + total, "%s", " ");
			total += len;
		}
	}

	return total;
}

static ssize_t
bfad_im_active_fc4s_show(struct class_device *cdev, char *buf)
{
	struct Scsi_Host *shost = class_to_shost(cdev);
	struct bfad_im_port_s *im_port =
			(struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_port_s    *port = im_port->port;
	u8 fc4s[32];
	int i, len = 0, total = 0;

	memset(fc4s, 0, sizeof(fc4s));
	if (port->
	    supported_fc4s & (BFA_LPORT_ROLE_FCP_IM | BFA_LPORT_ROLE_FCP_TM))
		/*
		 * For FCP type 0x08
		 */
		fc4s[2] = 1;
	if (port->supported_fc4s | BFA_LPORT_ROLE_FCP_IPFC)
		/*
		 * For LLC/SNAP type 0x05
		 */
		fc4s[3] = 0x20;
	/*
	 * For fibre channel services type 0x20
	 */
	fc4s[7] = 1;

	for (i = 0; i < 32; i++) {
		len = sprintf(buf + total, "0x%02x", fc4s[i]);
		total += len;
		if (i != 31) {
			len = sprintf(buf + total, "%s", " ");
			total += len;
		}
	}

	return total;
}

static ssize_t
bfad_im_supported_speeds_show(struct class_device *cdev, char *buf)
{
	return snprintf(buf, PAGE_SIZE, "%s\n",
			"1 Gbit, 2 Gbit, 4 Gbit," " 8 Gbit");
}

static ssize_t
bfad_im_speed_show(struct class_device *cdev, char *buf)
{
	struct Scsi_Host *shost = class_to_shost(cdev);
	struct bfad_im_port_s *im_port =
			(struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_s *bfad = im_port->bfad;
	struct bfa_port_attr_s attr;
	int speed;

	bfa_fcport_get_attr(&bfad->bfa, &attr);
	speed = attr.speed;
	return snprintf(buf, PAGE_SIZE, "%d Gbit\n", speed);
}

static ssize_t
bfad_im_maxframe_size_show(struct class_device *cdev, char *buf)
{
	struct Scsi_Host *shost = class_to_shost(cdev);
	struct bfad_im_port_s *im_port =
			(struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_s *bfad = im_port->bfad;
	struct bfa_port_attr_s attr;
	uint32_t maxfrsize;

	bfa_fcport_get_attr(&bfad->bfa, &attr);
	maxfrsize = attr.pport_cfg.maxfrsize;
	return snprintf(buf, PAGE_SIZE, "%d\n", maxfrsize);
}

static ssize_t
bfad_im_fabric_name_show(struct class_device *cdev, char *buf)
{
	struct Scsi_Host *shost = class_to_shost(cdev);
	struct bfad_im_port_s *im_port =
			(struct bfad_im_port_s *) shost->hostdata[0];
	struct bfad_port_s    *port = im_port->port;
	wwn_t fabric_nwwn = 0;

	/*
	 * TBD: Set to fabric nwwn if in switch enviroment. Set to local port
	 * nwwn if private loop or P2P.
	 */
	fabric_nwwn = bfa_fcs_lport_get_fabric_name(port->fcs_port);
	return snprintf(buf, PAGE_SIZE, "0x%llx\n", bfa_os_htonll(fabric_nwwn));
}
#endif

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 26))

static DEVICE_ATTR(serial_number, S_IRUGO,
		bfad_im_serial_num_show, NULL);
static DEVICE_ATTR(model, S_IRUGO, bfad_im_model_show, NULL);
static DEVICE_ATTR(model_description, S_IRUGO,
		bfad_im_model_desc_show, NULL);
static DEVICE_ATTR(node_name, S_IRUGO, bfad_im_node_name_show, NULL);
static DEVICE_ATTR(symbolic_name, S_IRUGO,
		bfad_im_symbolic_name_show, NULL);
static DEVICE_ATTR(hardware_version, S_IRUGO,
		bfad_im_hw_version_show, NULL);
static DEVICE_ATTR(driver_version, S_IRUGO,
		bfad_im_drv_version_show, NULL);
static DEVICE_ATTR(option_rom_version, S_IRUGO,
		bfad_im_optionrom_version_show, NULL);
static DEVICE_ATTR(firmware_version, S_IRUGO,
		bfad_im_fw_version_show, NULL);
static DEVICE_ATTR(number_of_ports, S_IRUGO,
		bfad_im_num_of_ports_show, NULL);
static DEVICE_ATTR(driver_name, S_IRUGO, bfad_im_drv_name_show, NULL);
static DEVICE_ATTR(number_of_discovered_ports, S_IRUGO,
		bfad_im_num_of_discovered_ports_show, NULL);
static DEVICE_ATTR(irq, S_IRUGO, bfad_im_irq_show, NULL);
#if (LINUX_VERSION_CODE <= KERNEL_VERSION(2, 6, 31))
static DEVICE_ATTR(vport_create, S_IWUSR, NULL, bfad_im_vport_create);
static DEVICE_ATTR(vport_delete, S_IWUSR, NULL, bfad_im_vport_delete);
#endif
struct device_attribute *bfad_im_host_attrs[] = {
	&dev_attr_serial_number,
	&dev_attr_model,
	&dev_attr_model_description,
	&dev_attr_node_name,
	&dev_attr_symbolic_name,
	&dev_attr_hardware_version,
	&dev_attr_driver_version,
	&dev_attr_option_rom_version,
	&dev_attr_firmware_version,
	&dev_attr_number_of_ports,
	&dev_attr_driver_name,
	&dev_attr_number_of_discovered_ports,
	&dev_attr_irq,
#if (LINUX_VERSION_CODE <= KERNEL_VERSION(2, 6, 31))
	&dev_attr_vport_create,
	&dev_attr_vport_delete,
#endif
	NULL,
};

struct device_attribute *bfad_im_vport_attrs[] = {
	&dev_attr_serial_number,
	&dev_attr_model,
	&dev_attr_model_description,
	&dev_attr_node_name,
	&dev_attr_symbolic_name,
	&dev_attr_hardware_version,
	&dev_attr_driver_version,
	&dev_attr_option_rom_version,
	&dev_attr_firmware_version,
	&dev_attr_number_of_ports,
	&dev_attr_driver_name,
	&dev_attr_number_of_discovered_ports,
	NULL,
};

#else

static CLASS_DEVICE_ATTR(serial_number, S_IRUGO,
				bfad_im_serial_num_show, NULL);
static CLASS_DEVICE_ATTR(model, S_IRUGO, bfad_im_model_show, NULL);
static CLASS_DEVICE_ATTR(model_description, S_IRUGO,
				bfad_im_model_desc_show, NULL);
static CLASS_DEVICE_ATTR(node_name, S_IRUGO, bfad_im_node_name_show, NULL);
static CLASS_DEVICE_ATTR(symbolic_name, S_IRUGO,
				bfad_im_symbolic_name_show, NULL);
static CLASS_DEVICE_ATTR(hardware_version, S_IRUGO,
				bfad_im_hw_version_show, NULL);
static CLASS_DEVICE_ATTR(driver_version, S_IRUGO,
				bfad_im_drv_version_show, NULL);
static CLASS_DEVICE_ATTR(option_rom_version, S_IRUGO,
				bfad_im_optionrom_version_show, NULL);
static CLASS_DEVICE_ATTR(firmware_version, S_IRUGO,
				bfad_im_fw_version_show, NULL);
static CLASS_DEVICE_ATTR(number_of_ports, S_IRUGO,
				bfad_im_num_of_ports_show, NULL);
static CLASS_DEVICE_ATTR(driver_name, S_IRUGO, bfad_im_drv_name_show, NULL);
static CLASS_DEVICE_ATTR(number_of_discovered_ports, S_IRUGO,
				bfad_im_num_of_discovered_ports_show, NULL);
static CLASS_DEVICE_ATTR(irq, S_IRUGO, bfad_im_irq_show, NULL);
static CLASS_DEVICE_ATTR(vport_create, S_IWUSR, NULL, bfad_im_vport_create);
static CLASS_DEVICE_ATTR(vport_delete, S_IWUSR, NULL, bfad_im_vport_delete);

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 10))
static CLASS_DEVICE_ATTR(port_name, S_IRUGO, bfad_im_port_name_show, NULL);
static CLASS_DEVICE_ATTR(port_id, S_IRUGO, bfad_im_port_id_show, NULL);
static CLASS_DEVICE_ATTR(port_type, S_IRUGO, bfad_im_port_type_show, NULL);
static CLASS_DEVICE_ATTR(port_state, S_IRUGO, bfad_im_port_state_show, NULL);
static CLASS_DEVICE_ATTR(supported_classes, S_IRUGO,
				bfad_im_supported_classes_show, NULL);
static CLASS_DEVICE_ATTR(supported_fc4s, S_IRUGO,
				bfad_im_supported_fc4s_show, NULL);
static CLASS_DEVICE_ATTR(active_fc4s, S_IRUGO, bfad_im_active_fc4s_show, NULL);
static CLASS_DEVICE_ATTR(supported_speeds, S_IRUGO,
				bfad_im_supported_speeds_show, NULL);
static CLASS_DEVICE_ATTR(speed, S_IRUGO, bfad_im_speed_show, NULL);
static CLASS_DEVICE_ATTR(maxframe_size, S_IRUGO,
				bfad_im_maxframe_size_show, NULL);
static CLASS_DEVICE_ATTR(fabric_name, S_IRUGO, bfad_im_fabric_name_show, NULL);
#endif

struct class_device_attribute *bfad_im_host_attrs[] = {
	&class_device_attr_serial_number,
	&class_device_attr_model,
	&class_device_attr_model_description,
	&class_device_attr_node_name,
	&class_device_attr_symbolic_name,
	&class_device_attr_hardware_version,
	&class_device_attr_driver_version,
	&class_device_attr_option_rom_version,
	&class_device_attr_firmware_version,
	&class_device_attr_number_of_ports,
	&class_device_attr_driver_name,
	&class_device_attr_number_of_discovered_ports,
	&class_device_attr_vport_create,
	&class_device_attr_vport_delete,
	&class_device_attr_irq,
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 10))
	&class_device_attr_port_name,
	&class_device_attr_port_id,
	&class_device_attr_port_type,
	&class_device_attr_port_state,
	&class_device_attr_supported_classes,
	&class_device_attr_supported_fc4s,
	&class_device_attr_active_fc4s,
	&class_device_attr_supported_speeds,
	&class_device_attr_speed,
	&class_device_attr_maxframe_size,
	&class_device_attr_fabric_name,
#endif
	NULL,
};

struct class_device_attribute *bfad_im_vport_attrs[] = {
	&class_device_attr_serial_number,
	&class_device_attr_model,
	&class_device_attr_model_description,
	&class_device_attr_node_name,
	&class_device_attr_symbolic_name,
	&class_device_attr_hardware_version,
	&class_device_attr_driver_version,
	&class_device_attr_option_rom_version,
	&class_device_attr_firmware_version,
	&class_device_attr_number_of_ports,
	&class_device_attr_driver_name,
	&class_device_attr_number_of_discovered_ports,
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 10))
	&class_device_attr_port_name,
	&class_device_attr_port_id,
	&class_device_attr_port_type,
	&class_device_attr_port_state,
	&class_device_attr_supported_classes,
	&class_device_attr_supported_fc4s,
	&class_device_attr_active_fc4s,
	&class_device_attr_supported_speeds,
	&class_device_attr_speed,
	&class_device_attr_maxframe_size,
	&class_device_attr_fabric_name,
#endif
	NULL,
};

#endif /* kernel version < 2.6.26 */

/**
 * @}
 */
