/*
 * Copyright (c) 2005-2014 Brocade Communications Systems, Inc.
 * Copyright (c) 2014 QLogic Corporation.
 * All rights reserved
 * www.qlogic.com
 *
 * See LICENSE.bfa for copyright and licensing details.
 */

/**
 * @brief
 * Contains base driver definitions.
 */

/**
 * @file bfa_drv.h Linux driver data structures.
 */

#ifndef __BFAD_DRV_H__
#define __BFAD_DRV_H__

#include "bfa_os_inc.h"
#include "bfa_version.h"

#include <bfa/bfa_modules.h>
#include <fcs/bfa_fcs.h>
#include <defs/bfa_defs_fcs.h>

#include <bfa/bfa_plog.h>
#include "aen/bfa_aen.h"
#include <log/bfa_log_linux.h>
#include <cs/bfa_cs.h>

#ifdef CONFIG_PCIEAER
#define  SUPPORT_PCI_AER
#include <linux/aer.h>
#endif

#ifndef VMKLNX_DDI_VERSION_MAJOR
#define VMKLNX_DDI_VERSION_MAJOR VMKLNX_API_VERSION_MAJOR_NUM
#endif
#ifndef VMKLNX_DDI_VERSION_MINOR
#define VMKLNX_DDI_VERSION_MINOR VMKLNX_API_VERSION_MINOR_NUM
#endif
#define MKVERS(maj, min, update) (((maj) << 16) | (min << 8) | update)
#define ESX40_VERS		MKVERS(9, 0, 0)
#define ESX41_VERS		MKVERS(9, 1, 0)
#define ESX50_VERS		MKVERS(9, 2, 0)
#define ESX51_VERS		MKVERS(9, 2, 1)
#define ESX55_VERS		MKVERS(9, 2, 2)
#ifdef VMKLNX_DDI_VERSION_UPDATE
#define ESX_VERS MKVERS(VMKLNX_DDI_VERSION_MAJOR, \
	VMKLNX_DDI_VERSION_MINOR, VMKLNX_DDI_VERSION_UPDATE)
#else
#define ESX_VERS MKVERS(VMKLNX_DDI_VERSION_MAJOR, VMKLNX_DDI_VERSION_MINOR, 0)
#endif

#if (defined(__VMKLNX__) && (ESX_VERS > ESX50_VERS))
#include <vmklinux_9/vmklinux_scsi.h>
#include <vmklinux_9/vmklinux_iodm.h>
#endif

#define BFAD_DRIVER_NAME	"bfa"
#define BFAD_DRIVER_VERSION    BFA_VERSION


#if defined(CONFIG_PCIEAER)
#define BFAD_ENABLE_PCIE_AER(x) pci_enable_pcie_error_reporting(x)
#define BFAD_DISABLE_PCIE_AER(x) pci_disable_pcie_error_reporting(x)
#define BFAD_CLEANUP_PCIE_AER(x) pci_cleanup_aer_uncorrect_error_status(pdev);
#else
#define BFAD_ENABLE_PCIE_AER(x) {}
#define BFAD_DISABLE_PCIE_AER(x) {}
#define BFAD_CLEANUP_PCIE_AER(x) {}
#endif

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 11))
#define pm_message_t u32
#endif

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 18))
#define BFAD_IRQ_FLAGS SA_SHIRQ
#else
#define BFAD_IRQ_FLAGS IRQF_SHARED
#endif

#ifndef FC_PORTSPEED_8GBIT
#define FC_PORTSPEED_8GBIT 0x10
#endif
#ifndef FC_PORTSPEED_16GBIT
#define FC_PORTSPEED_16GBIT 0x20
#endif

#ifndef DMA_BIT_MASK
#define DMA_BIT_MASK(n) (((n) == 64) ? ~0ULL : ((1ULL<<(n))-1))
#endif
/*
 * BFAD flags
 */
#define BFAD_MSIX_ON				0x00000001
#define BFAD_HAL_INIT_DONE			0x00000002
#define BFAD_DRV_INIT_DONE			0x00000004
#define BFAD_CFG_PPORT_DONE			0x00000008
#define BFAD_HAL_START_DONE			0x00000010
#define BFAD_PORT_ONLINE			0x00000020
#define BFAD_RPORT_ONLINE			0x00000040
#define BFAD_FCS_INIT_DONE			0x00000080
#define BFAD_HAL_INIT_FAIL			0x00000100
#define BFAD_FC4_PROBE_DONE			0x00000200
#define BFAD_INTX_ON				0x00000400
#define BFAD_EEH_BUSY				0x00000800
#define BFAD_EEH_PCI_CHANNEL_IO_PERM_FAILURE	0x00001000

/* bfad_port flags */
#define BFAD_PORT_DELETE			0x00000001

/*
 * BFAD related definition
 */
#define SCSI_SCAN_DELAY		HZ
#define BFAD_STOP_TIMEOUT	30
#define BFAD_SUSPEND_TIMEOUT	BFAD_STOP_TIMEOUT

/*
 * BFAD configuration parameter default values
 */
#define BFAD_LUN_QUEUE_DEPTH 		32
#define BFAD_IO_MAX_SGE 		SG_ALL
#define BFAD_MIN_SECTORS 		128    /* 64k   */
#define BFAD_MAX_SECTORS 		0xFFFF /* 32 MB */

#if (defined(__VMKLNX__) && (ESX_VERS > ESX41_VERS))
typedef irqreturn_t(*bfad_isr_t) (int, void *);
#else
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 19)
#define bfad_isr_t irq_handler_t
#else
typedef irqreturn_t(*bfad_isr_t) (int, void *, struct pt_regs *);
#endif
#endif

#if defined(__VMKERNEL_MODULE__)
#ifndef CONFIG_PCI_MSI
#define CONFIG_PCI_MSI
#endif
#endif

#ifdef CONFIG_PCI_MSI
#define MAX_MSIX_ENTRY 22

struct bfad_msix_s {
	struct bfad_s *bfad;
	struct msix_entry msix;
	char name[32];
};
#endif

/*
 * Only append to the enums defined here to avoid any versioning
 * needed between trace utility and driver version
 */
enum {
	/* 2.6 Driver */
	BFA_TRC_LDRV_BFAD		= 1,
	BFA_TRC_LDRV_BFAD_2_6		= 2,
	BFA_TRC_LDRV_BFAD_2_6_9		= 3,
	BFA_TRC_LDRV_BFAD_2_6_10	= 4,
	BFA_TRC_LDRV_IOCTL		= 5,
	BFA_TRC_LDRV_OS			= 6,
	BFA_TRC_LDRV_IM			= 7,
	BFA_TRC_LDRV_IM_2_6		= 8,
	BFA_TRC_LDRV_IM_2_6_9		= 9,
	BFA_TRC_LDRV_IM_2_6_10		= 10,
	BFA_TRC_LDRV_TM			= 11,
	BFA_TRC_LDRV_IPFC		= 12,
	BFA_TRC_LDRV_IM_2_4		= 13,
	BFA_TRC_LDRV_IM_VMW		= 14,
	BFA_TRC_LDRV_IM_LT_2_6_10	= 15,
	BFA_TRC_LDRV_BSG		= 16,
	BFA_TRC_LDRV_IOCMD		= 17,
};

enum bfad_port_pvb_type {
	BFAD_PORT_PHYS_BASE = 0,
	BFAD_PORT_PHYS_VPORT = 1,
	BFAD_PORT_VF_BASE = 2,
	BFAD_PORT_VF_VPORT = 3,
};

/*
 * PORT data structure
 */
struct bfad_port_s {
	struct list_head list_entry;
	struct bfad_s	*bfad;
	struct bfa_fcs_lport_s *fcs_port;
	uint32_t	roles;
	int32_t		flags;
	uint32_t	supported_fc4s;
	uint8_t		ipfc_flags;
	enum bfad_port_pvb_type pvb_type;
	struct bfad_im_port_s *im_port;	/* IM specific data */
	struct bfad_tm_port_s *tm_port;	/* TM specific data */
	struct bfad_ipfc_port_s *ipfc_port;	/* IPFC specific data */
	 /* port debugfs specific data */
	struct dentry *port_debugfs_root;
#if (defined(__VMKLNX__) && (ESX_VERS > ESX50_VERS))
	bfa_boolean_t multi_q_support;
	uint8_t multi_q_num_qs;
#endif
};

/*
 * VPORT data structure
 */
struct bfad_vport_s {
	struct bfad_port_s     drv_port;
	struct bfa_fcs_vport_s fcs_vport;
	struct completion *comp_del;
	struct list_head list_entry;
};

/*
 * VF data structure
 */
struct bfad_vf_s {
	bfa_fcs_vf_t    fcs_vf;
	struct bfad_port_s    base_port;	/* base port for vf */
	struct bfad_s   *bfad;
};

struct bfad_cfg_param_s {
	uint32_t	rport_del_timeout;
	uint32_t	ioc_queue_depth;
	uint32_t	lun_queue_depth;
	uint32_t	io_max_sge;
	uint32_t	binding_method;
};

union bfad_tmp_buf {
	/* From struct bfa_adapter_attr_s */
	char		manufacturer[BFA_ADAPTER_MFG_NAME_LEN];
	char		serial_num[BFA_ADAPTER_SERIAL_NUM_LEN];
	char		model[BFA_ADAPTER_MODEL_NAME_LEN];
	char		fw_ver[BFA_VERSION_LEN];
	char		optrom_ver[BFA_VERSION_LEN];

	/* From struct bfa_ioc_pci_attr_s */
	uint8_t		chip_rev[BFA_IOC_CHIP_REV_LEN];  /*!< chip revision */

	wwn_t		wwn[BFA_FCS_MAX_LPORTS];
};

/*
 * BFAD (PCI function) data structure
 */
struct bfad_s {
	bfa_sm_t	sm;	/* state machine */
	struct list_head list_entry;
	struct bfa_s	bfa;
	struct bfa_fcs_s bfa_fcs;
	struct pci_dev *pcidev;
	const char *pci_name;
	struct bfa_pcidev_s hal_pcidev;
	struct bfa_ioc_pci_attr_s pci_attr;
	unsigned long   pci_bar0_map;
	void __iomem   *pci_bar0_kva;
	unsigned long   pci_bar2_map;
	void __iomem   *pci_bar2_kva;
	struct completion comp;
	struct completion suspend;
	struct completion enable_comp;
	struct completion disable_comp;
	bfa_boolean_t   disable_active;
	struct bfad_port_s     pport;	/* physical port of the BFAD */
	struct bfa_meminfo_s meminfo;
	struct bfa_iocfc_cfg_s   ioc_cfg;
	uint32_t	inst_no;	/* BFAD instance number */
	uint32_t	bfad_flags;
	spinlock_t      bfad_lock;
	struct task_struct *bfad_tsk;
	struct bfad_cfg_param_s cfg_data;
#ifdef CONFIG_PCI_MSI
	struct bfad_msix_s msix_tab[MAX_MSIX_ENTRY];
	int		nvec;
#endif

#if CONFIG_PM && (LINUX_VERSION_CODE <= KERNEL_VERSION(2, 6, 9))
	uint32_t	pci_state[64];	/* save PCI config space */
#endif
	char 		adapter_name[BFA_ADAPTER_SYM_NAME_LEN];
	char		port_name[BFA_ADAPTER_SYM_NAME_LEN];
	struct timer_list hal_tmo;
	unsigned long   hs_start;
	struct bfad_im_s *im;		/* IM specific data */
	struct bfad_tm_s *tm;		/* TM specific data */
	struct bfad_ipfc_s *ipfc;	/* IPFC specific data */
	atomic_t ipfc_ops_flag;		/* IPFC ops registered or not */
	struct bfa_log_mod_s   log_data;
	struct bfa_trc_mod_s  *trcmod;
	struct bfa_log_mod_s  *logm;
	struct bfa_aen_s	*aen;
	struct bfa_aen_s	aen_buf;
	void		*file_map[BFA_AEN_MAX_APP];
	struct bfa_plog_s      plog_buf;
	int		ref_count;
	bfa_boolean_t	ipfc_enabled;
	union bfad_tmp_buf tmp_buf;
#if (LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 10))
	struct fc_host_statistics link_stats;
#endif
	struct list_head pbc_vport_list;
	/* debugfs specific data */
	char *regdata;
	u32 reglen;
	struct dentry *bfad_dentry_files[5];
	struct list_head vport_list;
	struct bfa_trs_cfg_s	trs_cfg;
};

/* BFAD state machine events */
enum bfad_sm_event {
	BFAD_E_CREATE			= 1,
	BFAD_E_KTHREAD_CREATE_FAILED	= 2,
	BFAD_E_INIT			= 3,
	BFAD_E_INIT_SUCCESS		= 4,
	BFAD_E_HAL_INIT_FAILED		= 5,
	BFAD_E_INIT_FAILED		= 6,
	BFAD_E_FCS_EXIT_COMP		= 7,
	BFAD_E_EXIT_COMP		= 8,
	BFAD_E_STOP			= 9
};

/*
 * RPORT data structure
 */
struct bfad_rport_s {
	struct bfa_fcs_rport_s fcs_rport;
};

struct bfad_buf_info {
	void		*virt;
	dma_addr_t      phys;
	uint32_t	size;
};

struct bfad_fcxp {
	struct bfad_port_s    *port;
	struct bfa_rport_s *bfa_rport;
	bfa_status_t    req_status;
	uint16_t	tag;
	uint16_t	rsp_len;
	uint16_t	rsp_maxlen;
	uint8_t		use_ireqbuf;
	uint8_t		use_irspbuf;
	uint32_t	num_req_sgles;
	uint32_t	num_rsp_sgles;
	fchs_t		fchs;
	void		*reqbuf_info;
	void		*rspbuf_info;
	struct bfa_sge_s  *req_sge;
	struct bfa_sge_s  *rsp_sge;
	fcxp_send_cb_t  send_cbfn;
	void		*send_cbarg;
	void		*bfa_fcxp;
	struct completion comp;
};

struct bfad_hal_comp {
	bfa_status_t    status;
	struct completion comp;
};

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 14)
#ifndef kzalloc
#define kzalloc bfad_kzalloc
static inline void *
bfad_kzalloc(size_t size, int flags)
{
	void	*ptr = kmalloc(size, flags);
	if (ptr)
		memset(ptr, 0, size);
	return ptr;
}
#endif
#endif

#if (defined(__VMKLNX__) && ESX_VERS > ESX41_VERS) ||	\
	LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 18)
#define BFAD_WORK_HANDLER(name) void name(struct work_struct *work)
#define BFAD_INIT_WORK(x, work, func) INIT_WORK(&(x)->work, func)
#else
#define BFAD_WORK_HANDLER(name) void name(void *work)
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 0))
#define BFAD_INIT_WORK(x, work, func) 			\
	INIT_WORK(&(x)->work, func, &(x)->work)
#else
#define BFAD_INIT_WORK(x, work, func) 			\
	INIT_TQUEUE(&(x)->work, func, &(x)->work)
#endif
#endif

#define list_remove_head(list, entry, type, member)       	\
do {                                                    	\
	entry = NULL;                                           \
	if (!list_empty(list)) {                                \
		entry = list_entry((list)->next, type, member); 	\
		list_del_init(&entry->member);                  	\
	}								\
} while (0)

#define list_get_first(list, type, member)				\
((list_empty(list)) ? NULL :						\
	list_entry((list)->next, type, member))

#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 31)
#define bfad_get_bufptr(__iocmd, __bufsz, __iocmd_hdrsz) \
	((char *)(__iocmd) + (__iocmd_hdrsz))
#define bfad_free_bufptr(__bufptr) {}
#define bfad_copy_buf_to_user(__usrbuf, __kbuf, __bufsz) {}
#define bfad_copy_buf_from_user(__kbuf, __usrbuf, __bufsz) {}
#define bfad_chk_iocmd_sz(__payload_len, __hdrsz, __bufsz)	\
	(((__payload_len) != ((__hdrsz) + (__bufsz))) ?	\
			BFA_STATUS_FAILED : BFA_STATUS_OK)
#else
#define bfad_get_bufptr(__iocmd, __bufsz, __iocmd_hdrsz) \
	((void *)vmalloc(__bufsz))
#define bfad_free_bufptr(__bufptr)	vfree(__bufptr)
#define bfad_copy_buf_to_user(__usrbuf, __kbuf, __bufsz) do {	\
	if (copy_to_user((void __user *)(size_t)(__usrbuf),	\
			 (void *)(__kbuf), (__bufsz))) {	\
		bfad_free_bufptr(__kbuf);			\
		bfa_trc(bfad, 0x55551);				\
		return EIO;					\
	}							\
} while (0)
#define bfad_copy_buf_from_user(__kbuf, __usrbuf, __bufsz) do {	\
	if (copy_from_user((void *)(__kbuf),			\
		(void __user *)(size_t)(__usrbuf), (__bufsz))) {\
		bfad_free_bufptr(__kbuf);			\
		return EIO;					\
	}							\
} while (0)
#define bfad_chk_iocmd_sz(__payload_len, __hdrsz, __bufsz) BFA_STATUS_OK
#endif

#define bfa_cb_trs_get_cfg(_bfad) bfad_trs_get_cfg(_bfad)
bfa_trs_cfg_t * bfad_trs_get_cfg(struct bfad_s *bfad);
bfa_status_t bfa_cb_trs_enqueue_entry(struct bfad_s *bfad,
				wwn_t lpwwn, wwn_t rpwwn);
void bfa_cb_trs_dequeue_entry(struct bfad_s *bfad, wwn_t lpwwn, wwn_t rpwwn);
void bfa_cb_trs_clear_list(struct bfad_s *bfad);

bfa_boolean_t	bfad_is_ready(void);
bfa_status_t	bfad_vport_create(struct bfad_s *bfad, uint16_t vf_id,
				struct bfa_lport_cfg_s *port_cfg,
				struct device *dev);
bfa_status_t	bfad_vf_create(struct bfad_s *bfad, uint16_t vf_id,
			       struct bfa_lport_cfg_s *port_cfg);
bfa_status_t	bfad_cfg_pport(struct bfad_s *bfad, enum bfa_lport_role role);
bfa_status_t	bfad_fc4_probe(struct bfad_s *bfad);
bfa_status_t	bfad_drv_init(struct bfad_s *bfad);
bfa_status_t	bfad_start_ops(struct bfad_s *bfad);
struct bfad_s	*bfad_find_bfad_by_inst_no(int inst_no);
void		bfad_drv_start(struct bfad_s *bfad);
void		bfad_uncfg_pport(struct bfad_s *bfad);
void		bfad_stop(struct bfad_s *bfad);
void		bfad_fcs_stop(struct bfad_s *bfad);
void		bfad_remove_intr(struct bfad_s *bfad);
void		bfad_fc4_probe_undo(struct bfad_s *bfad);
void		bfad_hal_mem_release(struct bfad_s *bfad);
void		bfad_hcb_comp(void *arg, bfa_status_t status);

int		bfad_os_ioctl_init(void);
int		bfad_os_ioctl_exit(void);
int		bfad_setup_intr(struct bfad_s *bfad);
void		bfad_remove_intr(struct bfad_s *bfad);
int		bfad_os_pci_register_driver(struct pci_driver *drv);
void		bfad_os_pci_unregister_driver(struct pci_driver *drv);
void		bfad_os_device_sysfs_create(struct bfad_s *);
void		bfad_os_device_sysfs_remove(struct bfad_s *);
void		bfad_os_pci_set_mwi(struct pci_dev *pdev);
void		bfad_os_idr_init(struct idr *idr);
void		bfad_os_idr_destroy(struct idr *idr);
void 		*bfad_os_dma_alloc(struct bfad_s *bfad, struct bfa_mem_dma_s
				*meminfo_elem, dma_addr_t *phys_addr);
void 		bfad_os_dma_free(struct bfad_s *bfad, struct bfa_mem_dma_s
					*meminfo_elem);
void 		bfad_os_idr_remove(struct idr *idp, int id);
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 9))
int 		bfad_os_idr_pre_get(struct idr *idp, int gfp_mask);
#else
int 		bfad_os_idr_pre_get(struct idr *idp, gfp_t gfp_mask);
#endif
void 		bfad_os_iounmap(struct pci_dev *pdev, struct bfad_s *bfad);

void		bfad_update_hal_cfg(struct bfa_iocfc_cfg_s *bfa_cfg);
bfa_status_t	bfad_hal_mem_alloc(struct bfad_s *bfad);
void		bfad_bfa_tmo(unsigned long data);
void		bfad_init_timer(struct bfad_s *bfad);
int		bfad_pci_init(struct pci_dev *pdev, struct bfad_s *bfad);
void		bfad_pci_uninit(struct pci_dev *pdev, struct bfad_s *bfad);
void		bfad_fcs_port_cfg(struct bfad_s *bfad);
void		bfad_drv_log_level_set(struct bfad_s *bfad);
bfa_status_t	bfad_fc4_module_init(void);
void		bfad_fc4_module_exit(void);
int 		bfad_ipfc_probe(struct bfad_s *bfad);
int 		bfad_ipfc_probe_undo(struct bfad_s *bfad);
void 		bfad_ipfc_probe_post(struct bfad_s *bfad);
int 		bfad_ipfc_module_init(void);
void 		bfad_ipfc_module_exit(void);
int			bfad_ipfc_port_online(struct bfad_s *bfad,
				struct bfad_port_s *port);
int			bfad_ipfc_port_offline(struct bfad_s *bfad,
				struct bfad_port_s *port);
int			bfad_ipfc_port_new(struct bfad_s *bfad,
				struct bfad_port_s *port, int port_type);
int			bfad_ipfc_port_delete(struct bfad_s *bfad,
				struct bfad_port_s *port);

void		bfad_os_kthread_create(struct bfad_s *bfad);
void		bfad_os_kthread_stop(struct bfad_s *bfad);
void		bfad_os_kthread_wakeup(struct bfad_s *bfad);
int		bfad_os_kthread_should_stop(void);
int		bfad_worker(void *ptr);

void bfad_pci_remove(struct pci_dev *pdev);
int bfad_pci_probe(struct pci_dev *pdev, const struct pci_device_id *pid);
void bfad_os_rport_online_wait(struct bfad_s *bfad);
int bfad_os_get_linkup_delay(struct bfad_s *bfad);
uint32_t bfa_os_get_instance_id(struct bfad_s *bfad);
int bfad_install_msix_handler(struct bfad_s *bfad);
int bfad_get_num_pbc_vports(struct bfad_s *bfad);
void bfad_debugfs_init(struct bfad_port_s *port);
void bfad_debugfs_exit(struct bfad_port_s *port);
void bfad_flags_set(struct bfad_s *bfad, uint32_t flags);
void bfad_flags_clear(struct bfad_s *bfad, uint32_t flags);
int  bfad_os_pci_save_state(struct pci_dev *pdev);
void bfad_os_pci_restore_state(struct pci_dev *pdev);
#if (defined(__VMKLNX__) && (ESX_VERS > ESX50_VERS))
void bfad_os_scsi_ioqueue_register(struct bfad_port_s *bfad_drv);
void bfad_os_iodm_enable_events(struct Scsi_Host *shost);
void bfad_os_iodm_disable_events(struct Scsi_Host *shost);
void bfad_os_iodm_event(struct Scsi_Host *shost, unsigned int event,
			void *addr, unsigned long data);
#else
#define bfad_os_scsi_ioqueue_register(__bfad_port)
#define bfad_os_iodm_enable_events(__shost)
#define bfad_os_iodm_disable_events(__shost)
#define bfad_os_iodm_event(__shost, __id, __addr, __data)
#endif

/* IOCMD definitions */
int bfad_iocmd_flash_mfg_info(struct bfad_s *bfad, void *cmd,
			      unsigned int payload_len);
int bfad_iocmd_handler(struct bfad_s *bfad, unsigned int cmd,
		void *iocmd, unsigned long arg, unsigned int payload_len);
uint64_t bfad_fcxp_get_req_sgaddr_cb(void *bfad_fcxp, int sgeid);
uint32_t bfad_fcxp_get_req_sglen_cb(void *bfad_fcxp, int sgeid);
uint64_t bfad_fcxp_get_rsp_sgaddr_cb(void *bfad_fcxp, int sgeid);
uint32_t bfad_fcxp_get_rsp_sglen_cb(void *bfad_fcxp, int sgeid);
void bfad_send_fcpt_cb(void *bfad_fcxp, struct bfa_fcxp_s *fcxp, void *cbarg,
		bfa_status_t req_status, uint32_t rsp_len, uint32_t resid_len,
		fchs_t *rsp_fchs);

extern struct idr bfad_im_port_index;
extern struct pci_driver *bfad_pci_driver_p;

extern struct pci_device_id bfad_id_table[];
extern struct semaphore bfad_sem;
extern int bfad_scan_done;
extern struct list_head bfad_list;
extern char	*os_name;
extern char	*os_patch;
extern char	*host_name;
extern int	num_rports;
extern int	num_ioims;
extern int	num_tios;
extern int	num_tms;
extern int	num_fcxps;
extern int	num_ufbufs;
extern int	reqq_size;
extern int	rspq_size;
extern int	num_sgpgs;
extern int      rport_del_timeout;
extern int      bfa_lun_queue_depth;
extern int      bfa_io_max_sge;
extern int      log_level;
extern int      ioc_auto_recover;
extern int      ipfc_mtu;
extern int      linkup_delay;
extern int      msix_disable_cb;
extern int      msix_disable_ct;
extern int      fdmi_enable;
extern int      supported_fc4s;
extern int	pcie_max_read_reqsz;
extern int	max_xfer_size;
extern atomic_t tm_ops_flag;
extern int	max_rport_logins;
#if (defined(__VMKLNX__) && (ESX_VERS > ESX50_VERS))
extern int	vmklnx_multiq;
#endif

struct bfad_ipfc_ops {
	int     (*port_online)(struct bfad_s *bfad, struct bfad_port_s *port);
	int     (*port_offline)(struct bfad_s *bfad, struct bfad_port_s *port);
	int     (*probe)(struct bfad_s *bfad);
	int     (*probe_undo)(struct bfad_s *bfad);
	void    (*probe_post)(struct bfad_s *bfad);
	void    (*process_farp)(void *lport, void *buf);
	uint16_t (*pktlen)(bfad_pkt_t *dpkt);
	void    (*sge)(bfad_pkt_t *dpkt, int sge_index, bfa_sge_t *sge);
	int     (*nsgles)(bfad_pkt_t *dpkt);
	void    (*pkt_recv)(void *dipfc, bfad_pkt_t *dpkt, bfa_ipfc_status_t
		status, uint32_t pktlen, uint32_t residue);
};

void bfad_register(struct bfad_ipfc_ops *ops);
void bfad_unregister(void);

typedef struct bfad_tm_ops_s {
	bfa_status_t (*lport_new)(struct bfad_s *, struct bfad_port_s *);
	void (*lport_delete)(struct bfad_s *bfad, struct bfad_port_s *port);
	void (*lport_online)(struct bfad_s *bfad, struct bfad_port_s *port);
	void (*lport_offline)(struct bfad_s *bfad, struct bfad_port_s *port);
	bfa_status_t (*bfat_ioctl)(struct bfad_s *bfad, void *cmd);
}bfad_tm_ops_t;

void bfad_tm_register(struct bfad_tm_ops_s *bfad_tm_ops);
void bfad_tm_unregister(void);
bfa_status_t bfad_target_mode_ioctl(struct bfad_s *bfad, void *arg);

/* Interfaces needed by bfat */
void bfad_get_port_instances(struct list_head **bfad_port_list);
void bfad_acquire_lock(struct bfad_s *bfad, unsigned long *flags);
void bfad_release_lock(struct bfad_s *bfad, unsigned long flags);
void bfad_acquire_semaphore(void);
void bfad_release_semaphore(void);

#endif /* __BFAD_DRV_H__ */
