/*
 * Copyright (c) 2005-2014 Brocade Communications Systems, Inc.
 * Copyright (c) 2014 QLogic Corporation.
 * All rights reserved
 * www.qlogic.com
 *
 * See LICENSE.bfa for copyright and licensing details.
 */

/**
 * @file bfad_im_2.6.10.c Linux driver IM module.
 * code for Linux >= 2.6.10
 */

#include "bfad_drv.h"
#include "bfad_im.h"
#include "bfa_cb_ioim_macros.h"

BFA_TRC_FILE(LDRV, IM_2_6_10);

BFAD_WORK_HANDLER(bfad_im_port_delete_handler);
BFAD_WORK_HANDLER(bfad_im_itnim_work_handler);

static void bfad_im_fc_rport_add(struct bfad_im_port_s  *im_port,
					struct bfad_itnim_s *itnim);

struct bfad_itnim_s *
bfad_os_get_itnim(struct bfad_im_port_s *im_port, int id)
{
	struct bfad_itnim_s   *itnim = NULL;

	/* Search the mapped list for this target ID */
	list_for_each_entry(itnim, &im_port->itnim_mapped_list, list_entry) {
		if (id == itnim->scsi_tgt_id)
			return itnim;
	}

	return NULL;
}

void
bfad_os_init_work(struct bfad_im_port_s *im_port)
{
	BFAD_INIT_WORK(im_port, port_delete_work, bfad_im_port_delete_handler);

	return;
}

void
bfad_os_itnim_tov(struct bfad_itnim_s *itnim)
{
}

#if (LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 31))
/*
 * Function is invoked from the SCSI Host Template slave_alloc() entry point.
 * Has the logic to query the LUN Mask database to check if this LUN needs to
 * be made visible to the SCSI mid-layer or not.
 *
 * Returns BFA_STATUS_OK if this LUN needs to be added to the OS stack.
 * Returns -ENXIO to notify SCSI mid-layer to not add this LUN to the OS stack.
 */
static int
bfad_im_check_if_make_lun_visible(struct scsi_device *sdev,
				  struct fc_rport *rport)
{
	struct bfad_itnim_data_s *itnim_data =
			(struct bfad_itnim_data_s *) rport->dd_data;
	struct bfa_s *bfa = itnim_data->itnim->bfa_itnim->bfa;
	struct bfa_rport_s *bfa_rport = itnim_data->itnim->bfa_itnim->rport;
	struct bfa_lun_mask_s *lun_list = bfa_get_lun_mask_list(bfa);
	int i = 0, ret = -ENXIO;

	for (i = 0; i < MAX_LUN_MASK_CFG; i++) {
		if (lun_list[i].state == BFA_IOIM_LUN_MASK_ACTIVE &&
		    lun_list[i].lun == bfad_int_to_lun(sdev->lun) &&
		    lun_list[i].rp_tag == bfa_rport->rport_tag &&
		    lun_list[i].lp_tag == (u8)bfa_rport->rport_info.lp_tag) {
			ret = BFA_STATUS_OK;
			break;
		}
	}
	return ret;
}
#endif

/**
 * @brief
 * Scsi_Host template entry slave_alloc
 */
int
bfad_im_slave_alloc(struct scsi_device *sdev)
{
	struct fc_rport *rport = starget_to_rport(scsi_target(sdev));

	if (!rport || fc_remote_port_chkready(rport))
		return -ENXIO;
	else {
#if (LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 31))
		struct bfad_itnim_data_s * itnim_data =
				(struct bfad_itnim_data_s *) rport->dd_data;
		struct bfa_s *bfa = itnim_data->itnim->bfa_itnim->bfa;

		if (bfa_get_lun_mask_status(bfa) == BFA_LUNMASK_ENABLED) {
			/*
			 * We should not mask LUN 0 - since this will translate
			 * to no LUN / TARGET for SCSI ml resulting no scan.
		 	 */
			if (sdev->lun == 0) {
				sdev->sdev_bflags |= BLIST_NOREPORTLUN |
						     BLIST_SPARSELUN;
				goto done;
			}

			/*
			 * Query LUN Mask configuration - to expose this LUN
			 * to the SCSI mid-layer or to mask it.
			 */
			if (bfad_im_check_if_make_lun_visible(sdev, rport) !=
								BFA_STATUS_OK)
				return -ENXIO;
		}
done:
#endif
		sdev->hostdata = rport->dd_data;
	}

	return 0;
}

uint32_t
bfad_im_supported_speeds(struct bfa_s *bfa)
{
	uint32_t supported_speed = 0;
	uint8_t max_speed;
	uint32_t card_type;

	max_speed = bfa_get_max_speed(bfa);
	card_type = bfa_get_card_type(bfa);
	if (max_speed == BFA_PORT_SPEED_16GBPS) {
		supported_speed |=  FC_PORTSPEED_16GBIT | FC_PORTSPEED_8GBIT |
				FC_PORTSPEED_4GBIT | FC_PORTSPEED_2GBIT;
	} else if (max_speed == BFA_PORT_SPEED_8GBPS) {
		if (bfa_mfg_is_mezz(card_type)) {
			supported_speed |= FC_PORTSPEED_8GBIT | FC_PORTSPEED_4GBIT |
				FC_PORTSPEED_2GBIT | FC_PORTSPEED_1GBIT;
		} else {
			supported_speed |= FC_PORTSPEED_8GBIT | FC_PORTSPEED_4GBIT |
				FC_PORTSPEED_2GBIT;
		}
	} else if (max_speed == BFA_PORT_SPEED_4GBPS) {
		supported_speed |=  FC_PORTSPEED_4GBIT | FC_PORTSPEED_2GBIT |
				FC_PORTSPEED_1GBIT;
	} else if (max_speed == BFA_PORT_SPEED_10GBPS) {
		supported_speed |= FC_PORTSPEED_10GBIT;
	}
	return	supported_speed; 
}

void
bfad_os_fc_host_init(struct bfad_im_port_s *im_port)
{
	struct Scsi_Host *host = im_port->shost;
	struct bfad_s         *bfad = im_port->bfad;
	struct bfad_port_s    *port = im_port->port;

	fc_host_node_name(host) =
		bfa_os_htonll((bfa_fcs_lport_get_nwwn(port->fcs_port)));
	fc_host_port_name(host) =
		bfa_os_htonll((bfa_fcs_lport_get_pwwn(port->fcs_port)));

	bfad_os_set_npiv(im_port);
#if (LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 31))
	fc_host_max_npiv_vports(host) = bfa_lps_get_max_vport(&bfad->bfa);
#endif
	fc_host_supported_classes(host) = FC_COS_CLASS3;

	memset(fc_host_supported_fc4s(host), 0,
	       sizeof(fc_host_supported_fc4s(host)));
	if (supported_fc4s & (BFA_LPORT_ROLE_FCP_IM | BFA_LPORT_ROLE_FCP_TM))
		/* For FCP type 0x08 */
		fc_host_supported_fc4s(host)[2] = 1;
	if (supported_fc4s | BFA_LPORT_ROLE_FCP_IPFC)
		/* For LLC/SNAP type 0x05 */
		fc_host_supported_fc4s(host)[3] = 0x20;
	/* For fibre channel services type 0x20 */
	fc_host_supported_fc4s(host)[7] = 1;

	sprintf(fc_host_symbolic_name(host), "%s",
			bfa_fcs_lport_get_symname(&bfad->bfa_fcs.fabric.bport));

	fc_host_supported_speeds(host) = bfad_im_supported_speeds(&bfad->bfa);
	fc_host_maxframe_size(host) = bfa_fcport_get_maxfrsize(&bfad->bfa);
}

static void
bfad_im_fc_rport_add(struct bfad_im_port_s *im_port, struct bfad_itnim_s *itnim)
{
	struct fc_rport_identifiers rport_ids;
	struct fc_rport *fc_rport;
	struct bfad_itnim_data_s *itnim_data;

	rport_ids.node_name =
		bfa_os_htonll(bfa_fcs_itnim_get_nwwn(&itnim->fcs_itnim));
	rport_ids.port_name =
		bfa_os_htonll(bfa_fcs_itnim_get_pwwn(&itnim->fcs_itnim));
	rport_ids.port_id =
		bfa_os_hton3b(bfa_fcs_itnim_get_fcid(&itnim->fcs_itnim));
	rport_ids.roles = FC_RPORT_ROLE_UNKNOWN;

	itnim->fc_rport = fc_rport =
		fc_remote_port_add(im_port->shost, 0, &rport_ids);

	if (!fc_rport)
		return;

	fc_rport->maxframe_size =
		bfa_fcs_itnim_get_maxfrsize(&itnim->fcs_itnim);
	fc_rport->supported_classes = bfa_fcs_itnim_get_cos(&itnim->fcs_itnim);

	itnim_data = fc_rport->dd_data;
	itnim_data->itnim = itnim;

	rport_ids.roles |= FC_RPORT_ROLE_FCP_TARGET;

	if (rport_ids.roles != FC_RPORT_ROLE_UNKNOWN)
		fc_remote_port_rolechg(fc_rport, rport_ids.roles);

	if ((fc_rport->scsi_target_id != -1)
	    && (fc_rport->scsi_target_id < MAX_FCP_TARGET))
		itnim->scsi_tgt_id = fc_rport->scsi_target_id;

	itnim->channel = fc_rport->channel;

	return;
}

/**
 * @brief
 * Work queue handler using FC transport service
* Context: kernel
 */
BFAD_WORK_HANDLER(bfad_im_itnim_work_handler)
{	/* work struct is the function arg */
	struct bfad_itnim_s   *itnim = container_of(work, struct bfad_itnim_s,
							itnim_work);
	struct bfad_im_s      *im = itnim->im;
	struct bfad_s         *bfad = im->bfad;
	struct bfad_im_port_s *im_port;
	unsigned long   flags;
	struct fc_rport *fc_rport;
	wwn_t wwpn;
	uint32_t fcid;
	char wwpn_str[32], fcid_str[16];

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	im_port = itnim->im_port;
	bfa_trc(bfad, itnim->state);
	switch (itnim->state) {
	case ITNIM_STATE_ONLINE:
		if (!itnim->fc_rport) {
			spin_unlock_irqrestore(&bfad->bfad_lock, flags);
			bfad_im_fc_rport_add(im_port, itnim);
			spin_lock_irqsave(&bfad->bfad_lock, flags);
			wwpn = bfa_fcs_itnim_get_pwwn(&itnim->fcs_itnim);
			fcid = bfa_fcs_itnim_get_fcid(&itnim->fcs_itnim);
			bfad_wwn_to_str(wwpn, wwpn_str);
			bfad_fcid_to_str(fcid, fcid_str);
			list_add_tail(&itnim->list_entry,
				&im_port->itnim_mapped_list);
			bfa_log(bfad->logm, BFA_LOG_LINUX_ITNIM_ONLINE,
				im_port->shost->host_no,
				itnim->scsi_tgt_id,
				fcid_str, wwpn_str);
		} else {
			printk(KERN_WARNING
				"%s: itnim %llx is already in online state\n",
				__FUNCTION__,
				bfa_fcs_itnim_get_pwwn(&itnim->fcs_itnim));
		}

		break;
	case ITNIM_STATE_OFFLINE_PENDING:
		itnim->state = ITNIM_STATE_OFFLINE;
		if (itnim->fc_rport) {
			fc_rport = itnim->fc_rport;
			((struct bfad_itnim_data_s *)
				fc_rport->dd_data)->itnim = NULL;
			itnim->fc_rport = NULL;
			if (!(im_port->port->flags & BFAD_PORT_DELETE)) {
				spin_unlock_irqrestore(&bfad->bfad_lock, flags);
				fc_rport->dev_loss_tmo =
					bfa_fcpim_path_tov_get(&bfad->bfa) + 1;
				fc_remote_port_delete(fc_rport);
				spin_lock_irqsave(&bfad->bfad_lock, flags);
			}
			wwpn = bfa_fcs_itnim_get_pwwn(&itnim->fcs_itnim);
			fcid = bfa_fcs_itnim_get_fcid(&itnim->fcs_itnim);
			bfad_wwn_to_str(wwpn, wwpn_str);
			bfad_fcid_to_str(fcid, fcid_str);
			list_del(&itnim->list_entry);
			bfa_log(bfad->logm, BFA_LOG_LINUX_ITNIM_OFFLINE,
				im_port->shost->host_no,
				itnim->scsi_tgt_id,
				fcid_str, wwpn_str);
		}
		break;
	case ITNIM_STATE_FREE:
		if (itnim->fc_rport) {
			fc_rport = itnim->fc_rport;
			((struct bfad_itnim_data_s *)
				fc_rport->dd_data)->itnim = NULL;
			itnim->fc_rport = NULL;
			if (!(im_port->port->flags & BFAD_PORT_DELETE)) {
				spin_unlock_irqrestore(&bfad->bfad_lock, flags);
				fc_rport->dev_loss_tmo =
					bfa_fcpim_path_tov_get(&bfad->bfa) + 1;
				fc_remote_port_delete(fc_rport);
				spin_lock_irqsave(&bfad->bfad_lock, flags);
			}
			list_del(&itnim->list_entry);
		}

		kfree(itnim);
		break;
	default:
		bfa_assert(0);
		break;
	}

	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
}

#ifdef DEF_SCSI_QCMD
int bfad_im_queuecommand_lck(struct scsi_cmnd *cmnd,
		void (*done) (struct scsi_cmnd *));
DEF_SCSI_QCMD(bfad_im_queuecommand)
#else
#define bfad_im_queuecommand_lck bfad_im_queuecommand
#endif

/**
 * @brief
 * Scsi_Host template entry, queue a SCSI command to the BFAD.
 */
int
bfad_im_queuecommand_lck(struct scsi_cmnd *cmnd,
				void (*done) (struct scsi_cmnd *))
{
	struct bfad_im_port_s *im_port =
		(struct bfad_im_port_s *) cmnd->device->host->hostdata[0];
	struct bfad_s         *bfad = im_port->bfad;
	struct bfad_itnim_data_s *itnim_data = cmnd->device->hostdata;
	struct bfad_itnim_s   *itnim;
	struct bfa_ioim_s *hal_io;
	unsigned long   flags;
	int             rc;
	int       sg_cnt = 0;
	struct fc_rport *rport = starget_to_rport(scsi_target(cmnd->device));

	rc = fc_remote_port_chkready(rport);
	if (rc) {
		cmnd->result = rc;
		done(cmnd);
		return 0;
	}

	if (bfad->bfad_flags & BFAD_EEH_BUSY) {
		if (bfad->bfad_flags & BFAD_EEH_PCI_CHANNEL_IO_PERM_FAILURE)
			cmnd->result = DID_NO_CONNECT << 16;
		else
			cmnd->result = DID_REQUEUE << 16;
		done(cmnd);
		return 0;
	}

	rc = bfad_os_im_dma_map(bfad, cmnd, &sg_cnt);
	if (rc)
		return SCSI_MLQUEUE_HOST_BUSY;

	cmnd->scsi_done = done;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	if (!(bfad->bfad_flags & BFAD_HAL_START_DONE)) {
		printk(KERN_WARNING
			"bfad%d, queuecommand %p %x failed, BFA stopped\n",
		       bfad->inst_no, cmnd, cmnd->cmnd[0]);
		cmnd->result = ScsiResult(DID_NO_CONNECT, 0);
		goto out_fail_cmd;
	}


	itnim = itnim_data->itnim;
	if (!itnim) {
		cmnd->result = ScsiResult(DID_IMM_RETRY, 0);
		goto out_fail_cmd;
	}

	hal_io = bfa_ioim_alloc(&bfad->bfa, (struct bfad_ioim_s *) cmnd,
				    itnim->bfa_itnim, sg_cnt);
	if (!hal_io) {
		bfa_trc(bfad, 0);
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		bfad_os_im_dma_unmap(bfad, cmnd);
		return SCSI_MLQUEUE_HOST_BUSY;
	}

	cmnd->host_scribble = (char *)hal_io;
	bfa_trc_fp(bfad, hal_io->iotag);
	bfa_ioim_start(hal_io, BFA_IOIM_WITH_SG);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	bfad_im_io_polling_wait(bfad, cmnd);

	return 0;

out_fail_cmd:
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	bfad_os_im_dma_unmap(bfad, cmnd);
	if (done)
		done(cmnd);

	return 0;
}
