/*
 * Copyright (c) 2005-2014 Brocade Communications Systems, Inc.
 * Copyright (c) 2014 QLogic Corporation.
 * All rights reserved
 * www.qlogic.com
 *
 * See LICENSE.bfa for copyright and licensing details.
 */

/**
 * @file bfad_im_lt_2.6.10.c Linux driver IM module.
 * code for Linux ver <= 2.6.9
 */

#include "bfad_drv.h"
#include "bfad_im.h"

BFA_TRC_FILE(LDRV, IM_LT_2_6_10);

struct bfad_itnim_s *
bfad_os_get_itnim(struct bfad_im_port_s  *im_port, int id)
{
        struct bfad_itnim_data_s *itnim_data;

	bfa_assert (id < MAX_FCP_TARGET);

        itnim_data = &im_port->itnim_data_table[id];

	if (itnim_data)
        	return itnim_data->itnim;
	else
		return NULL;
}

void
bfad_os_itnim_tov(struct bfad_itnim_s *itnim)
{
	bfad_im_itnim_unmap(itnim->im_port, itnim);
}

struct bfad_fcp_binding *
bfad_im_find_binding(struct bfad_im_port_s  *im_port, struct bfad_itnim_s *itnim)
{
	struct bfad_fcp_binding *bp = NULL;

	list_for_each_entry(bp, &im_port->binding_list, list_entry) {
		switch (bp->binding_type) {
		case FCP_PWWN_BINDING:
			if (bp->pwwn ==
			    bfa_fcs_itnim_get_pwwn(&itnim->fcs_itnim))
				return bp;
			break;
		case FCP_NWWN_BINDING:
			if (bp->nwwn ==
			    bfa_fcs_itnim_get_nwwn(&itnim->fcs_itnim))
				return bp;
			break;
		case FCP_FCID_BINDING:
			if (bp->fc_id ==
			    bfa_fcs_itnim_get_fcid(&itnim->fcs_itnim))
				return bp;
			break;
		default:
			break;
		}
	}

	return NULL;
}

struct bfad_fcp_binding *
bfad_im_create_binding(struct bfad_im_port_s  *im_port, struct bfad_itnim_s *itnim)
{
	struct bfad_fcp_binding *bp = NULL;
	struct bfad_s         *bfad = im_port->bfad;

	bp = kzalloc(sizeof(struct bfad_fcp_binding), GFP_ATOMIC);

	if (!bp) {
		/*
		 * Log info and return
		 */
		printk(KERN_WARNING "%s: kzalloc failure\n", __FUNCTION__);
		return bp;
	}

	bp->binding_type = bfad->cfg_data.binding_method;
	bp->scsi_target_id = itnim->scsi_tgt_id;
	bp->fc_id = bfa_fcs_itnim_get_fcid(&itnim->fcs_itnim);
	bp->pwwn = bfa_fcs_itnim_get_pwwn(&itnim->fcs_itnim);
	bp->nwwn = bfa_fcs_itnim_get_nwwn(&itnim->fcs_itnim);

	list_add_tail(&bp->list_entry, &im_port->binding_list);

	return bp;
}

int
bfad_im_check_binding(struct bfad_im_port_s  *im_port)
{
	struct bfad_fcp_binding *bp;

	list_for_each_entry(bp, &im_port->binding_list, list_entry) {
		if (im_port->cur_scsi_id == bp->scsi_target_id)
			return 1;
	}

	return 0;
}

void
bfad_im_itnim_unmap(struct bfad_im_port_s  *im_port, struct bfad_itnim_s *itnim)
{
	struct bfad_itnim_data_s *itnim_data;

	if (itnim->flags & ITNIM_MAPPED) {
		itnim->flags &= ~ITNIM_MAPPED;
		itnim_data = &im_port->itnim_data_table[itnim->scsi_tgt_id];
		if (itnim_data->itnim == itnim)
			itnim_data->itnim = NULL;
	}
}

bfa_status_t
bfad_im_itnim_map(struct bfad_im_port_s  *im_port, struct bfad_itnim_s *itnim)
{
	bfa_status_t    rc = BFA_STATUS_OK;
	struct bfad_fcp_binding *bp;
	struct bfad_itnim_data_s *itnim_data;

	bp = bfad_im_find_binding(im_port, itnim);
	if (bp) {
		bp->fc_id = bfa_fcs_itnim_get_fcid(&itnim->fcs_itnim);
		itnim->scsi_tgt_id = bp->scsi_target_id;
		itnim->flags |= ITNIM_MAPPED;
	} else {
		while (1) {
			if (im_port->cur_scsi_id >= MAX_FCP_TARGET) {
				rc = BFA_STATUS_FAILED;
				goto ext;
			}
			/*
			 * First check if the cur_scsi_id is already in
			 * binding list
			 */
			if (bfad_im_check_binding(im_port))
				im_port->cur_scsi_id++;
			else {
				itnim->scsi_tgt_id = im_port->cur_scsi_id++;

				if (bfad_im_create_binding(im_port, itnim)
				    != NULL) {
					itnim->flags |= ITNIM_MAPPED;
					break;
				} else {
					/*
					 * Failed to create binding, leave it
					 * in the add list, and try later
					 */
					im_port->cur_scsi_id--;
					rc = BFA_STATUS_FAILED;
					goto ext;
				}

				break;
			}
		}
	}

	itnim_data = &im_port->itnim_data_table[itnim->scsi_tgt_id];
	itnim_data->itnim = itnim;
ext:
	return rc;
}

void
bfad_os_delay_workq(struct bfad_im_s *im, struct bfad_im_port_s  *im_port);

/**
 * @brief
 * itnim work queue handler, NO FC transport support
* Context: kernel
 */
BFAD_WORK_HANDLER(bfad_im_itnim_work_handler)
{	/* work is the function arg */
	struct bfad_itnim_s   *itnim = container_of(work, struct bfad_itnim_s, itnim_work);
	struct bfad_im_s      *im = itnim->im;
	struct bfad_s         *bfad = im->bfad;
	struct bfad_im_port_s  *im_port;
	unsigned long   flags;
	wwn_t wwpn;
	uint32_t fcid;
	char wwpn_str[32], fcid_str[16];
	

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	im_port = itnim->im_port;
	switch (itnim->state) {
	case ITNIM_STATE_ONLINE: {
		int             rc;
		rc = bfad_im_itnim_map(im_port, itnim);
		if (itnim->flags & ITNIM_MAPPED) {
			wwpn = bfa_fcs_itnim_get_pwwn(&itnim->fcs_itnim);
			fcid = bfa_fcs_itnim_get_fcid(&itnim->fcs_itnim);
			bfad_wwn_to_str(wwpn, wwpn_str);
			bfad_fcid_to_str(fcid, fcid_str);
			bfa_log(bfad->logm, BFA_LOG_LINUX_ITNIM_ONLINE,
					im_port->shost->host_no,
					itnim->scsi_tgt_id,
					fcid_str, wwpn_str);
		}
		break;
	}
	case ITNIM_STATE_OFFLINE_PENDING:
		itnim->state = ITNIM_STATE_OFFLINE;
		if (itnim->flags & ITNIM_MAPPED) {
			wwpn = bfa_fcs_itnim_get_pwwn(&itnim->fcs_itnim);
			fcid = bfa_fcs_itnim_get_fcid(&itnim->fcs_itnim);
			bfad_wwn_to_str(wwpn, wwpn_str);
			bfad_fcid_to_str(fcid, fcid_str);
			bfa_log(bfad->logm, BFA_LOG_LINUX_ITNIM_OFFLINE,
					im_port->shost->host_no,
					itnim->scsi_tgt_id,
					fcid_str, wwpn_str);
		}

		break;
	case ITNIM_STATE_FREE:
		bfad_im_itnim_unmap(im_port, itnim);
		kfree(itnim);
		break;
	default:
		bfa_assert(0);
		break;
	}
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
}

/**
 * @brief
 * Scsi_Host template entry, queue a SCSI command to the BFAD.
 */
int
bfad_im_queuecommand(struct scsi_cmnd *cmnd, void (*done) (struct scsi_cmnd *))
{
	struct Scsi_Host *shost = cmnd->device->host;
	struct bfad_im_port_s  *im_port = (struct bfad_im_port_s  *) shost->hostdata[0];
	struct bfad_s         *bfad = im_port->bfad;
	struct bfad_itnim_data_s *itnim_data;
	struct bfad_itnim_s   *itnim;
	struct bfa_ioim_s *hal_io = NULL;
	unsigned long   flags;
	int             rc;
	int        sg_cnt = 0;

	/* Map DMA */
	rc = bfad_os_im_dma_map(bfad, cmnd, &sg_cnt);
	if (rc)
		return SCSI_MLQUEUE_HOST_BUSY;

	cmnd->scsi_done = done;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	if (!(bfad->bfad_flags & BFAD_HAL_START_DONE)) {
		printk(KERN_WARNING
			"bfad%d, queuecommand %p %x failed, BFA stopped\n",
			bfad->inst_no, cmnd, cmnd->cmnd[0]);
		cmnd->result = ScsiResult(DID_NO_CONNECT, 0);
		goto out_fail_cmd;
	}

	itnim_data = bfad_os_itnim_data(im_port, cmnd);
	if (!itnim_data) {
		cmnd->result = ScsiResult(DID_NO_CONNECT, 0);
		goto out_fail_cmd;
	}

	itnim = itnim_data->itnim;
	if (!itnim) {
		cmnd->result = ScsiResult(DID_NO_CONNECT, 0);
		goto out_fail_cmd;
	}

	if (!bfad_os_im_itnim_is_online(itnim)) {
                spin_unlock_irqrestore(&bfad->bfad_lock, flags);
                bfad_os_im_dma_unmap(bfad, cmnd);
                return SCSI_MLQUEUE_HOST_BUSY;
	}

	hal_io = bfa_ioim_alloc(&bfad->bfa, (struct bfad_ioim_s *) cmnd,
				    itnim->bfa_itnim, sg_cnt);
	if (!hal_io) {
		bfa_trc(bfad, 0);
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		bfad_os_im_dma_unmap(bfad, cmnd);
		return SCSI_MLQUEUE_HOST_BUSY;
	}

	cmnd->host_scribble = (char *)hal_io;
	bfa_trc_fp(bfad, hal_io->iotag);
	bfa_ioim_start(hal_io, BFA_IOIM_WITH_SG);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	return 0;

out_fail_cmd:
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	bfad_os_im_dma_unmap(bfad, cmnd);
	if (done)
		done(cmnd);

	return 0;
}

void 
bfad_os_rport_online_wait(struct bfad_s *bfad)
{
	int i;
	int rport_delay = 10;

	for (i = 0; !(bfad->bfad_flags & BFAD_PORT_ONLINE)
		&& i < linkup_delay; i++) {
		set_current_state(TASK_UNINTERRUPTIBLE);
		schedule_timeout(HZ);
	}

	if (bfad->bfad_flags & BFAD_PORT_ONLINE) {
		rport_delay = rport_delay < linkup_delay ?
				 rport_delay : linkup_delay;
		for (i = 0; !(bfad->bfad_flags & BFAD_RPORT_ONLINE)
			 && i < rport_delay; i++) {
			set_current_state(TASK_UNINTERRUPTIBLE);
			schedule_timeout(HZ);
		}

		if (rport_delay > 0 && (bfad->bfad_flags & BFAD_RPORT_ONLINE)) {
			set_current_state(TASK_UNINTERRUPTIBLE);
			schedule_timeout(rport_delay * HZ);
		}
	}
}

int
bfad_os_get_linkup_delay(struct bfad_s *bfad)
{

	uint8_t         nwwns = 0;
	wwn_t           wwns[BFA_PREBOOT_BOOTLUN_MAX];
	int             linkup_delay;

	/*
	 * Querying for the boot target port wwns
	 * -- read from boot information in flash.
	 * If nwwns > 0 => boot over SAN and set linkup_delay = 30
	 * else => local boot machine set linkup_delay = 10
	 */

	bfa_iocfc_get_bootwwns(&bfad->bfa, &nwwns, wwns);

	if (nwwns > 0) {
		/* If boot over SAN; linkup_delay = 30sec */
		linkup_delay = 30;
	} else {
		/* If local boot; linkup_delay = 10sec */
		linkup_delay = 10;
	}

	return linkup_delay;
}


