/*
 * Copyright (c) 2005-2014 Brocade Communications Systems, Inc.
 * Copyright (c) 2014 QLogic Corporation.
 * All rights reserved
 * www.qlogic.com
 *
 * See LICENSE.bfa for copyright and licensing details.
 */

#include <asm/uaccess.h>
#include "bfad_ioctl_cna.h"
#include "bfad_ioctl.h"
#include "bfad_drv.h"
#include "bfad_aen.h"
#include "bfad_im.h"
#include "cna/bfa_diag.h"
#include "fcs/bfa_fcs.h"
#include <linux/module.h>

BFA_TRC_FILE(LDRV, IOCTL);


struct bfad_fcxp    *
bfad_fcxp_alloc(struct bfad_s *bfad)
{
	void           *ptr;

	ptr = kmalloc(sizeof(struct bfad_fcxp), GFP_KERNEL);
	if (ptr)
		memset(ptr, 0, sizeof(struct bfad_fcxp));
	return ptr;
}

void
bfad_fcxp_free(struct bfad_fcxp *fcxp)
{
	kfree(fcxp);
}

struct bfad_scsi_passthru {
	struct completion comp;	/* cv for completion */
	bfa_ioctl_scsi_t iocmd;
};

void
bfad_send_scsi_cb(void *arg)
{
	struct scsi_cmnd *scsi_cmd = arg;
	struct bfad_scsi_passthru *pthru =
		(struct bfad_scsi_passthru *)scsi_cmd->SCp.ptr;

	/*
	 * Wake up the waiting thread
	 */
	complete(&pthru->comp);
}

void
bfad_scsi_sg_data_free(struct bfad_s *bfad, struct scatterlist *sg_base, uint32_t num_sgles)
{
	int i;
	struct scatterlist  *sg_info = sg_base;
	void *cpu_addr;


	for (i = 0; i < num_sgles; sg_info++, i++) {
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 26))
		cpu_addr = (void *)sg_info->page_link;
#else
#if (defined(__VMKLNX__) && (ESX_VERS > ESX41_VERS))
		cpu_addr = (void *)sg_virt(sg_info);
#else
		cpu_addr = (void *)sg_info->page;
#endif
#endif
		if (cpu_addr != NULL) {
#if defined(__VMKLNX__) || (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
			dma_free_coherent(&bfad->pcidev->dev, sg_dma_len(sg_info),
					  cpu_addr, sg_dma_address(sg_info));
#else
			pci_free_consistent(bfad->pcidev, sg_dma_len(sg_info),
					    cpu_addr, sg_dma_address(sg_info));
#endif
		}
	}

	kfree(sg_base);
}

int
bfad_scsi_data_setup(struct bfad_s *bfad, struct scsi_cmnd *scsi_cmd, char *data_buf,
		     int write, int size)
{
	int sge_num = 0, cnt = 0, offset = 0;
	struct scatterlist *sg_info;
	int	rc = 0;
	void *request_buffer;
	uint16_t sg_count;
	unsigned request_bufferlen;
	void *cpu_addr;

	if (size == 0) {
		bfa_trc(bfad, 0);
		return 0;
	}

	/*
	 * We try allocating "size" blocks of contiguous memory.
	 * If it fails then we allocate the memory using scatterlist.
	 */

	request_buffer =
#if defined(__VMKLNX__)
			dma_alloc_coherent(&bfad->pcidev->dev, size,
					   (void *)&(scsi_cmd->request_bufferMA),
					   GFP_KERNEL);
#elif (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
			dma_alloc_coherent(&bfad->pcidev->dev, size,
					   &(scsi_cmd->SCp.dma_handle),
					   GFP_KERNEL);
#else
			pci_alloc_consistent(bfad->pcidev, size,
					     &(scsi_cmd->SCp.dma_handle));
#endif

	if (!request_buffer) {
		bfa_trc(bfad, 0);
		goto scatterlist_mem_alloc;
	}

	request_bufferlen = size;
	sg_count = 0;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 26))
	scsi_cmd->sdb.table.sgl = (struct scatterlist *)request_buffer;
	scsi_cmd->sdb.length = request_bufferlen;
	scsi_cmd->sdb.table.nents = sg_count;
#else
	scsi_cmd->request_buffer = (void *)request_buffer;
	scsi_cmd->request_bufflen = request_bufferlen;
	scsi_cmd->use_sg = sg_count;
#endif

	if (write) {
		if (copy_from_user
			((uint8_t *) request_buffer,
			 (void __user *)data_buf, size)) {

#if defined(__VMKLNX__)
			dma_free_coherent(&bfad->pcidev->dev, size,
					  request_buffer,
					  scsi_cmd->request_bufferMA);
#elif (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
			dma_free_coherent(&bfad->pcidev->dev, size,
					  request_buffer,
					  scsi_cmd->SCp.dma_handle);
#else
			pci_free_consistent(bfad->pcidev, size,
					  request_buffer,
					  scsi_cmd->SCp.dma_handle);
#endif
			bfa_trc(bfad, 0);
			return EIO;
		}
	}
	goto done;

scatterlist_mem_alloc:

	/*
	 * Allocating memory using scatterlist
	 */

	sge_num = ((size % PAGE_SIZE) ? (size / PAGE_SIZE +
			1) : (size / PAGE_SIZE));
	request_buffer =
			kzalloc(sizeof(struct scatterlist[sge_num]),
					GFP_KERNEL);
	if (!request_buffer) {
		bfa_trc(bfad, 0);
		return ENOMEM;
	}

	request_bufferlen = size;
	sg_count = sge_num;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 26))
	scsi_cmd->sdb.table.sgl = (struct scatterlist *)request_buffer;
	sg_info = (struct scatterlist *)request_buffer;
	scsi_cmd->sdb.length = request_bufferlen;
	scsi_cmd->sdb.table.nents = sg_count;
#else
	scsi_cmd->request_buffer = (void *)request_buffer;
	sg_info = scsi_cmd->request_buffer;
	scsi_cmd->request_bufflen = request_bufferlen;
	scsi_cmd->use_sg = sg_count;
#endif

	while (size > 0) {
		cnt = (size > PAGE_SIZE) ? PAGE_SIZE : size;

#if defined(__VMKLNX__) || (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
#if (defined(__VMKLNX__) && (ESX_VERS > ESX41_VERS))
		cpu_addr =
			dma_alloc_coherent(&bfad->pcidev->dev, cnt,
					   (void*)sg_dma_address(sg_info),
					   GFP_KERNEL);
#else
		cpu_addr =
			dma_alloc_coherent(&bfad->pcidev->dev, cnt,
					   &sg_dma_address(sg_info),
					   GFP_KERNEL);
#endif
#else
		cpu_addr =
			pci_alloc_consistent(bfad->pcidev, cnt,
					     &sg_dma_address(sg_info));
#endif
		if (!cpu_addr) {
			bfa_trc(bfad, 0);
			rc = ENOMEM;
			goto out_free_mem;
		}

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 26))
		sg_info->page_link = (unsigned long) cpu_addr;
#else
#if (defined(__VMKLNX__) && (ESX_VERS > ESX41_VERS))
		sg_info->cursgel->addr      = (unsigned long)cpu_addr;
#else
		sg_info->page      = (struct page*) cpu_addr;
#endif
#endif

		if (write) {
			if (copy_from_user
				((uint8_t *) cpu_addr,
				 (void __user *)(data_buf + offset), cnt)) {
				bfa_trc(bfad, 0);
				rc = EIO;
				goto out_free_mem;
			}
		}

#if (defined(__VMKLNX__) && (ESX_VERS > ESX41_VERS))
		sg_info->cursgel->length = cnt;
#else
		sg_dma_len(sg_info) = cnt;
#endif
		size -= cnt;
		offset += cnt;
		sg_info++;
	}

done:
	return 0;

out_free_mem:
	bfad_scsi_sg_data_free(bfad, request_buffer, sge_num);
	return rc;
}

int
bfad_scsi_data_copyout(struct bfad_s *bfad, char *data_buf,
		       struct scsi_cmnd *scsi_cmd, uint32_t size)
{
	int             cnt, offset = 0;
	struct scatterlist *sg_info;
	void *request_buffer;
	uint16_t sg_count;
	void *cpu_addr;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 26))
	request_buffer = (void *)scsi_sglist(scsi_cmd);
	sg_count = scsi_sg_count(scsi_cmd);
#else
	request_buffer = (void *)scsi_cmd->request_buffer;
	sg_count = scsi_cmd->use_sg;
#endif

	if (sg_count == 0 && data_buf != NULL) {

		/*
		 * Copy data to user space
		 */
		if (copy_to_user
			((void __user *)(data_buf),
			 (uint8_t *) request_buffer, size)) {

			bfa_trc(bfad, 0);
			return EIO;
		}

	} else if (sg_count > 0 && data_buf != NULL) {

		sg_info = (struct scatterlist *)request_buffer;

		while (size) {

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 26))
			cpu_addr = (void *)sg_info->page_link;
#else
#if (defined(__VMKLNX__) && (ESX_VERS > ESX41_VERS))
			cpu_addr = (void *)sg_virt(sg_info);
#else
			cpu_addr = (void *)sg_info->page;
#endif
#endif

			cnt = (size > PAGE_SIZE) ? PAGE_SIZE : size;

			/*
			 * pci_dma_sync_single_for_device(bfad->pcidev,
			 * sg_info->dma_address, cnt, PCI_DMA_FROMDEVICE);
			 */

			/*
			 * Copy data to user space
			 */
			if (copy_to_user
				((void __user *)(data_buf + offset),
				 (uint8_t *) cpu_addr, cnt)) {
				bfa_trc(bfad, 0);
				return EIO;
			}

			sg_info++;
			offset += cnt;
			size -= cnt;
		}
	}

	return 0;
}

unsigned int 
bfad_scsilun_to_int(uint8_t *scsilun)
{
        int i;
        unsigned int lun;

        lun = 0;
        for (i = 0; i < sizeof(lun); i += 2)
                lun = lun | (((scsilun[i] << 8) | scsilun[i + 1]) << (i * 8));
        return lun;
}

void
bfad_scsi_cmd_setup(struct bfad_s *bfad, struct scsi_cmnd *scsi_cmd,
		    bfa_ioctl_scsi_t *ci, struct bfad_scsi_passthru *pthru)
{
	/*
	 * Setup data direction
	 */
	if (ci->write && ci->iosz > 0)
		scsi_cmd->sc_data_direction = DMA_TO_DEVICE;
	else if (ci->iosz > 0)
		scsi_cmd->sc_data_direction = DMA_FROM_DEVICE;
	else
		scsi_cmd->sc_data_direction = DMA_NONE;

	memcpy(scsi_cmd->cmnd, ci->cdb.scsi_cdb, sizeof(ci->cdb));
	scsi_cmd->cmd_len = sizeof(ci->cdb);
	scsi_cmd->device->tagged_supported = 0;
	scsi_cmd->device->lun = bfad_scsilun_to_int((uint8_t *)&ci->lun);

	scsi_cmd->scsi_done = (void (*)(struct scsi_cmnd *))bfad_send_scsi_cb;
	scsi_cmd->SCp.ptr = (char *)pthru;

}

static int
bfad_send_scsi_passthru(unsigned long arg)
{
	struct scsi_cmnd *scsi_cmd;
	struct scsi_device *scsi_device;
	struct bfad_scsi_passthru *pthru;
	bfa_ioctl_scsi_t *ci;
	int             rc = 0, total_size;
	uint16_t        sg_cnt = 0;
	unsigned long   flags;
	struct bfa_ioim_s *hal_io = NULL;
	struct bfa_fcs_lport_s *fcs_port;
	struct bfa_fcs_itnim_s *fcs_itnim;
	struct bfa_itnim_s *bfa_itnim;
	struct bfad_s         *bfad;
	uint16_t host_status, scsi_status;
	void *request_buffer;
	uint16_t sg_count;
	unsigned request_bufferlen;

	total_size =
		sizeof(struct scsi_cmnd) + sizeof(struct scsi_device) +
		sizeof(struct bfad_scsi_passthru);
	scsi_cmd = kzalloc(total_size, GFP_KERNEL);
	if (scsi_cmd == NULL)
		return ENOMEM;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 26))

	/* scsi_cmnd buffer param's */
	#define MAX_COMMAND_SIZE	16
	#define SCSI_SENSE_BUFFERSIZE	96

	scsi_cmd->cmnd = kzalloc(MAX_COMMAND_SIZE,GFP_KERNEL);
	if (scsi_cmd->cmnd == NULL) {
		kfree(scsi_cmd);
		return ENOMEM;
	}

	scsi_cmd->sense_buffer = kzalloc(SCSI_SENSE_BUFFERSIZE,GFP_KERNEL);
	if (scsi_cmd->sense_buffer == NULL) {
		kfree(scsi_cmd->cmnd);
		kfree(scsi_cmd);
		return ENOMEM;
	}

#endif

	scsi_device = (struct scsi_device *)(scsi_cmd + 1);
	pthru = (struct bfad_scsi_passthru *)(scsi_device + 1);
	ci = &pthru->iocmd;
	scsi_cmd->device = scsi_device;

	if (copy_from_user
	    ((uint8_t *) ci, (void __user *)arg, sizeof(bfa_ioctl_scsi_t))) {
		rc = EIO;
		goto out_free_scsi_cmd;
	}

	bfad = bfad_find_bfad_by_inst_no(ci->bfad_num);
	if (bfad == NULL) {
		rc = EINVAL;
		goto out_free_scsi_cmd;
	}

	bfa_trc(bfad, 0);

	/*
	 * Setup of the data fields in scsi_cmd
	 */
	rc = bfad_scsi_data_setup(bfad, scsi_cmd,
				  (char *)(size_t) ci->iobuf_ptr, ci->write,
				  ci->iosz);
	if (rc != 0) {
		bfa_trc(bfad, 0);
		goto out_free_scsi_cmd;
	}

	/*
	 * Setup Misc scsi_cmd/scsi_device field needed for callback macros
	 */
	bfad_scsi_cmd_setup(bfad, scsi_cmd, ci, pthru);

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 26))
        if (scsi_sg_count(scsi_cmd))
                sg_cnt = scsi_sg_count(scsi_cmd);
        else if (scsi_bufflen(scsi_cmd))
                sg_cnt = 1;
#else
	if (scsi_cmd->use_sg)
		sg_cnt = scsi_cmd->use_sg;
	else if (scsi_cmd->request_bufflen)
		sg_cnt = 1;
#endif

	spin_lock_irqsave(&bfad->bfad_lock, flags);

	fcs_port = bfa_fcs_lookup_port(&bfad->bfa_fcs, ci->vf_id, ci->lpwwn);
	if (!fcs_port) {
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		ci->status = BFA_STATUS_UNKNOWN_LWWN;
		bfa_trc(bfad, 0);
		goto out_free_scsi_data;
	}

	fcs_itnim = bfa_fcs_itnim_lookup(fcs_port, ci->rpwwn);
	if (!fcs_itnim) {
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		ci->status = BFA_STATUS_UNKNOWN_RWWN;
		bfa_trc(bfad, 0);
		goto out_free_scsi_data;
	}

	if (fcs_itnim->itnim_drv->state != ITNIM_STATE_ONLINE) {
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		ci->status = BFA_STATUS_NO_FCPIM_NEXUS;
		bfa_trc(bfad, 0);
		goto out_free_scsi_data;
	}

	bfa_itnim = bfa_fcs_itnim_get_halitn(fcs_itnim);
	hal_io = bfa_ioim_alloc(&bfad->bfa, (struct bfad_ioim_s *) scsi_cmd,
				    bfa_itnim, sg_cnt);
	if (!hal_io) {
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		ci->status = BFA_STATUS_ENOMEM;
		bfa_trc(bfad, 0);
		goto out_free_scsi_data;
	}

	scsi_cmd->host_scribble = (char *)hal_io;
	init_completion(&pthru->comp);
	bfa_ioim_start(hal_io, BFA_IOIM_NO_SG);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	wait_for_completion(&pthru->comp);

	bfa_trc(bfad, ci->iosz);
	bfa_trc(bfad, ci->write);

	host_status = scsi_cmd->result >> 16;
	scsi_status = scsi_cmd->result & 0xFF;
	if (host_status != DID_OK) {
		ci->status = BFA_STATUS_FAILED;
		bfa_trc(bfad, host_status);
		goto out_free_scsi_data;
	} else if (ci->iosz > 0 && ci->write == 0) {
		rc = bfad_scsi_data_copyout(bfad,
					    (char *)(size_t) ci->
					    iobuf_ptr, scsi_cmd,
					    ci->iosz);
		if (rc != 0) {
			bfa_trc(bfad, 0);
			goto out_free_scsi_data2;
		}
	}

	/* Copy sense data */
	if (ci->snssz > 0 &&
	    copy_to_user((void __user *)(size_t)ci->snsbuf_pt,
		(uint8_t *) scsi_cmd->sense_buffer,
		 (ci->snssz > SCSI_SENSE_BUFFERSIZE ? SCSI_SENSE_BUFFERSIZE :
			ci->snssz))) {
		bfa_trc(bfad, 0);
		rc = EIO;
		goto out_free_scsi_data2; 
	}

	ci->status = BFA_STATUS_OK;
	ci->scsi_status = scsi_status;

out_free_scsi_data:
	if (copy_to_user
		((void __user *)arg, (uint8_t *) ci, sizeof(bfa_ioctl_scsi_t))) {
		bfa_trc(bfad, 0);
		rc = EIO;
	}

out_free_scsi_data2:
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 26))
	request_buffer       = (void *)scsi_sglist(scsi_cmd);
	sg_count             = scsi_sg_count(scsi_cmd);
	request_bufferlen    = scsi_bufflen(scsi_cmd);
#else
	request_buffer       = (void *)scsi_cmd->request_buffer;
	sg_count             = scsi_cmd->use_sg;
	request_bufferlen    = scsi_cmd->request_bufflen;
#endif

	if (request_buffer && sg_count)
		bfad_scsi_sg_data_free(bfad, (struct scatterlist *)request_buffer, sg_count);
	else if (request_buffer)
#if defined(__VMKLNX__)
		dma_free_coherent(&bfad->pcidev->dev, scsi_cmd->request_bufflen,
					scsi_cmd->request_buffer,
					scsi_cmd->request_bufferMA);
#elif (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
		dma_free_coherent(&bfad->pcidev->dev, request_bufferlen,
					request_buffer,
					scsi_cmd->SCp.dma_handle);
#else
		pci_free_consistent(bfad->pcidev, scsi_cmd->request_bufflen,
					scsi_cmd->request_buffer,
					scsi_cmd->SCp.dma_handle);
#endif

out_free_scsi_cmd:
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 26))
	kfree(scsi_cmd->cmnd);
	kfree(scsi_cmd->sense_buffer);
#endif
	kfree(scsi_cmd);
	return rc;
}

void
bfad_fcxp_mem_free(struct bfad_s *bfad, struct bfad_buf_info *buf_base,
		   uint32_t num_sgles)
{
	int             i;
	struct bfad_buf_info *buf_info = buf_base;

	for (i = 0; i < num_sgles; buf_info++, i++) {
		if (buf_info->virt != NULL) {
#if defined(__VMKLNX__) || (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
			dma_free_coherent(&bfad->pcidev->dev, buf_info->size,
					  buf_info->virt, buf_info->phys);
#else
			pci_free_consistent(bfad->pcidev, buf_info->size,
					  buf_info->virt, buf_info->phys);
#endif
		}
	}

	kfree(buf_base);
}

struct bfad_buf_info *
bfad_fcxp_mem_alloc(struct bfad_s *bfad, char *inbuf_ptr, uint32_t size,
		    uint32_t *num_sgles)
{
	int             sge_num, total_size;
	struct bfad_buf_info *buf_base, *buf_info;
	struct bfa_sge_s  *sg_table;

	sge_num = 1;
	total_size = sizeof(struct bfad_buf_info) + sizeof(struct bfa_sge_s);
	buf_base = kzalloc(total_size * sge_num, GFP_KERNEL);
	sg_table = (struct bfa_sge_s *) (buf_base + sge_num);

	if (buf_base == NULL)
		return NULL;

	buf_info = buf_base;
#if defined(__VMKLNX__) || (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
	buf_info->virt =
		dma_alloc_coherent(&bfad->pcidev->dev, size,
			&(buf_info->phys), GFP_KERNEL);
#else
	buf_info->virt =
		pci_alloc_consistent(bfad->pcidev, size,
			&(buf_info->phys));
#endif
	buf_info->size = size;
	if (!buf_info->virt) {
		return NULL;
	}

	if (inbuf_ptr) {
		if (copy_from_user
			((uint8_t *) buf_info->virt,
			(void __user *)(inbuf_ptr), size))
			goto out;
	} else {
		memset(buf_info->virt, 0, size);
	}

	/*
	 * Setup SG table
	 */
	sg_table->sg_len = size;
	sg_table->sg_addr = (void *)(size_t) buf_info->phys;

	*num_sgles = sge_num;

	return buf_base;

out:
	bfad_fcxp_mem_free(bfad, buf_base, sge_num);
	return NULL;
}

int
bfad_rsp_data_copy(struct bfad_s *bfad, char *obuf_ptr, struct bfad_buf_info *buf_info,
		   uint32_t size)
{
	if (obuf_ptr) {
		/*
		 * pci_dma_sync_single_for_device(bfad->pcidev,
		 * buf_info->phys, cnt, PCI_DMA_FROMDEVICE);
		 */

		/*
		 * Copy data to user space
		 */
		if (copy_to_user
			((void __user *)(obuf_ptr),
			(uint8_t *) buf_info->virt, size))
			return 1;
	}
	return 0;
}

int
bfad_fcxp_send(struct bfad_fcxp *drv_fcxp, bfa_ioctl_fcpt_t *ci)
{
	struct bfa_fcxp_s *hal_fcxp;
	struct bfad_s         *bfad = drv_fcxp->port->bfad;
	bfa_fcxp_get_sgaddr_t req_sgaddr_cb;
	bfa_fcxp_get_sglen_t req_sglen_cb;
	bfa_fcxp_get_sgaddr_t rsp_sgaddr_cb;
	bfa_fcxp_get_sglen_t rsp_sglen_cb;
	int             num_req_sgles, num_rsp_sgles;
	unsigned long   flags;
	uint8_t         lp_tag;

	if (drv_fcxp->use_ireqbuf && drv_fcxp->use_irspbuf) {
		bfa_trc(bfad, 0);
		num_req_sgles = 0;
		num_rsp_sgles = 0;
		req_sgaddr_cb = NULL;
		req_sglen_cb = NULL;
		rsp_sgaddr_cb = NULL;
		rsp_sglen_cb = NULL;
	} else if (drv_fcxp->use_ireqbuf && !drv_fcxp->use_irspbuf) {
		bfa_trc(bfad, 0);
		num_req_sgles = 0;
		num_rsp_sgles = drv_fcxp->num_rsp_sgles;
		req_sgaddr_cb = NULL;
		req_sglen_cb = NULL;
		rsp_sgaddr_cb = bfad_fcxp_get_rsp_sgaddr_cb;
		rsp_sglen_cb = bfad_fcxp_get_rsp_sglen_cb;
	} else if (!drv_fcxp->use_ireqbuf && drv_fcxp->use_irspbuf) {
		bfa_trc(bfad, 0);
		num_req_sgles = drv_fcxp->num_req_sgles;
		num_rsp_sgles = 0;
		req_sgaddr_cb = bfad_fcxp_get_req_sgaddr_cb;
		req_sglen_cb = bfad_fcxp_get_req_sglen_cb;
		rsp_sgaddr_cb = NULL;
		rsp_sglen_cb = NULL;
	} else {
		bfa_trc(bfad, 0);
		num_req_sgles = drv_fcxp->num_req_sgles;
		num_rsp_sgles = drv_fcxp->num_rsp_sgles;
		req_sgaddr_cb = bfad_fcxp_get_req_sgaddr_cb;
		req_sglen_cb = bfad_fcxp_get_req_sglen_cb;
		rsp_sgaddr_cb = bfad_fcxp_get_rsp_sgaddr_cb;
		rsp_sglen_cb = bfad_fcxp_get_rsp_sglen_cb;
	}
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	hal_fcxp =
		bfa_fcxp_alloc(drv_fcxp, &bfad->bfa, num_req_sgles,
				   num_rsp_sgles, req_sgaddr_cb, req_sglen_cb,
				   rsp_sgaddr_cb, rsp_sglen_cb);
	if (!hal_fcxp) {
		bfa_trc(bfad, 0);
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		return BFA_STATUS_ENOMEM;
	}

	drv_fcxp->bfa_fcxp = hal_fcxp;
	if (drv_fcxp->use_ireqbuf) {
		drv_fcxp->reqbuf_info = bfa_fcxp_get_reqbuf(hal_fcxp);

		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		if (copy_from_user
		    ((uint8_t *) drv_fcxp->reqbuf_info,
		     (void __user *)(size_t) ci->reqbuf_ptr, ci->reqsz))
			return EIO;
		spin_lock_irqsave(&bfad->bfad_lock, flags);
	}

	if (drv_fcxp->use_irspbuf)
		drv_fcxp->rspbuf_info = bfa_fcxp_get_rspbuf(hal_fcxp);

	lp_tag = bfa_lps_get_tag_from_pid(&bfad->bfa, ci->fchs.s_id);
	bfa_fcxp_send(hal_fcxp, drv_fcxp->bfa_rport, ci->vf_id, lp_tag, ci->cts,
			  ci->cos, ci->reqsz, &ci->fchs, bfad_send_fcpt_cb,
			  bfad, ci->rspsz, ci->tsecs);

	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	return 0;
}

static int
bfad_send_fc_passthru(unsigned long arg)
{
	bfa_ioctl_fcpt_t fc_req, *ci;
	int             rc = 0;
	struct bfad_fcxp    *drv_fcxp;
	struct bfa_fcs_lport_s *fcs_port;
	struct bfa_fcs_rport_s *fcs_rport;
	struct bfa_rport_s *bfa_rport = NULL;
	struct bfad_s         *bfad;
	unsigned long   flags;

	ci = &fc_req;

	if (copy_from_user((uint8_t *) ci, (void __user *)arg,
			    sizeof(bfa_ioctl_fcpt_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(ci->bfad_num);
	if (bfad == NULL)
		return EINVAL;

	bfa_trc(bfad, ci->vf_id);
	bfa_trc(bfad, ci->lpwwn);
	bfa_trc(bfad, ci->dpwwn);
	bfa_trc(bfad, ci->tsecs);
	bfa_trc(bfad, ci->cts);
	bfa_trc(bfad, ci->cos);
	bfa_trc(bfad, ci->reqsz);
	bfa_trc(bfad, ci->rspsz);
	bfa_trc(bfad, ci->fchs.d_id);
	bfa_trc(bfad, ci->fchs.s_id);
	bfa_trc(bfad, ci->nport_nexus_needed);

	drv_fcxp = bfad_fcxp_alloc(bfad);
	if (drv_fcxp == NULL) {
		bfa_trc(bfad, 0);
		return ENOMEM;
	}

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	fcs_port = bfa_fcs_lookup_port(&bfad->bfa_fcs, ci->vf_id, ci->lpwwn);
	if (fcs_port == NULL) {
		ci->status = BFA_STATUS_UNKNOWN_LWWN;
		bfa_trc(bfad, 0);
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		goto out_free_fcxp;
	}

	/* Check if the port is online before sending FC Passthru cmd */
	if (!bfa_fcs_lport_is_online(fcs_port)) {
		ci->status = BFA_STATUS_PORT_OFFLINE;
		bfa_trc(bfad, 0);
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		goto out_free_fcxp;
	}

	bfa_trc(bfad, fcs_port->pid);

	drv_fcxp->port = bfa_fcs_lport_get_drvport(fcs_port);

	/*
	 * TODO - drv_fcxp->port->bfad is null now.
	 */
	if (drv_fcxp->port->bfad == 0) {
		drv_fcxp->port->bfad = bfad;
		bfa_trc(bfad, 0);
	}

	if (ci->nport_nexus_needed) {
		fcs_rport = bfa_fcs_rport_lookup(fcs_port, ci->dpwwn);

		if (fcs_rport == NULL) {
			ci->status = BFA_STATUS_UNKNOWN_RWWN;
			bfa_trc(bfad, 0);
			spin_unlock_irqrestore(&bfad->bfad_lock, flags);
			goto out_free_fcxp;
		}

		bfa_rport = bfa_fcs_rport_get_halrport(fcs_rport);
	} else
		drv_fcxp->bfa_rport = NULL;

	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	if (ci->reqsz > BFA_FCXP_MAX_IBUF_SZ) {
		bfa_trc(bfad, 0);
		/*
		 * Allocate DMA memory for req
		 */
		drv_fcxp->reqbuf_info =
			bfad_fcxp_mem_alloc(bfad,
					    (char *)(size_t) ci->reqbuf_ptr,
					    ci->reqsz,
					    &drv_fcxp->num_req_sgles);
		if (!drv_fcxp->reqbuf_info) {
			bfa_trc(bfad, 0);
			rc = ENOMEM;
			goto out_free_fcxp;
		}
		drv_fcxp->req_sge =
			(struct bfa_sge_s  *)((struct bfad_buf_info *)drv_fcxp->reqbuf_info + drv_fcxp->num_req_sgles);
	} else {
		bfa_trc(bfad, ci->reqsz);
		drv_fcxp->use_ireqbuf = 1;
	}

	if (ci->rspsz > BFA_FCXP_MAX_IBUF_SZ) {
		bfa_trc(bfad, 0);
		/*
		 * Allocate DMA memory for rsp
		 */
		drv_fcxp->rspbuf_info =
			bfad_fcxp_mem_alloc(bfad, NULL, ci->rspsz,
					    &drv_fcxp->num_rsp_sgles);
		if (!drv_fcxp->rspbuf_info) {
			bfad_fcxp_mem_free(bfad, drv_fcxp->reqbuf_info,
					   drv_fcxp->num_req_sgles);
			bfa_trc(bfad, 0);
			rc = ENOMEM;
			goto out_free_fcxp;
		}
		drv_fcxp->rsp_sge =
			(struct bfa_sge_s  *)((struct bfad_buf_info *)drv_fcxp->rspbuf_info + drv_fcxp->num_rsp_sgles);
	} else {
		bfa_trc(bfad, ci->rspsz);
		drv_fcxp->use_irspbuf = 1;
	}

	init_completion(&drv_fcxp->comp);
	rc = bfad_fcxp_send(drv_fcxp, ci);
	if (rc == 0) {
		wait_for_completion(&drv_fcxp->comp);
		if (drv_fcxp->req_status == BFA_STATUS_OK) {
			if (drv_fcxp->use_irspbuf) {
				bfa_trc(bfad, 0);
				if (copy_to_user
				    ((void __user *)(size_t) ci->rspbuf_ptr,
				     (uint8_t *) drv_fcxp->rspbuf_info,
				     drv_fcxp->rsp_len)) {
					rc = EIO;
					goto out_mem_free;
				}
			} else {
				rc = bfad_rsp_data_copy(bfad,
							(char *)(size_t) ci->
							rspbuf_ptr,
							drv_fcxp->rspbuf_info,
							ci->rspsz);
				if (rc != 0) {
					rc = EIO;
					goto out_mem_free;
				}
			}
		}
		ci->status = drv_fcxp->req_status;
	}

out_mem_free:
	if (!drv_fcxp->use_ireqbuf)
		bfad_fcxp_mem_free(bfad, drv_fcxp->reqbuf_info,
				   drv_fcxp->num_req_sgles);
	if (!drv_fcxp->use_irspbuf)
		bfad_fcxp_mem_free(bfad, drv_fcxp->rspbuf_info,
				   drv_fcxp->num_rsp_sgles);

out_free_fcxp:
	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) ci, sizeof(bfa_ioctl_fcpt_t)))
		rc = EIO;

	bfad_fcxp_free(drv_fcxp);
	bfa_trc(bfad, 0);
	return rc;

}

/*
 * VPORT
 */
int
bfad_ioctl_vport_create(unsigned long arg, void *cmd)
{
	bfa_ioctl_vport_create_t *iocmd = (bfa_ioctl_vport_create_t *)cmd;
	struct bfad_s         *bfad;


	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return EINVAL;

	bfa_trc(bfad, iocmd->port_cfg.pwwn);
	bfa_trc(bfad, iocmd->port_cfg.nwwn);
	iocmd->status = bfad_vport_create(bfad, iocmd->vf_id, &iocmd->port_cfg, 0);

	if (iocmd->status != BFA_STATUS_OK) {
		BFA_DEV_PRINTF(bfad, BFA_ERR,
			       "Failed to create vport "
			       "wwnn %llx wwpn %llx for vf %d\n",
			       bfa_os_htonll(iocmd->port_cfg.nwwn),
			       bfa_os_htonll(iocmd->port_cfg.pwwn), iocmd->vf_id);
	}


	return 0;
}

int
bfad_ioctl_vport_del(unsigned long arg, void *cmd)
{
	bfa_ioctl_vport_t *iocmd = (bfa_ioctl_vport_t *)cmd;
	struct bfad_vport_s   *bfad_vport;
	struct bfa_fcs_vport_s *fcs_vport;
	unsigned long   flags;
	struct bfad_s         *bfad;
	struct completion fcomp;
	struct Scsi_Host *shost;


	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	fcs_vport =
		bfa_fcs_vport_lookup(&bfad->bfa_fcs, iocmd->vf_id, iocmd->vpwwn);
	if (fcs_vport == NULL) {
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		iocmd->status = BFA_STATUS_UNKNOWN_VWWN;
		goto out;
	}

	bfad_vport = fcs_vport->vport_drv;
	shost = bfad_vport->drv_port.im_port->shost;
	spin_lock(shost->host_lock);
	if (shost->shost_state == SHOST_RECOVERY) {
		spin_unlock(shost->host_lock);
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		iocmd->status = BFA_STATUS_DEVBUSY;
		goto out;
	}
        spin_unlock(shost->host_lock);
	bfad_vport->drv_port.flags |= BFAD_PORT_DELETE;

	bfad_vport->comp_del = &fcomp;
	init_completion(bfad_vport->comp_del);

	iocmd->status = bfa_fcs_vport_delete(fcs_vport);
	bfa_trc(bfad, iocmd->status);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	wait_for_completion(bfad_vport->comp_del);

	if (iocmd->status != BFA_STATUS_OK) {
		BFA_DEV_PRINTF(bfad, BFA_ERR,
			       "Failed to delete vport "
			       "wwpn %llx for vf %d\n", bfa_os_htonll(iocmd->vpwwn),
			       iocmd->vf_id);
	}

	bfa_trc(bfad, bfad->inst_no);
#if defined(__VMKLNX__) || (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
	bfad_os_scsi_host_free(bfad, bfad_vport->drv_port.im_port);
#endif
	kfree(bfad_vport);

out:

	return 0;
}

/*
 * DIAG
 */
int
bfad_ioctl_diag_regrd(unsigned long arg, void *cmd)
{
	bfa_ioctl_diag_regread_t *iocmd = (bfa_ioctl_diag_regread_t *)cmd;
	uint32_t       *buf;
	struct bfad_s         *bfad;
	unsigned long   flags;


	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return EINVAL;

	buf = kzalloc(iocmd->nwords << 2, GFP_KERNEL);
	if (!buf)
		return ENOMEM;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd->status =
		bfa_diag_reg_read(BFA_DIAG_MOD(&bfad->bfa), iocmd->off, iocmd->nwords,
				      buf, iocmd->force);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	if (iocmd->status != BFA_STATUS_OK)
		goto out;

	if (copy_to_user
	    ((void __user *)(size_t) iocmd->buf_ptr, (uint8_t *) buf,
	     iocmd->nwords << 2)) {
		kfree(buf);
		return EIO;
	}

out:
	kfree(buf);

	return 0;
}

int
bfad_ioctl_diag_regwr(unsigned long arg, void *cmd)
{
	bfa_ioctl_diag_regwrite_t *iocmd = (bfa_ioctl_diag_regwrite_t *)cmd;
	struct bfad_s         *bfad;
	unsigned long   flags;


	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd->status =
		bfa_diag_reg_write(BFA_DIAG_MOD(&bfad->bfa), iocmd->off, 1,
				       iocmd->word, iocmd->force);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);


	return 0;
}

/*
 * DEBUG
 */
int
bfad_ioctl_debug_drv_trace(unsigned long arg, void *cmd)
{
	bfa_ioctl_debug_t *iocmd = (bfa_ioctl_debug_t *)cmd;
	struct bfad_s         *bfad;


	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return EINVAL;

	if (iocmd->bufsz < sizeof(struct bfa_trc_mod_s)) {
		iocmd->status = BFA_STATUS_EINVAL;
		goto out;
	}

	if (copy_to_user
	    ((void __user *)(size_t) iocmd->buf_ptr, (uint8_t *) bfad->trcmod,
	     iocmd->bufsz))
		return EIO;

	iocmd->status = BFA_STATUS_OK;
out:

	return 0;
}

int
bfad_ioctl_debug_fw_trace(unsigned long arg, bfa_boolean_t saved, void *cmd)
{
	bfa_ioctl_debug_t *iocmd = (bfa_ioctl_debug_t *)cmd;
	void           *buf;
	struct bfad_s         *bfad;
	unsigned long   flags;


	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return EINVAL;

	/*
	 * TBD: Verify the size of the buffer
	 */
	bfa_trc(bfad, iocmd->bufsz);
	if (iocmd->bufsz < sizeof(struct bfa_trc_mod_s)) {
		bfa_trc(bfad, sizeof(struct bfa_trc_mod_s));
		iocmd->status = BFA_STATUS_EINVAL;
		goto out;
	}

	buf = vmalloc(iocmd->bufsz);
	if (!buf) {
		bfa_trc(bfad, 0x4444);
		return ENOMEM;
	}

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	if (!saved)
		iocmd->status = bfa_debug_fwtrc(&bfad->bfa, buf, &iocmd->bufsz);
	else
		iocmd->status = bfa_debug_fwsave(&bfad->bfa, buf, &iocmd->bufsz);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	if (iocmd->status != BFA_STATUS_OK) {
		bfa_trc(bfad, 0x5555);
		goto out_free_buf;
	}

	bfa_trc(bfad, iocmd->bufsz);
	if (copy_to_user
	    ((void __user *)(size_t) iocmd->buf_ptr, (uint8_t *) buf,
	     iocmd->bufsz)) {
		vfree(buf);
		bfa_trc(bfad, 0x55551);
		return EIO;
	}

out_free_buf:
	vfree(buf);
out:

	bfa_trc(bfad, 0x6666);

	return 0;
}

int
bfad_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
	   unsigned long arg)
{
	int rc = EINVAL, size = _IOC_SIZE(cmd);
	void *iocmd = NULL;

	if (!arg)
		return -rc;

	if (!bfad_is_ready()) {
		rc = EAGAIN;
		return -rc;
	}

	if (cmd == IOCMD_IOC_GET_VERSION) {
		rc = bfad_ioctl_ioc_get_version(arg);
		goto ext;
	}

	if (cmd == IOCMD_IOC_GET_INSTS) {
		rc = bfad_ioctl_ioc_get_inst(arg);
		goto ext;
	}

	/* SCSI PASSTHRU */
        if (cmd == IOCMD_SEND_SCSI) {
                rc = bfad_send_scsi_passthru(arg);
		goto ext;
	}
        /* FC PASSTHRU */
        if (cmd == IOCMD_SEND_FC) {
                rc = bfad_send_fc_passthru(arg);
		goto ext;
	}

	/* EVENT IOCTLs */
	if (cmd == IOCMD_AEN_GET) {
		rc = bfad_ioctl_aen_get(file, arg);
                goto ext;
	}
	if (cmd == IOCMD_AEN_UPDATE) {
		rc = bfad_ioctl_aen_update(file, arg);
                goto ext;
	}

	iocmd = bfad_cp_from_user(size, arg);
	if (!iocmd) {
		return -EIO;
	}

	/*
	 * IOCTL requests handler - will handle requests cases that are not
	 * common across BSG & IOCTL;
	 * For all common rqsts - calls bfad_iocmd_handler for rqst service.
	 */
	switch (cmd) {

	/* VPORT IOCTLs */
	case IOCMD_VPORT_CREATE:
		rc = bfad_ioctl_vport_create(arg, iocmd);
		break;
	case IOCMD_VPORT_DEL:
		rc = bfad_ioctl_vport_del(arg, iocmd);
		break;
	
	/* DIAG IOCTLs */
	case IOCMD_DIAG_REGRD:
		rc = bfad_ioctl_diag_regrd(arg, iocmd);
		break;
	case IOCMD_DIAG_REGWR:
		rc = bfad_ioctl_diag_regwr(arg, iocmd);
		break;
	/* DEBUG IOCTLs */
	case IOCMD_DEBUG_DRV_TRACE:
		rc = bfad_ioctl_debug_drv_trace(arg, iocmd);
		break;
	case IOCMD_DEBUG_FW_TRACE:
		rc = bfad_ioctl_debug_fw_trace(arg, BFA_FALSE, iocmd);
		break;
	case IOCMD_DEBUG_FW_STATE:
		rc = bfad_ioctl_debug_fw_trace(arg, BFA_TRUE, iocmd);
		break;

	default:
		/* Invoke common IOCMD request handler
		 * For IOCTL path - we get bfad from iocmd; so pass NULL
		 * bfad_iocmd_handler functions will fetch bfad.
		 */
		rc = bfad_iocmd_handler(NULL, cmd, iocmd, arg, 0);
		break;
	}

	if(bfad_cp_to_user(size, arg, iocmd)) {
		rc = EIO;
	}

ext:
	return -rc;
}

