/*
 * Copyright (c) 2005-2014 Brocade Communications Systems, Inc.
 * Copyright (c) 2014 QLogic Corporation.
 * All rights reserved
 * www.qlogic.com
 *
 * See LICENSE.bfa for copyright and licensing details.
 */

#include <linux/module.h>

#include "bfad_ioctl_cna.h"
#include "bfad_ioctl.h"
#include "bfad_drv.h"

#include "bfad_aen.h"

static int      bfad_open(struct inode *inode, struct file *file);
static int      bfad_close(struct inode *inode, struct file *file);
int		bfad_ioctl(struct inode *inode, struct file *file,
			   unsigned int cmd, unsigned long arg);
static unsigned int bfad_poll(struct file *file,
			      struct poll_table_struct *wait);
static int      bfad_fasync(int fd, struct file *file, int on);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 13)
static struct class *bfad_class;
#else
static struct class_simple *bfad_class;
#endif

static int      bfad_major;

struct bfad_s	*bfad_find_bfad_by_inst_no(int inst_no);

static char    *bfad_dev_name = BFAD_DRIVER_NAME;
static struct file_operations bfad_fops;

int
bfad_os_class_create(struct module *owner, char *name)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 13)
	bfad_class = class_create(owner, name);
#else
	bfad_class = class_simple_create(owner, name);
#endif
	if (IS_ERR(bfad_class)) {
		printk(KERN_WARNING "Unable to create sysfs class for bfa.\n");
		bfad_class = NULL;
		return 1;
	}

	return 0;
}

void
bfad_os_class_destroy(void)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 13)
	class_destroy(bfad_class);
#else
	class_simple_destroy(bfad_class);
#endif
	bfad_class = NULL;
}

int
bfad_os_class(void)
{
	return (bfad_class != 0);
}

int
bfad_os_ioctl_init(void)
{

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 26)
	struct device *cd;
#else
	struct class_device *cd;
#endif
	int ret = 0;

	if (bfad_os_class_create(THIS_MODULE, bfad_dev_name)) {
		printk(KERN_WARNING "Unable to create sysfs class for bfa.\n");
		return 1;
	}

	ret = bfad_os_ioctl32_register();
	if (ret < 0) {
		/*
		 * The IOCTL32 registration failed. Unregister
		 * all the IOCTL32 registrations.
		 */
		bfad_os_ioctl32_unregister();
	}

	bfad_major = register_chrdev(0, bfad_dev_name, &bfad_fops);
	if (bfad_major < 0) {
		printk(KERN_WARNING "Unable to register chrdev for bfa.\n");
		bfad_os_class_destroy();
		return bfad_major;
	}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 27)
	cd = device_create(bfad_class, NULL, MKDEV(bfad_major, 0), NULL,
							"bfa");
#elif LINUX_VERSION_CODE == KERNEL_VERSION(2, 6, 26)
	cd = device_create(bfad_class, NULL, MKDEV(bfad_major, 0), "bfa");
#elif LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 13)
	cd = class_simple_device_add(bfad_class, MKDEV(bfad_major, 0), NULL,
							"bfa");
#elif LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 15)
	cd = class_device_create(bfad_class, MKDEV(bfad_major, 0), NULL, "bfa");
#else
	cd = class_device_create(bfad_class, NULL, MKDEV(bfad_major, 0), NULL,
							"bfa");
#endif

	if (IS_ERR(cd))
		return BFA_STATUS_FAILED;

	return BFA_STATUS_OK;
}

int
bfad_os_ioctl_exit(void)
{
	if (!bfad_os_class())
		return 1;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 26)
	device_destroy(bfad_class, MKDEV(bfad_major, 0));
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 13)
	class_device_destroy(bfad_class, MKDEV(bfad_major, 0));
#else
	class_simple_device_remove(MKDEV(bfad_major, 0));
#endif

	unregister_chrdev(bfad_major, bfad_dev_name);
	bfad_os_ioctl32_unregister();
	bfad_os_class_destroy();
	return 0;
}

#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 14)
static long
bfad_compat_ioctl(struct file *file, unsigned int cmd, unsigned long arg)
{
	return (bfad_ioctl(file->f_dentry->d_inode, file, cmd, arg));
}
#endif

static int
bfad_open(struct inode *inode, struct file *file)
{
	struct bfad_s   *bfad;
	int32_t		bfad_minor;

	if (!bfad_is_ready())
		return (-EAGAIN);

	bfad_minor = MINOR(inode->i_rdev);
	bfad = bfad_find_bfad_by_inst_no(bfad_minor);
		// for bfa, minor will be always 0
	if (bfad == NULL) {
		printk(KERN_WARNING
			"%s:%d: Invalid minor device=%d for file=%p\n",
			__FUNCTION__, __LINE__, bfad_minor, file);
		return (-EINVAL);
	}

	down(&bfad_sem);
	bfad->ref_count++;
	up(&bfad_sem);

	return 0;
}

static int
bfad_close(struct inode *inode, struct file *file)
{
	struct bfad_s   *bfad;
	int32_t		bfad_minor;

	if (!bfad_is_ready())
		return (-EPERM);

	bfad_minor = MINOR(inode->i_rdev);
	bfad = bfad_find_bfad_by_inst_no(bfad_minor);
	if (bfad == NULL) {
		printk(KERN_WARNING
			"%s:%d: Invalid minor device=%d for file=%p\n",
			__FUNCTION__, __LINE__, bfad_minor, file);
		return (-EINVAL);
	}

	down(&bfad_sem);
	if (--bfad->ref_count) {
		up(&bfad_sem);
		return (0);
	}
	up(&bfad_sem);
	/*
	 * No more apps using this file handle
	 * Remove this file from the asynchronously notified file's
	 * fasync_helper() may sleep, so release the lock
	 */
	bfad_aen_fasync(-1, file, 0);

	return (0);
}

static unsigned int
bfad_poll(struct file *file, struct poll_table_struct *wait)
{
	return bfad_aen_poll(file, wait);
}

static int
bfad_fasync(int fd, struct file *file, int on)
{
	return bfad_aen_fasync(fd, file, on);
}

/**
 * Define the BFA driver file operations.
 */
static struct file_operations bfad_fops = {
	.owner = THIS_MODULE,
	.ioctl = bfad_ioctl,
#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 14)
	.compat_ioctl = bfad_compat_ioctl,
#endif
	.open = bfad_open,
	.release = bfad_close,
	.poll = bfad_poll,
	.fasync = bfad_fasync
};
