/*
 * Copyright (c) 2005-2014 Brocade Communications Systems, Inc.
 * Copyright (c) 2014 QLogic Corporation.
 * All rights reserved
 * www.qlogic.com
 * See LICENSE.bfa for copyright and licensing details.
 */
/**
 * @file bfad_ip.c IP over FC module implementation.
 */

#include <linux/device.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>

#ifdef IPV6_SUPPORT
#include <net/ndisc.h>
#endif

#if (LINUX_VERSION_CODE > KERNEL_VERSION(3, 0, 0))
#ifndef HAVE_NET_DEVICE_OPS
#define HAVE_NET_DEVICE_OPS
#endif
#endif

#include <linux/ip.h>
#include <linux/ipv6.h>
#include <linux/in.h>
#include <linux/if.h>
#include <linux/if_arp.h>
#include <linux/if_ether.h>
#include <linux/netdevice.h>
#include <linux/inetdevice.h>
#include <linux/etherdevice.h>
#include <linux/ethtool.h>
#include <linux/skbuff.h>
#include <linux/device.h>
#include <linux/sysfs.h>

#include <cs/bfa_cs.h>

#include "bfad_drv.h"
#include <fcs/bfa_fcs.h>
#include <fcs/bfa_fcs_ipfc.h>
#include "bfad_im.h"
#include "bfad_tm.h"
#include "bfad_ipfc.h"
#include <linux/module.h>
#include <ioctl/bfa_ioctl_ipfc.h>
#include "bfad_ip_drv.h"

int ipfc_mtu = -1;
int ipfc_unloadpkt_count; /* IPFC unload packet count */

module_param(ipfc_mtu, int, S_IRUGO | S_IWUSR);
MODULE_PARM_DESC(ipfc_mtu,
	"MTU value of IPoFC interfaces, default=2024, Range[68-65280]");

BFA_HT_INIT(bfad_ip_pid)
BFA_HT_LOOKUP(bfad_ip_pid, bfad_iprp_t *, bfad_ip_pid_hash,	\
		bfad_ip_key_cmp_pid)
BFA_HT_ADD(bfad_ip_pid, bfad_iprp_t *, bfad_ip_pid_hash)
BFA_HT_REMOVE(bfad_ip_pid, bfad_iprp_t *)

BFA_TRC_FILE(LDRV, IPFC);

/*
 * XML Handler for logging messages
 */
	bfa_log_mod_t   bfad_ip_logm;

/*
 * Save the MTU to be used while allocating the RX buffers
 */
	int		bfad_ip_mtu = BFAD_IP_DEFAULT_MTU_SIZE;

/*
 * sg list used during transmit
 */
	bfa_sge_t	bfad_ip_sge[BFAD_IP_MAX_SGES + 1];

/*
 *!<Running index for vports
 */
	static uint32_t bfad_ip_port_index;

/**
 * @addtogroup ip_private
 * @{
 */

/**
 * @brief Send out fcarp request [RFC 4338]
 * @param[in] dev  - device
 * @param[in] type - Request/Response
 * @param[in] ip   - ip address of the destination
 * @param[in] wwpn - wwpn of the destination
 * @retval  0 on success
 * @retval -1 on failure
 */
static int
bfad_ip_send_fcarp(struct net_device *dev, int type,
	uint32_t *sip, uint32_t *dip, uint8_t *wwpn,
	bfad_ip_fcarp_t *in_fcarp)
{
	bfad_ip_wwpn_t  out_wwpn;
	bfad_iprp_t     arp, arpc;
	char		message[100];
	bfad_ip_fcarp_t *fcarp;
	struct sk_buff *out_skb;
	bfad_ip_private_t *private = netdev_priv(dev);

	bfa_trc(&private->ioc->bfa_fcs, *sip);
	bfa_trc(&private->ioc->bfa_fcs, *dip);

	memset(&arp, 0x00, sizeof(bfad_iprp_t));

	/** Allocate SKB */
	out_skb = dev_alloc_skb(sizeof(bfad_ip_fcarp_t));
	if (out_skb == NULL) {
		sprintf(message, "fcarp (%s) drop packet\n", dev->name);
		bfa_log(&bfad_ip_logm, BFA_LOG_LINUX_MEM_ALLOC_FAILED, message);
		return -1;
	}

	skb_put(out_skb, sizeof(bfad_ip_fcarp_t));

	/**
	 * Fill the fcarp structure. If it is fcarp request, the pid will be
	 * broadcast pid.
	 * Set destination wwpn to 0 in case of request 4338 [7]
	 */
	if (type == BFAD_IP_FCARP_CODE_REQ)
		memset(&out_wwpn, 0x00, sizeof(bfad_ip_wwpn_t));
	else
		memcpy(&out_wwpn, wwpn, sizeof(bfad_ip_wwpn_t));

	out_skb->dev = dev;

	fcarp = (bfad_ip_fcarp_t *) (out_skb->data);

	private->my_pid = bfa_fcs_lport_get_fcid(
			bfa_fcs_get_base_port(&private->ioc->bfa_fcs));

	if (type == BFAD_IP_FCARP_CODE_REQ) {
		memcpy(fcarp->src_hw_addr.pname.raw, &private->my_wwpn,
			BFAD_IP_WWPN_LEN);
		fcarp->src_hw_addr.reserved = 0;
		fcarp->src_hw_addr.pid = bfa_os_hton3b(private->my_pid);
		fcarp->src_ip.s_addr = bfa_os_htonl(*sip);
		memcpy(fcarp->dest_hw_addr.pname.raw, &out_wwpn,
			BFAD_IP_WWPN_LEN);
		fcarp->dest_hw_addr.reserved = 0;
		fcarp->dest_hw_addr.pid = 0;
		fcarp->dest_ip.s_addr = bfa_os_htonl(*dip);
		arp.pid = bfa_os_hton3b(FC_BROADCAST_SERVER);
	} else {
		memcpy(fcarp, in_fcarp, sizeof (bfad_ip_fcarp_t));
		fcarp->dest_hw_addr.pname.raw[0] = 0x10;
		memcpy(fcarp->dest_hw_addr.pname.raw, &out_wwpn,
			BFAD_IP_WWPN_LEN);
		fcarp->dest_hw_addr.pid = bfa_os_hton3b(private->my_pid);
		if (private->my_ip == 0)
			bfad_ip_get_ip(dev, (struct in_addr *)&private->my_ip);
		fcarp->dest_ip.s_addr = bfa_os_htonl(private->my_ip);
		arp.pid = fcarp->src_hw_addr.pid;

		/* process and update local ARP cache */
		memset(&arpc, 0x00, sizeof(bfad_iprp_t));
		memcpy(arpc.wwpn.raw, fcarp->src_hw_addr.pname.raw,
		       BFAD_IP_WWPN_LEN);
		arpc.pid = bfa_os_ntoh3b(fcarp->src_hw_addr.pid);
		arpc.ip_addr_type = BFAD_IP_ADDR_TYPE_IPV4;
		arpc.ip_addr.ipv4_addr.s_addr =
			bfa_os_ntohl(fcarp->src_ip.s_addr);
		bfad_ip_update_arp_cache(dev, &arpc, BFAD_IP_ARP_STATE_INIT);
	}

	memcpy(arp.wwpn.raw, out_wwpn.raw, BFAD_IP_WWPN_LEN);
	fcarp->hw_type = bfa_os_htons(BFAD_IP_FCARP_HW_TYPE);
	fcarp->protocol = (uint16_t) bfa_os_htons(ETH_P_IP);
	fcarp->hw_len = BFAD_IP_FCARP_HW_LEN;
	fcarp->prot_len = BFAD_IP_PROTOCOL_LEN;
	fcarp->opcode = bfa_os_htons(type);

	/** Ship the packet out */
	bfad_ip_add_headers_ship_out(out_skb, BFAD_IP_FC4_TYPE, &arp,
				     ETH_P_ARP);
	return 0;
}

void
bfad_ip_init_fchs(fchs_t *fchs, int fc_type)
{
	memset(fchs, 0x00, sizeof(fchs_t));

	fchs->rx_id = FC_RXID_ANY;

	fchs->cs_ctl = 0;
	fchs->rx_id = FC_RXID_ANY;

	if (fc_type == FC_TYPE_ELS) {
		fchs->routing = FC_RTG_EXT_LINK;
		fchs->cat_info = FC_CAT_LD_REQUEST;
		fchs->type = FC_TYPE_ELS;
		fchs->f_ctl =
			bfa_os_hton3b(FCTL_FS_EXCH | FCTL_END_SEQ |
				      FCTL_SI_XFER);
		fchs->df_ctl = 0x00;	/* ? */
	} else {
		fchs->routing = FC_RTG_FC4_DEV_DATA;
		fchs->cat_info = 0x04;
		fchs->type = FC_TYPE_IP;
		fchs->f_ctl =
			bfa_os_hton3b(FCTL_FS_EXCH | FCTL_LS_EXCH |
				      FCTL_END_SEQ);
		fchs->df_ctl = 0x20;
	}
}


/**
 * @brief Send out farp request [RFC 2625] (for backward compatibility)
 * @param[in] code - farp packet type, request/reply
 * @param[in] dev  - net device structure
 * @param[in] fchs - fchs
 * @param[in] wwpn - destination wwpn
 * @param[in] buf  - buffer = NULL [ REQ ], buffer = request [REPLY]
 */
static void
bfad_ip_send_farp(int code, struct net_device *dev, fchs_t *in_fchs,
		  bfad_ip_wwpn_t *wwpn, uint8_t *buf, bfad_iprp_t *arp,
		struct bfa_fcs_lport_s *lport)
{
	bfad_ip_els_farp_t *farp;
	fchs_t  bfad_ip_fchs;
	bfad_ip_private_t *pri = netdev_priv(dev);
	bfa_fcxp_t     *fcxp = NULL;

	bfa_trc(&pri->ioc->bfa_fcs, code);

	fcxp = bfa_fcxp_alloc(NULL, (lport->fcs)->bfa, 0, 0, NULL, NULL, NULL,
			NULL);
	if (!fcxp)
		return;

	farp = bfa_fcxp_get_reqbuf(fcxp);
	memset(farp, 0x00, sizeof(bfad_ip_els_farp_t));

	bfad_ip_init_fchs(&bfad_ip_fchs, FC_TYPE_ELS);

	if (code == BFAD_IP_FARP_CMD_REQ) {
		farp->code = code;
		/**
		 * If the packet if farp request, tell the remote to
		 * match the wwpn. Other parameters are optional.
		 */
		farp->match_addr = BFAD_IP_FARP_MATCH_WW_PN;
		farp->s_id = pri->my_pid;
		farp->d_id = 0xFFFFFF;
		memcpy(farp->s_wwpn.raw, pri->my_wwpn, BFAD_IP_WWPN_LEN);
		farp->s_wwpn.raw[0] = 0x10;
		farp->s_wwpn.raw[1] = 0x00;

		memcpy(farp->d_wwpn.raw, wwpn->raw, BFAD_IP_WWPN_LEN);
		farp->resp_flags =
			BFAD_IP_FARP_INIT_REPLY;
		memcpy(arp->wwpn.raw, wwpn->raw, BFAD_IP_WWPN_LEN);

		bfad_ip_fchs.s_id = pri->my_pid;
		bfad_ip_fchs.d_id = 0xFFFFFF;

	} else if (code == BFAD_IP_FARP_CMD_REPLY) {
		/**
		 * If the packet is response, copy all the information from the
		 * incoming request packet before updating other fields.
		 */
		memcpy(farp, buf, sizeof(bfad_ip_els_farp_t));
		farp->resp_flags = 0;
		farp->code = code;
		farp->s_id = in_fchs->s_id;
		farp->d_id = pri->my_pid;
		memcpy(farp->d_wwpn.raw, pri->my_wwpn, BFAD_IP_WWPN_LEN);

		bfad_ip_fchs.d_id = in_fchs->s_id;
		bfad_ip_fchs.s_id = pri->my_pid;
		bfad_ip_fchs.ox_id = in_fchs->ox_id;

	}


	bfa_fcxp_send(fcxp, NULL, lport->fabric->vf_id, lport->lp_tag,
			BFA_FALSE, FC_CLASS_3, sizeof(bfad_ip_els_farp_t),
			&bfad_ip_fchs, NULL, NULL,
			FC_MAX_PDUSZ, 0);

}

/**
 * @brief Function to findout if the packet is of ipv6/mcast
 * @param[in] buf - buffer starting from ethernet Header
 * @retval  1 if packet is multicast
 * @retval -1 if packet is not multicast
 */
static int
bfad_ip_is_ipv6_mcast(uint8_t *buf)
{
	struct ipv6_pinfo *ipv6;

	ipv6 = (struct ipv6_pinfo *)(buf + sizeof(struct ethhdr));
	/*
	 * ff-ff-ff-ff-ff-ff is a special multicast address
	 */
	if (ipv6->daddr.in6_u.u6_addr8[0] == 0xff)
		return 1;
	/*
	 * Test for least-significant bit of the first byte
	 */
	/*
	 * The packets typically look like 33-33-xx-xx-xx-xx
	 */
	else if (buf[0] & 0x01)
		return 1;
	else
		return -1;
}

/**
 * @brief Function to do lookup from a ulp[arp] to the actual wwpn
 * @param[in]  dev - network device
 * @param[in]  mac - the mac address from the ulp
 * @param[out] res - the wwpn resulted from lookup
 * @retval  0 on success
 * @retval -1 on failure
 */
static int
bfad_ip_wwpn_lookup(struct net_device *dev, unsigned char *mac,
		    bfad_iprp_t **res)
{
	bfad_iprp_t     arp;
	bfad_ip_private_t *pri = netdev_priv(dev);

	/** continue to lookup in the hash table */
	arp.ip_addr.ipv4_addr.s_addr = *(uint32_t *) (mac + 2);
	*res = bfad_ip_pid_hash_lookup(pri->hash_pid_table, &arp);
	if (*res) {
		/** If lookup is successful, update timestamp and return 0 */
		(*res)->timestamp = jiffies;
		return 0;
	} else {
		/** If lookup fails, return -1 */
		return -1;
	}
}

/**
 * @brief Function to update local ARP cache
 * @param[in] dev   - Network device
 * @param[in] arp   - the entry to be updated
 * @param[in] state - state of the arp entry
 * @return - index of the entry added
 */
int
bfad_ip_update_arp_cache(struct net_device *dev, bfad_iprp_t *arp,
			 BFAD_IP_ARP_STATE state)
{
	bfad_iprp_t    *narp, *res_pid;
	bfad_ip_private_t *pri = netdev_priv(dev);
	unsigned long   flags;

	bfa_trc(&pri->ioc->bfa_fcs, arp->pid);

	/** First look up in the arp cache table */
	res_pid = bfad_ip_pid_hash_lookup(pri->hash_pid_table, arp);
	if (res_pid) {
	/** if entry already exists, update fields: state/ipv4-6 address */
#ifdef IPV6_SUPPORT
		res_pid->state = state;
		if (arp->ip_addr_type != BFAD_IP_ADDR_TYPE_NONE) {
			memcpy(&res_pid->ip_addr.ipv6_addr,
			       &arp->ip_addr.ipv6_addr,
			       sizeof(struct in6_addr));
		}
		res_pid->fcs_iprp = arp->fcs_iprp;
#endif
		spin_lock_irqsave(&pri->ioc->bfad_lock, flags);
		res_pid->timestamp = jiffies;
		res_pid->pid = arp->pid;
		spin_unlock_irqrestore(&pri->ioc->bfad_lock, flags);
	} else {
		/** If look up fails, which means a new entry has to be added */
		narp = kmalloc(sizeof(bfad_iprp_t), GFP_ATOMIC);
		if (narp == NULL) {
			bfa_assert(0);
			return -1;
		}
		arp->state = BFAD_IP_ARP_STATE_INIT;
		arp->timestamp = jiffies;
		bfa_q_init(narp);

		memcpy(narp, arp, sizeof(bfad_iprp_t));

		spin_lock_irqsave(&pri->ioc->bfad_lock, flags);
		bfad_ip_pid_hash_add(pri->hash_pid_table, narp);
		spin_unlock_irqrestore(&pri->ioc->bfad_lock, flags);

	}

	return arp->pid;
}

/**
 * @brief bfad_ip_out_translate_dest_mac
 *  Function that converts mac address coming from ulp into a real wwpn
 *  to be used to form network header. RFC 4338 [10]
 * @param[in] type - in coming packet type
 * @param[in]  skb - in coming packet from the ulp
 * @param[out] arp - the arp entry found in the local cache
 * @retval  0 on success
 * @retval -1 on failure
 */
static int
bfad_ip_out_translate_dest_broadcast(BFAD_IP_PKT_TYPE type, struct sk_buff *skb,
				     bfad_iprp_t *arp)
{
	int		ret = -1;
	uint8_t		bcast_wwpn[] =
		{ 0x10, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff };
	uint8_t		mcast_wwpn_ipv6[] =
		{ 0x10, 0x00, 0x33, 0x33, 0x00, 0x00, 0x00, 0x00 };
	uint8_t		mcast_wwpn[] =
		{ 0x10, 0x00, 0x01, 0x00, 0x5e, 0x00, 0x00, 0x00 };
	bfad_ip_private_t *pri = netdev_priv(skb->dev);

	memset(arp, 0, sizeof(bfad_iprp_t));

	bfa_trc(&pri->ioc->bfa_fcs, type);
	bfa_trc(&pri->ioc->bfa_fcs, ((struct ethhdr *)(skb->data))->h_dest[0]);

	/** Check, if the packet is a broadcast packet */
	if (((struct ethhdr *)(skb->data))->h_dest[0] == 0xFF) {
		bfa_trc(&pri->ioc->bfa_fcs, type);
		memcpy(arp->wwpn.raw, bcast_wwpn, BFAD_IP_WWPN_LEN);
		arp->pid = bfa_os_hton3b(FC_BROADCAST_SERVER);
		ret = 0;
	}
	/** Check, if the packet is a IPv6 multicast packet */
	else if (type == BFAD_IP_PKT_TYPE_IPV6 &&
		 bfad_ip_is_ipv6_mcast((unsigned char *)skb->data)) {
		struct ipv6hdr *ip6h =
			(struct ipv6hdr *)(skb->data + sizeof(struct ethhdr));
		bfa_trc(&pri->ioc->bfa_fcs, type);
		memcpy(arp->wwpn.raw, mcast_wwpn_ipv6, BFAD_IP_WWPN_LEN);
		/** Copy 4 least significant octets from the destination IP */
		memcpy(&arp->wwpn.raw[3], (void *)&ip6h->saddr.s6_addr32[0], 4);
		arp->pid = bfa_os_hton3b(FC_BROADCAST_SERVER);
		ret = 0;
	}
	/** Check, if the packet is a IPv4 multicast packet */
	else if (type == BFAD_IP_PKT_TYPE_IPV4 &&
		 /*
		  * Test for least-significant bit of the first byte
		  */
		 (((struct ethhdr *)(skb->data))->h_dest[0] & 0x01)) {
		uint32_t	*t = (uint32_t *) & arp->wwpn.raw[3];
		struct iphdr	*ip =
			(struct iphdr *)(skb->data + sizeof(struct ethhdr));
		bfa_trc(&pri->ioc->bfa_fcs, type);
		memcpy(arp->wwpn.raw, mcast_wwpn, BFAD_IP_WWPN_LEN);
		/* Copy 23 least significant bits from the IP Address */
		*t = (*t >> 23) << 23;	/*!< Clear last 23 bits from wwpn */
		*t |= ((ip->daddr << 9) >> 9);	/*!< Copy 23 bits from */
						/* destination IP */
		arp->pid = bfa_os_hton3b(FC_BROADCAST_SERVER);
		ret = 0;
	}
	bfa_trc(&pri->ioc->bfa_fcs, ret);
	return ret;
}

/**
 * @brief Callback function called by fcxp to query an sge address
 * @param [in] bfad_fcxp  - Packet passed in as argument
 * @param [in] sgeid - index of the sge currently being queried
 * @retval dma address of the current sge being queried
 */
uint64_t
bfad_ip_hcb_fcxp_req_sge(void *bfad_fcxp, int sgeid)
{
	bfad_pkt_t     *pkt = (bfad_pkt_t *) bfad_fcxp;
	bfa_assert(sgeid == 0);
	return (uint64_t)(size_t)pkt->sge[sgeid].sg_addr;
}

/**
 * @brief Callback function called by fcxp to query an sge length
 * @param [in] bfad_fcxp  - Packet passed in as argument
 * @param [in] sgeid - index of the sge currently being queried
 * @retval length of the current sge being queried
 */
uint32_t
bfad_ip_hcb_fcxp_req_sgelen(void *bfad_fcxp, int sgeid)
{
	bfad_pkt_t     *pkt = (bfad_pkt_t *) bfad_fcxp;
	return (uint32_t) pkt->sge[sgeid].sg_len;
}

/**
 * @brief The function sends out the buffer on the FC port
 * @param[in] fchs - Pointer to FC Header
 * @param[in] skb  - skb currently being transmitted
 * @param[in] cts  - cts
 * @param[in] cbfn - Callback function, called after transmission is completed
 * @param[in] arp  - contains arp information
 * @param[in] pkt  - All information including sge for current packet to xmit
 * @retval none
 */
static void
bfad_ip_fcxp_send_req(fchs_t *fchs, struct sk_buff *skb, uint8_t cts,
		      bfa_cb_fcxp_send_t cbfn, bfad_iprp_t *arp,
		      bfad_pkt_t *pkt)
{
	bfad_ip_private_t *pri = netdev_priv(skb->dev);
	char		message[100];
	bfa_fcxp_t     *fcxp = NULL;
	bfa_rport_t    *rport = NULL;
	uint32_t	len, i;
	uint16_t	vf_id = pri->ioc->bfa_fcs.fabric.vf_id;
	uint8_t	        lp_tag = bfa_lps_get_tag(pri->ioc->bfa_fcs.fabric.lps);
	unsigned long   flags;

	bfa_trc(&pri->ioc->bfa_fcs, fchs->d_id);
	bfa_trc(&pri->ioc->bfa_fcs, vf_id);

	spin_lock_irqsave(&pri->ioc->bfad_lock, flags);

	/** Setup fcxp */
	fcxp = bfa_fcxp_alloc(pkt, &pri->ioc->bfa, pkt->nsges, 0,
				bfad_ip_hcb_fcxp_req_sge,    /* Callback func */
				bfad_ip_hcb_fcxp_req_sgelen, /* Callback func */
				NULL, NULL);
	if (fcxp == NULL) {
		spin_unlock_irqrestore(&pri->ioc->bfad_lock, flags);
		sprintf(message, "fcxp_send_req (%s) dropping packet.\n",
			skb->dev->name);
		bfa_log(&bfad_ip_logm, BFA_LOG_LINUX_MEM_ALLOC_FAILED, message);
		/** Unmap DMA memory */
		for (i = 0; i < pkt->nsges; i++)
			pci_unmap_single(pri->ioc->pcidev,
					 (dma_addr_t)(ssize_t)pkt->sge[i].
					 sg_addr, pkt->sge[i].sg_len,
					 PCI_DMA_TODEVICE);

		/** Free resources */
		dev_kfree_skb(pkt->skb_payload);
		kfree(pkt);
		return;
	}

	/** Send packet out via. fcxp */
	len = skb->len + ((skb->next) ? ((skb->next)->len) : 0);

	bfa_trc(&pri->ioc->bfa_fcs, len);
	if (arp->fcs_iprp) {
		rport = arp->fcs_iprp->rport->bfa_rport, vf_id =
			arp->fcs_iprp->rport->port->fabric->vf_id;
		lp_tag = arp->fcs_iprp->rport->port->lp_tag;
	}

	bfa_fcxp_send(fcxp, rport, vf_id, lp_tag, BFA_FALSE, FC_CLASS_3, len,
			fchs, cbfn, pkt, 0, 0);

	spin_unlock_irqrestore(&pri->ioc->bfad_lock, flags);

}

/**
 * @brief callback function called after data is sent
 * @param[in] bfad       - bfad
 * @param[in] fcxp       - fcxp to be freed
 * @param[in] cbarg      - pkt came in as argument after transmit
 * @param[in] req_status - Status of the request completion
 * @param[in] rsp_len    - rsp_len not used in ipfc
 * @param[in] resid_len  - resid_len not used in ipfc
 * @param[in] rsp_fchs   - rsp_fchs not used in ipfc
 * @return - none
 */
void
bfad_ip_send_complete(void *fcsarg, bfa_fcxp_t *fcxp, void *cbarg,
		      bfa_status_t req_status, uint32_t rsp_len,
		      uint32_t resid_len, fchs_t *rsp_fchs)
{
	bfad_pkt_t     *pkt = (bfad_pkt_t *) cbarg;

	bfad_ip_private_t *pri = netdev_priv(pkt->dev);
	struct bfad_s  *bfad = pri->ioc;
	int	i;

#ifdef DEBUG
	printk(KERN_INFO "IP Send callback function called\n");
#endif
	bfa_trc(&bfad->bfa_fcs, req_status);

	/** update transmit statistics */
	pri->stats.tx_packets++;
	pri->stats.tx_bytes += pkt->buflen;

#ifdef USE_MULTIPLE_SGES
	if (pkt->fc_type == BFAD_IP_FC4_TYPE) {
		bfad->ipfc->buffer_in++;
		if (bfad->ipfc->buffer_in >= BFAD_IP_MAX_TX_BUFS)
			bfad->ipfc->buffer_in = 0;
	}
	bfa_trc(&bfad->bfa_fcs, bfad->ipfc->buffer_in);
	bfa_trc(&bfad->bfa_fcs, bfad->ipfc->buffer_out);
#endif

	/** Unmap DMA memory */
	for (i = 0; i < pkt->nsges; i++)
		pci_unmap_single(pri->ioc->pcidev,
				 (dma_addr_t)(ssize_t)pkt->sge[i].sg_addr,
				 pkt->sge[i].sg_len, PCI_DMA_TODEVICE);

	/** Free resources */
	dev_kfree_skb(pkt->skb_payload);
	kfree(pkt);

}

/**
 * @brief function to update network header with destination wwpn
 * @param[in] dev - network device
 * @param[in] buf - buffer to be filled with network header
 * @param[in] wwpn- destination world wide port name
 * @return - none
 */
static void
bfad_ip_add_network_header(struct net_device *dev, uint8_t *buf,
			   bfad_ip_wwpn_t *wwpn)
{
	bfad_ip_private_t *pri = netdev_priv(dev);

	/** Copy the destination wwpn on the buffer */
	bfad_ip_net_hdr_t *net;
	net = (bfad_ip_net_hdr_t *) buf;
	memcpy((void *)net->dest_addr.raw, wwpn->raw, BFAD_IP_WWPN_LEN);
	memcpy((void *)net->src_addr.raw, (void *)pri->my_wwpn,
	       BFAD_IP_WWPN_LEN);
}

/**
 * @brief Add LLC/SNAP header RFC 4338 [4.6]
 * @param[in] buf 		- buffer to be filled with llc/snap header
 * @param[in] pkt_type 	- Packet type
 * @return - none
 */
static void
bfad_ip_add_snap_header(uint8_t *buf, BFAD_IP_PKT_TYPE pkt_type)
{
	bfad_ip_snap_hdr_t *snap;
	snap = (bfad_ip_snap_hdr_t *) buf;

	/** Fill the buffer fields [dsap/ssap/ctrl/oui] with values. RFC 4338 */
	snap->dsap = BFAD_IP_LLC_SNAP_DSAP;
	snap->ssap = BFAD_IP_LLC_SNAP_SSAP;
	snap->ctrl = BFAD_IP_LLC_SNAP_CTRL;
	snap->oui_hi = BFAD_IP_LLC_SNAP_OUI_HI;
	snap->oui_low = BFAD_IP_LLC_SNAP_OUT_LOW;
	/** Assign pid which is the packet type */
	snap->pid = bfa_os_htons(pkt_type);
}

/**
 * @brief add fc/network/snap headers and ship the packet out
 * @param[in] skb	- packet coming from ulp
* @param[in] fc_type	- Type of the packet eg: ELS/IPoFC
 * @param[in] arp	- local arp
 * @param[in] pkt_type	- packet type
 * @return - none
 */
void
bfad_ip_add_headers_ship_out(struct sk_buff *skb, int fc_type,
			     bfad_iprp_t *arp, BFAD_IP_PKT_TYPE pkt_type)
{
	uint8_t		*ptr;
	struct sk_buff *header_skb = NULL;
	bfad_ip_private_t *private = netdev_priv(skb->dev);
	bfad_pkt_t     *pkt = NULL;
	struct bfad_s  *ioc = private->ioc;
	int		nsges = 0;
	fchs_t		bfad_ip_fchs;

	bfa_trc(&private->ioc->bfa_fcs, fc_type);

	if (fc_type == BFAD_IP_FC4_TYPE) {

#if 0
		bfa_trc(&private->ioc->bfa_fcs, arp->pid);
		/** If the packet is a uni-cast packet, login if not done yet */
		if (arp->pid != bfa_os_hton3b(FC_BROADCAST_SERVER)) {
			bfa_trc(&private->ioc->bfa_fcs, arp->pid);
			if (arp->state != BFAD_IP_ARP_STATE_LOGIN) {
				bfa_trc(&private->ioc->bfa_fcs, arp->state);
				bfa_assert(arp->fcs_iprp != NULL);
				bfa_fcs_iprp_login(arp->fcs_iprp->rport->port,
						   (wwn_t *)arp->wwpn.raw,
						   arp->pid, arp->fcs_iprp,
						   arp);

				dev_kfree_skb(skb);
				bfa_trc(&private->ioc->bfa_fcs, arp->pid);
				return;
			}
		}
#endif

		bfa_trc(&private->ioc->bfa_fcs, arp->pid);

#ifdef USE_MULTIPLE_SGES
		uint8_t	index;
		/*
		 * get one of the transmit buffers to add headers
		 */
		if (ioc->ipfc->buffer_in != ioc->ipfc->buffer_out) {

			spin_lock_irqsave(&private->ioc->bfad_lock, flags);
			if (ioc->ipfc->buffer_out >= BFAD_IP_MAX_TX_BUFS)
				ioc->ipfc->buffer_out = 0;
			index = ioc->ipfc->buffer_out;
			ioc->ipfc->buffer_out++;
			spin_unlock_irqrestore(&private->ioc->bfad_lock, flags);
			bfa_trc(&private->ioc->bfa_fcs, arp->pid);
			pkt = ioc->ipfc->send_pkt[index];

			pkt->skb->dev = skb->dev;

			header_skb = (struct sk_buff *)
				(ioc->ipfc->send_pkt[index]->skb);
			header_skb->next = skb;
			header_skb->dev = skb->dev;
			ptr = (uint8_t *) header_skb->data;
			skb->next = NULL;
		} else {
			/*
			 * could not get tx buffer, transmitter may have choked
			 *!!
			 */
			private->stats.tx_errors++;
			private->stats.tx_fifo_errors++;

			bfa_trc(&private->ioc->bfa_fcs,
				private->stats.tx_errors);
			bfa_trc(&private->ioc->bfa_fcs,
				private->stats.tx_fifo_errors);

			netif_stop_queue(skb->dev);
			dev_kfree_skb(skb);
			return;
		}
#else
		pkt = kmalloc(sizeof(bfad_pkt_t), GFP_ATOMIC);
		if (pkt == NULL) {
			private->stats.tx_errors++;
			private->stats.tx_fifo_errors++;
			netif_stop_queue(skb->dev);
			dev_kfree_skb(skb);
			return;
		}
#endif
	} else {
		bfa_trc(&private->ioc->bfa_fcs, arp->pid);
		pkt = kmalloc(sizeof(bfad_pkt_t), GFP_ATOMIC);
		pkt->skb = skb;
		pkt->buflen = skb->len;
		skb->next = NULL;
		ptr = skb->data;
	}

	pkt->fc_type = fc_type;

	/**
	 * setup fc header
	 */
	private->my_pid =
		bfa_fcs_lport_get_fcid(bfa_fcs_get_base_port(&ioc->bfa_fcs));

	bfad_ip_init_fchs(&bfad_ip_fchs, fc_type);

	bfad_ip_fchs.d_id = bfa_os_hton3b(arp->pid);
	bfad_ip_fchs.s_id = private->my_pid;

	skb->dev->trans_start = jiffies;
	pkt->dev = skb->dev;

	/** Set up sges */
	pkt->nsges = 0;
#ifdef USE_MULTIPLE_SGES
	if (fc_type == BFAD_IP_FC4_TYPE) {
		bfad_ip_add_network_header(header_skb->dev, ptr, &arp->wwpn);
		bfad_ip_add_snap_header((ptr + sizeof(bfad_ip_net_hdr_t)),
					pkt_type);
	}

	if (header_skb &&
	    (nsges =
		bfad_ip_map_dma(ioc, header_skb, pkt, 0,
				PCI_DMA_TODEVICE)) == -1) {
		bfa_trc(&private->ioc->bfa_fcs, arp->pid);
		return;
	}

	nsges =
	     bfad_ip_map_dma(ioc, skb, pkt, nsges, PCI_DMA_TODEVICE);
	if (nsges == -1) {
		bfa_trc(&private->ioc->bfa_fcs, arp->pid);
		return;
	}
#else /*  USE SINGLE SGE as a TEMP SOLUTION */
	if (fc_type == BFAD_IP_FC4_TYPE) {

		struct sk_buff *out_skb;
		/** Allocate SKB */
		out_skb =
		     dev_alloc_skb(NET_SNAP_HDR_SZ + skb->len + 2);
		if (out_skb == NULL) {
			dev_kfree_skb(skb);
			return;
		}

		skb_reserve(out_skb, 2);

		skb_put(out_skb, NET_SNAP_HDR_SZ + skb->len);
		out_skb->next = NULL;
		out_skb->dev = skb->dev;

		bfad_ip_add_network_header(out_skb->dev, out_skb->data,
					   &arp->wwpn);
		bfad_ip_add_snap_header((out_skb->data +
					 sizeof(bfad_ip_net_hdr_t)), pkt_type);
		memcpy((out_skb->data + NET_SNAP_HDR_SZ), skb->data, skb->len);

		nsges =
		     bfad_ip_map_dma(ioc, out_skb, pkt, 0, PCI_DMA_TODEVICE);
		if (nsges == -1) {
			bfa_trc(&private->ioc->bfa_fcs, arp->pid);
			return;
		}

		pkt->buflen = out_skb->len;

		header_skb = out_skb;
		dev_kfree_skb(skb);

		pkt->skb = header_skb;
		pkt->skb_payload = header_skb;

	} else {
		nsges =
		     bfad_ip_map_dma(ioc, skb, pkt, nsges, PCI_DMA_TODEVICE);
		if (nsges == -1) {
			bfa_trc(&private->ioc->bfa_fcs, arp->pid);
			return;
		}

		pkt->skb_payload = skb;
	}
#endif

	pkt->ioc = ioc;
	bfa_trc(&private->ioc->bfa_fcs, arp->pid);

	/** Send the packet out */
	bfad_ip_fcxp_send_req(&bfad_ip_fchs,
(fc_type == BFAD_IP_FC4_TYPE) ? header_skb : skb,
			      1, (bfa_cb_fcxp_send_t) bfad_ip_send_complete,
			      arp, pkt);

}

/**
 * @brief Given packet, determines what type of packet it is i.e arp/ipv4/ipv6
 * @param[in] buf - in coming buffer starting with FC Header
 * @return - BFAD_IP_PKT_TYPE value
 */
static	BFAD_IP_PKT_TYPE
bfad_ip_packet_type_in(uint8_t *buf)
{
	struct arphdr  *arphdr;
	bfad_ip_snap_hdr_t *snap;
	BFAD_IP_PKT_TYPE pkt_type = BFAD_IP_PKT_TYPE_UNKNOWN;

	/** Find the packet type from the snap header (pid) field */
	snap = (bfad_ip_snap_hdr_t *) (buf + sizeof(fchs_t) +
				       sizeof(bfad_ip_net_hdr_t));
	switch (bfa_os_htons(snap->pid)) {
	case ETH_P_IP:
		pkt_type = BFAD_IP_PKT_TYPE_IPV4;
		break;

	case ETH_P_IPV6:
		pkt_type = BFAD_IP_PKT_TYPE_IPV6;
		break;

	case ETH_P_ARP:
		arphdr = (struct arphdr *)((uint8_t *) snap +
					   sizeof(bfad_ip_snap_hdr_t));
		if (arphdr->ar_hrd == bfa_os_htons(BFAD_IP_FCARP_HW_TYPE)) {
			pkt_type = BFAD_IP_PKT_TYPE_FCARP;
		} else {
			pkt_type =
				BFAD_IP_PKT_TYPE_ARP_REQ - ARPOP_REQUEST +
				bfa_os_htons(arphdr->ar_op);
		}
		break;
	default:
		bfa_assert(0);
	}

	return pkt_type;
}

/**
 * @brief Given packet, determines what type of packet it is i.e arp/ipv4/ipv6
 * @param[in] buf - buffer from ulp starting with ethernet header
 * @return - BFAD_IP_PKT_TYPE value
 */
static	BFAD_IP_PKT_TYPE
bfad_ip_packet_type_out(struct net_device *dev, unsigned char *buf,
	uint32_t *vlan)
{
	struct ethhdr  *eth = (struct ethhdr *)buf;
	bfad_iprp_t     arpc;
	uint32_t	*ip;

	if (bfa_os_htons(eth->h_proto) == ETH_P_8021Q) {
		*vlan = 4;
		/* pid is a place holder for vlan tag */
		arpc.vlan_tag = *(uint16_t *)(buf + sizeof (struct ethhdr));
		eth = (struct ethhdr *)(buf + 4);
	} else
		*vlan = 0;

	/** From the ethernet protocol field, findout the packet type */
	if (bfa_os_htons(eth->h_proto) == ETH_P_ARP) {
		struct arphdr  *arp =
			(struct arphdr *)((unsigned char *)eth +
			sizeof(struct ethhdr));
		if (bfa_os_htons(arp->ar_op) == ARPOP_REQUEST) {
			if (*vlan) {
				/* save vlan tag of the source ip */
				ip = (uint32_t *)((unsigned char *)arp +
						sizeof (struct arphdr) + 6);
				arpc.ip_addr_type = BFAD_IP_ADDR_TYPE_IPV4;
				arpc.ip_addr.ipv4_addr.s_addr = *ip;
				bfad_ip_update_arp_cache(dev, &arpc,
						BFAD_IP_ARP_STATE_INIT);
			}

			return BFAD_IP_PKT_TYPE_ARP_REQ;
		} else if (bfa_os_htons(arp->ar_op) == ARPOP_REPLY)
			return BFAD_IP_PKT_TYPE_ARP_RES;
		else if (bfa_os_htons(arp->ar_op) == ARPOP_RREQUEST)
			return BFAD_IP_PKT_TYPE_RARP_REQ;
		else if (bfa_os_htons(arp->ar_op) == ARPOP_RREPLY)
			return BFAD_IP_PKT_TYPE_RARP_RES;
		else if (bfa_os_htons(arp->ar_op) == ARPOP_InREQUEST)
			return BFAD_IP_PKT_TYPE_IARP_REQ;
		else if (bfa_os_htons(arp->ar_op) == ARPOP_InREPLY)
			return BFAD_IP_PKT_TYPE_IARP_RES;
		else if (bfa_os_htons(arp->ar_op) == ARPOP_NAK)
			return BFAD_IP_PKT_TYPE_ARP_NAK;
	} else if (bfa_os_htons(eth->h_proto) == ETH_P_IP) {
		return BFAD_IP_PKT_TYPE_IPV4;
	} else if (bfa_os_htons(eth->h_proto) == ETH_P_IPV6) {
		return BFAD_IP_PKT_TYPE_IPV6;
	}

	return BFAD_IP_PKT_TYPE_UNKNOWN;
}

/**
 * @brief function gets the WWN type
 * @param[in] wwn - world wide port name
 * @return - BFAD_IP_WWN_TYPE value
 */
BFAD_IP_WWN_TYPE
bfad_ip_get_wwn_type(bfad_ip_wwpn_t *wwn)
{
	/** The first 4 bits of wwn would tell which format the wwn is in */
	if ((wwn->raw[0] & 0xF0) == 0x10)
		return BFAD_IP_WWN_TYPE_FMT1;
	else if ((wwn->raw[0] & 0xF0) == 0x20)
		return BFAD_IP_WWN_TYPE_FMT2;
	else if ((wwn->raw[0] & 0xF0) == 0x50)
		return BFAD_IP_WWN_TYPE_FMT5;
	else if ((wwn->raw[0] & 0xF0) == 0xC0)
		return BFAD_IP_WWN_TYPE_FMTCF;

	return BFAD_IP_WWN_TYPE_UNKNOWN;
}

/**
 * @brief bfad_ip_send
 *  Function strips ethernet header, adds network header&LLC/SNAP header
 * 	before shipping out on the FC port via. FCXP driver
 * @param[in] skb - In coming packet from ulp
 * @param[in] dev - Network device
 * @retval  0 on success
 * @retval -1 on failure
 */
static int
bfad_ip_send(struct sk_buff *skb, struct net_device *dev)
{
	struct arphdr  *earp;
	bfad_iprp_t     arp, *arp_res;
	int		res, out_pkt_type = 0;
	BFAD_IP_PKT_TYPE type;
	bfad_ip_private_t *private = netdev_priv(dev);
	unsigned long   flags;
	struct bfad_port_s *port = &private->ioc->pport;
	uint32_t	vlan = 0;

	if (!(port->ipfc_flags & BFAD_IP_IFF_ONLINE)) {
		skb->dev->trans_start = jiffies;
		dev_kfree_skb(skb);
		private->stats.tx_dropped++;
		return 0;
	}

	type = bfad_ip_packet_type_out(dev, skb->data, &vlan);

#ifdef DEBUG
	uint8_t	*p = skb->data, i;
	printk("[%d]Packet to transmit type: %d len: %d proto: %x\n",
		vlan, type, skb->len, skb->protocol);
	for (i = 0; i < 20; i++)
		printk("%x ", p[i]);
	printk("\n");
#endif
	/**
	 * Translate the destination MAC address to actual WWN
	 * If we get uni-cast packet, surely the translation must succeed as
	 * the ARP cache of ULP & the local ARP cache will be in sync.
	 */
	spin_lock_irqsave(&private->ioc->bfad_lock, flags);
	res = bfad_ip_out_translate_dest_broadcast(type, skb, &arp);
	spin_unlock_irqrestore(&private->ioc->bfad_lock, flags);
	if (res == -1) {
		spin_lock_irqsave(&private->ioc->bfad_lock, flags);
		res = bfad_ip_wwpn_lookup(skb->dev,
					  ((struct ethhdr *)(skb->data))->
					  h_dest, &arp_res);
		spin_unlock_irqrestore(&private->ioc->bfad_lock, flags);

		if (res != -1) {
			if (arp_res == NULL)
				res = -1;
			else
				memcpy(&arp, arp_res, sizeof(bfad_iprp_t));
		}
	}

	if (type == BFAD_IP_PKT_TYPE_ARP_REQ ||
	    type == BFAD_IP_PKT_TYPE_ARP_RES || res == -1) {

		bfa_trc(&private->ioc->bfa_fcs, type);
		/**
		 * If the packet from ulp is an ARP REQ, send FCARP
		 * @img arp_out.jpg
		 */
		if (type == BFAD_IP_PKT_TYPE_ARP_REQ || res == -1) {
			bfad_ip_send_fcarp(skb->dev,
				BFAD_IP_FCARP_CODE_REQ,
				(uint32_t *) ((uint8_t *)
				skb->data + sizeof(struct ethhdr) + vlan +
				sizeof(struct arphdr) + 6),
				(uint32_t *) ((uint8_t *)
				skb->data + sizeof(struct ethhdr) + vlan +
				sizeof(struct arphdr) + 16),
				NULL, NULL);

			private->drv_stats.arp_sent++;
			if (res == -1)
				return 0;
		} else
			private->drv_stats.arp_reply_sent++;

		if (bfad_ip_get_wwn_type
		    ((bfad_ip_wwpn_t *)&private->my_wwpn)
		    != BFAD_IP_WWN_TYPE_FMT1) {
			bfa_trc(&private->ioc->bfa_fcs, type);
			/**
			 * Drop the packet as we can't advertise our
			 * full PWWN in ARP packet. In this case
			 * when fc arp response comes
			 * back, we need to generate ARP response to ULP
			 * If we have non FMT1 WWN, we can't
			 * interoperate with RFC 2625 clients.
			 */
			dev_kfree_skb(skb);
			return 0;
		} else {
			arp.pid = bfa_os_hton3b(arp.pid);
			skb_pull(skb, vlan + (sizeof(struct ethhdr)));
			earp = (struct arphdr *)((uint8_t *) skb->data);
			if (type == BFAD_IP_PKT_TYPE_ARP_RES)
				memcpy(((uint8_t *) earp +
					sizeof(struct arphdr) + 10),
				       &arp.wwpn.raw[2], 6);
			*(uint16_t *) (skb->data) =
				bfa_os_htons(0x0006);
			*(uint16_t *) (skb->data + 2) =
				bfa_os_htons(0x0800);
			bfa_trc(&private->ioc->bfa_fcs, type);
			/* ARP request: Fall through ... */
		}
	} else {
		arp.pid = bfa_os_hton3b(arp.pid);
		arp.state = BFAD_IP_ARP_STATE_LOGIN;
		skb_pull(skb, vlan + (sizeof(struct ethhdr)));
	}

	/**
	 *  If the packet is a IPV6 Discovery packet,
	 *  we need to alter the packet to change the link-layer address
	 *  length from 6 to 8 bytes and substibute acutal address
	 *  @img ipv6_out.jpg
	 */
	if (type == BFAD_IP_PKT_TYPE_IPV6) {
#ifndef IPV6_SUPPORT
		bfa_trc(&private->ioc->bfa_fcs, type);
		skb->dev->trans_start = jiffies;
		dev_kfree_skb(skb);
		return 0;
#else
		struct ipv6hdr *ip6h;
		int	options_offset, options_length;
		ip6h = (struct ipv6hdr *)(skb->data + sizeof(struct ethhdr));
		if (bfad_ip_ipv6_pkt_has_ll_address
		    (ip6h, &options_offset, &options_length)) {

			struct sk_buff *skb_out;
			skb_out = dev_alloc_skb(skb->len);
			skb_put(skb_out, skb->len);
			skb_out->dev = skb->dev;
			skb_out->len = skb->len;
			bfad_ip_copy_ipv6(skb_out->dev,
					  (uint8_t *) (skb_out->data +
						       sizeof(struct ethhdr)),
					  (uint8_t *) ip6h,
					  BFAD_IP_PACKET_DIR_OUT, 0,
					  options_offset, options_length);

			dev_kfree_skb(skb);
			skb = skb_out;
		}
#endif
	}

	/*
	 * IPV4 packets do not required altering them.
	 * @img ipv4_out.jpg
	 */

	if (type >= BFAD_IP_PKT_TYPE_ARP_REQ && type <= BFAD_IP_PKT_TYPE_FCARP)
		out_pkt_type = ETH_P_ARP;
	else if (type == BFAD_IP_PKT_TYPE_IPV4 ||
		 type == BFAD_IP_PKT_TYPE_8021Q)
		out_pkt_type = ETH_P_IP;
	else if (type == BFAD_IP_PKT_TYPE_IPV6)
		out_pkt_type = ETH_P_IPV6;

	bfa_trc(&private->ioc->bfa_fcs, out_pkt_type);

	/**
	 * If we reached here, we have ARP resolved and ready to ship the
	 * packet
	 */
	bfad_ip_add_headers_ship_out(skb, BFAD_IP_FC4_TYPE, &arp, out_pkt_type);

	return 0;
}

/**
 * @brief Process FC ARP packet
 * @param[in] dev - net device
 * @param[in] buf - buffer pointer starting at fcarp packet
 * @return - void
 */
static int
bfad_ip_process_fcarp(struct net_device *dev, uint8_t *buf)
{
	bfad_ip_private_t *private = netdev_priv(dev);
	bfad_ip_fcarp_t *fcarp;
	bfad_iprp_t     arp;

	fcarp = (bfad_ip_fcarp_t *) buf;

	bfa_trc(&private->ioc->bfa_fcs, fcarp->src_ip.s_addr);

	/** If the packet is a request packet, send the reply */
	if (fcarp->opcode == bfa_os_htons(BFAD_IP_FCARP_CODE_REQ)) {
		bfad_ip_send_fcarp(dev, BFAD_IP_FCARP_CODE_RES,
				   &fcarp->src_ip.s_addr,
				   &fcarp->dest_ip.s_addr,
				   fcarp->src_hw_addr.pname.raw, fcarp);
	}
	/** If the packet is a response packet, process and update local ARP */
	else if (fcarp->opcode == bfa_os_htons(BFAD_IP_FCARP_CODE_RES)) {

		/* process and update local ARP cache */
		memset(&arp, 0x00, sizeof(bfad_iprp_t));
		memcpy(arp.wwpn.raw, fcarp->src_hw_addr.pname.raw,
		       BFAD_IP_WWPN_LEN);
		arp.pid = bfa_os_ntoh3b(fcarp->dest_hw_addr.pid);
		arp.ip_addr_type = BFAD_IP_ADDR_TYPE_IPV4;
		arp.ip_addr.ipv4_addr.s_addr =
			bfa_os_ntohl(fcarp->dest_ip.s_addr);
		bfad_ip_update_arp_cache(dev, &arp, BFAD_IP_ARP_STATE_INIT);

	}

	return 1;
}

/**
 * @brief Process ARP packet ( backward compatibility with RFC 2625 )
 * @param[in] skb  - sk_buff with data pointer pointing to arp packet
 * @param[in] s_id - source fc id, used to add arp cache entry
 * @retval 0 - if packet is not forwarded to upper layer
 * @retval 1 - if packet is forwarded to upper layer
 * @img arp_in.jpg
 */
static int
bfad_ip_process_arp(struct sk_buff *skb, int s_id)
{
	struct arphdr  *arp;
	struct ethhdr  *eth;
	bfad_iprp_t     arpc, *arp_res, arpin;
	fchs_t		*fchs;
	bfad_ip_net_hdr_t *neth;
	bfad_ip_snap_hdr_t *snap;
	bfad_ip_private_t *pri = netdev_priv(skb->dev);
	struct bfad_s	*bfad = pri->ioc;

	/** If we received ARP packet, and our' is not NAA1 - ignore packet */
	if (bfad_ip_get_wwn_type((bfad_ip_wwpn_t *) pri->my_wwpn)
	    != BFAD_IP_WWN_TYPE_FMT1) {
		bfa_trc(&pri->ioc->bfa_fcs, 0);
		return 0;
	}

	fchs = (fchs_t *) skb->data;
	neth = (bfad_ip_net_hdr_t *) ((uint8_t *) fchs + sizeof(fchs_t));
	snap = (bfad_ip_snap_hdr_t *) ((uint8_t *) neth +
				       sizeof(bfad_ip_net_hdr_t));
	arp = (struct arphdr *)((uint8_t *) snap + sizeof(bfad_ip_snap_hdr_t));

	/* check if the dest ip has vlan tag */
	arpin.ip_addr.ipv4_addr.s_addr =
		*(uint32_t *) ((uint8_t *) arp + sizeof(struct arphdr) + 16);
	arp_res = bfad_ip_pid_hash_lookup(pri->hash_pid_table, &arpin);
	if (arp_res)
		arp_res->timestamp = jiffies;

	/**
	 * If in coming packet is ARP, Update our cache, send out FARP request
	 * to resolve WWN to PID
	 */
	if (arp->ar_op == bfa_os_htons(ARPOP_REQUEST) ||
	    arp->ar_op == bfa_os_htons(ARPOP_REPLY)) {
		/* process and update local ARP cache */
		memset(&arpc, 0x00, sizeof(bfad_iprp_t));
		memcpy(arpc.wwpn.raw, neth->src_addr.raw, BFAD_IP_WWPN_LEN);

		arpc.pid = s_id;
		arpc.ip_addr_type = BFAD_IP_ADDR_TYPE_IPV4;
		arpc.ip_addr.ipv4_addr.s_addr =
			*(uint32_t *) ((uint8_t *) arp + sizeof(struct arphdr) +
				       6);

		bfad_ip_update_arp_cache(skb->dev, &arpc,
					 BFAD_IP_ARP_STATE_INIT);

		/* Send out FARP request to resolve WWPN to PID */
		bfad_ip_send_farp(BFAD_IP_FARP_CMD_REQ, skb->dev, 0,
			&arpc.wwpn, NULL, &arpc, bfad->pport.fcs_port);
	}

	skb_pull(skb,
		 (sizeof(fchs_t) + sizeof(bfad_ip_net_hdr_t) +
		  sizeof(bfad_ip_snap_hdr_t)) - sizeof(struct ethhdr) -
		 ((arp_res)?4:0));

	eth = (struct ethhdr *)(skb->data);
	memset(eth, 0x00, sizeof(struct ethhdr));

	memcpy(eth->h_dest, skb->dev->dev_addr, ETH_ALEN);
	memcpy(&eth->h_source[2], (uint8_t *) arp + sizeof(struct arphdr) + 6,
	       sizeof(struct in_addr));

	memset(((uint8_t *) arp + sizeof(struct arphdr)), 0x00, 2);
	memcpy(((uint8_t *) arp + sizeof(struct arphdr) + 2),
	       (uint8_t *) arp + sizeof(struct arphdr) + 6,
	       sizeof(struct in_addr));

	if (arp_res) {
		eth->h_proto = bfa_os_htons(ETH_P_8021Q);
		/* add vlan tag */
		*(uint16_t *)((unsigned char *)eth +
			sizeof(struct ethhdr)) = arp_res->vlan_tag;
		*(uint16_t *)((unsigned char *)eth +
			sizeof(struct ethhdr) + 2) = bfa_os_htons(ETH_P_ARP);
	} else
		eth->h_proto = bfa_os_htons(ETH_P_ARP);

	arp->ar_hrd = bfa_os_htons(ARPHRD_ETHER);

	skb->protocol = eth_type_trans(skb, skb->dev);
	skb->pkt_type = PACKET_HOST;

	netif_rx(skb);
	skb->dev->last_rx = jiffies;

	return 1;
}

/**
 * @brief Process IPv4 Packet
 * @param[in] skb - sk_buff
 * @retval 1 if skb is deleted inside the function
 * @img ipv4_in.jpg
 */
static int
bfad_ip_process_ipv4(struct sk_buff *skb, int pktlen)
{
	fchs_t	*fchs;
	bfad_ip_net_hdr_t *neth;
	bfad_ip_snap_hdr_t *snap;
	struct iphdr   *ip4h;
	struct ethhdr  *eth;
	bfad_iprp_t     *arp_res, arp;
	bfad_ip_private_t *pri = netdev_priv(skb->dev);

	fchs = (fchs_t *)skb->data;
	neth = (bfad_ip_net_hdr_t *) ((uint8_t *) fchs + sizeof(fchs_t));
	snap = (bfad_ip_snap_hdr_t *) ((uint8_t *) neth +
				       sizeof(bfad_ip_net_hdr_t));
	ip4h = (struct iphdr *)((uint8_t *) snap + sizeof(bfad_ip_snap_hdr_t));

	/* check if the dest ip has vlan tag */
	arp.ip_addr.ipv4_addr.s_addr = ip4h->daddr;
	arp_res = bfad_ip_pid_hash_lookup(pri->hash_pid_table, &arp);
	if (arp_res)
		arp_res->timestamp = jiffies;

	skb_pull(skb,
		 (sizeof(fchs_t) + sizeof(bfad_ip_net_hdr_t) +
		  sizeof(bfad_ip_snap_hdr_t)) - sizeof(struct ethhdr) -
		 ((arp_res)?4:0));

	/** Add ethernet header with source MAC as the ip address itself */
	eth = (struct ethhdr *)(skb->data);
	memset((void *)eth, 0x00, sizeof(struct ethhdr));
	memcpy(eth->h_dest, skb->dev->dev_addr, ETH_ALEN);
	memcpy(&eth->h_source[2], &ip4h->saddr, 4);

	if(arp_res) {
		eth->h_proto = bfa_os_htons(ETH_P_8021Q);
		/* add vlan tag */
		*(uint16_t *)((unsigned char *)eth +
			sizeof(struct ethhdr)) = arp_res->vlan_tag;
		*(uint16_t *)((unsigned char *)eth +
			sizeof(struct ethhdr) + 2) = bfa_os_htons(ETH_P_IP);
	} else
		eth->h_proto = bfa_os_htons(ETH_P_IP);

	skb->protocol = eth_type_trans(skb, skb->dev);

	/** Send the packet to upper layer stack */
	netif_rx(skb);
	skb->dev->last_rx = jiffies;

	return 1;
}

/**
 * @brief Process FARP packet ( backward compatibility with RFC 2625 )
 * @param[in] dev - net device
 * @param[in] buf - buffer starting with FC Header
 * @return - none
 */
void
bfad_ip_process_farp(void *in_lport, void *in_buf)
{
	struct bfad_s	*bfad =
			((struct bfa_fcs_lport_s *)in_lport)->bfad_port->bfad;
	struct net_device	*dev = bfad->pport.ipfc_port->dev;
	fchs_t	*fchs = (fchs_t *)in_buf;
	bfad_ip_els_farp_t *farp =
		(bfad_ip_els_farp_t *) (in_buf + sizeof(fchs_t));
	bfad_ip_private_t *pri = netdev_priv(dev);
	bfad_iprp_t     arp;

	bfa_trc(&pri->ioc->bfa_fcs, farp->code);
	bfa_trc(&pri->ioc->bfa_fcs, fchs->s_id);

	/**
	 * If the packet is FARP reqest, send the FARP reply
	 */
	if (farp->code == BFAD_IP_FARP_CMD_REQ &&
	    (memcmp(farp->d_wwpn.raw, pri->my_wwpn, BFAD_IP_WWPN_LEN) ==
		0)) {
		memset(&arp, 0, sizeof(bfad_iprp_t));
		arp.pid = fchs->s_id;
		bfad_ip_send_farp(BFAD_IP_FARP_CMD_REPLY, dev, fchs,
				&farp->d_wwpn, (uint8_t *) farp, &arp,
				in_lport);

	} /* If the packet is FARP reply, process and update our local ARP */
	else if (farp->code == BFAD_IP_FARP_CMD_REPLY) {
		bfad_iprp_t *res;
		memcpy((void *)arp.wwpn.raw, farp->d_wwpn.raw,
		       BFAD_IP_WWPN_LEN);
		res = bfad_ip_pid_hash_lookup(pri->hash_pid_table, &arp);
		if (res) {
		    arp.pid = fchs->s_id;
		    bfad_ip_update_arp_cache(dev, &arp, BFAD_IP_ARP_STATE_INIT);
		}
	}
}

/**
 * @brief function checks if the IPv6 has link layer address in the options
 * @param [in] ip6h - pointer to the ipv6 header
 * @param [out] options_offset - offset to the options from ipv6 header
 * @param [out] options_length - length of the options
 * @retval 0 - if the packet does not have link layer address
 * @retval 1 - if the packet has link layer address
 */
#ifdef IPV6_SUPPORT
int
bfad_ip_ipv6_pkt_has_ll_address(struct ipv6hdr *ip6h, int *options_offset,
				int *options_length)
{
	struct nd_msg  *msg;
	struct nd_opt_hdr *opt_hdr;
	int		lrem = bfa_os_htons(ip6h->payload_len);
	uint8_t		*iptr = (uint8_t *) ip6h;

	msg = (struct nd_msg *)((uint8_t *) ip6h + sizeof(struct ipv6hdr));

	if (ip6h->nexthdr != IPPROTO_ICMPV6)
		return 0;

	*options_offset = sizeof(struct ipv6hdr);
	if (msg->icmph.icmp6_type == NDISC_NEIGHBOUR_SOLICITATION ||
	    msg->icmph.icmp6_type == NDISC_NEIGHBOUR_ADVERTISEMENT) {
		lrem -= (sizeof(struct icmp6hdr) + sizeof(struct in6_addr));
		*options_offset +=
			(sizeof(struct icmp6hdr) + sizeof(struct in6_addr));
	} else if (msg->icmph.icmp6_type == NDISC_ROUTER_SOLICITATION) {
		lrem -= sizeof(struct icmp6hdr);
		*options_offset += sizeof(struct icmp6hdr);
	} else if (msg->icmph.icmp6_type == NDISC_ROUTER_ADVERTISEMENT) {
		lrem -= (sizeof(struct icmp6hdr) + 2*sizeof(uint32_t));
		*options_offset +=
			(sizeof(struct icmp6hdr) + 2*sizeof(uint32_t));
	} else if (msg->icmph.icmp6_type == NDISC_REDIRECT) {
		lrem -= (sizeof(struct icmp6hdr) + 2*sizeof(struct in6_addr));
		*options_offset +=
			(sizeof(struct icmp6hdr) + 2*sizeof(struct in6_addr));
	} else
		return 0;

	*options_length = lrem;

	if (!lrem)
		return 0;

	iptr += *options_offset;
	while (lrem) {
		opt_hdr = (struct nd_opt_hdr *)iptr;
		if (opt_hdr->nd_opt_type == ND_OPT_TARGET_LL_ADDR ||
		    opt_hdr->nd_opt_type == ND_OPT_SOURCE_LL_ADDR) {
			return 1;
		} else {
			/*
			 * nd_opt_len is in multiples of 8 bytes
			 */
			lrem -= (sizeof(struct nd_opt_hdr) +
				 opt_hdr->nd_opt_len*8);
		}
	}

	return 0;
}

/**
 * @brief Copy IPv6 Packet into optr from iptr. Depending on the direction
 * of the packet, the packet may be expanded/compressed if it is a
 * neighbour-advertisement or neighbour-solicitation packet
 * @param[in]  dev 	- network device
 * @param[out] optr - points to buffer copied to
 * @param[in]  iptr - points to buffer copied from
 * @param[in]  dir 	- direction 0-in 1-out
 * @param[in]  pid 	- fcid [ will be 0 for out going packets ]
 * @param[in]  options_offset - offset to options field from ipv6 header
 * @param[in]  lrem - length of the options fields
 * @return - none
 */
void
bfad_ip_copy_ipv6(struct net_device *dev, uint8_t *optr, uint8_t *iptr,
		  int dir, int pid, int options_offset, int lrem)
{
	bfad_iprp_t    *res = NULL, arp;
	struct ipv6hdr *ip6h = (struct ipv6hdr *)iptr;
	struct icmp6hdr *icmp6h;
	struct nd_opt_hdr *opt_hdr;
	bfad_ip_private_t *pri = netdev_priv(dev);
	int		index;
	uint8_t		found = 0;
	uint16_t	len;

	bfa_trc(&pri->ioc->bfa_fcs, dir);
	bfa_trc(&pri->ioc->bfa_fcs, pid);

	/**
	 * The routine will process the packet only if the ->nexthdr is ICMPV6
	 * and if the type is NEIGHBOR/ROUTER SOLICITATION/ADVERTISEMENT
	 * or if the type is REDIRECT message.
	 * If the packet direction is in coming, the routine replaces 8 byte
	 * hw address to 6 byte address with is locally generated.
	 * The the packet direction is out going, the routine replaces 6 byte
	 * hw address to 8 bytes which is the actual wwpn.
	 */

	/*
	 * Copy data until options
	 */
	memcpy(optr, (void *)ip6h, options_offset);
	optr += options_offset;
	iptr += options_offset;

	/*
	 * RFC 2461
	 * OPTION FORMAT
	 * <TYPE : 1BYTE> - <LENGTH : 1BYTE in terms 8 octets> -
	 * <VALUE of size LENGTH*8>
	 */
	while (lrem) {

		opt_hdr = (struct nd_opt_hdr *)iptr;
		if (opt_hdr->nd_opt_type == ND_OPT_TARGET_LL_ADDR ||
		    opt_hdr->nd_opt_type == ND_OPT_SOURCE_LL_ADDR) {

			/*
			 * Copy current option header
			 */
			memcpy(optr, iptr, sizeof(struct nd_opt_hdr));

			/*
			 * Advance pointers over option header
			 */
			lrem -= sizeof(struct nd_opt_hdr);
			iptr += sizeof(struct nd_opt_hdr);
			optr += sizeof(struct nd_opt_hdr);

			/*
			 * copy index
			 */
			if (dir == BFAD_IP_PACKET_DIR_IN) {
				if (memcmp
				    ((void *)iptr, pri->my_wwpn,
				     BFAD_IP_WWPN_LEN)
				    == 0) {	/* If this is my address, */
						/* substitute my mac addr */
					memset(optr, 0x00, 2);
					memcpy((optr + 2),
					       (void *)dev->dev_addr, ETH_ALEN);
				} else {	/* If the wwpn is peer', */
						/* substitute the locally */
						/* generated index */
					memcpy((void *)arp.wwpn.raw, iptr,
					       BFAD_IP_WWPN_LEN);
					arp.ip_addr_type =
						BFAD_IP_ADDR_TYPE_IPV6;
					arp.pid = pid;
					memcpy((void *)&arp.ip_addr.ipv6_addr,
					       (void *)&ip6h->saddr.s6_addr,
					       sizeof(struct in6_addr));

					/*
					 * If the pid does not exist,
					 * add it to local arp cache
					 */
					index = bfad_ip_update_arp_cache(dev,
						&arp, BFAD_IP_ARP_STATE_INIT);

					/*
					 * initialize first 4 bytes of the
					 * address and copy index
					 */
					memset(optr, 0x00, 4);
					memcpy((optr + 4), (void *)&index, 4);
				}
			} else {
				/* BFAD_IP_PACKET_DIR_OUT */
				if (memcmp
				    ((void *)iptr, dev->dev_addr, ETH_ALEN)
				    == 0) {
					/*
					 * If this is my address,
					 * substitute my wwpn
					 */
					memcpy(optr, (void *)pri->my_wwpn,
					       BFAD_IP_WWPN_LEN);
				}
				if (bfad_ip_wwpn_lookup(dev, iptr, &res) == 0) {
					memcpy(optr, (void *)res->wwpn.raw,
					       BFAD_IP_WWPN_LEN);
				}
			}

			/*
			 * Advance pointers over option value: nd_opt_len is in
			 * multiples of 8 bytes
			 */
			lrem -= (opt_hdr->nd_opt_len*8);
			iptr += (opt_hdr->nd_opt_len*8);
			optr += (opt_hdr->nd_opt_len*8);

			found = 1;

		} else {

			/*
			 * Copy option header & values
			 */
			len = sizeof(struct nd_opt_hdr) +
				opt_hdr->nd_opt_len*8;
			memcpy(optr, iptr, len);

			/*
			 * Advance pointers
			 */
			lrem -= len;
			iptr += len;
			optr += len;
		}
	}

	if (found) {
		bfa_trc(&pri->ioc->bfa_fcs, ip6h->payload_len);
		/*
		 * Re-calculate ICMP6 checksum
		 */
		len = bfa_os_htons(ip6h->payload_len) - sizeof(struct ipv6hdr);
		icmp6h = (struct icmp6hdr *)((uint8_t *) ip6h +
					     sizeof(struct ipv6hdr));
		icmp6h->icmp6_cksum =
			csum_ipv6_magic(&ip6h->saddr, &ip6h->daddr, len,
					IPPROTO_ICMPV6,
					csum_partial((uint8_t *) icmp6h, len,
						     0));
	} else if (lrem) {
		/* unlikely to get into this condition */
		memcpy(optr, iptr, lrem);
	}

}

/**
 * @brief Process IPv6 Packet.
 * @param[in] skb - sk_buff
 * @retval 1 if skb is deleted within the function
 * @img ipv6_in.jpg
 */
static int
bfad_ip_process_ipv6(struct sk_buff *skb)
{
	fchs_t		*fchs;
	bfad_ip_net_hdr_t *neth;
	bfad_ip_snap_hdr_t *snap;
	struct ipv6hdr *ip6h;
	bfad_iprp_t     arp;
	struct ethhdr  *eth;
	int		index;
	struct net_device *dev = skb->dev;
	bfad_ip_private_t *pri = netdev_priv(dev);
	int	options_offset, options_length;

	fchs = (fchs_t *)skb->data;
	neth = (bfad_ip_net_hdr_t *) ((uint8_t *) fchs + sizeof(fchs_t));
	snap = (bfad_ip_snap_hdr_t *) ((uint8_t *) neth +
				       sizeof(bfad_ip_net_hdr_t));
	ip6h = (struct ipv6hdr *)((uint8_t *) snap +
				  sizeof(bfad_ip_snap_hdr_t));

	memcpy((void *)arp.wwpn.raw, (void *)&neth->src_addr, BFAD_IP_WWPN_LEN);
	arp.ip_addr_type = BFAD_IP_ADDR_TYPE_IPV6;
	arp.pid = fchs->s_id;
	memcpy((void *)&arp.ip_addr.ipv6_addr, (void *)ip6h->saddr.s6_addr,
	       sizeof(struct in6_addr));

	bfa_trc(&pri->ioc->bfa_fcs, fchs->s_id);

	/**
	 * Check if the pid exists in the local ARP cache.
	 * If entry does not exist in local ARP, add it.
	 */
	index = bfad_ip_update_arp_cache(dev, &arp, BFAD_IP_ARP_STATE_INIT);

	skb_pull(skb,
		 sizeof(fchs_t) + sizeof(bfad_ip_net_hdr_t) +
		 sizeof(bfad_ip_snap_hdr_t) - sizeof(struct ethhdr));

	/*
	 * Add ethernet header with destination MAC as 1(us) and source MAC as
	 * (I)
	 */
	eth = (struct ethhdr *)skb->data;
	memset((void *)eth, 0x00, sizeof(struct ethhdr));
	memcpy((uint8_t *) &eth->h_source[2], (void *)&index, 4);
	memcpy((uint8_t *) &eth->h_dest, (void *)dev->dev_addr, ETH_ALEN);
	eth->h_proto = ETH_P_IPV6;

	/*
	 * Check if the packet needs to be altered ( if it has link layer addr)
	 */
	if (bfad_ip_ipv6_pkt_has_ll_address
	    (ip6h, &options_offset, &options_length)) {
		/*
		 * copy the ipv6 packet
		 */
		uint8_t	*buf =
			kmalloc(sizeof(struct ipv6hdr) +
				bfa_os_htons(ip6h->payload_len),
				GFP_ATOMIC);
		memcpy(buf, (skb->data + sizeof(struct ethhdr)),
		       sizeof(struct ipv6hdr) +
		       bfa_os_htons(ip6h->payload_len));
		bfad_ip_copy_ipv6(skb->dev, (skb->data + sizeof(struct ethhdr)),
				  buf, BFAD_IP_PACKET_DIR_IN, fchs->d_id,
				  options_offset, options_length);
		kfree(buf);
	}

	skb->protocol = eth_type_trans(skb, skb->dev);

	/** Give is upto the upper layer stack */
	netif_rx(skb);
	skb->dev->last_rx = jiffies;

	return 1;
}
#endif

/**
 * @brief Gets pointer to the location that contains statistics
 * @param[in] dev - net device
 * @return - net_device structure that contains statistics
 */
static struct net_device_stats *
bfad_ip_get_stats(struct net_device *dev)
{
	bfad_ip_private_t *private = netdev_priv(dev);

	/**
	 * return the pointer to the statistics structure in the private data
	 */
	return &private->stats;
}

/*
* @Brief: ipfc_do_ioctl
 *      Handles ioctl calls from ulp Stack
 * @retval 0 on success
 * @retval -EOPNOTSUPP on failure
 */
static int
bfad_ip_do_ioctl(struct net_device *dev, struct ifreq *req, int cmd)
{

	switch (cmd) {
	case ETHTOOL_GDRVINFO:
		strcpy(req->ifr_name, (char *)BFAD_IP_DRIVER_NAME);
		req->ifr_mtu = dev->mtu;
		break;

	default:
		return -EOPNOTSUPP;
	}

	return 0;
}

/**
 * @brief Callback function which is a response for bfa_ipfc_detach
 *  In thie function all the rx sges are unmapped and all rx skbs are freed
 * @param[in] dev - network device
 * @retval none
 */
void
bfa_hcb_ipfc_detach(void *dipfc)
{
	struct net_device *dev = (struct net_device *)dipfc;
	bfad_ip_private_t *pri = netdev_priv(dev);
	int		i, j;

	bfa_trc(&pri->ioc->bfa_fcs, 0);

	/** Deallocate receive buffers */
	for (i = 0; i < BFAD_IP_MAX_RX_BUFS; i++) {
	     if (pri->ioc->ipfc->recv_pkt[i]) {
	         /** Traverse through number of sges and unmap all of them */
		 for (j = 0; j < pri->ioc->ipfc->recv_pkt[i]->nsges; j++) {
		      pci_unmap_single(pri->ioc->pcidev,
				 (uint32_t)(ssize_t) pri->ioc->
				 ipfc->recv_pkt[i]->sge[j].sg_addr,
				 pri->ioc->ipfc->recv_pkt[i]-> sge[j].sg_len,
				 PCI_DMA_FROMDEVICE);
		 }
		 dev_kfree_skb(pri->ioc->ipfc->recv_pkt[i]->skb);
		 kfree(pri->ioc->ipfc->recv_pkt[i]);
	     }
	}

}

/**
 * @brief De-allocate send/receive buffers during driver exit
 * @return - none
 */
static void
bfad_ip_deallocate_buffers(struct bfad_s *ioc)
{
	struct bfad_port_s *port = &ioc->pport;
	struct net_device *dev = port->ipfc_port->dev;
	spinlock_t	*lock = &ioc->bfad_lock;
	unsigned long	flags;
	int	i;

	bfa_trc(&ioc->bfa_fcs, 0);

	if ((port->ipfc_flags & BFAD_IP_IFF_CREATED) && dev)
		netif_stop_queue(dev);

#ifdef USE_MULTIPLE_SGES
	/** Deallocate transmit buffers */
	for (i = 0; i < BFAD_IP_MAX_TX_BUFS; i++) {
		dev_kfree_skb(ioc->ipfc->send_pkt[i]->skb);
		kfree(ioc->ipfc->send_pkt[i]);
	}
#endif

	/** flush all the receive buffers. pkt is freed bfad_cb_ipfc_pktrecv */
	ipfc_unloadpkt_count = 0;
	spin_lock_irqsave(lock, flags);
	for (i = 0; i < BFAD_IP_MAX_RX_BUFS; i++) {
		bfa_ipfc_flushpkt(ioc->ipfc->recv_pkt[i]->bpkt);
		ipfc_unloadpkt_count++;
	}
	spin_unlock_irqrestore(lock, flags);

	i = 0;
	while (ipfc_unloadpkt_count && i++ <= 5) {
		set_current_state(TASK_UNINTERRUPTIBLE);
		schedule_timeout(HZ);
	}

}

/**
 * @brief ipfc_close()
 * 	Shutdown the interface
 * @return - 0 on success
 */
static int
bfad_ip_close(struct net_device *dev)
{

	netif_stop_queue(dev);

	return 0;
}

/**
 * @brief Function maps kva to dma
 * @param[in] ioc - ioc
 * @param[in] skb - skb
 * @param[in] pkt - packet contains buffer lenght to allocate
 * @param[in] start_sge_index - index of sges[] array to be filled in.
 *	  It is 0 first the first invocation of the function for a pkt.
 * @param[in] direction - Direction of the packet
 * @retval N - Index of the last sges[] array after the mapping is done
 * @retval -1 on failure
 */
int
bfad_ip_map_dma(struct bfad_s *ioc, struct sk_buff *skb, bfad_pkt_t *pkt,
		int start_sge_index, int direction)
{
	dma_addr_t      phys_addr;

	/** try to map the kva of skb to single block of dma */
	phys_addr =
		(dma_addr_t) pci_map_single(ioc->pcidev, (void *)skb->data,
				(size_t) skb->len, (int)direction);
	if (phys_addr == 0 ||
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 27))
		dma_mapping_error(phys_addr)) {
#else
		dma_mapping_error(&ioc->pcidev->dev, phys_addr)) {
#endif
#ifdef USE_MULTIPLE_SGES
		int	page, remain, len;
		/** if it fails, allocate multiple blocks, each of PAGE_SIZE */
		bfa_assert(0);
		page = start_sge_index;
		remain = pkt->buflen;
		pkt->nsges = 0;
		do {
			page++;
			len = (remain >= PAGE_SIZE) ? PAGE_SIZE : remain;
			phys_addr =
				pci_map_single(ioc->pcidev,
					       (void *)(skb->data +
							(PAGE_SIZE *
							 (page - 1))),
					       (size_t)len,
					       (int)PCI_DMA_FROMDEVICE);
			if (phys_addr == 0 ||
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 27))
			    dma_mapping_error(phys_addr))
#else
			    dma_mapping_error(&ioc->pcidev->dev, phys_addr))
#endif
				return -1;

			pkt->sge[page - 1].sg_addr = (void *)(size_t)phys_addr;
			pkt->sge[page - 1].sg_len = len;
			remain -= len;
			pkt->nsges++;
		} while (remain > 0);
		return page;
#else
		return -1;
#endif
	} else {
		pkt->sge[start_sge_index].sg_addr = (void *)(size_t)phys_addr;
		pkt->sge[start_sge_index].sg_len = skb->len;
#if 1
		pkt->nsges = 1;
#else
		pkt->nsges++;
#endif
		start_sge_index++;
		return start_sge_index;
	}

}

/**
 * @brief function to allocate packet
 * @param [in] pkt   - packet
 * @retval 0 on success
 * @retval -1 on failure
 */
static int
bfad_ip_allocate_packet(struct bfad_s *ioc, bfad_pkt_t *pkt)
{
	struct sk_buff *skb;
	unsigned long   flags;

	/** calculate size of buffers to allocate */
	pkt->buflen =
		bfad_ip_mtu - sizeof(struct ethhdr) + sizeof(fchs_t) +
		sizeof(bfad_ip_net_hdr_t) + sizeof(bfad_ip_snap_hdr_t) +
		sizeof(struct iphdr) + 100;

	pkt->mtu = bfad_ip_mtu;

	/** Allocate SKB */
	skb = dev_alloc_skb(pkt->buflen + 2);
	if (skb == NULL) {
		bfa_assert(skb != NULL);
		return -1;
	}

	skb_reserve(skb, 2);	/* for 16 byte alignment */

	skb_put(skb, pkt->buflen);

	/** Map the kva to dma */
	if (bfad_ip_map_dma(ioc, skb, pkt, 0, PCI_DMA_FROMDEVICE) == -1) {
		dev_kfree_skb(skb);
		bfa_assert(0);
		return -1;
	}

	/** Save the addresses and skb in the packet */
	pkt->skb = skb;
	pkt->kva = skb->data;

	/** Post the buffer */
	spin_lock_irqsave(&ioc->bfad_lock, flags);
	pkt->bpkt = bfa_ipfc_postpkt(&ioc->bfa, pkt);
	spin_unlock_irqrestore(&ioc->bfad_lock, flags);

	return 0;
}

/**
 * @brief function to allocate rx buffers
 * @param [in] ioc - ioc
 * @param [in] num - number of buffers to allocate
 * @retval 0 on success
 * @retval -1 on failure
 */
static int
bfad_ip_allocate_rx_buffers(struct bfad_s *ioc)
{
	struct bfad_port_s *port = &ioc->pport;
	bfa_meminfo_t  *mem_info = &port->ipfc_port->ip_meminfo;
	bfad_pkt_t     *pkt;
	int		i;

	/** Query memory requirement for BFA ipfc module */
	memset(mem_info, 0, sizeof(bfa_meminfo_t));
	bfa_ipfc_meminfo(BFAD_IP_MAX_RX_BUFS + 100, mem_info);

	/** Allocate bfa resources */
	if (bfad_ip_alloc_bfa_resources(mem_info) != BFA_STATUS_OK) {
		bfa_log(&bfad_ip_logm, BFA_LOG_LINUX_MEM_ALLOC_FAILED,
		"IOC resources: drop packet");
		return -1;
	}

	/** Attach and initialize BFA component */
	bfa_ipfc_attach(&ioc->bfa, ioc, BFAD_IP_MAX_RX_BUFS + 100, mem_info);

	/** For each buffer */
	for (i = 0; i < BFAD_IP_MAX_RX_BUFS; i++) {
		pkt = ioc->ipfc->recv_pkt[i] =
			kmalloc(sizeof(bfad_pkt_t), GFP_ATOMIC);
		if (pkt == NULL)
			return -1;

		pkt->qe.next = pkt->qe.prev = NULL;
		pkt->index = i;
		if (bfad_ip_allocate_packet(ioc, pkt) == -1)
			return -1;
	}

	return 0;
}

/**
 * @brief function to allocate bfa resources for ipfc module
 * @param [in] mem_info - mem_info that contains memory requirements
 * @retval 0 on success
 * @retval -1 on failure
 */
bfa_status_t
bfad_ip_alloc_bfa_resources(bfa_meminfo_t *mem_info)
{
	bfa_mem_kva_t	*kva_elem = &mem_info->kva_info;

	/*
	 * Allocate the required Memory blocks
	 */
	kva_elem->kva = kmalloc(kva_elem->mem_len, GFP_ATOMIC);
	kva_elem->kva_curp = kva_elem->kva;

	return 0;
}

/**
 * @brief function to deallocate bfa resources for ipfc module
 * @param [in] ioc - IOC
 * @retval 0 on success
 * @retval -1 on failure
 */
bfa_status_t
bfad_ip_free_bfa_resources(struct bfad_s *ioc)
{
	struct bfad_port_s *port = &ioc->pport;
	bfa_meminfo_t  *mem_info = &port->ipfc_port->ip_meminfo;
	bfa_mem_kva_t   *kva_elem = &mem_info->kva_info;

	kfree(kva_elem->kva);

	return 0;
}

/**
 * @brief bfad_ip_allocate_tx_buffers
 * Allocate send buffers during driver initialization
 * The transmit buffers will contain only the headers. The data is passed
 * onto fcxp from the incoming buffer from the ulp.
 * @retval  0 on success
 * @retval -1 on failure
 */
static int
bfad_ip_allocate_tx_buffers(struct bfad_s *ioc)
{
#ifdef USE_MULTIPLE_SGES
	int		i;
	struct sk_buff *skb;
	char		message[100];

	/**
	 * Allocate tranmit buffers and set the values that wont change.
	 * like - source wwn in the network header.
	 * 	- and everything except the pid field in the LLC/SNAP header
	 */
	for (i = 0; i < BFAD_IP_MAX_TX_BUFS; i++) {
		ioc->ipfc->send_pkt[i] =
			kmalloc(sizeof(bfad_pkt_t), GFP_ATOMIC);
		skb = ioc->ipfc->send_pkt[i]->skb =
			dev_alloc_skb(sizeof(bfad_ip_net_hdr_t) +
				      sizeof(bfad_ip_snap_hdr_t));
		if (skb == NULL) {
			sprintf(message,
				"(Alloc Tx Buffers) dropping packet.\n");
			bfa_log(&bfad_ip_logm, BFA_LOG_LINUX_MEM_ALLOC_FAILED,
				message);
			return -1;
		}

		skb_put(skb,
			(sizeof(bfad_ip_net_hdr_t) +
			 sizeof(bfad_ip_snap_hdr_t)));

		bfad_ip_add_snap_header(skb->data + sizeof(bfad_ip_net_hdr_t),
					0);

		ioc->ipfc->buffer_in++;
	}
#endif

	return 0;
}

/**
 * @brief prepare the interface to use
 * @param[in] dev - device to open
 * @retval 0 on success
 */
static int
bfad_ip_open(struct net_device *dev)
{
	return 0;
}

/**
 * @brief Function to compute the IPv6 link-local address
 * RFC 4338 [Section 5&6]
 * @param[in] wwpn 		- world wide port name
 * @param[out] out_addr - link-local address
 * @retval  0 on success
 * @retval -1 on failure
 */
static int
bfad_ip_compute_ipv6_linklocal_address(bfad_ip_wwpn_t *wwpn,
				       struct in6_addr *out_addr)
{
	int		i;
	bfad_ip_wwpn_t  addr;
	BFAD_IP_WWN_TYPE type;

	/**
	 * Get the type of the wwn.
	 * Depending on type, extract the link-local address as per RFC 4338 [5]
	 */
	type = bfad_ip_get_wwn_type(wwpn);

	switch (type) {
	case BFAD_IP_WWN_TYPE_FMT1:	/* RFC 4338 [5.2] */
		*(uint16_t *) &addr.raw[0] = wwpn->fmt_1.oui_hi;
		addr.raw[2] = wwpn->fmt_1.oui_lo;
		addr.raw[3] = 0x10 | ((uint8_t) wwpn->fmt_2.vsid) >> 4;
		addr.raw[4] = wwpn->raw[5] << 4 | wwpn->raw[6] >> 4;
		addr.raw[5] = wwpn->raw[6] << 4 | wwpn->raw[7] >> 4;
		addr.raw[6] = wwpn->raw[7] << 4 | 0xf0;
		addr.raw[7] = 0x00;
		break;

	case BFAD_IP_WWN_TYPE_FMT2:	/* RFC 4338 [5.3] */
		*(uint16_t *) &addr.raw[0] = wwpn->fmt_1.oui_hi;
		addr.raw[2] = wwpn->fmt_1.oui_lo;
		addr.raw[3] = 0x20 | ((uint8_t) wwpn->fmt_2.vsid) >> 4;
		addr.raw[4] = wwpn->raw[5] << 4 | wwpn->raw[6] >> 4;
		addr.raw[5] = wwpn->raw[6] << 4 | wwpn->raw[7] >> 4;
		addr.raw[6] = wwpn->raw[7] << 4 | (wwpn->raw[0] & 0x0F);
		addr.raw[7] = wwpn->raw[1];
		break;

	case BFAD_IP_WWN_TYPE_FMT5:	/* RFC 4338 [5.4] */
		addr.raw[0] = wwpn->raw[0] << 4 | wwpn->raw[1] >> 4;
		addr.raw[1] = wwpn->raw[1] << 4 | wwpn->raw[2] >> 4;
		addr.raw[2] = wwpn->raw[2] << 4 | wwpn->raw[3] >> 4;
		addr.raw[3] = (wwpn->raw[3] & 0x0F) | 0x50;
		for (i = 4; i < BFAD_IP_WWPN_LEN; i++)
			addr.raw[i] = wwpn->raw[i];
		break;

	case BFAD_IP_WWN_TYPE_FMTCF:	/* RFC 4338 [5.5] */
		for (i = 0; i < BFAD_IP_WWPN_LEN; i++)
			addr.raw[i] = wwpn->raw[i];
		addr.raw[0] <<= 2;
		addr.raw[0] |= 0x02;
		break;

	default:
		memset((void *)out_addr, 0x00, sizeof(struct in6_addr));
		return 1;
	}

	/*
	 * RFC 4338 [6]
	 */
	out_addr->s6_addr[0] = 0xFE;
	out_addr->s6_addr[1] = 0x80;
	for (i = 2; i < 8; i++)
		out_addr->s6_addr[i] = 0x00;
	memcpy((void *)&out_addr->s6_addr[8], (void *)&addr,
	       sizeof(bfad_ip_wwpn_t));

	return 0;

}

/*
 * @brief Function to change the mtu
 * @param[in] dev - net device
 * @param[in] mtu - new mtu
 * @retval  0 on success
 * @retval -EINVAL on failure
 */
static int
bfad_ip_change_mtu(struct net_device *dev, int new_mtu)
{
	unsigned long   flags;
	bfad_ip_private_t *pri = netdev_priv(dev);
	spinlock_t     *lock = &pri->ioc->bfad_lock;
	int i;

	bfa_trc(&pri->ioc->bfa_fcs, new_mtu);

	/*
	 * check ranges
	 */
	if ((new_mtu < 68) || (new_mtu > BFAD_IP_MAX_MTU_SIZE))
		return -EINVAL;
	/*
	 * Do anything you need, and the accept the value
	 */
	spin_lock_irqsave(lock, flags);
	dev->mtu = new_mtu;
	spin_unlock_irqrestore(lock, flags);

	bfad_ip_mtu = new_mtu;

	netif_stop_queue(dev);
	/**
	 * Flush all the previously posted buffers. As the buffers we post
	 * are of mtu size all the time.
	 */
	spin_lock_irqsave(lock, flags);
	for (i = 0; i < BFAD_IP_MAX_RX_BUFS; i++)
		bfa_ipfc_flushpkt(pri->ioc->ipfc->recv_pkt[i]->bpkt);
	spin_unlock_irqrestore(lock, flags);

	netif_start_queue(dev);

	return 0;
}

/*
 * @brief Function called every BFAD_IP_ARP_AGE_TIMER secs to age arp entries
 * @param[in] arg - not used
 * @return - none
 */
void
bfad_ip_age_arp(unsigned long arg)
{
	bfa_q_t		*q, *qe, *tmp;
	struct net_device *dev = (struct net_device *)arg;
	bfad_ip_private_t *pri = netdev_priv(dev);
	spinlock_t     *lock = &pri->ioc->bfad_lock;
	struct in_addr  addr;
	int		h_val;
	unsigned long   flags;
	struct bfad_port_s *port = &pri->ioc->pport;

	/**
	 * Note: There is an IP module requirement that mtu for all Linux
	 * interfaces be the same.
	 * If MTU is changed for any interface, change it for all the interfaces
	 */
	if (dev->mtu != bfad_ip_mtu) {
		spin_lock_irqsave(lock, flags);
		dev->mtu = bfad_ip_mtu;
		spin_unlock_irqrestore(lock, flags);
	}

	/** Get my IPv4 address if configured */
	if (pri->my_ip == 0 && bfad_ip_get_ip(dev, &addr) == 0) {
		bfa_trc(&pri->ioc->bfa_fcs, addr.s_addr);
		memcpy(&pri->my_ip, &addr, sizeof(struct in_addr));
	}
#ifdef IPV6_SUPPORT
	/** Updated IPv6 link local computed address, if not already done */
	if (!pri->ip6_updated)
		bfad_ip_change_ipv6_addr(dev);
#endif

	/**
	 * 	- Check the timestamp.
	 * 	- If the entry is not used for BFAD_IP_ARP_AGE_TIMEOUT secs
	 * 	`	- call bfa fcs function to logout
	 * 		- delete the entry
	 */
	for (h_val = 0; h_val < BFAD_IP_ARP_HASH_TABLE_SZ; h_val++) {
		q = &pri->hash_pid_table[h_val];
		qe = bfa_q_next(q);

		/*
		 * Walk through the list in the current bucket
		 */
		while (qe != q) {
			/*
			 * If the entry is not used in the last
			 * BFAD_IP_ARP_AGE_TIMEOUT sec, delete it
			 */
			if (((jiffies - ((bfad_iprp_t *) qe)->timestamp) >
			    BFAD_IP_ARP_AGE_TIMEOUT*HZ) ||
				(!(port->ipfc_flags & BFAD_IP_IFF_ONLINE))) {
				tmp = qe;

				/*
				 * Advance the pointer
				 */
				qe = bfa_q_next(qe);

				/* do not age out vlan tags */
				if (((bfad_iprp_t *) qe)->vlan_tag != 0)
					continue;

				spin_lock_irqsave(lock, flags);
				bfa_q_qe_deq(tmp);
				spin_unlock_irqrestore(lock, flags);

				kfree(tmp);
			} else
				qe = bfa_q_next(qe);
		}
	}

	mod_timer(&dev->watchdog_timer, jiffies + BFAD_IP_ARP_AGE_TIMER*HZ);
}

/**
 * @brief Function used to get IP Address of the interface
 * @param[in] dev - net device
 * @param[out] addr - IP Address
 * @retval  0 on success
 * @retval -1 on failure
 */
int
bfad_ip_get_ip(struct net_device *dev, struct in_addr *addr)
{
	struct in_device *in_dev = (struct in_device *)dev->ip_ptr;
	struct in_ifaddr *ifa;

	/** Get the first entry in the ifa_list of the device */
	if (in_dev != NULL) {
		ifa = (struct in_ifaddr *)in_dev->ifa_list;
		for (ifa=in_dev->ifa_list; ifa != NULL; ifa=ifa->ifa_next) {
			if (ifa->ifa_address != 0) {
				memcpy(addr, &ifa->ifa_address,
						sizeof (struct in_addr));
				return 0;
			}
		}
	}

	return -ENOENT;
}

#ifdef IPV6_SUPPORT
/**
 * @brief Change the default link local address to the computed address
 * according to RFC 4336 [5]
 * @param[in] dev - network device
 * @return - none
 */
void
bfad_ip_change_ipv6_addr(struct net_device *dev)
{
	bfad_ip_private_t *pri = netdev_priv(dev);

	/**
	 * Loop through the ipv6 addresses and change the default address
	 * to the one computed earlier.
	 */
	if (dev->ip6_ptr) {
		struct inet6_dev *idev = dev->ip6_ptr;
		struct inet6_ifaddr *ifp;
		for (ifp = idev->addr_list; ifp; ifp = ifp->if_next) {
			memcpy(&ifp->addr.s6_addr[0], pri->my_ip6,
			       sizeof(struct in6_addr));
			pri->ip6_updated = 1;
		}

	}
}
#endif

#ifdef SYSFS
/**
 * @brief Function called when cat/more used on sysfs entry
 *  Each invocation of the function fills buffer of size only PAGE_SIZE
 *  and remembers which entries to fill in the next invocation of the function
 * @param[in]	in_dev - device
 * @param[in]	attr   - device attributes
 * @param[out] 	buf	   - buffer contains data to be displayed
 * @return - length of the buffer returned for display
 */
static	ssize_t
bfad_ip_show_arp_cache(struct device *in_dev, struct device_attribute *attr,
		       char *buf)
{
	struct net_device *dev = (struct net_device *)in_dev->driver_data;
	bfad_ip_private_t *pri = netdev_priv((struct net_device *)dev);
	bfad_iprp_t    *arp;
	bfa_q_t		*q, *qe;
	uint8_t		*w;
	char		*p = buf;
	int		h_val, k, index = 0, more = 0, *g_index =
			&pri->arp_debug_index;

	/**
	 * For each invocation of this function, only a buffer of size
	 * PAGE_SIZE will be returned, the sub-sequent call will take
	 * from where we left.
	 * arp_debug_index will keep track of where we are.
	 */

	int	len = 0;
	len += snprintf(p, PAGE_SIZE, "ARP Cache: %s\n", dev->name);
	len += snprintf(p + len, PAGE_SIZE - len, "---------------\n");

	if (in_dev == NULL || dev == NULL)
		return len;

	for (h_val = 0; h_val < BFAD_IP_ARP_HASH_TABLE_SZ; h_val++) {
		q = &pri->hash_pid_table[h_val];
		qe = bfa_q_next(q);

		while (qe != q) {
			arp = ((bfad_iprp_t *) qe);
			w = ((bfad_iprp_t *) qe)->wwpn.raw;

			if (++index >= *g_index) {
				len += snprintf(p + len, PAGE_SIZE - len,
					"pid: %d timestamp: %d WWPN:",
						arp->pid, arp->timestamp);
				for (k = 0; k < BFAD_IP_WWPN_LEN; k++) {
					len += snprintf(p + len,
							PAGE_SIZE - len,
							"%02x:", w[k]);
				}
				len += snprintf(p + len, PAGE_SIZE - len, "\n");
			}

			if (len >= PAGE_SIZE) {
				more = 1;
				break;
			}

			qe = bfa_q_next(qe);

			*g_index++;
		}

		if (len >= PAGE_SIZE) {
			more = 1;
			break;
		}
	}

	/*
	 * Reset the global counter if no more entries exist
	 */
	if (!more)
		*g_index = 0;

	return len;

}
#endif

/**
 * @}
 */

/**
 * @dg fcs2ip_api functions called by bfa fcs
 * @{
 */

/**
 * @brief Function called by bfa fcs to allocate bfad_iprp_s
 * @param[in] ioc  		- Base driver PCI/IOC instance
 * @param[in] iprp 		- Bfa Fcs IPFC pointer
 * @param[out] iprp_drv	- IPFC pointer
 * @retval BFA_STATUS_SUCCESS on success
 * @retval BFA_STATUS_FAILURE on failure
 */
int
bfa_fcb_iprp_alloc(struct bfad_s *ioc, bfa_fcs_iprp_t **iprp,
		   struct bfad_iprp_s **iprp_drv)
{
	/** Allocate memory for iprp */
	*iprp_drv = kmalloc(sizeof(bfad_iprp_t), GFP_ATOMIC);

	if (*iprp_drv == NULL)
		return BFA_STATUS_FAILED;

	memset(*iprp_drv, 0, sizeof(bfad_iprp_t));

	/** Save ioc pointer & IPRP */
	(*iprp_drv)->ioc = ioc;
	(*iprp_drv)->fcs_iprp = *iprp;

	return BFA_STATUS_OK;
}

/**
 * @brief Function called by bfa to report a new rport
 * @param [in] iprp - remote port instance
 * @return - none
 */
void
bfa_fcb_iprp_online(bfa_fcs_iprp_t *iprp)
{
	bfad_iprp_t     arp;
	struct bfad_port_s *port;
	bfa_fcs_rport_t *rport = iprp->rport;
	struct net_device *dev;
	bfad_ip_private_t *pri;

	port = bfa_fcs_ipfc_get_drvport(iprp);
	dev = port->ipfc_port->dev;
	pri = netdev_priv(dev);

	bfa_assert(dev != NULL);

	bfa_trc(&pri->ioc->bfa_fcs, rport->pid);

	/** process and update local ARP cache */
	memset(&arp, 0x00, sizeof(bfad_iprp_t));
	memcpy(arp.wwpn.raw, &rport->pwwn, sizeof(wwn_t));
	arp.pid = rport->pid;

	arp.ip_addr_type = BFAD_IP_ADDR_TYPE_NONE;
	arp.fcs_iprp = iprp;
	bfad_ip_update_arp_cache(dev, &arp, BFAD_IP_ARP_STATE_LOGIN);

}

/**
 * @brief Function called by bfa to notify an rport removal
 * @param [in] iprp - remote port instance
 * @return - none
 */
void
bfa_fcb_iprp_offline(bfa_fcs_iprp_t *iprp)
{
	bfad_iprp_t     arp;
	struct bfad_port_s *port;
	bfa_fcs_rport_t *rport = iprp->rport;
	struct net_device *dev;
	unsigned long   flags;
	bfad_ip_private_t *pri;

	port = bfa_fcs_ipfc_get_drvport(iprp);
	dev = port->ipfc_port->dev;
	pri = netdev_priv(dev);

	bfa_assert(dev != NULL);
	bfa_trc(&pri->ioc->bfa_fcs, rport->pid);

	/** process and remove the entry from local ARP cache */
	memset(&arp, 0x00, sizeof(bfad_iprp_t));
	memcpy(arp.wwpn.raw, &rport->pwwn, sizeof(wwn_t));
	arp.pid = rport->pid;
	arp.ip_addr_type = BFAD_IP_ADDR_TYPE_NONE;

	spin_lock_irqsave(&pri->ioc->bfad_lock, flags);
	bfad_ip_pid_hash_remove(pri->hash_pid_table, &arp);
	spin_unlock_irqrestore(&pri->ioc->bfad_lock, flags);

}

/**
 * @brief function called by BFA to query number of sges
 * @param [in] dpkt - packet contains details such as address and lenght
 * @retval size - number of sges
 */
int
bfad_cb_ipfc_nsgles(bfad_pkt_t *dpkt)
{
	return dpkt->nsges;
}

/**
 * @brief function called by BFA to query buffler length
 * @param [in] dpkt - packet contains details such as address and length
 * @retval len - lenght of the buffer
 */
uint16_t
bfad_cb_ipfc_pktlen(bfad_pkt_t *dpkt)
{
	return dpkt->buflen;
}

/**
 * @brief function called by BFA to query an sge of a packet
 * @param [in] dpkt - packet contains details such as address and length
 * @param [in] sge_index - indicates the index of sge that has to be returned
 * by the function
 * @param [out] sge - sge of a particular index of packet to be filled in
 * @retval none
 */
void
bfad_cb_ipfc_sge(bfad_pkt_t *dpkt, int sge_index, bfa_sge_t *sge)
{

	bfa_assert(sge_index < dpkt->nsges);
	bfa_assert(sge_index == 0);
	memcpy(sge, &dpkt->sge[sge_index], sizeof(bfa_sge_t));
}

/**
 * @todo Define the callback function name/arguments
 * @brief The routine gets called by bfa when a packet is destined to
 * ipfc module. This function processes the packet and sends it up to the ulp
 * stack.
 * @param[in] dipfc   - parameter passed in the bfa_ipfc_attach(...)
 * @param[in] dpkt    - Holds information regarding the receive data including
 * skb
 * @param[in] status  - status of the receive buffer
 * @param[in] pktlen  - length of the receive buffer
 * @param[in] residue - If packet is partially received, tells how much was
 * thrown out.
 * @return - none
 */
void
bfad_cb_ipfc_pktrecv(void *dipfc, bfad_pkt_t *dpkt, bfa_ipfc_status_t status,
		     uint32_t pktlen, uint32_t residue)
{
	struct bfad_s  *ioc = (struct bfad_s *)dipfc;
	fchs_t		*fchs;
	bfad_ip_net_hdr_t *neth;
	bfad_ip_snap_hdr_t *snap;
	uint8_t		*payload, new_buf_required = 0, del_skb = 0, i;
	struct bfad_port_s *port = &ioc->pport;
	struct net_device *dev = port->ipfc_port->dev;
	bfad_ip_private_t *pri = netdev_priv(dev);
	uint16_t	type;
	unsigned long   flags;

	if (dev == NULL) {
		printk(KERN_INFO "port not ready to receive data\n");
		dev_kfree_skb(dpkt->skb);
		bfad_ip_allocate_packet(ioc, dpkt);
		return;
	}

	dpkt->skb->dev = dev;

#ifdef DEBUG
	printk(KERN_INFO "*#*##*#*#*PKT RECEIVED ON FC PORT len: %d\n",
	       pktlen);
	printk(KERN_INFO "*#*##*#*#*DEV: 0x%x MTU: %d\n",
		dpkt->skb->dev, dev->mtu);
	uint8_t	*ptr = (uint8_t *) dpkt->skb->data;
	printk(KERN_INFO "*#*##*#*#*Data[0x%p]: ", (void *)dpkt);
	for (i = 0; i < 20; i++)
		printk(KERN_INFO "%x ", ptr[i]);
	printk("\n");
#endif

	bfa_trc(&ioc->bfa_fcs, pktlen);

	/**
	 * If dpkt->buffer size is less than the mtu of the device currently
	 * We need to allocate the new buffer and post it again.
	 * Note: There is an IP module requirement that mtu for all Linux
	 * interfaces be the same.
	 */
	if (dpkt->mtu != dev->mtu || status == BFA_IPFC_STATUS_ERROR) {
		new_buf_required = 1;
		del_skb = 1;

		/*
		 * If this is not a valid packet, drop the packet,
		 * reallocate buffer
		 */
		if (!(pktlen && !residue) || status == BFA_IPFC_STATUS_ERROR) {
			bfa_assert(dpkt->nsges == 1);

			for (i = 0; i < dpkt->nsges; i++)
				pci_unmap_single(ioc->pcidev,
						 (uint32_t)(ssize_t)dpkt->
						 sge[i].sg_addr,
						 dpkt->sge[i].sg_len,
						 PCI_DMA_FROMDEVICE);

			bfa_assert(dpkt->skb != NULL);
			if (dpkt->skb != NULL);
			dev_kfree_skb(dpkt->skb);

			if (port->ipfc_flags & BFAD_IP_IFF_UNLOADING) {
				kfree(dpkt);
				ipfc_unloadpkt_count--;
			} else {
				 if (bfad_ip_allocate_packet(ioc, dpkt) == -1)
					bfa_assert(0);
			}

			return;
		}
	}

	/**
	 * If it has 0 length OR partial buffer , post it again
	 */
	if (pktlen == 0 || residue) {
		bfa_assert(0);
		spin_lock_irqsave(&pri->ioc->bfad_lock, flags);
		dpkt->bpkt = bfa_ipfc_postpkt(&ioc->bfa, dpkt);
		spin_unlock_irqrestore(&pri->ioc->bfad_lock, flags);
		pri->stats.rx_dropped++;
		return;
	}

	/** Unmap the DMA memory */
	for (i = 0; i < dpkt->nsges; i++)
		pci_unmap_single(ioc->pcidev,
				 (uint32_t)(ssize_t) dpkt->sge[i].sg_addr,
				 dpkt->sge[i].sg_len, PCI_DMA_FROMDEVICE);

	fchs = (fchs_t *)(dpkt->skb->data);
	neth = (bfad_ip_net_hdr_t *) ((uint8_t *) fchs + sizeof(fchs_t));
	snap = (bfad_ip_snap_hdr_t *) ((uint8_t *) neth +
				       sizeof(bfad_ip_net_hdr_t));

	payload = (uint8_t *) ((uint8_t *) snap + sizeof(bfad_ip_snap_hdr_t));

	pri->stats.rx_packets++;
	pri->stats.rx_bytes += pktlen;

#ifdef DEBUG
	uint8_t	*ptr = (uint8_t *) neth;
	printk(KERN_INFO "Data: ");
	for (i = 0; i < 20; i++)
		printk(KERN_INFO "%x ", ptr[i]);
	printk("\n");
#endif

	/** Determine the packet type */
	type = bfad_ip_packet_type_in((uint8_t *) fchs);

	skb_trim(dpkt->skb, pktlen);

	/**
	 * Call appropriate function to process FCARP/ARP/FARP/IPV4/IPV6
	 * packets.
	 */
	switch (type) {
	case BFAD_IP_PKT_TYPE_FCARP:
		new_buf_required = bfad_ip_process_fcarp(dev, payload);
		del_skb = 1;
		break;
	case BFAD_IP_PKT_TYPE_ARP_REQ:
		pri->drv_stats.arp_recv++;
	case BFAD_IP_PKT_TYPE_ARP_RES:
		pri->drv_stats.arp_reply_recv++;
	case BFAD_IP_PKT_TYPE_RARP_REQ:
	case BFAD_IP_PKT_TYPE_RARP_RES:
	case BFAD_IP_PKT_TYPE_ARP_NAK:
		new_buf_required = bfad_ip_process_arp(dpkt->skb, fchs->s_id);
		break;
		break;
	case BFAD_IP_PKT_TYPE_IPV4:
		new_buf_required = bfad_ip_process_ipv4(dpkt->skb, pktlen);
		break;
#ifdef IPV6_SUPPORT
	case BFAD_IP_PKT_TYPE_IPV6:
		new_buf_required = bfad_ip_process_ipv6(dpkt->skb);
		break;
#endif
	default:
		printk(KERN_INFO "Unknown packet type: %d\n", type);
		del_skb = new_buf_required = 1;
		break;
	}

	if (del_skb)
		dev_kfree_skb(dpkt->skb);

	/**
	 * Post a new buffer to bfa if required
	 */
	if (new_buf_required)
		bfad_ip_allocate_packet(ioc, dpkt);
}

/**
 * @}
 */

/**
 * @dg base2ip_api functions called by base driver
 * @{
 */

/**
* @brief ipfc_init: IPFC Driver Initialization function
 * 	The routing initializes the linux interface for the HBA to support
 * 	IP over FC. The function will be called by the base driver
 * @retval 0
 */
int
bfad_ipfc_module_init(void)
{
	struct bfad_ipfc_ops ops;

	ops.probe = bfad_ipfc_probe;
	ops.probe_undo = bfad_ipfc_probe_undo;
	ops.probe_post = bfad_ipfc_probe_post;
	ops.port_online = bfad_ipfc_port_online;
	ops.port_offline = bfad_ipfc_port_offline;
	ops.process_farp = bfad_ip_process_farp;
	ops.pktlen = bfad_cb_ipfc_pktlen;
	ops.sge = bfad_cb_ipfc_sge;
	ops.nsgles = bfad_cb_ipfc_nsgles;
	ops.pkt_recv = bfad_cb_ipfc_pktrecv;

	if (ipfc_mtu != -1) {
		if (ipfc_mtu < BFAD_IP_MIN_MTU_SIZE)
			ipfc_mtu = BFAD_IP_MIN_MTU_SIZE;

		if (ipfc_mtu > BFAD_IP_MAX_MTU_SIZE)
			ipfc_mtu = BFAD_IP_MAX_MTU_SIZE;
	} else
		ipfc_mtu = BFAD_IP_DEFAULT_MTU_SIZE;

	bfad_ip_mtu = ipfc_mtu;

	bfad_register(&ops);

	/** Initialize the log message handler */
	bfa_log_init(&bfad_ip_logm, "IPFC", (bfa_log_cb_t) bfa_os_printf);

	bfad_ip_os_ioctl_init();

	return 0;
}

/**
 * @brief ipfc_exit
 * 	Function called by base driver while unloading the driver
 * @return - none
 */
void
bfad_ipfc_module_exit(void)
{
	bfad_ip_os_ioctl_exit();
	bfad_unregister();
}

/**
 * @brief Function called by base driver during probe
 * @param [in] bfad - ioc
 * @retval 0
 */
int
bfad_ipfc_probe(struct bfad_s *bfad)
{
	char		message[100];
	bfad_ipfc_t    *ipfc;
	bfad_ipfc_port_t *ipfc_port;

	printk(KERN_INFO "IPFC PROBE Called\n");
	bfa_trc(&bfad->bfa_fcs, 1111);

	ipfc = kmalloc(sizeof(bfad_ipfc_t), GFP_ATOMIC);
	if (ipfc == NULL) {
		sprintf(message, "IP IOC instance data\n");
		bfa_log(&bfad_ip_logm, BFA_LOG_LINUX_MEM_ALLOC_FAILED, message);
		return -ENOMEM;
	}

	/** Initialize the transmit buffer pointers */
	bfad->ipfc = ipfc;
	bfad->ipfc->buffer_in = bfad->ipfc->buffer_out = 0;

	/** Port specific parameters */
	ipfc_port = kmalloc(sizeof(bfad_ipfc_port_t), GFP_ATOMIC);
	if (ipfc_port == NULL) {
		sprintf(message, "IP port instance data\n");
		bfa_log(&bfad_ip_logm, BFA_LOG_LINUX_MEM_ALLOC_FAILED, message);
		return -ENOMEM;
	}

	memset(ipfc_port, 0x00, sizeof(bfad_ipfc_port_t));
	bfad->pport.ipfc_port = ipfc_port;

	/** Allocate buffers */
	bfad_ip_allocate_rx_buffers(bfad);

	/** Allocate Transmit buffers */
	bfad_ip_allocate_tx_buffers(bfad);

	bfad_ip_enable_ipfc(bfad, &bfad->pport, BFA_TRUE);

	return 0;
}

/**
 * @brief Function called by base driver to delete host interfaces.
 * @param [in] bfad - ioc
 * @retval 0
 */
int
bfad_ipfc_probe_undo(struct bfad_s *bfad)
{
	struct bfad_port_s *port = &bfad->pport;
	struct net_device *dev;

	dev = (struct net_device *)port->ipfc_port;

	printk("PROBE UNDO is called bfad:%p port:%p dev:%p\n",
		bfad, port, dev);

	bfa_assert(bfad != NULL);
	bfa_assert(port != NULL);
	bfa_assert(dev != NULL);

	port->ipfc_flags |= BFAD_IP_IFF_UNLOADING;

	bfad_ip_deallocate_buffers(bfad);

	port->ipfc_flags &= ~BFAD_IP_IFF_UNLOADING;

	if (port->ipfc_flags & BFAD_IP_IFF_CREATED)
		bfad_ipfc_port_delete(bfad, &bfad->pport);

	bfad_ip_free_bfa_resources(bfad);

	kfree(bfad->ipfc);

	kfree(bfad->pport.ipfc_port);

	return 0;
}

/**
 * @brief Function called by base driver
 * @param [in] bfad - ioc
 * @retval none
 */
void
bfad_ipfc_probe_post(struct bfad_s *bfad)
{
	printk(KERN_INFO "PROBE POST is called\n");
	bfa_trc(&bfad->bfa_fcs, 9999);
}

/**
 * @brief Function called by base driver to create host interfaces
 * @param [in] bfad - ioc
 * @param [in] port - port
 * @param [in] port_type - type of the port
 *		0-base port of base fabric
 *		1-vport of base fabric
 *		2-base port of virtual fabric
 * 		3-vport of virtual fabric
 * @retval - 0 on success
 * @retval - negative value on failure
 */
int
bfad_ipfc_port_new(struct bfad_s *bfad, struct bfad_port_s *port, int port_type)
{
	port->ipfc_flags = BFAD_IP_IFF_UNKNOWN;
	return BFA_STATUS_OK;
}

static int
bfad_ip_get_settings(struct net_device *netdev, struct ethtool_cmd *cmd)
{

	cmd->autoneg = AUTONEG_DISABLE;
	cmd->supported |= SUPPORTED_FIBRE;
	cmd->advertising |= ADVERTISED_FIBRE;
	cmd->port = PORT_FIBRE;
	cmd->phy_address = 0;

	if (netif_carrier_ok(netdev)) {
		cmd->speed = 8000; /* There is no macro for 8G */
		cmd->duplex = DUPLEX_FULL;
	} else {
		cmd->speed = -1;
		cmd->duplex = -1;
	}
	cmd->transceiver = XCVR_EXTERNAL;
	cmd->maxtxpkt = 0;
	cmd->maxrxpkt = 0;

	return 0;
}

static void
bfad_ip_get_drvinfo(struct net_device *netdev, struct ethtool_drvinfo *drvinfo)
{
	bfad_ip_private_t *pri = netdev_priv(netdev);
	struct bfad_s *bfad = pri->ioc;
	struct bfa_ioc_attr_s *ioc_attr;
	unsigned long flags;

	strcpy(drvinfo->driver, BFAD_IP_DRIVER_NAME);
	strcpy(drvinfo->version, BFAD_DRIVER_VERSION);

	ioc_attr = kzalloc(sizeof(*ioc_attr), GFP_KERNEL);
	if (ioc_attr) {
		memset(ioc_attr, 0, sizeof(*ioc_attr));
		spin_lock_irqsave(&pri->ioc->bfad_lock, flags);
		bfa_get_attr(&pri->ioc->bfa, ioc_attr);
		spin_unlock_irqrestore(&pri->ioc->bfad_lock, flags);

		strncpy(drvinfo->fw_version, ioc_attr->adapter_attr.fw_ver,
			sizeof(drvinfo->fw_version) - 1);
		kfree(ioc_attr);
	}

	strncpy(drvinfo->bus_info, pci_name(bfad->pcidev), ETHTOOL_BUSINFO_LEN);
}

static void
bfad_ip_get_ethtool_stats(struct net_device *netdev,
		struct ethtool_stats *stats, u64 *buf)
{
	bfad_ip_private_t *pri = netdev_priv(netdev);
	unsigned long flags, *net_stats;
	int i;

	memset(buf, 0, stats->n_stats * sizeof(u64));

	spin_lock_irqsave(&pri->ioc->bfad_lock, flags);
	net_stats = (unsigned long *)&pri->stats;
	for (i = 0; i < sizeof(struct net_device_stats) / sizeof(unsigned long);
		i++)
		buf[i] = net_stats[i];
	spin_unlock_irqrestore(&pri->ioc->bfad_lock, flags);

}

#define BFAD_IP_ETHTOOL_STATS_NUM	\
	(sizeof(struct net_device_stats) / sizeof(unsigned long))

static char *bfad_ip_net_stats_strings[BFAD_IP_ETHTOOL_STATS_NUM] = {
	"rx_packets",
	"tx_packets",
	"rx_bytes",
	"tx_bytes",
	"rx_errors",
	"tx_errors",
	"rx_dropped",
	"tx_dropped",
	"multicast",
	"collisions",

	"rx_length_errors",
	"rx_over_errors",
	"rx_crc_errors",
	"rx_frame_errors",
	"rx_fifo_errors",
	"rx_missed_errors",
	"tx_aborted_errors",
	"tx_carrier_errors",
	"tx_fifo_errors",
	"tx_heartbeat_errors",
	"tx_window_errors",

	"rx_compressed",
	"tx_compressed"
};


static void
bfad_ip_get_strings(struct net_device *netdev, u32 stringset, u8 * string)
{
	int i;

	switch (stringset) {
	case ETH_SS_STATS:
		for (i = 0; i < BFAD_IP_ETHTOOL_STATS_NUM; i++) {
			bfa_assert(strlen(bfad_ip_net_stats_strings[i]) <
				ETH_GSTRING_LEN);
			memcpy(string, bfad_ip_net_stats_strings[i],
				ETH_GSTRING_LEN);
			string += ETH_GSTRING_LEN;
		}
		break;

	default:
		bfa_assert(0);
	}

}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 29)
static int
bfad_ip_get_stats_count(struct net_device *netdev)
{
	return BFAD_IP_ETHTOOL_STATS_NUM;
}
#else
static int
bfad_ip_get_sset_count(struct net_device *netdev, int sset)
{
	switch (sset) {
	case ETH_SS_STATS:
		return BFAD_IP_ETHTOOL_STATS_NUM;
	default:
		return -EOPNOTSUPP;
	}
}
#endif

static struct ethtool_ops bfad_ethtool_ops = {
	.get_settings = bfad_ip_get_settings,
	.get_drvinfo = bfad_ip_get_drvinfo,
	.get_strings = bfad_ip_get_strings,
	.get_ethtool_stats = bfad_ip_get_ethtool_stats,
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 29)
	.get_stats_count = bfad_ip_get_stats_count
#else
	.get_sset_count = bfad_ip_get_sset_count
#endif

};

void
bfad_set_ethtool_ops(struct net_device *netdev)
{
	SET_ETHTOOL_OPS(netdev, &bfad_ethtool_ops);
}

#if defined(HAVE_NET_DEVICE_OPS)
static const struct net_device_ops netdev_ops =
{
	.ndo_open = bfad_ip_open,
	.ndo_stop = bfad_ip_close,
	.ndo_start_xmit = bfad_ip_send,
	.ndo_get_stats = bfad_ip_get_stats,
	.ndo_change_mtu = bfad_ip_change_mtu,
	.ndo_do_ioctl = bfad_ip_do_ioctl,
};
#endif

int
bfad_ipfc_port_new_real(struct bfad_s *bfad, struct bfad_port_s *port,
			int port_type)
{
	struct net_device *dev;
	bfad_ip_private_t *pri;
	struct in6_addr out_addr;
	struct in_addr  addr;
	char		name[50];
	char		message[100];
	bfa_port_attr_t pattr;

	/**
	 * Create linux interface name based on the type of the port
	 */
	if (port_type == 0)	/* Base port of base fabric */
		sprintf(name, "fc0%d", (bfad->inst_no));
	else if (port_type == 1)	/* vport of base fabric */
		sprintf(name, "fc0%d-v%d", bfad->inst_no,
			(bfad_ip_port_index++ % 0xfffffffe));
	else if (port_type == 2)	/* base port of VF */
		sprintf(name, "fcvf%d", port->fcs_port->fabric->vf_id);
	else if (port_type == 3)	/* vport port of VF */
		sprintf(name, "fcvf%d-v%d", port->fcs_port->fabric->vf_id,
			(bfad_ip_port_index++ % 0xfffffffe));

	/**
	 * Create the interface as ethernet device.
	 * Allocate additional space of size bfad_ip_private_t to store
	 * spin locks and other private data
	 */
	dev = alloc_etherdev(sizeof(bfad_ip_private_t));
	if (!dev) {
		sprintf(message, "etherdev (%s) port add failure .\n", name);
		bfa_log(&bfad_ip_logm, BFA_LOG_LINUX_MEM_ALLOC_FAILED, message);
		return -ENOMEM;
	}

	dev->irq = bfad->pcidev->irq;

	pri = (bfad_ip_private_t *) netdev_priv(dev);
	pri->ioc = bfad;

	/**
	 * Set the default MTU
	 */
	dev->mtu = bfad_ip_mtu;
	dev->flags &= ~IFF_MULTICAST;
	dev->flags |= IFF_NOTRAILERS;
	dev->features |= NETIF_F_HIGHDMA;
	dev->tx_queue_len = BFAD_IP_MAX_TX_BUFS;

	bfad_set_ethtool_ops(dev);

	/** Initialize the hardware address with wwpn last 6 bytes */
	bfa_fcport_get_attr(&bfad->bfa, &pattr);
	memcpy(dev->dev_addr, ((uint8_t *) &pattr.pwwn + 2), ETH_ALEN);

	/**
	 * Initialize the device entry points
	 */
#if !defined(HAVE_NET_DEVICE_OPS)
	dev->open = bfad_ip_open;
	dev->stop = bfad_ip_close;
	dev->hard_start_xmit = bfad_ip_send;
	dev->do_ioctl = bfad_ip_do_ioctl;
	dev->get_stats = bfad_ip_get_stats;
	dev->change_mtu = bfad_ip_change_mtu;
#else
	dev->netdev_ops = &netdev_ops;
#endif

	/**
	 * Change the device name from eth<xx> to fc<xx>
	 */
	strncpy(dev->name, name, sizeof(dev->name) - 1);

	/**
	 * Register the device
	 */
	if (register_netdev(dev)) {
		sprintf(message, "(%s) new port add failed \n", dev->name);
		bfa_log(&bfad_ip_logm, BFA_LOG_LINUX_DRIVER_REGISTRATION_FAILED,
			message);
		free_netdev(dev);
		return -EIO;
	}

	memcpy(pri->my_wwpn, &pattr.pwwn, BFAD_IP_WWPN_LEN);

#ifndef IPV6_SUPPORT
	/** Add IPv6 link-local address */
	bfad_ip_compute_ipv6_linklocal_address((bfad_ip_wwpn_t *)
			&pri->my_wwpn, &out_addr);
	memcpy(pri->my_ip6, &out_addr, sizeof(struct in6_addr));
#endif

	/** Get my IPv4 address */
	if (bfad_ip_get_ip(dev, &addr) == 0)
		memcpy(&pri->my_ip, &addr, sizeof(struct in_addr));

	/** Init hash tables */
	if (bfad_ip_pid_hash_init
	    (pri->hash_pid_table, BFAD_IP_ARP_HASH_TABLE_SZ) < 0) {
		bfa_log(&bfad_ip_logm, BFA_LOG_LINUX_HASH_INIT_FAILED, name);
		unregister_netdev(dev);
		free_netdev(dev);
		return -ENOMEM;
	}

	pri->my_pid =
		bfa_fcs_lport_get_fcid(bfa_fcs_get_base_port(&bfad->bfa_fcs));

	/** Save the device to be used in delete/online/offline functions */
	port->ipfc_port->dev = dev;

	/** Initialize watchdog timer */
	dev->watchdog_timer.function = bfad_ip_age_arp;
	dev->watchdog_timer.data = (unsigned long)dev;
	dev->watchdog_timeo = BFAD_IP_ARP_AGE_TIMER*HZ;
	init_timer (&dev->watchdog_timer);
	mod_timer(&dev->watchdog_timer, jiffies + BFAD_IP_ARP_AGE_TIMER*HZ);

#ifdef SYSFS
	bfad_ip_start_sysfs(dev);
#endif

	port->ipfc_flags |= BFAD_IP_IFF_CREATED;
	return 0;
}

/**
 * @brief Function called by base driver to delete host interfaces
 * @return - 0 on success
 */
int
bfad_ipfc_port_delete(struct bfad_s *bfad, struct bfad_port_s *port)
{
	struct net_device *dev = port->ipfc_port->dev;

	printk(KERN_INFO "BFAD_IPFC_PORT_DELETE port: %p 0x%x\n",
		port, port->ipfc_flags);

	if (!(port->ipfc_flags & BFAD_IP_IFF_ENABLED))
		return 0;

#ifdef SYSFS
	bfad_ip_stop_sysfs(dev);
#endif

	del_timer(&dev->watchdog_timer);
	unregister_netdev(dev);
	free_netdev(dev);

	port->ipfc_flags &= ~BFAD_IP_IFF_CREATED;

	return 0;
}

/**
 * @brief Function called by base driver when a port is offline
 * @return - 0 on success
 */
int
bfad_ipfc_port_offline(struct bfad_s *bfad, struct bfad_port_s *port)
{
	struct net_device *dev = port->ipfc_port->dev;

	printk(KERN_INFO "BFAD_IPFC_PORT_OFFLINE: %x mtu: %d\n",
		port->ipfc_flags, dev->mtu);

	port->ipfc_flags &= ~BFAD_IP_IFF_ONLINE;
	bfa_trc(&bfad->bfa_fcs, port->fcs_port->pid);

	if (!(port->ipfc_flags & BFAD_IP_IFF_ENABLED))
		return 0;

	if (dev == NULL)
		return -1;

	/**
	 *  Stop device queue so the interface is not available
	 */
	netif_stop_queue(dev);

	bfa_trc(&bfad->bfa_fcs, port->fcs_port->pid);

	return 0;
}

/**
 * @brief Function called by base driver when a port is online
 * @return - 0 on success
 */
int
bfad_ipfc_port_online(struct bfad_s *bfad, struct bfad_port_s *port)
{
	struct net_device *dev = port->ipfc_port->dev;
	bfad_ip_private_t *pri = (bfad_ip_private_t *) netdev_priv(dev);

	printk(KERN_INFO "BFAD_IPFC_PORT_ONLINE dev: %p mtu: %d flags: %x\n",
			dev, dev->mtu, port->ipfc_flags);

	pri->my_pid =
		bfa_fcs_lport_get_fcid(bfa_fcs_get_base_port(&bfad->bfa_fcs));

	port->ipfc_flags |= BFAD_IP_IFF_ONLINE;

	bfa_trc(&bfad->bfa_fcs, port->fcs_port->pid);

	if (!(port->ipfc_flags & BFAD_IP_IFF_ENABLED))
		return 0;

	bfa_trc(&bfad->bfa_fcs, port->fcs_port->fabric->vf_id);

	if (dev == NULL)
		return -1;

	/**
	 *  Start device queue so the interface is available
	 */
	netif_start_queue(dev);

	return 0;
}

/**
 * @}
 */

/**
 * @addtogroup ip_hash_private
 * @{
 */
/**
 * @brief function that creates a new index while adding new member in hash
 * table. Primary key would be the pid
 * @param[in] arp 	- An arp entry
 * @return 			- hash value
 */
int
bfad_ip_pid_hash(bfad_iprp_t *arp)
{
	uint8_t		hash_val;
	uint32_t	id;

	bfa_assert(arp != NULL);
	id = arp->ip_addr.ipv4_addr.s_addr,
	/** Generate the hash value from the pid */
		id &= 0xFFFFFFFF;
	hash_val =
		((((id & 0xFF000000) >> 24) + ((id & 0x00FF0000) >> 16) +
		  ((id & 0x0000FF00) >> 8) +
		  ((id & 0xFF))) % BFAD_IP_ARP_HASH_TABLE_SZ);

	return (hash_val);
}

/**
 * @brief function used while searching for an entry in hash table
 * @param[in] arp1 - arp record supplied for search
 * @param[in] arp2 - arp record being compared against while searching
 * @retval 1 if search is successful
 * @retval 0 if the search fails
 */
int
bfad_ip_key_cmp_pid(bfad_iprp_t *arp1, bfad_iprp_t *arp2)
{
	if (arp1->ip_addr.ipv4_addr.s_addr == arp2->ip_addr.ipv4_addr.s_addr)
		return 1;
	else
		return 0;
}

/**
 * @}
 */

int
bfad_ip_enable_ipfc(struct bfad_s *bfad, struct bfad_port_s *port,
		    bfa_boolean_t flag)
{
	struct net_device *dev;

	if (flag == BFA_TRUE) {
		port->ipfc_flags |= BFAD_IP_IFF_ENABLED;

		if (!(port->ipfc_flags & BFAD_IP_IFF_CREATED))
			bfad_ipfc_port_new_real(bfad, port, 0);

		if (!(port->ipfc_flags & BFAD_IP_IFF_CREATED))
			bfa_assert(0);

		dev = port->ipfc_port->dev;

		if ((port->ipfc_flags & BFAD_IP_IFF_ONLINE) && dev)
			netif_start_queue(dev);
	} else {
		dev = port->ipfc_port->dev;

		if (port->ipfc_flags & BFAD_IP_IFF_CREATED) {
			if (dev)
				netif_stop_queue(dev);
			bfad_ipfc_port_delete(bfad, port);
		}

		port->ipfc_flags &= ~BFAD_IP_IFF_ENABLED;
	}

	return 0;
}

int
bfad_ip_get_drv_stats(struct bfad_s *bfad, bfa_ipfc_stats_t *stats)
{
	struct net_device *dev = bfad->pport.ipfc_port->dev;
	bfad_ip_private_t *pri = netdev_priv(dev);

	if (!(bfad->pport.ipfc_flags & BFAD_IP_IFF_ENABLED))
		return 1;

	memcpy(stats, &pri->drv_stats, sizeof(bfa_ipfc_stats_t));

	return (0);
}

int
bfad_ip_reset_drv_stats(struct bfad_s *bfad)
{
	struct net_device 	*dev;
	bfad_ip_private_t	*pri;

	if (!(bfad->pport.ipfc_flags & BFAD_IP_IFF_ENABLED))
		return 1;

	dev = bfad->pport.ipfc_port->dev;
	pri = netdev_priv(dev);
	if (dev)
		memset(&pri->drv_stats, 0x00, sizeof(bfa_ipfc_stats_t));

	return (0);
}

int
bfad_ip_get_attr(struct bfad_s *bfad, bfa_ipfc_attr_t *attr)
{
	struct bfad_port_s *port;
	struct net_device 	*dev;

	port = (struct bfad_port_s *)&bfad->pport;
	attr->enabled = (port->ipfc_flags & BFAD_IP_IFF_ENABLED) ? 1 : 0;

	if (attr->enabled) {
		attr->mtu = ipfc_mtu;
		dev = bfad->pport.ipfc_port->dev;
		if (dev)
			sprintf(attr->eth_name, "%s", dev->name);
	}

	return 0;
}

module_init(bfad_ipfc_module_init);
module_exit(bfad_ipfc_module_exit);
MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("QLogic BR-series Fibre Channel HBA/IPFC Driver");
MODULE_AUTHOR("QLogic Corporation");
MODULE_VERSION(BFAD_DRIVER_VERSION);
