/** 
 * Copyright (c) 2005-2014 Brocade Communications Systems, Inc.
 * Copyright (c) 2014 QLogic Corporation.
 * All rights reserved
 * www.qlogic.com
 *
 * See LICENSE.bfa for copyright and licensing details.
 */

/**
 * @file bfad_ip_drv.h  Include file for IP over FC module
 */

#ifndef __IPFC_H__
#define __IPFC_H__

#define	BFAD_IP_DRIVER_NAME			"bfaipfc"	/*!< Driver Name 			*/
#define BFAD_IP_MIN_MTU_SIZE		68			/*!< Minimum MTU size 		*/
#define BFAD_IP_MAX_MTU_SIZE		65280		/*!< Maximum MTU size 		*/
#define BFAD_IP_DEFAULT_MTU_SIZE  	2024		/*!< Default MTU size 		*/
#define BFAD_IP_WWPN_LEN			8			/*!< Length of WWN    		*/

#define BFAD_IP_REQ_TIMEOUT  		2     		/*!< Timeout value 			*/

#define BFAD_IP_FC4_TYPE			0x05  		/*!< IP over FC Type    	*/
#define BFAD_IP_FC_NET_HDR      	0x20


#define BFAD_IP_BCAST_IPV4_ADDR 	0xffffffff /*!< Broadcast IPV4 IP Address */
#define BFAD_IP_ENET_ARP_HDR_SZ		(sizeof (struct arphdr) + 20)
#define BFAD_IP_ARP_HASH_TABLE_SZ 	256			/*!< ARP Hash table size 	*/
#define BFAD_IP_ARP_AGE_TIMER 		10			/*!< arp age timer 	interval */
#define BFAD_IP_ARP_AGE_TIMEOUT 	10			/*!< arp age timeout period */
#define BFAD_IP_MAX_TX_BUFS 		64		/*!< Max transmit buffers allocated */
#define BFAD_IP_MAX_RX_BUFS 		64		/*!< Max receive  buffers allocated */
#define BFAD_IP_MAX_SGES 			17		/*!< Max sges to map a recv buffer  */

/**
 * Packet direction, going out or coming in
 */
enum {
	BFAD_IP_PACKET_DIR_IN = 0,
	BFAD_IP_PACKET_DIR_OUT,
};

/**
 * WWN Types
 */
typedef enum {
	BFAD_IP_WWN_TYPE_UNKNOWN = 0,
	BFAD_IP_WWN_TYPE_FMT1 = 1,
	BFAD_IP_WWN_TYPE_FMT2 = 2,
	BFAD_IP_WWN_TYPE_FMT5 = 5,
	BFAD_IP_WWN_TYPE_FMTCF = 0xC,
} BFAD_IP_WWN_TYPE;

/**
 * Structure for WWPN for different formats
 */

typedef union bfad_ip_wwpn_s {
	uint8_t		raw[BFAD_IP_WWPN_LEN];		/*!< Raw WWPN */
	struct _fmt_1 {					/* Format 1 WWPN */
		uint32_t		naa_id:4,
						zeros:12,
						oui_hi:16;
		uint32_t		oui_lo:8,
						vsid:24;
	} fmt_1;
	struct _fmt_2 {					/* Format 2 WWPN */
		uint32_t 		naa_id:4,
						vendor:12,
						oui_hi:16;
		uint32_t		oui_lo:8,
						vsid:24;
	} fmt_2;
	struct _fmt_5 {					/* Format 5 WWPN */
		uint32_t 		naa_id:4,
						oui:24,
						vsid_hi:4;
		uint32_t		vsid_lo;
	} fmt_5;
	struct _fmt_cf {				/* Format 0xC-0xF WWPN */
		uint32_t 		naa_id:2,
						oui_0_5:6,
						oui_8_23:12,
						vsid_hi:8;
		uint32_t		vsid_lo;
	} fmt_cf;

} bfad_ip_wwpn_t;

/**
 * @brief IPFC port
 */
typedef struct bfad_ipfc_port_s {
	struct net_device *dev;
	bfa_meminfo_t	  ip_meminfo;
} bfad_ipfc_port_t;

/**
 * @brief Network Header
 */
typedef struct bfad_ip_net_hdr_s {
	bfad_ip_wwpn_t		dest_addr;
	bfad_ip_wwpn_t		src_addr;
} bfad_ip_net_hdr_t;

/**
 *LLC/SNAP Header
 */
#define BFAD_IP_LLC_SNAP_DSAP		0xAA
#define BFAD_IP_LLC_SNAP_SSAP		0xAA
#define BFAD_IP_LLC_SNAP_CTRL		0x03
#define BFAD_IP_LLC_SNAP_OUI_HI		0x00
#define BFAD_IP_LLC_SNAP_OUT_LOW	0x0000

typedef struct bfad_ip_snap_hdr {
	uint8_t 	dsap;		/*!< Will have value 0xAA */
	uint8_t 	ssap;		/*!< Will have value 0xAA */
	uint8_t 	ctrl;		/*!< Will have value 0x03 */
	uint8_t 	oui_hi;		/*!< Will have value 0x00 */
	uint16_t 	oui_low;	/*!< Will have value 0x0000 */
	uint16_t 	pid;		/*!< IPv4=0x0800, IPv6=86DD, ARP=0806 */
} bfad_ip_snap_hdr_t;

#define NET_SNAP_HDR_SZ (sizeof (bfad_ip_net_hdr_t) + sizeof (bfad_ip_snap_hdr_t))

/**
 * @brief packet structure to be filled while posting receive buffers
 *  These values will be returned when data is received on FC port
 *  in the callback function
 */
struct bfad_pkt_s {
	struct bfa_q_s 		qe;
	void*			kva;	/*!< kernel virtual address 				*/
	struct sk_buff*	skb;	/*!< SKB holding kva						*/
	struct sk_buff*	skb_payload;	/*!< SKB transmitted 						*/
	struct bfad_s	*ioc;
	struct net_device
					*dev;
	uint16_t		buflen;	/*!< Length of buffer allocated 			*/
	int				mtu;	/*!< MTU of the buffer at alloc time */
	int				index;	/*!< Index of the buffer in rx buffer ring 	*/
	int				nsges;	/*!< Number of sges to map the kva			*/
	int				fc_type;
	void			*bpkt;
	struct			bfa_sge_s sge[BFAD_IP_MAX_SGES];
							/*!< DMA Address/length of each sge in list */
};

typedef struct bfad_ipfc_s {
	bfad_pkt_t*	send_pkt[BFAD_IP_MAX_TX_BUFS];
	bfad_pkt_t*	recv_pkt[BFAD_IP_MAX_RX_BUFS];
	int			buffer_in;  /*!< Indicates total buffers available any */
							/*	 time for transmit */
	int			buffer_out; /*!< Points to buffer available for next xmit */
							/*	 buffer */
} bfad_ipfc_t;

/**
 *ipfc device private structure
 */
typedef struct bfad_ip_private_s {
	struct net_device_stats	stats;
	bfa_ipfc_stats_t		drv_stats;
	uint8_t					send_buffer;
	uint8_t					sent_buffer;
	struct bfa_q_s			hash_pid_table[BFAD_IP_ARP_HASH_TABLE_SZ];
	struct bfad_s*			ioc;
	int						my_pid;
	uint32_t				my_ip;
#ifndef __VMKERNEL_MODULE__
	uint8_t					my_ip6[sizeof(struct in6_addr)];
#endif
	uint8_t					ip6_updated;
	uint8_t					my_wwpn[BFAD_IP_WWPN_LEN];
	struct timer_list		watchdog;
	int						arp_debug_index;
	bfa_boolean_t			iff_online;
} bfad_ip_private_t;

enum {
	BFAD_IP_IFF_UNKNOWN = 0,
	BFAD_IP_IFF_ENABLED = 0x01,
	BFAD_IP_IFF_ONLINE  = 0x02,
	BFAD_IP_IFF_CREATED = 0x04,
	BFAD_IP_IFF_UNLOADING = 0x08,
};

/**
 * Packet type
 */
typedef enum {
	BFAD_IP_PKT_TYPE_UNKNOWN = 0,
	BFAD_IP_PKT_TYPE_ARP_REQ,
	BFAD_IP_PKT_TYPE_ARP_RES,
	BFAD_IP_PKT_TYPE_RARP_REQ,
	BFAD_IP_PKT_TYPE_RARP_RES,
	BFAD_IP_PKT_TYPE_IARP_REQ,
	BFAD_IP_PKT_TYPE_IARP_RES,
	BFAD_IP_PKT_TYPE_ARP_NAK,
	BFAD_IP_PKT_TYPE_FCARP,
	BFAD_IP_PKT_TYPE_IPV4,
	BFAD_IP_PKT_TYPE_8021Q,
	BFAD_IP_PKT_TYPE_IPV6_DISC,
	BFAD_IP_PKT_TYPE_IPV6,
	BFAD_IP_PKT_TYPE_FARP,
} BFAD_IP_PKT_TYPE;

/**
 * @brief ARP entry state in the hash table
 */

typedef enum {
	BFAD_IP_ARP_STATE_INIT = 0,		/*!< Arp query Initiated or Login Not Done */
	BFAD_IP_ARP_STATE_QUERY_FCARP,	/*!< Arp query(FC ARP) in progress for this entry */
	BFAD_IP_ARP_STATE_QUERY_FARP,	/*!< Arp query(FARP) in progress for this entry */
	BFAD_IP_ARP_STATE_LOGIN,		/*!< Arp entry valid. Login done */
	BFAD_IP_ARP_STATE_INVALID,		/*!< Arp entry invalid. */
} BFAD_IP_ARP_STATE;

/**
 * @brief ARP entry that is stored in a hash table
 */
#define BFAD_IP_ADDR_TYPE_NONE 0
#define BFAD_IP_ADDR_TYPE_IPV4 1
#define BFAD_IP_ADDR_TYPE_IPV6 2
struct bfad_iprp_s {
	struct bfa_q_s 			q;
	bfad_ip_wwpn_t		wwpn;				/*!< Address of WWPN 		*/
	uint32_t			ip_addr_type;		/*!< Address type: IPv4/IPv6 */
	uint32_t			pid;				/*!< Destination pid 		*/
	uint16_t		vlan_tag;			/*!< vlan_tag */
	union {
		struct in_addr	ipv4_addr;			/*!< IPv4 Address 			*/
		struct in6_addr	ipv6_addr;			/*!< IPv6 Address 			*/
	} ip_addr;
	struct sk_buff*		skb;				/*!< Address of the skbuff 	*/
	BFAD_IP_ARP_STATE	state;				/*!< State of the ARP entry */
	unsigned long		timestamp;
	struct bfa_fcs_iprp_s *fcs_iprp;		/*!< bfa fcs 				*/
	struct bfad_s* 		ioc;				/*!< pointer to ioc 		*/
};

/**
 * @dg ip_fcarp_private FCARP data structures
 * @{
 */
/**
 *@brief  FC ARP structure
 */
#define BFAD_IP_FCARP_HW_TYPE	0x0012
#define BFAD_IP_FCARP_HW_LEN 	12
#define BFAD_IP_FCARP_CODE_REQ 	0x0001
#define BFAD_IP_FCARP_CODE_RES 	0x0002
#define BFAD_IP_PROTOCOL_LEN	4

typedef struct bfad_ip_hw_addr_s {
	bfad_ip_wwpn_t		pname;
	uint32_t			reserved:8,
						pid:24;
} bfad_ip_hw_addr_t;

typedef struct bfad_ip_fcarp_s {
	uint16_t			hw_type;
	uint16_t			protocol;
	uint8_t 			hw_len;
	uint8_t				prot_len;
	uint16_t			opcode;
	bfad_ip_hw_addr_t	src_hw_addr;
	struct in_addr		src_ip;
	bfad_ip_hw_addr_t	dest_hw_addr;
	struct in_addr		dest_ip;
} bfad_ip_fcarp_t;

/*
 * @}
 */

/**
 * @dg ip_farp_private FARP data structures
 * @{
 */
/**
 * @brief FARP Structure
 */
typedef enum els_code {
	BFAD_IP_FARP_CMD_REQ	= 0x54,
	BFAD_IP_FARP_CMD_REPLY	= 0x55
} els_code_t;

typedef struct bfad_ip_els_farp_s {
	els_code_t		code;  			/*!< FARP ELS code - 0x54/0x55 			*/
	uint32_t		match_addr:8,	/*!< match addr. code points 			*/
					s_id:24; 		/*!< Requester Port_ID 					*/
	uint32_t		resp_flags:8,	/*!< Responder flags 					*/
					d_id:24; 		/*!< Responder Port_ID 					*/
	bfad_ip_wwpn_t	s_wwpn; 		/*!< wwpn of Requester 					*/
	bfad_ip_wwpn_t	s_wwnn; 		/*!< wwnn of Requester:OPTIONAL 		*/
	bfad_ip_wwpn_t	d_wwpn; 		/*!< wwpn of Responder 					*/
	bfad_ip_wwpn_t	d_wwnn; 		/*!< wwnn of Responder:OPTIONAL 		*/
	uint8_t			s_ip[16]; 		/*!< IP address or Requester:OPTIONAL 	*/
	uint8_t			d_ip[16];		/*!< IP address or Responder:OPTIONAL 	*/
} bfad_ip_els_farp_t;

/* Match address code points */
#define BFAD_IP_FARP_MATCH_RSVD				0x00
#define BFAD_IP_FARP_MATCH_WW_PN			0x01
#define BFAD_IP_FARP_MATCH_WW_NN			0x02
#define BFAD_IP_FARP_MATCH_WW_PN_NN			0x03
#define BFAD_IP_FARP_MATCH_IPv4				0x04
#define BFAD_IP_FARP_MATCH_WW_PN_IPv4		0x05
#define BFAD_IP_FARP_MATCH_WW_NN_IPv4		0x06
#define BFAD_IP_FARP_MATCH_WW_PN_NN_IPv4	0x07

/* Responder flags */
#define BFAD_IP_FARP_INIT_P_LOGI			0x1
#define BFAD_IP_FARP_INIT_REPLY				0x2
/*
 * @}
 */

/**
 * @dg ip_private private functions
 * @{
 */

/* Proto Types */
void bfad_ip_add_headers_ship_out(struct sk_buff *skb, int fc_type,
			bfad_iprp_t *arp, BFAD_IP_PKT_TYPE pkt_type);
#ifdef IPV6_SUPPORT
static void bfad_ip_copy_ipv6(struct net_device *dev, uint8_t *optr,
			uint8_t *iptr, int dir, int pid, int options_offset,
			int options_length);
static void bfad_ip_change_ipv6_addr(struct net_device *dev);
#endif
#ifdef SYSFS
static ssize_t
			bfad_ip_show_arp_cache(struct device *in_dev, struct
			device_attribute *attr, char *buf);
#endif
int 		bfad_ip_map_dma(struct bfad_s *ioc, struct sk_buff *skb,
			bfad_pkt_t *pkt, int start_sge_index, int direction);
int 		bfad_ip_get_ip(struct net_device *dev, struct in_addr *addr);
int 		bfad_ip_ipv6_pkt_has_ll_address(struct ipv6hdr *ip6h,
			int *options_offset, int *options_length);
bfa_status_t 	bfad_ip_alloc_bfa_resources(bfa_meminfo_t *mem_info);
int 		bfad_ip_start_sysfs(struct net_device *dev);
void 		bfad_ip_stop_sysfs(struct net_device *dev);
int 		bfad_ipfc_port_new_real(struct bfad_s *bfad, struct bfad_port_s
			*port, int port_type);
int			bfad_ip_os_ioctl_init(void);
int			bfad_ip_os_ioctl_exit(void);
int			bfad_ip_update_arp_cache(struct net_device *dev, bfad_iprp_t *arp,
			 BFAD_IP_ARP_STATE);



/**
 * @dg ip_hash_private hash table functions
 * @{
 */
int bfad_ip_pid_hash(bfad_iprp_t *arp);
int bfad_ip_key_cmp_pid(bfad_iprp_t *arp1, bfad_iprp_t *arp2);
/**
 * @}
 */

/**
 * @}
 */

/* vim:set ts=4 sw=4 noet ff=unix ft=c tw=78 ai: */
#endif /* __IPFC_H__ */
