/*
 * Copyright (c) 2005-2014 Brocade Communications Systems, Inc.
 * Copyright (c) 2014 QLogic Corporation.
 * All rights reserved
 * www.qlogic.com
 * See LICENSE.bfa for copyright and licensing details.
 */

#include <linux/device.h>

#include <linux/ip.h>
#include <linux/ipv6.h>
#include <linux/in.h>

#include <fcs/bfa_fcs_ipfc.h>

#include "bfad_ioctl_cna.h"
#include "bfad_ioctl.h"
#include "bfad_drv.h"
#include "bfad_ip_drv.h"
#include "bfad_ipfc.h"

static int      bfad_ip_open(struct inode *inode, struct file *file);
static int      bfad_ip_close(struct inode *inode, struct file *file);
int		bfad_ip_ioctl(struct inode *inode, struct file *file,
			   unsigned int cmd, unsigned long arg);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 13)
static struct class *bfad_ip_class;
#else
static struct class_simple *bfad_ip_class;
#endif

static int      bfad_ip_major;

static char    *bfad_ip_dev_name = BFAD_IP_DRIVER_NAME;
static struct file_operations bfad_ip_fops;

int
bfad_ioctl_ipfc_get_drv_stats(unsigned long arg, void *cmd)
{
	bfa_ioctl_ipfc_stats_t *iocmd = (bfa_ioctl_ipfc_stats_t *)cmd;
	struct bfad_s	*bfad;
	unsigned long   flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd->status = bfad_ip_get_drv_stats(bfad, &iocmd->stats);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	return 0;
}

int
bfad_ioctl_ipfc_reset_drv_stats(unsigned long arg, void *cmd)
{
	bfa_ioctl_gen_t *iocmd = (bfa_ioctl_gen_t *)cmd;
	struct bfad_s   *bfad;


	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return EINVAL;

	iocmd->status = bfad_ip_reset_drv_stats(bfad);

	return 0;
}

int
bfad_ioctl_ipfc_get_attr(unsigned long arg, void *cmd)
{
	bfa_ioctl_ipfc_attr_t   *iocmd = (bfa_ioctl_ipfc_attr_t *)cmd;
	struct bfad_s	*bfad;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return EINVAL;

	bfad_ip_get_attr(bfad, &iocmd->attr);

	iocmd->status = BFA_STATUS_OK;

	return 0;
}

int
bfad_ip_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
	   unsigned long arg)
{
	int rc = EINVAL, size = _IOC_SIZE(cmd);
	void *iocmd = NULL;

	if (!arg)
		return -rc;

	if (!bfad_is_ready()) {
		rc = EAGAIN;
		return -rc;
	}
	if (cmd == IOCMD_IOC_GET_VERSION) {
		rc = bfad_ioctl_ioc_get_version(arg);
		goto ext;
	}

	if (cmd == IOCMD_IOC_GET_INSTS) {
		rc = bfad_ioctl_ioc_get_inst(arg);
		goto ext;
	}

	iocmd = bfad_cp_from_user(size, arg);
	if (!iocmd) {
		return -EIO;
	}
	switch (cmd) {
	/*
	 * IOC IOCTLs
	 */
	case IOCMD_IPFC_GET_STATS:
		rc = bfad_ioctl_ipfc_get_drv_stats(arg, iocmd);
		break;
	case IOCMD_IPFC_RESET_STATS:
		rc = bfad_ioctl_ipfc_reset_drv_stats(arg, iocmd);
		break;
	case IOCMD_IPFC_GET_ATTR:
		rc = bfad_ioctl_ipfc_get_attr(arg, iocmd);
		break;

	default:
		rc = EINVAL;
		break;
	}

	if (bfad_cp_to_user(size, arg, iocmd)) {
		rc = EIO;
	}

ext:
	return -rc;
}

int
bfad_ip_os_class_create(struct module *owner, char *name)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 13)
	bfad_ip_class = class_create(owner, name);
#else
	bfad_ip_class = class_simple_create(owner, name);
#endif
	if (IS_ERR(bfad_ip_class)) {
		printk(KERN_WARNING "Unable to create class for bfaipfc.\n");
		bfad_ip_class = NULL;
		return 1;
	}

	return 0;
}

void
bfad_ip_os_class_destroy(void)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 13)
	class_destroy(bfad_ip_class);
#else
	class_simple_destroy(bfad_ip_class);
#endif
	bfad_ip_class = NULL;
}

int
bfad_ip_os_class(void)
{
	return (bfad_ip_class != 0);
}

int
bfad_ip_os_ioctl_init(void)
{

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 26)
	struct device *cd;
#else
	struct class_device *cd;
#endif

	if (bfad_ip_os_class_create(THIS_MODULE, bfad_ip_dev_name)) {
		printk(KERN_WARNING "Unable to create sysfs class for bfa.\n");
		return 1;
	}

	bfad_ip_major = register_chrdev(0, bfad_ip_dev_name, &bfad_ip_fops);
	if (bfad_ip_major < 0) {
		printk(KERN_WARNING "Unable to register chrdev for bfa.\n");
		bfad_ip_os_class_destroy();
		return bfad_ip_major;
	}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 27)
	cd = device_create(bfad_ip_class, NULL, MKDEV(bfad_ip_major, 0), NULL,
							"bfaipfc");
#elif LINUX_VERSION_CODE == KERNEL_VERSION(2, 6, 26)
	cd = device_create(bfad_ip_class, NULL, MKDEV(bfad_ip_major, 0),
							"bfaipfc");
#elif LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 13)
	cd = class_simple_device_add(bfad_ip_class, MKDEV(bfad_ip_major, 0),
						NULL, "bfaipfc");
#elif LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 15)
	cd = class_device_create(bfad_ip_class, MKDEV(bfad_ip_major, 0),
						NULL, "bfa");
#else
	cd = class_device_create(bfad_ip_class, NULL, MKDEV(bfad_ip_major, 0),
						NULL, "bfaipfc");
#endif

	if (IS_ERR(cd))
		return BFA_STATUS_FAILED;

	return BFA_STATUS_OK;
}

int
bfad_ip_os_ioctl_exit(void)
{
	if (!bfad_ip_os_class())
		return 1;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 26)
	device_destroy(bfad_ip_class, MKDEV(bfad_ip_major, 0));
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 13)
	class_device_destroy(bfad_ip_class, MKDEV(bfad_ip_major, 0));
#else
	class_simple_device_remove(MKDEV(bfad_ip_major, 0));
#endif

	unregister_chrdev(bfad_ip_major, bfad_ip_dev_name);
	bfad_ip_os_class_destroy();
	return 0;
}

#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 14)
static long
bfad_ip_compat_ioctl(struct file *file, unsigned int cmd, unsigned long arg)
{
	return (bfad_ip_ioctl(file->f_dentry->d_inode, file, cmd, arg));
}
#endif

static int
bfad_ip_open(struct inode *inode, struct file *file)
{
	struct bfad_s   *bfad;
	int32_t	bfad_ip_minor;

	if (!bfad_is_ready())
		return (-EAGAIN);

	bfad_ip_minor = MINOR(inode->i_rdev);
	bfad = bfad_find_bfad_by_inst_no(bfad_ip_minor);
	if (bfad == NULL) {
		printk(KERN_WARNING
			"%s:%d: Invalid minor device=%d for file=%p\n",
			__FUNCTION__, __LINE__, bfad_ip_minor, file);
		return (-EINVAL);
	}

	return 0;
}

static int
bfad_ip_close(struct inode *inode, struct file *file)
{
	struct bfad_s   *bfad;
	int32_t	bfad_ip_minor;

	if (!bfad_is_ready())
		return (-EAGAIN);

	bfad_ip_minor = MINOR(inode->i_rdev);
	bfad = bfad_find_bfad_by_inst_no(bfad_ip_minor);
	if (bfad == NULL) {
		printk(KERN_WARNING
			"%s:%d: Invalid minor device=%d for file=%p\n",
			__FUNCTION__, __LINE__, bfad_ip_minor, file);
		return (-EINVAL);
	}

	return (0);
}

/**
 * Define the IPFC driver file operations.
 */
static struct file_operations bfad_ip_fops = {
	.owner = THIS_MODULE,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 36)
	.unlocked_ioctl = bfad_ip_compat_ioctl,
#else
	.ioctl = bfad_ip_ioctl,
#endif
#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 14)
	.compat_ioctl = bfad_ip_compat_ioctl,
#endif
	.open = bfad_ip_open,
	.release = bfad_ip_close,
};
