/*
 * Copyright (c) 2005-2014 Brocade Communications Systems, Inc.
 * Copyright (c) 2014 QLogic Corporation.
 * All rights reserved
 * www.qlogic.com
 *
 * See LICENSE.bfa for copyright and licensing details.
 */
#include <bfa/bfa_modules.h>
#include <bfa/bfa_dconf.h>

BFA_TRC_FILE(HAL, DCONF);
BFA_MODULE(dconf);

/*
 * forward declarations
 */

void bfa_dconf_cbfn(void *dconf, bfa_status_t status);
static void bfa_dconf_timer(void *cbarg);
bfa_status_t bfa_dconf_flash_write(bfa_dconf_mod_t *dconf);
void bfa_dconf_init_cb(void *arg, bfa_status_t status);
/**
 * DCONF state machine events
 */
typedef enum {
	BFA_DCONF_SM_INIT		= 1,	/* !< Init pBind */
	BFA_DCONF_SM_FLASH_COMP		= 2,	/* !< read/write to flash */
	BFA_DCONF_SM_WR			= 3,	/* !< binding change, map */
	BFA_DCONF_SM_TIMEOUT		= 4,	/* !< Start timer */
	BFA_DCONF_SM_EXIT		= 5,	/* !< exit pBind module */
	BFA_DCONF_SM_IOCDISABLE		= 6,	/* !< exit pBind module */
} bfa_dconf_event_t;

/*
 * forward declaration of DCONF state machine
 */
static void bfa_dconf_sm_uninit(bfa_dconf_mod_t *dconf,
	 bfa_dconf_event_t event);
static void bfa_dconf_sm_flash_read(bfa_dconf_mod_t *dconf,
	 bfa_dconf_event_t event);
static void bfa_dconf_sm_ready(bfa_dconf_mod_t *dconf,
	 bfa_dconf_event_t event);
static void bfa_dconf_sm_dirty(bfa_dconf_mod_t *dconf, bfa_dconf_event_t event);
static void bfa_dconf_sm_sync(bfa_dconf_mod_t *dconf, bfa_dconf_event_t event);
static void bfa_dconf_sm_final_sync(bfa_dconf_mod_t *dconf,
	 bfa_dconf_event_t event);
static void bfa_dconf_sm_iocdown_dirty(bfa_dconf_mod_t *dconf,
	 bfa_dconf_event_t event);

/**
 * @brief
 * Begining state of dconf module. Waiting for an event to start.
 */
static void
bfa_dconf_sm_uninit(bfa_dconf_mod_t *dconf, bfa_dconf_event_t event)
{
	bfa_status_t bfa_status;
	bfa_trc(dconf->bfa, event);

	switch (event) {
		case BFA_DCONF_SM_INIT:

		if (dconf->min_cfg) {
			bfa_trc(dconf->bfa, dconf->min_cfg);
			bfa_iocfc_dconf_cb(dconf->bfa, BFA_STATUS_OK);
			return;
		}
		bfa_sm_set_state(dconf, bfa_dconf_sm_flash_read);
		bfa_timer_start(dconf->bfa, &dconf->timer,
			 bfa_dconf_timer, dconf, (2 * BFA_DCONF_UPDATE_TOV));
		bfa_status = bfa_flash_read_part(BFA_FLASH(dconf->bfa),
				BFA_FLASH_PART_DRV, dconf->instance,
				dconf->dconf,
				sizeof(bfa_dconf_t), 0, bfa_dconf_init_cb,
				dconf->bfa, 1);

		if (bfa_status != BFA_STATUS_OK) {
			bfa_timer_stop(&dconf->timer);
			bfa_dconf_init_cb(dconf->bfa, BFA_STATUS_FAILED);
			bfa_sm_set_state(dconf, bfa_dconf_sm_uninit);
			return;
		}

		break;
	case BFA_DCONF_SM_EXIT:
		bfa_iocfc_dconf_cb(dconf->bfa, BFA_STATUS_OK);
	case BFA_DCONF_SM_IOCDISABLE:
	case BFA_DCONF_SM_WR:
	case BFA_DCONF_SM_FLASH_COMP:
		break;

	default:
		bfa_sm_fault(dconf->bfa, event);
	}
}

/**
 * @brief
 * Read flash for dconf entries and make a call back to the driver once done.
 */
static void
bfa_dconf_sm_flash_read(bfa_dconf_mod_t *dconf,
	 bfa_dconf_event_t event)
{
	bfa_trc(dconf->bfa, event);

	switch (event) {
	case BFA_DCONF_SM_FLASH_COMP:
		bfa_timer_stop(&dconf->timer);
		bfa_sm_set_state(dconf, bfa_dconf_sm_ready);
		break;

	case BFA_DCONF_SM_TIMEOUT:
		bfa_sm_set_state(dconf, bfa_dconf_sm_ready);
		bfa_iocfc_dconf_cb(dconf->bfa, BFA_STATUS_FAILED);
		break;

	case BFA_DCONF_SM_EXIT:
		bfa_timer_stop(&dconf->timer);
		bfa_sm_set_state(dconf, bfa_dconf_sm_uninit);
		bfa_iocfc_dconf_cb(dconf->bfa, BFA_STATUS_OK);
		break;

	case BFA_DCONF_SM_IOCDISABLE:
		bfa_timer_stop(&dconf->timer);
		bfa_sm_set_state(dconf, bfa_dconf_sm_uninit);
		break;
	default:
		bfa_sm_fault(dconf->bfa, event);
	}
}

/**
 * @brief
 * DCONF Module is in ready state. Has completed the initialization.
 */

static void
bfa_dconf_sm_ready(bfa_dconf_mod_t *dconf, bfa_dconf_event_t event)
{
	bfa_trc(dconf->bfa, event);

	switch (event) {
	case BFA_DCONF_SM_WR:
		bfa_timer_start(dconf->bfa, &dconf->timer,
			bfa_dconf_timer, dconf, BFA_DCONF_UPDATE_TOV);
		bfa_sm_set_state(dconf, bfa_dconf_sm_dirty);
		break;

	case BFA_DCONF_SM_EXIT:
		bfa_sm_set_state(dconf, bfa_dconf_sm_uninit);
		bfa_iocfc_dconf_cb(dconf->bfa, BFA_STATUS_OK);
		break;

	case BFA_DCONF_SM_INIT:
	case BFA_DCONF_SM_IOCDISABLE:
		break;

	default:
		bfa_sm_fault(dconf->bfa, event);
	}
}

/**
 * @brief
 * pBind entries are dirty, write back to the flash.
 */

static void
bfa_dconf_sm_dirty(bfa_dconf_mod_t *dconf, bfa_dconf_event_t event)
{
	bfa_trc(dconf->bfa, event);

	switch (event) {
	case BFA_DCONF_SM_TIMEOUT:
		bfa_sm_set_state(dconf, bfa_dconf_sm_sync);
		bfa_dconf_flash_write(dconf);
		break;

	case BFA_DCONF_SM_WR:
		bfa_timer_stop(&dconf->timer);
		bfa_timer_start(dconf->bfa, &dconf->timer,
			 bfa_dconf_timer, dconf, BFA_DCONF_UPDATE_TOV);
		break;

	case BFA_DCONF_SM_EXIT:
		bfa_timer_stop(&dconf->timer);
		bfa_timer_start(dconf->bfa, &dconf->timer,
			 bfa_dconf_timer, dconf, BFA_DCONF_UPDATE_TOV);
		bfa_sm_set_state(dconf, bfa_dconf_sm_final_sync);
		bfa_dconf_flash_write(dconf);
		break;

	case BFA_DCONF_SM_FLASH_COMP:
		break;
	case BFA_DCONF_SM_IOCDISABLE:
		bfa_timer_stop(&dconf->timer);
		bfa_sm_set_state(dconf, bfa_dconf_sm_iocdown_dirty);
		break;
	default:
		bfa_sm_fault(dconf->bfa, event);
	}
}

/**
 * @brief
 * Sync the dconf entries to the flash.
 */

static void
bfa_dconf_sm_final_sync(bfa_dconf_mod_t *dconf,
	 bfa_dconf_event_t event)
{
	bfa_trc(dconf->bfa, event);

	switch (event) {
	case BFA_DCONF_SM_IOCDISABLE:
	case BFA_DCONF_SM_FLASH_COMP:
		bfa_timer_stop(&dconf->timer);
	case BFA_DCONF_SM_TIMEOUT:
		bfa_sm_set_state(dconf, bfa_dconf_sm_uninit);
		bfa_iocfc_dconf_cb(dconf->bfa, BFA_STATUS_OK);
		break;

	default:
		bfa_sm_fault(dconf->bfa, event);
	}
}
static void
bfa_dconf_sm_sync(bfa_dconf_mod_t *dconf, bfa_dconf_event_t event)
{
	bfa_trc(dconf->bfa, event);

	switch (event) {
	case BFA_DCONF_SM_FLASH_COMP:
		bfa_sm_set_state(dconf, bfa_dconf_sm_ready);
		break;

	case BFA_DCONF_SM_WR:
		bfa_timer_start(dconf->bfa, &dconf->timer,
			 bfa_dconf_timer, dconf, BFA_DCONF_UPDATE_TOV);
		bfa_sm_set_state(dconf, bfa_dconf_sm_dirty);
		break;

	case BFA_DCONF_SM_EXIT:
		bfa_timer_start(dconf->bfa, &dconf->timer,
			 bfa_dconf_timer, dconf, BFA_DCONF_UPDATE_TOV);
		bfa_sm_set_state(dconf, bfa_dconf_sm_final_sync);
		break;
	case BFA_DCONF_SM_IOCDISABLE:
		bfa_sm_set_state(dconf, bfa_dconf_sm_iocdown_dirty);
		break;

	default:
		bfa_sm_fault(dconf->bfa, event);
	}
}

static void
bfa_dconf_sm_iocdown_dirty(bfa_dconf_mod_t *dconf, bfa_dconf_event_t event)
{
	bfa_trc(dconf->bfa, event);

	switch (event) {
	case BFA_DCONF_SM_INIT:
		bfa_timer_start(dconf->bfa, &dconf->timer,
			bfa_dconf_timer, dconf, BFA_DCONF_UPDATE_TOV);
		bfa_sm_set_state(dconf, bfa_dconf_sm_dirty);
		break;
	case BFA_DCONF_SM_EXIT:
		bfa_sm_set_state(dconf, bfa_dconf_sm_uninit);
		bfa_iocfc_dconf_cb(dconf->bfa, BFA_STATUS_OK);
		break;
	case BFA_DCONF_SM_IOCDISABLE:
		break;
	default:
		bfa_sm_fault(dconf->bfa, event);
	}
}
/**
 * @brief
 * Compute and return memory needed by DRV_CFG module.
 */

static void
bfa_dconf_meminfo(struct bfa_iocfc_cfg_s *cfg, struct bfa_meminfo_s *meminfo,
		  struct bfa_s *bfa)
{
	struct bfa_mem_kva_s *dconf_kva = bfa ? BFA_MEM_DCONF_KVA(bfa) : NULL;

	if (cfg->drvcfg.min_cfg)
		bfa_mem_kva_setup(meminfo, dconf_kva, sizeof(bfa_dconf_hdr_t));
	else
		bfa_mem_kva_setup(meminfo, dconf_kva, sizeof(bfa_dconf_t));
}

/**
 * @brief
 */
static void
bfa_dconf_attach(bfa_t *bfa, void *bfad, struct bfa_iocfc_cfg_s *cfg,
		 struct bfa_pcidev_s *pcidev)
{
	bfa_dconf_mod_t *dconf = BFA_DCONF_MOD(bfa);

	dconf->bfad = bfad;
	dconf->bfa = bfa;
	dconf->instance = bfa_port_id_get(bfa);

	bfa_trc(bfa, dconf->instance);

	dconf->dconf = (bfa_dconf_t *) bfa_mem_kva_curp(dconf);
	if (cfg->drvcfg.min_cfg) {
		bfa_mem_kva_curp(dconf) += sizeof(bfa_dconf_hdr_t);
		dconf->min_cfg = BFA_TRUE;
	} else {
		dconf->min_cfg = BFA_FALSE;
		bfa_mem_kva_curp(dconf) += sizeof(bfa_dconf_t);
	}

	bfa_dconf_read_data_valid(bfa) = BFA_FALSE;
	bfa_sm_set_state(dconf, bfa_dconf_sm_uninit);
}
void
bfa_dconf_init_cb(void *arg, bfa_status_t status)
{
	bfa_t *bfa = arg;
	bfa_dconf_mod_t *dconf = BFA_DCONF_MOD(bfa);
	if (status == BFA_STATUS_OK) {
		bfa_dconf_read_data_valid(bfa) = BFA_TRUE;
		if (dconf->dconf->hdr.signature != BFI_DCONF_SIGNATURE)
			dconf->dconf->hdr.signature = BFI_DCONF_SIGNATURE;
		if (dconf->dconf->hdr.version != BFI_DCONF_VERSION)
			dconf->dconf->hdr.version = BFI_DCONF_VERSION;
	}
	bfa_sm_send_event(dconf, BFA_DCONF_SM_FLASH_COMP);
	bfa_iocfc_dconf_cb(bfa, BFA_STATUS_OK);
}

void
bfa_dconf_modinit(bfa_t *bfa)
{
	bfa_dconf_mod_t *dconf = BFA_DCONF_MOD(bfa);
	bfa_sm_send_event(dconf, BFA_DCONF_SM_INIT);
}
static void
bfa_dconf_start(bfa_t *bfa)
{
}

static void
bfa_dconf_stop(bfa_t *bfa)
{
}


static void bfa_dconf_timer(void *cbarg)
{
	bfa_dconf_mod_t *dconf = cbarg;
	bfa_sm_send_event(dconf, BFA_DCONF_SM_TIMEOUT);
}
static void
bfa_dconf_iocdisable(bfa_t *bfa)
{
	bfa_dconf_mod_t *dconf = BFA_DCONF_MOD(bfa);
	bfa_sm_send_event(dconf, BFA_DCONF_SM_IOCDISABLE);
}

static void
bfa_dconf_detach(bfa_t *bfa)
{
}

bfa_status_t
bfa_dconf_flash_write(bfa_dconf_mod_t *dconf)
{
	bfa_status_t bfa_status;
	bfa_trc(dconf->bfa, 0);
	bfa_status = bfa_flash_update_part(BFA_FLASH(dconf->bfa),
		BFA_FLASH_PART_DRV, dconf->instance,
		dconf->dconf,  sizeof(bfa_dconf_t), 0,
		bfa_dconf_cbfn, dconf, 1);
	if (bfa_status != BFA_STATUS_OK) {
		bfa_assert(!bfa_status);
	}
	bfa_trc(dconf->bfa, bfa_status);
	return bfa_status;
}

bfa_status_t
bfa_dconf_update(struct bfa_s *bfa)
{
	bfa_dconf_mod_t *dconf = BFA_DCONF_MOD(bfa);
	bfa_trc(dconf->bfa, 0);
	if (bfa_sm_cmp_state(dconf, bfa_dconf_sm_iocdown_dirty)) {
		return BFA_STATUS_FAILED;
	}
	if (dconf->min_cfg) {
		bfa_trc(dconf->bfa, dconf->min_cfg);
		return BFA_STATUS_FAILED;
	}
	bfa_sm_send_event(dconf, BFA_DCONF_SM_WR);
	return BFA_STATUS_OK;
}

void
bfa_dconf_cbfn(void *arg, bfa_status_t status)
{
	bfa_dconf_mod_t *dconf = arg;
	bfa_assert(!status);
	bfa_sm_send_event(dconf, BFA_DCONF_SM_FLASH_COMP);
}

void
bfa_dconf_modexit(bfa_t *bfa)
{
	bfa_dconf_mod_t *dconf = BFA_DCONF_MOD(bfa);
	bfa_sm_send_event(dconf, BFA_DCONF_SM_EXIT);
}
