/*
 * Copyright (c) 2005-2014 Brocade Communications Systems, Inc.
 * Copyright (c) 2014 QLogic Corporation.
 * All rights reserved
 * www.qlogic.com
 *
 * See LICENSE.bfa for copyright and licensing details.
 */

#include <bfa/bfa_modules.h>
#include <bfa/bfa_fcpim.h>

BFA_MODULE(fcp);
BFA_TRC_FILE(HAL, FCP);

/**
 * @dg hal_fcp_mod BFA FCP module - common module for FCPIM and FCPTM
 * @{
 */

static void bfa_iotag_attach(struct bfa_fcp_mod_s *fcp);
static void bfa_iotag_detach(struct bfa_fcp_mod_s *fcp);

static void
bfa_fcp_meminfo(struct bfa_iocfc_cfg_s *cfg, struct bfa_meminfo_s *minfo,
		struct bfa_s *bfa)
{
	struct bfa_fcp_mod_s *fcp = bfa ? BFA_FCP_MOD(bfa) : NULL;
	struct bfa_mem_kva_s *fcp_kva = bfa ? BFA_MEM_FCP_KVA(bfa) : NULL;
	uint32_t km_len = 0;
	uint16_t num_io_req;

	/*
	 * ZERO for num_ioim_reqs and num_fwtio_reqs is allowed config value.
	 * So if the values are non zero, adjust them appropriately.
	 */
	if (cfg->fwcfg.num_ioim_reqs &&
			cfg->fwcfg.num_ioim_reqs < BFA_IOIM_MIN)
		cfg->fwcfg.num_ioim_reqs = BFA_IOIM_MIN;
	else if (cfg->fwcfg.num_ioim_reqs > BFA_IOIM_MAX)
		cfg->fwcfg.num_ioim_reqs = BFA_IOIM_MAX;

	/*
	 * num_fwtio_reqs < BFA_FWTIO_MIN (Now it is zero) check can be added
	 * when complete target functionality is available.
	 */
	if (cfg->fwcfg.num_fwtio_reqs &&
			cfg->fwcfg.num_fwtio_reqs < BFA_FWTIO_MIN)
		cfg->fwcfg.num_fwtio_reqs = BFA_FWTIO_MIN;
	else if (cfg->fwcfg.num_fwtio_reqs > BFA_FWTIO_MAX)
		cfg->fwcfg.num_fwtio_reqs = BFA_FWTIO_MAX;

	num_io_req = (cfg->fwcfg.num_ioim_reqs + cfg->fwcfg.num_fwtio_reqs);
	if (num_io_req > BFA_IO_MAX) {
		if (cfg->fwcfg.num_ioim_reqs && cfg->fwcfg.num_fwtio_reqs) {
			cfg->fwcfg.num_ioim_reqs = BFA_IO_MAX/2;
			cfg->fwcfg.num_fwtio_reqs = BFA_IO_MAX/2;
		} else if (cfg->fwcfg.num_fwtio_reqs) {
			cfg->fwcfg.num_fwtio_reqs = BFA_FWTIO_MAX;
		} else {
			cfg->fwcfg.num_ioim_reqs = BFA_IOIM_MAX;
		}
	}

	bfa_fcpim_meminfo(cfg, &km_len);

	num_io_req = (cfg->fwcfg.num_ioim_reqs + cfg->fwcfg.num_fwtio_reqs);
	km_len += num_io_req * sizeof(struct bfa_iotag_s);
	km_len += cfg->fwcfg.num_rports * sizeof(struct bfa_itn_s);

	/* dma memory */
	if (!fcp)
		bfa_mem_dma_setup(minfo, NULL, num_io_req * BFI_IOIM_SNSLEN);
	else {
		struct bfa_mem_dma_s *seg_ptr;
		uint16_t nsegs, idx, per_seg_ios;

		nsegs = BFI_MEM_DMA_NSEGS(num_io_req, BFI_IOIM_SNSLEN);
		per_seg_ios = BFI_MEM_NREQS_SEG(BFI_IOIM_SNSLEN);

		bfa_mem_dma_seg_iter(fcp, seg_ptr, nsegs, idx) {
			if (num_io_req >= per_seg_ios) {
				num_io_req -= per_seg_ios;
				bfa_mem_dma_setup(minfo, seg_ptr,
					per_seg_ios * BFI_IOIM_SNSLEN);
			} else
				bfa_mem_dma_setup(minfo, seg_ptr,
					num_io_req * BFI_IOIM_SNSLEN);
		}
	}

	/* kva memory */
	bfa_mem_kva_setup(minfo, fcp_kva, km_len);
}

static void
bfa_fcp_attach(struct bfa_s *bfa, void *bfad, struct bfa_iocfc_cfg_s *cfg,
		struct bfa_pcidev_s *pcidev)
{
	struct bfa_fcp_mod_s *fcp = BFA_FCP_MOD(bfa);
	struct bfa_mem_dma_s *seg_ptr;
	uint16_t idx, nsegs, num_io_req;

	fcp->max_ioim_reqs = cfg->fwcfg.num_ioim_reqs;
	fcp->num_ioim_reqs = cfg->fwcfg.num_ioim_reqs;
	fcp->num_fwtio_reqs  = cfg->fwcfg.num_fwtio_reqs;
	fcp->num_itns	   = cfg->fwcfg.num_rports;
	fcp->bfa = bfa;

	/**
	 * Setup the pool of snsbase addr's, that is passed to fw as
	 * part of bfi_iocfc_cfg_s.
	 */
	num_io_req = (cfg->fwcfg.num_ioim_reqs + cfg->fwcfg.num_fwtio_reqs);
	nsegs = BFI_MEM_DMA_NSEGS(num_io_req, BFI_IOIM_SNSLEN);

	bfa_mem_dma_seg_iter(fcp, seg_ptr, nsegs, idx) {

		if (!bfa_mem_dma_virt(seg_ptr))
			break;

		fcp->snsbase[idx].pa = bfa_mem_dma_phys(seg_ptr);
		fcp->snsbase[idx].kva = bfa_mem_dma_virt(seg_ptr);
		bfa_iocfc_set_snsbase(bfa, idx, fcp->snsbase[idx].pa);
	}

	fcp->throttle_update_required = 1;

	bfa_fcpim_attach(fcp, bfad, cfg, pcidev);

	bfa_iotag_attach(fcp);

	fcp->itn_arr = (struct bfa_itn_s *) bfa_mem_kva_curp(fcp);
	bfa_mem_kva_curp(fcp) = (uint8_t *)fcp->itn_arr +
			(fcp->num_itns * sizeof(struct bfa_itn_s));
	bfa_os_memset(fcp->itn_arr, 0,
			(fcp->num_itns * sizeof(struct bfa_itn_s)));
}

static void
bfa_fcp_detach(struct bfa_s *bfa)
{
	struct bfa_fcp_mod_s *fcp = BFA_FCP_MOD(bfa);

	bfa_iotag_detach(fcp);
	bfa_fcpim_detach(fcp);
}

static void
bfa_fcp_start(struct bfa_s *bfa)
{
	struct bfa_fcp_mod_s *fcp = BFA_FCP_MOD(bfa);

	bfa_fcpim_start(fcp);
}

static void
bfa_fcp_stop(struct bfa_s *bfa)
{
	struct bfa_fcp_mod_s *fcp = BFA_FCP_MOD(bfa);

	bfa_fcpim_stop(fcp);
}

static void
bfa_fcp_iocdisable(struct bfa_s *bfa)
{
	struct bfa_fcp_mod_s *fcp = BFA_FCP_MOD(bfa);

	bfa_fcpim_iocdisable(fcp);

	if (bfa->tm_ops->iocdisable)
		bfa->tm_ops->iocdisable(bfa);
}

/**
 * @ig bfa_fcp_public
 * @{
 */

void
bfa_itn_create(struct bfa_s *bfa, struct bfa_rport_s *rport,
		void (*isr)(struct bfa_s *bfa, struct bfi_msg_s *m))
{
	struct bfa_fcp_mod_s *fcp = BFA_FCP_MOD(bfa);
	struct bfa_itn_s *itn;

	itn =  BFA_ITN_FROM_TAG(fcp, rport->rport_tag);
	itn->isr = isr;
}

/**
 * @brief
 *	Itn interrupt processing.
 */
void
bfa_itn_isr(struct bfa_s *bfa, struct bfi_msg_s *m)
{
	struct bfa_fcp_mod_s *fcp = BFA_FCP_MOD(bfa);
	union bfi_itn_i2h_msg_u msg;
	struct bfa_itn_s *itn;

	msg.msg = m;
	itn =  BFA_ITN_FROM_TAG(fcp, msg.create_rsp->bfa_handle);

	if (itn->isr)
		itn->isr(bfa, m);
	else
		bfa_assert(0);
}

bfa_status_t
bfa_fcp_alloc_iotag(struct bfa_s *bfa, itn_role_t role, uint16_t *tag)
{
	struct bfa_fcp_mod_s *fcp = BFA_FCP_MOD(bfa);
	struct bfa_iotag_s *iotag = NULL;

	if (role == BFI_ITN_ROLE_IM) {
		bfa_q_deq(&fcp->iotag_ioim_free_q, &iotag);
	} else if (role == BFI_ITN_ROLE_TM) {
		bfa_q_deq(&fcp->iotag_tio_free_q, &iotag);
	} else {
		bfa_trc(bfa, role);
		bfa_assert(0);
	}

	if (!iotag)
		return BFA_STATUS_FAILED;

	*tag = iotag->tag;
	return BFA_STATUS_OK;
}

void
bfa_fcp_release_iotag(struct bfa_s *bfa, uint16_t tag)
{
	struct bfa_fcp_mod_s *fcp = BFA_FCP_MOD(bfa);
	struct bfa_iotag_s *iotag;

	bfa_assert(tag < (fcp->num_ioim_reqs + fcp->num_fwtio_reqs));
	iotag = BFA_IOTAG_FROM_TAG(fcp, tag);

	if (tag < fcp->num_ioim_reqs) {
		bfa_q_enq(&fcp->iotag_ioim_free_q, iotag);
	} else {
		bfa_q_enq(&fcp->iotag_tio_free_q, iotag);
	}
}

void
bfa_fcp_res_recfg(struct bfa_s *bfa, uint16_t num_ioim_fw, uint16_t max_ioim_fw)
{
	struct bfa_fcp_mod_s	*mod = BFA_FCP_MOD(bfa);
	struct bfa_q_s		*qe;
	int			i;

	/* Update io throttle value only once during driver load time */
	if (!mod->throttle_update_required)
		return;

	for (i = 0; i < (mod->num_ioim_reqs - num_ioim_fw); i++) {
		bfa_q_deq_tail(&mod->iotag_ioim_free_q, &qe);
		bfa_q_enq(&mod->iotag_unused_q, qe);
	}

	if (mod->num_ioim_reqs != num_ioim_fw) {
		bfa_trc(bfa, mod->num_ioim_reqs);
		bfa_trc(bfa, num_ioim_fw);
	}

	mod->max_ioim_reqs = max_ioim_fw;
	mod->num_ioim_reqs = num_ioim_fw;
	mod->throttle_update_required = 0;
}

/**
 * @}
 */

/**
 * @dg bfa_fcp_private
 * @{
 */

static void
bfa_iotag_attach(struct bfa_fcp_mod_s *fcp)
{
	struct bfa_iotag_s *iotag;
	uint16_t num_io_req, i;

	iotag = (struct bfa_iotag_s *) bfa_mem_kva_curp(fcp);
	fcp->iotag_arr = iotag;

	bfa_q_init(&fcp->iotag_ioim_free_q);
	bfa_q_init(&fcp->iotag_tio_free_q);
	bfa_q_init(&fcp->iotag_unused_q);

	num_io_req = fcp->num_ioim_reqs + fcp->num_fwtio_reqs;
	for (i = 0; i < num_io_req; i++, iotag++) {
		bfa_os_memset(iotag, 0, sizeof(struct bfa_iotag_s));
		iotag->tag = i;
		if (i < fcp->num_ioim_reqs) {
			bfa_q_enq(&fcp->iotag_ioim_free_q, iotag);
		} else {
			bfa_q_enq(&fcp->iotag_tio_free_q, iotag);
		}
	}

	bfa_mem_kva_curp(fcp) = (uint8_t *) iotag;
}


void
bfa_iotag_detach(struct bfa_fcp_mod_s *fcp)
{
}

/**
 * @}
 */
