/*
 * Copyright (c) 2005-2014 Brocade Communications Systems, Inc.
 * Copyright (c) 2014 QLogic Corporation.
 * All rights reserved
 * www.qlogic.com
 *
 * See LICENSE.bfa for copyright and licensing details.
 */
/**
 * @file bfa_ipfc.c BFA IP frame receive implementation
 */

#include <bfa/bfa_modules.h>
#include <bfi/bfi_ipfc.h>

BFA_TRC_FILE(HAL, IPFC);

#define BFA_IPFC_REQQ		(0)
#define BFA_IPFC_MOD(__bfa)	(&(__bfa)->modules.ipfc_mod)
#define BFA_IPFC_PKT(__ipfc_mod, __pkt_tag)	\
				(&(__ipfc_mod)->pkt_arr[__pkt_tag])

static bfa_boolean_t bfa_ipfc_post(bfa_ipfc_pkt_t *pkt);
static bfa_boolean_t bfa_ipfc_flush(bfa_ipfc_pkt_t *pkt);

/**
 * @ig ipfc_sm
 * @{
 */

typedef enum {
	BFA_IPFC_SM_POST     = 1,
	BFA_IPFC_SM_COMP     = 2,
	BFA_IPFC_SM_FLUSH    = 3,
	BFA_IPFC_SM_FLUSHED  = 4,
	BFA_IPFC_SM_RESUME   = 5,
	BFA_IPFC_SM_HWFAIL   = 6,
} bfa_ipfc_event_t;

static void bfa_ipfc_sm_uninit(bfa_ipfc_pkt_t *pkt, bfa_ipfc_event_t event);
static void bfa_ipfc_sm_wait(bfa_ipfc_pkt_t *pkt, bfa_ipfc_event_t event);
static void bfa_ipfc_sm_posted(bfa_ipfc_pkt_t *pkt, bfa_ipfc_event_t event);
static void bfa_ipfc_sm_flush_wait(bfa_ipfc_pkt_t *pkt, bfa_ipfc_event_t event);
static void bfa_ipfc_sm_flushing(bfa_ipfc_pkt_t *pkt, bfa_ipfc_event_t event);

static void
bfa_ipfc_free(bfa_ipfc_pkt_t *pkt)
{
	bfa_ipfc_mod_t	*mod = BFA_IPFC_MOD(pkt->bfa);

	bfa_q_qe_deq(pkt);
	bfa_q_enq(&mod->pkt_free_q, pkt);
}

static void
__bfa_ipfc_send_cbfn(void *cbarg, bfa_boolean_t complete)
{
	bfa_ipfc_pkt_t *pkt = cbarg;

	if (complete) {
		pkt->send_cbfn(pkt->dipfc, pkt->dpkt, pkt->status,
				pkt->pkt_len, pkt->residue_len);
	} else {
		bfa_ipfc_free(pkt);
	}
}

static void
bfa_ipfc_sm_uninit(bfa_ipfc_pkt_t *pkt, bfa_ipfc_event_t event)
{
	bfa_trc(pkt->bfa, event);
	switch (event) {
	case BFA_IPFC_SM_POST:
		bfa_sm_set_state(pkt, bfa_ipfc_sm_posted);
		if (!bfa_ipfc_post(pkt)) {
			bfa_sm_set_state(pkt, bfa_ipfc_sm_wait);
			bfa_reqq_wait(pkt->bfa, BFA_IPFC_REQQ, &pkt->reqq_wqe);
		}
		break;

	default:
		bfa_sm_fault(pkt->bfa, event);
	}
}

static void
bfa_ipfc_sm_wait(bfa_ipfc_pkt_t *pkt, bfa_ipfc_event_t event)
{

	bfa_trc(pkt->bfa, event);
	switch (event) {
	case BFA_IPFC_SM_RESUME:
		bfa_sm_set_state(pkt, bfa_ipfc_sm_posted);
		bfa_ipfc_post(pkt);
		break;

	case BFA_IPFC_SM_FLUSH:
		bfa_reqq_wcancel(&pkt->reqq_wqe);
		bfa_sm_set_state(pkt, bfa_ipfc_sm_uninit);
		pkt->pkt_len = pkt->residue_len = 0;
		pkt->status = BFA_IPFC_STATUS_OK;
		bfa_cb_queue(pkt->bfa, &pkt->hcb_qe,
				__bfa_ipfc_send_cbfn, pkt);
		break;

	case BFA_IPFC_SM_HWFAIL:
		bfa_reqq_wcancel(&pkt->reqq_wqe);
		bfa_sm_set_state(pkt, bfa_ipfc_sm_uninit);
		pkt->pkt_len = pkt->residue_len = 0;
		pkt->status = BFA_IPFC_STATUS_OK;
		bfa_cb_queue(pkt->bfa, &pkt->hcb_qe,
				__bfa_ipfc_send_cbfn, pkt);
		break;

	default:
		bfa_sm_fault(pkt->bfa, event);
	}
}

static void
bfa_ipfc_sm_posted(bfa_ipfc_pkt_t *pkt, bfa_ipfc_event_t event)
{
	bfa_ipfc_mod_t	*mod = BFA_IPFC_MOD(pkt->bfa);

	bfa_trc(pkt->bfa, event);
	switch (event) {
	case BFA_IPFC_SM_COMP:
		bfa_trc(pkt->bfa, pkt->pkt_len);
		bfa_sm_set_state(pkt, bfa_ipfc_sm_uninit);
		bfa_cb_queue(pkt->bfa, &pkt->hcb_qe,
				__bfa_ipfc_send_cbfn, pkt);
		break;

	case BFA_IPFC_SM_HWFAIL:
		bfa_sm_set_state(pkt, bfa_ipfc_sm_uninit);
		pkt->pkt_len = pkt->residue_len = 0;
		pkt->status = BFA_IPFC_STATUS_ERROR;
		bfa_cb_queue(pkt->bfa, &pkt->hcb_qe,
				__bfa_ipfc_send_cbfn, pkt);
		break;

	case BFA_IPFC_SM_FLUSH:
		bfa_sm_set_state(pkt, bfa_ipfc_sm_flushing);
		if (!bfa_ipfc_flush(pkt)) {
			bfa_sm_set_state(pkt, bfa_ipfc_sm_flush_wait);
			bfa_reqq_wait(pkt->bfa, BFA_IPFC_REQQ, &pkt->reqq_wqe);
		}
		bfa_q_qe_deq(pkt);
		bfa_q_enq(&mod->pkt_free_q, pkt);
		break;

	default:
		bfa_sm_fault(pkt->bfa, event);
	}
}

static void
bfa_ipfc_sm_flush_wait(bfa_ipfc_pkt_t *pkt, bfa_ipfc_event_t event)
{
	bfa_ipfc_mod_t	*mod = BFA_IPFC_MOD(pkt->bfa);

	bfa_trc(pkt->bfa, event);
	switch (event) {
	case BFA_IPFC_SM_COMP:
		bfa_sm_set_state(pkt, bfa_ipfc_sm_uninit);
		pkt->pkt_len = pkt->residue_len = 0;
		pkt->status = BFA_IPFC_STATUS_ERROR;
		bfa_cb_queue(pkt->bfa, &pkt->hcb_qe,
				__bfa_ipfc_send_cbfn, pkt);
		break;

	case BFA_IPFC_SM_HWFAIL:
		bfa_reqq_wcancel(&pkt->reqq_wqe);
		bfa_sm_set_state(pkt, bfa_ipfc_sm_uninit);
		pkt->pkt_len = pkt->residue_len = 0;
		pkt->status = BFA_IPFC_STATUS_ERROR;
		bfa_cb_queue(pkt->bfa, &pkt->hcb_qe,
				__bfa_ipfc_send_cbfn, pkt);
		break;

	case BFA_IPFC_SM_FLUSH:
		bfa_sm_set_state(pkt, bfa_ipfc_sm_flushing);
		bfa_ipfc_flush(pkt);
		bfa_q_qe_deq(pkt);
		bfa_q_enq(&mod->pkt_free_q, pkt);
		break;
	default:
		bfa_sm_fault(pkt->bfa, event);
	}
}

static void
bfa_ipfc_sm_flushing(bfa_ipfc_pkt_t *pkt, bfa_ipfc_event_t event)
{
	bfa_ipfc_mod_t	*mod = BFA_IPFC_MOD(pkt->bfa);

	bfa_trc(pkt->bfa, event);
	switch (event) {
	case BFA_IPFC_SM_COMP:
		bfa_q_qe_deq(pkt);
		bfa_q_enq(&mod->pkt_q, pkt);
		pkt->status = BFA_IPFC_STATUS_ERROR;
		bfa_sm_set_state(pkt, bfa_ipfc_sm_uninit);
		bfa_cb_queue(pkt->bfa, &pkt->hcb_qe,
				__bfa_ipfc_send_cbfn, pkt);
		break;

	case BFA_IPFC_SM_FLUSHED:
		bfa_sm_set_state(pkt, bfa_ipfc_sm_uninit);
		pkt->status = BFA_IPFC_STATUS_ERROR;
		bfa_cb_queue(pkt->bfa, &pkt->hcb_qe,
				__bfa_ipfc_send_cbfn, pkt);
		break;

	case BFA_IPFC_SM_HWFAIL:
		bfa_sm_set_state(pkt, bfa_ipfc_sm_uninit);
		pkt->pkt_len = pkt->residue_len = 0;
		pkt->status = BFA_IPFC_STATUS_ERROR;
		bfa_cb_queue(pkt->bfa, &pkt->hcb_qe,
				__bfa_ipfc_send_cbfn, pkt);
		break;

	default:
		bfa_sm_fault(pkt->bfa, event);
	}
}

/**
 * @}
 */

/**
 * @ig ipfc_private
 * @{
 */

/**
 * Called when command queue beccomes non-empty.
 */
static void
bfa_ipfc_qresume(void *cbarg)
{
	bfa_ipfc_pkt_t *pkt = cbarg;

	bfa_sm_send_event(pkt, BFA_IPFC_SM_RESUME);
}

/**
 * Called to handle IOC failures.
 * TODO: register IOC failure handler
 */
void
bfa_ipfc_iocdisable(struct bfa_s *bfa)
{
	bfa_ipfc_mod_t	*mod = BFA_IPFC_MOD(bfa);
	bfa_ipfc_pkt_t	*pkt;
	struct bfa_q_s 	*qe;

	bfa_q_iter(&mod->pkt_q, qe) {
		pkt = (bfa_ipfc_pkt_t *) qe;
		bfa_sm_send_event(pkt, BFA_IPFC_SM_HWFAIL);
	}
}

static bfa_boolean_t
bfa_ipfc_post(bfa_ipfc_pkt_t *pkt)
{
	bfi_ipfc_post_t 	*m;
	bfa_ipfc_mod_t		*mod = BFA_IPFC_MOD(pkt->bfa);
	struct bfa_sge_s 	sge;
	int			i;
	uint16_t		buf_len;

	m = bfa_reqq_next(pkt->bfa, BFA_IPFC_REQQ);
	if (!m)
		return BFA_FALSE;

	bfi_h2i_set(m->mh, BFI_MC_IPFC, BFI_IPFC_H2I_POST,
		bfa_fn_lpu(pkt->bfa));

	m->pkt_tag = pkt->pkt_tag;
	buf_len = bfa_cb_ipfc_pktlen(pkt->dpkt, pkt->bfa->bfad);
	m->buf_len = bfa_os_htons(buf_len);

	bfa_assert(bfa_cb_ipfc_nsgles(pkt->dpkt, pkt->bfa->bfad) == 1);

	for (i = 0; i < bfa_cb_ipfc_nsgles(pkt->dpkt, pkt->bfa->bfad); i++) {
		bfa_cb_ipfc_sge(pkt->dpkt, i, &sge, pkt->bfa->bfad);

		m->alen.al_len	= bfa_os_htonl(sge.sg_len);
		bfa_dma_be_addr_set(m->alen.al_addr, (ssize_t)sge.sg_addr);
	}

	bfa_reqq_produce(mod->bfa, BFA_IPFC_REQQ, m->mh);
	bfa_trc(mod->bfa, pkt->pkt_tag);
	return BFA_TRUE;
}

static bfa_boolean_t
bfa_ipfc_flush(bfa_ipfc_pkt_t *pkt)
{
	bfi_ipfc_flush_t 	*m;
	bfa_ipfc_mod_t		*mod = BFA_IPFC_MOD(pkt->bfa);

	m = bfa_reqq_next(pkt->bfa, BFA_IPFC_REQQ);
	if (!m)
		return BFA_FALSE;

	bfi_h2i_set(m->mh, BFI_MC_IPFC,
				BFI_IPFC_H2I_FLUSH, bfa_fn_lpu(pkt->bfa));

	m->pkt_tag =  pkt->pkt_tag;

	bfa_reqq_produce(mod->bfa, BFA_IPFC_REQQ, m->mh);

	return BFA_TRUE;
}

/**
 * @}
 */

/**
 * @ig ipfc_public
 * @{
 */

void
bfa_ipfc_isr(struct bfa_s *bfa, struct bfi_msg_s *m)
{
	bfa_ipfc_mod_t		*mod = BFA_IPFC_MOD(bfa);
	bfi_ipfc_i2h_msg_t	*msg = (bfi_ipfc_i2h_msg_t *) m;
	uint16_t		pkt_tag;
	bfa_ipfc_pkt_t		*pkt;

	bfa_trc(bfa, m->mhdr.msg_id);
	switch (m->mhdr.msg_id) {
	case BFI_IPFC_I2H_RECV:
		bfa_trc(bfa, bfa_os_ntohs(msg->recv.pkt_len));
		bfa_trc(bfa, msg->recv.pkt_tag);
		pkt_tag = msg->recv.pkt_tag;
		pkt = BFA_IPFC_PKT(mod, pkt_tag);
		if (pkt == NULL) {
			bfa_assert(0);
			return;
		}
		pkt->pkt_len = bfa_os_ntohs(msg->recv.pkt_len);
		pkt->residue_len = 0; /* HACK */
		pkt->status = BFA_IPFC_STATUS_OK; /* HACK */
		bfa_trc(bfa, pkt->pkt_len);
		bfa_sm_send_event(pkt, BFA_IPFC_SM_COMP);
		break;

	case BFI_IPFC_I2H_FLUSH_COMP:
		pkt_tag = msg->flush_rsp.pkt_tag;
		pkt = BFA_IPFC_PKT(mod, pkt_tag);
		if (pkt == NULL) {
			bfa_assert(0);
			return;
		}
		bfa_sm_send_event(pkt, BFA_IPFC_SM_FLUSHED);
		break;

	default:
		bfa_trc(bfa, m->mhdr.msg_id);
		bfa_assert(0);
	}
}

/**
 * @}
 */

/**
 * @ig ipfc_api
 * @{
 */

/**
 * @brief
 * Query memory requirement for BFA ipfc module.
 *
 * @param[in]	npkts	Max number of receive packets posted by driver.
 *
 * @return None
 */
void
bfa_ipfc_meminfo(int npkts, struct bfa_meminfo_s *meminfo)
{
	uint32_t	kva_len;
	bfa_mem_dma_t *dma_info = &meminfo->dma_info;
	bfa_mem_kva_t *kva_info = &meminfo->kva_info;

	kva_len = npkts * sizeof(bfa_ipfc_pkt_t);
	kva_info->mem_len = kva_len;
	dma_info->mem_len = 0;
}

/**
 * @brief
 * Attach and initialize IPFC bfa component.
 *
 * Once bfa IPFC component is initialized, packets can be posted by driver
 * to receive inbound IPFC packets. Synchronous call.
 *
 * @param[in]	bfa	instance handle
 * @param[in]	bfad	driver instance
 * @param[in]	npkts	Max number of receive packets posted by driver.
 * @param[in]	meminfo	Memory for managing IPFC packet pools.
 *
 * @return None
 */
void
bfa_ipfc_attach(struct bfa_s *bfa, void *bfad, int npkts,
	struct bfa_meminfo_s *meminfo)
{
	bfa_ipfc_mod_t	*mod = BFA_IPFC_MOD(bfa);
	bfa_ipfc_pkt_t	*pkt;
	bfa_mem_kva_t	*kva_elem;
	int		i;

	bfa_isr_bind(BFI_MC_IPFC, bfa_ipfc_isr);

	kva_elem = &meminfo->kva_info;
	mod->pkt_arr = (bfa_ipfc_pkt_t *) kva_elem->kva_curp;
	mod->npkts   = npkts;
	mod->bfa   = bfa;

	bfa_q_init(&mod->pkt_free_q);
	bfa_q_init(&mod->pkt_q);

	/**
	 * initialize pkt_free_q from array
	 */
	for (i = 0; i < mod->npkts; i++) {
		pkt = mod->pkt_arr + i;
		pkt->bfa = bfa;
		pkt->dipfc = bfad;
		pkt->pkt_tag = i;
		bfa_q_init(&pkt->qe);
		bfa_sm_set_state(pkt, bfa_ipfc_sm_uninit);

		bfa_q_enq(&mod->pkt_free_q, pkt);

		bfa_reqq_winit(&pkt->reqq_wqe, bfa_ipfc_qresume, pkt);
	}
}

/**
 * @brief
 *	Detach IPFC operations.
 */
void
bfa_ipfc_detach(struct bfa_s *bfa)
{
}

/**
 * @brief
 * Post a packet to receive IP type packets.
 *
 * Driver must post packets after attach to receive IPFC packets. Received
 * packets are notified by bfa_cb_ipfc_pktrecv() callback. Driver should
 * replenish BFA packet pool by posting more packets as part of receive
 * processing. Driver must not post more than the total number of packets
 * specified at attach time.
 *
 * @param[in]	bfa	instance handle
 * @param[in]	dpkt	Driver packet handle.
 *
 * @return None
 */
bfa_ipfc_pkt_t *
bfa_ipfc_postpkt(struct bfa_s *bfa, bfad_pkt_t *dpkt)
{
	bfa_ipfc_mod_t	*mod = BFA_IPFC_MOD(bfa);
	bfa_ipfc_pkt_t	*pkt;

	bfa_q_deq(&mod->pkt_free_q, &pkt);
	bfa_assert(pkt);
	if (pkt)
		bfa_q_enq(&mod->pkt_q, pkt);

	if (pkt == NULL)
		return NULL;

	pkt->send_cbfn = bfa_cb_ipfc_pktrecv;
	pkt->dpkt = dpkt;
	bfa_sm_send_event(pkt, BFA_IPFC_SM_POST);
	return pkt;
}

/**
 * @brief
 * Flush all packets posted to BFA.
 *
 * Packets posted to BFA are returned to driver using bfa_cb_ipfc_pktrecv()
 * callback for each outstanding packet. Flushed packets are returned with
 * status BFA_IPFC_STATUS_FLUSHED.
 *
 * @param[in]	pkt	BFA ipfc packet to be flushed
 *
 * @return None
 */
void
bfa_ipfc_flushpkt(bfa_ipfc_pkt_t *pkt)
{
	bfa_sm_send_event(pkt, BFA_IPFC_SM_FLUSH);
}

/**
 * @}
 */
