/*
 * Copyright (c) 2005-2014 Brocade Communications Systems, Inc.
 * Copyright (c) 2014 QLogic Corporation.
 * All rights reserved
 * www.qlogic.com
 *
 * See LICENSE.bfa for copyright and licensing details.
 */
#ifndef __BFA_IPFC_H__
#define __BFA_IPFC_H__

#include <bfa/bfa.h>

typedef struct bfad_pkt_s bfad_pkt_t;
typedef struct bfa_ipfc_pkt_s bfa_ipfc_pkt_t;

/**
 * Packet receive status
 */
typedef enum {
	BFA_IPFC_STATUS_OK = 0,		/*!< good packet		 */
	BFA_IPFC_STATUS_PARTIAL = 1,	/*!< only partial frame is recvd */
	BFA_IPFC_STATUS_ERROR = 2,	/*!< driver requested return	 */
} bfa_ipfc_status_t;

/**
 * @brief
 * 		Callback to process a received packet.
 *
 * @param[in]		dipfc		driver IPFC instance
 * @param[in]		dpkt		driver packet
 * @param[in]		status		Status of received packet.
 * @param[in]		pktlen		Length of packet received.
 * @param[in]		residue		Residual length in bytes of packet
 *					that was not copied due to lack of
 *					space.
 *
 * @return None
 */
typedef
void            (*send_cbfn_t)(void *dipfc, bfad_pkt_t *dpkt,
				     bfa_ipfc_status_t status,
				     uint32_t pktlen, uint32_t residue);

void		bfa_cb_ipfc_pktrecv(void *dipfc, bfad_pkt_t *dpkt,
				     bfa_ipfc_status_t status,
				     uint32_t pktlen, uint32_t residue);


struct bfa_ipfc_pkt_s {
        struct bfa_q_s  qe;
	bfa_sm_t	sm;
        struct bfa_s    *bfa;
        bfad_pkt_t      *dpkt;
        void      	*dipfc;
        uint16_t        pkt_tag;
        uint16_t        pkt_len;
        uint16_t        residue_len;
        bfa_ipfc_status_t  status;
        struct bfa_cb_qe_s     hcb_qe;
	send_cbfn_t	send_cbfn;
        struct bfa_reqq_wait_s reqq_wqe;
};

typedef struct {
	struct bfa_s 	*bfa;
        bfa_ipfc_pkt_t  *pkt_arr;
        struct bfa_q_s         pkt_q;
        struct bfa_q_s         pkt_free_q;
        int             npkts;
} bfa_ipfc_mod_t;

/*
 * bfa ipfc API functions
 */
void            bfa_ipfc_meminfo(int npkts, struct bfa_meminfo_s *meminfo);
void            bfa_ipfc_attach(struct bfa_s *bfa, void *dipfc, int npkts,
				    struct bfa_meminfo_s *meminfo);
void            bfa_ipfc_detach(struct bfa_s *bfa);
bfa_ipfc_pkt_t  *bfa_ipfc_postpkt(struct bfa_s *bfa, bfad_pkt_t *dpkt);
void            bfa_ipfc_flushpkt(bfa_ipfc_pkt_t *pkt);

/**
 * @brief
 * 		Completion callback for bfa_ipfc_detach().
 *
 * @param[in]		dipfc		driver IPFC instance
 *
 * @return None
 */
void            bfa_cb_ipfc_detach(void *dipfc);

/**
 * @brief
 * 		Get number of SG elements required to map receive packet buffer.
 *
 * @param[in]		dpkt		driver packet
 * @param[in]		bfad		bfad
 *
 * @return Number of SG elements for the packet.
 */
int             bfa_cb_ipfc_nsgles(bfad_pkt_t *dpkt, void *bfad);

/**
 * @brief
 * 		Get total packet buffer size in bytes.
 *
 * @param[in]		dpkt		driver packet
 * @param[in]		bfad		bfad
 *
 * @return Packet size in bytes.
 */
uint16_t        bfa_cb_ipfc_pktlen(bfad_pkt_t *dpkt, void *bfad);


/**
 * @brief
 * 		Get SG element of the mapped packet buffer given the SG index.
 *
 * @param[in]		dpkt		driver packet
 * @param[in]		sge_idx		SG element index
 * @param[out]		sge		SG element contents
 * @param[in]		bfad		bfad
 *
 * @return None
 */
void            bfa_cb_ipfc_sge(bfad_pkt_t *dpkt, int sge_idx,
				 struct bfa_sge_s *sge, void *bfad);

#endif /* __BFA_IPFC_H__ */
