/*
 * Copyright (c) 2005-2014 Brocade Communications Systems, Inc.
 * Copyright (c) 2014 QLogic Corporation.
 * All rights reserved
 * www.qlogic.com
 *
 * See LICENSE.bfa for copyright and licensing details.
 */

#ifndef __BFA_SVC_H__
#define __BFA_SVC_H__

#include <cs/bfa_cs.h>
#include <bfi/bfi_ms.h>

/**
 * Scatter-gather DMA related defines
 */
#define BFA_SGPG_MIN	(16U)
#define BFA_SGPG_MAX	(8192U)

/**
 * Alignment macro for SG page allocation
 */
#define BFA_SGPG_ROUNDUP(_l) (((_l) + (sizeof(struct bfi_sgpg_s) - 1))	\
			      & ~(sizeof(struct bfi_sgpg_s) - 1))

struct bfa_sgpg_wqe_s {
	struct bfa_q_s qe;	/*!< queue sg page element	*/
	int	nsgpg;		/*!< pages to be allocated	*/
	int	nsgpg_total;	/*!< total pages required	*/
	void	(*cbfn) (void *cbarg);	/*!< callback function	*/
	void	*cbarg;		/*!< callback arg		*/
	struct bfa_q_s sgpg_q;	/*!< queue of alloced sgpgs	*/
};

struct bfa_sgpg_s {
	struct bfa_q_s 	qe;		/*!< queue sg page element	*/
	struct bfi_sgpg_s *sgpg;	/*!< va of SG page		*/
	union bfi_addr_u sgpg_pa;	/*!< pa of SG page		*/
};

/**
 * Given number of SG elements, BFA_SGPG_NPAGE() returns the number of
 * SG pages required.
 */
#define BFA_SGPG_NPAGE(_nsges)  (((_nsges) / BFI_SGPG_DATA_SGES) + 1)

/* Max SGPG dma segs required */
#define BFA_SGPG_DMA_SEGS	\
	BFI_MEM_DMA_NSEGS(BFA_SGPG_MAX, (uint32_t)sizeof(struct bfi_sgpg_s))

struct bfa_sgpg_mod_s {
	struct bfa_s *bfa;
	int		num_sgpgs;	/*!< number of SG pages		*/
	int		free_sgpgs;	/*!< number of free SG pages	*/
	struct bfa_q_s 	sgpg_q;		/*!< queue of free SG pages	*/
	struct bfa_q_s 	sgpg_wait_q;	/*!< wait queue for SG pages	*/
	struct bfa_mem_dma_s dma_seg[BFA_SGPG_DMA_SEGS];
	struct bfa_mem_kva_s kva_seg;
};
#define BFA_SGPG_MOD(__bfa)	(&(__bfa)->modules.sgpg_mod)
#define BFA_MEM_SGPG_KVA(__bfa)	(&(BFA_SGPG_MOD(__bfa)->kva_seg))

bfa_status_t bfa_sgpg_malloc(struct bfa_s *bfa, struct bfa_q_s *sgpg_q,
			     int nsgpgs);
void bfa_sgpg_mfree(struct bfa_s *bfa, struct bfa_q_s *sgpg_q, int nsgpgs);
void bfa_sgpg_winit(struct bfa_sgpg_wqe_s *wqe,
		    void (*cbfn) (void *cbarg), void *cbarg);
void bfa_sgpg_wait(struct bfa_s *bfa, struct bfa_sgpg_wqe_s *wqe, int nsgpgs);
void bfa_sgpg_wcancel(struct bfa_s *bfa, struct bfa_sgpg_wqe_s *wqe);


/**
 * FCXP related defines
 */
#define BFA_FCXP_MIN     	(2U)
#define BFA_FCXP_MAX		(256U)
#define BFA_FCXP_MAX_IBUF_SZ	(2 * 1024 + 256)
#define BFA_FCXP_MAX_LBUF_SZ	(4 * 1024 + 256)

/* Max FCXP dma segs required */
#define BFA_FCXP_DMA_SEGS						\
	BFI_MEM_DMA_NSEGS(BFA_FCXP_MAX,					\
		(uint32_t)BFA_FCXP_MAX_IBUF_SZ + BFA_FCXP_MAX_LBUF_SZ)

struct bfa_fcxp_mod_s {
	struct bfa_s      *bfa;		/* backpointer to BFA */
	struct bfa_fcxp_s *fcxp_list;	/* array of FCXPs */
	uint16_t	num_fcxps;	/* max num FCXP requests */
	struct bfa_q_s  fcxp_req_free_q; /* free FCXPs used for sending req */
	struct bfa_q_s  fcxp_rsp_free_q; /* free FCXPs used for sending req */
	struct bfa_q_s  fcxp_active_q;	/* active FCXPs */
	struct bfa_q_s  req_wait_q;	/* wait queue for free req_fcxp */
	struct bfa_q_s  rsp_wait_q;	/* wait queue for free rsp_fcxp */
	struct bfa_q_s	fcxp_req_unused_q; /* unused req_fcxps */
	struct bfa_q_s	fcxp_rsp_unused_q; /* unused rsp_fcxps */
	uint32_t	req_pld_sz;
	uint32_t	rsp_pld_sz;
	struct bfa_mem_dma_s dma_seg[BFA_FCXP_DMA_SEGS];
	struct bfa_mem_kva_s kva_seg;
};

#define BFA_FCXP_MOD(__bfa)		(&(__bfa)->modules.fcxp_mod)
#define BFA_FCXP_FROM_TAG(__mod, __tag)	(&(__mod)->fcxp_list[__tag])
#define BFA_MEM_FCXP_KVA(__bfa)	(&(BFA_FCXP_MOD(__bfa)->kva_seg))

typedef void    (*fcxp_send_cb_t) (struct bfa_s *ioc, struct bfa_fcxp_s *fcxp,
				   void *cb_arg, bfa_status_t req_status,
				   uint32_t rsp_len, uint32_t resid_len,
				   fchs_t *rsp_fchs);

typedef uint64_t (*bfa_fcxp_get_sgaddr_t) (void *bfad_fcxp, int sgeid);
typedef uint32_t (*bfa_fcxp_get_sglen_t) (void *bfad_fcxp, int sgeid);
typedef void (*bfa_cb_fcxp_send_t) (void *bfad_fcxp, struct bfa_fcxp_s *fcxp,
				    void *cbarg, enum bfa_status req_status,
				    uint32_t rsp_len, uint32_t resid_len,
				    struct fchs_s *rsp_fchs);
typedef void (*bfa_fcxp_alloc_cbfn_t) (void *cbarg, struct bfa_fcxp_s *fcxp);



/**
 * Information needed for a FCXP request
 */
struct bfa_fcxp_req_info_s {
	struct bfa_rport_s *bfa_rport;
					/** Pointer to the bfa rport that was
					 * returned from bfa_rport_create().
					 * This could be left NULL for WKA or
					 * for FCXP interactions before the
					 * rport nexus is established
					 */
	fchs_t		fchs;	/* !< request FC header structure */
	uint8_t		cts;	/* !< continous sequence */
	uint8_t		class;	/* !< FC class for the request/response */
	uint16_t	max_frmsz;	/* !< max send frame size */
	uint16_t	vf_id;	/* !< vsan tag if applicable */
	uint8_t		lp_tag;	/* !< lport tag */
	uint32_t	req_tot_len;	/* !< request payload total length */
};

struct bfa_fcxp_rsp_info_s {
	fchs_t		rsp_fchs;
				/** !< Response frame's FC header will
				 * be sent back in this field */
	uint8_t		rsp_timeout;
				/** !< timeout in seconds, 0-no response
				 */
	uint8_t		rsvd2[3];
	uint32_t	rsp_maxlen;	/*!< max response length expected */
};

struct bfa_fcxp_s {
	struct bfa_q_s	qe;		/*!< fcxp queue element */
	bfa_sm_t	sm;		/*!< state machine */
	void		*caller;	/*!< driver or fcs */
	struct bfa_fcxp_mod_s *fcxp_mod;
	/*!< back pointer to fcxp mod */
	uint16_t	fcxp_tag;	/*!< internal tag */
	struct bfa_fcxp_req_info_s req_info;
	/*!< request info */
	struct bfa_fcxp_rsp_info_s rsp_info;
	/*!< response info */
	uint8_t 	use_ireqbuf;	/*!< use internal req buf */
	uint8_t		use_irspbuf;	/*!< use internal rsp buf */
	uint32_t	nreq_sgles;	/*!< num request SGLEs */
	uint32_t	nrsp_sgles;	/*!< num response SGLEs */
	struct bfa_q_s req_sgpg_q;	/*!< SG pages for request buf */
	struct bfa_q_s req_sgpg_wqe;	/*!< wait queue for req SG page */
	struct bfa_q_s rsp_sgpg_q;	/*!< SG pages for response buf */
	struct bfa_q_s rsp_sgpg_wqe;	/*!< wait queue for rsp SG page */

	bfa_fcxp_get_sgaddr_t req_sga_cbfn;
	/*!< SG elem addr user function */
	bfa_fcxp_get_sglen_t req_sglen_cbfn;
	/*!< SG elem len user function */
	bfa_fcxp_get_sgaddr_t rsp_sga_cbfn;
	/*!< SG elem addr user function */
	bfa_fcxp_get_sglen_t rsp_sglen_cbfn;
	/*!< SG elem len user function */
	bfa_cb_fcxp_send_t send_cbfn;   /*!< send completion callback */
	void		*send_cbarg;	/*!< callback arg */
	struct bfa_sge_s   req_sge[BFA_FCXP_MAX_SGES];
	/*!< req SG elems */
	struct bfa_sge_s   rsp_sge[BFA_FCXP_MAX_SGES];
	/*!< rsp SG elems */
	uint8_t		rsp_status;	/*!< comp: rsp status */
	uint32_t	rsp_len;	/*!< comp: actual response len */
	uint32_t	residue_len;	/*!< comp: residual rsp length */
	fchs_t		rsp_fchs;	/*!< comp: response fchs */
	struct bfa_cb_qe_s    hcb_qe;	/*!< comp: callback qelem */
	struct bfa_reqq_wait_s	reqq_wqe;
	bfa_boolean_t	reqq_waiting;
	bfa_boolean_t	req_rsp;	/*!< Used for sending req/rsp */
};
typedef struct bfa_fcxp_s bfa_fcxp_t;

struct bfa_fcxp_wqe_s {
	struct bfa_q_s		qe;
	bfa_fcxp_alloc_cbfn_t	alloc_cbfn;
	void			*alloc_cbarg;
	void 			*caller;
	struct bfa_s 		*bfa;
	int 			nreq_sgles;
	int 			nrsp_sgles;
	bfa_fcxp_get_sgaddr_t 	req_sga_cbfn;
	bfa_fcxp_get_sglen_t 	req_sglen_cbfn;
	bfa_fcxp_get_sgaddr_t 	rsp_sga_cbfn;
	bfa_fcxp_get_sglen_t 	rsp_sglen_cbfn;
};
typedef struct bfa_fcxp_wqe_s bfa_fcxp_wqe_t;

#define BFA_FCXP_REQ_PLD(_fcxp) 	(bfa_fcxp_get_reqbuf(_fcxp))

#define BFA_FCXP_RSP_FCHS(_fcxp) 	(&((_fcxp)->rsp_info.fchs))
#define BFA_FCXP_RSP_PLD(_fcxp) 	(bfa_fcxp_get_rspbuf(_fcxp))

#define BFA_FCXP_REQ_PLD_PA(_fcxp)					       \
	bfa_mem_get_dmabuf_pa((_fcxp)->fcxp_mod, (_fcxp)->fcxp_tag,	       \
	      (_fcxp)->fcxp_mod->req_pld_sz + (_fcxp)->fcxp_mod->rsp_pld_sz)

/* fcxp_buf = req_buf + rsp_buf :- add req_buf_sz to get to rsp_buf */
#define BFA_FCXP_RSP_PLD_PA(_fcxp)					       \
	bfa_mem_get_dmabuf_pa((_fcxp)->fcxp_mod, (_fcxp)->fcxp_tag,	       \
	      (_fcxp)->fcxp_mod->req_pld_sz + (_fcxp)->fcxp_mod->rsp_pld_sz) + \
	      (_fcxp)->fcxp_mod->req_pld_sz

void	bfa_fcxp_isr(struct bfa_s *bfa, struct bfi_msg_s *msg);

/**
 * RPORT related defines
 */
#define BFA_RPORT_MIN	4

struct bfa_rport_mod_s {
	struct bfa_rport_s *rps_list;	/*!< list of rports	*/
	struct bfa_q_s 	rp_free_q;	/*!< free bfa_rports	*/
	struct bfa_q_s 	rp_active_q;	/*!< free bfa_rports 	*/
	struct bfa_q_s	rp_unused_q;	/*!< unused bfa rports	*/
	uint16_t	num_rports;	/*!< number of rports	*/
	struct bfa_mem_kva_s	kva_seg;
};

#define BFA_RPORT_MOD(__bfa)	(&(__bfa)->modules.rport_mod)
#define BFA_MEM_RPORT_KVA(__bfa) (&(BFA_RPORT_MOD(__bfa)->kva_seg))

/**
 * Convert rport tag to RPORT
 */
#define BFA_RPORT_FROM_TAG(__bfa, _tag)				\
	(BFA_RPORT_MOD(__bfa)->rps_list +			\
	 ((_tag) & (BFA_RPORT_MOD(__bfa)->num_rports - 1)))

/*
 * protected functions
 */
void	bfa_rport_isr(struct bfa_s *bfa, struct bfi_msg_s *msg);
void	bfa_rport_res_recfg(struct bfa_s *bfa, uint16_t num_rport_fw);

/**
 * @brief
 * 		BFA rport information.
 */
struct bfa_rport_info_s {
	uint16_t	max_frmsz;	/*!< max rcv pdu size		    */
	uint32_t	pid : 24,	/*!< remote port ID		    */
		      lp_tag : 8;	/*!< tag			    */
	uint32_t	local_pid : 24,	/*!< local port ID		    */
			    cisc : 8;	/*!< CIRO supported		    */
	uint8_t		fc_class;	/*!< supported FC classes. fc_cos_t */
	uint8_t		vf_en;		/*!< virtual fabric enable	    */
	uint16_t	vf_id;		/*!< virtual fabric ID		    */
	enum bfa_port_speed speed;	/*!< Rport's current speed	    */
};
typedef struct bfa_rport_info_s bfa_rport_info_t;

static inline int
bfa_get_id_from_lun(lun_t lun)
{
	uint32_t *temp = (uint32_t *)&lun;
	return ((bfa_os_htonl((uint32_t)(*temp)) & 0x3fff0000) >> 16);
}

static inline lun_t
bfa_get_lun_from_id(uint32_t lunid)
{
	return (lun_t)(bfa_os_htonl((lunid) << 16));
}

/* for lun profile this stored the iopf index direct mapped to lun id */
struct bfa_rport_iopf_da_id_s {
	uint16_t iopf_id[BFA_IOPF_DA_MAX];
};

struct bfa_rport_iopf_nda_id_s {
	int max_id;
	lun_t lun[BFA_IOPF_MAX]; /* key of sorted array by insertion sort */
	uint16_t iopf_id[BFA_IOPF_MAX]; /* corresponding lunst id */
};


/**
 * BFA rport data structure
 */
struct bfa_rport_s {
	struct bfa_q_s	qe;		/*!< queue element		    */
	bfa_sm_t	sm; 		/*!< state machine		    */
	struct bfa_s	*bfa;		/*!< backpointer to BFA		    */
	void		*rport_drv;	/*!< fcs/driver rport object	    */
	uint16_t	fw_handle;	/*!< firmware rport handle	    */
	uint16_t	rport_tag;	/*!< BFA rport tag		    */
	uint8_t		lun_mask;	/*!< LUN mask flag		    */
	uint8_t		iopf_enabled;	/*!< Advance ITL IO profile flag    */
	uint8_t		iopf_count;	/*!< Advance ITL IO profile count    */
	uint8_t		rp_iopf_enabled; /*!< Advance IT IO profile flag    */
	struct bfa_rport_info_s rport_info; /*!< rport info from fcs/driver */
	struct bfa_reqq_wait_s reqq_wait; /*!< to wait for room in reqq     */
	struct bfa_cb_qe_s hcb_qe;	/*!< BFA callback qelem		    */
	struct bfa_rport_iopf_da_id_s lunid_arr; /* direct lun id indexed */
	struct bfa_rport_iopf_nda_id_s nda_lunid;
	/* non- direct lun id indexed, sorted lun array */
	struct bfa_rport_hal_stats_s stats; /*!< BFA rport statistics	    */
	struct bfa_rport_qos_attr_s qos_attr;
	union a {
		bfa_status_t	status;	/*!< f/w status */
		void		*fw_msg; /*!< QoS scn event		    */
	} event_arg;
};
typedef struct bfa_rport_s bfa_rport_t;
#define BFA_RPORT_FC_COS(_rport)	((_rport)->rport_info.fc_class)
#define BFA_RPORT_IOPF_DA_ID(_rport, __id)	((_rport)->lunid_arr[__id])
#define BFA_RPORT_IOPF(_rport)	((_rport)->iopf)


/**
 * UF - unsolicited receive related defines
 */

#define BFA_UF_MIN	(4U)
#define BFA_UF_MAX	(256U)

/**
 * @brief
 */
struct bfa_uf_s {
	struct bfa_q_s	qe;		/*!< queue element		*/
	struct bfa_s	*bfa;		/*!< bfa instance		*/
	uint16_t	uf_tag;		/*!< identifying tag fw msgs	*/
	uint16_t	vf_id;
	uint16_t	src_rport_handle;
	uint16_t	rsvd;
	uint8_t		*data_ptr;
	uint16_t	data_len;	/*!< actual receive length	*/
	uint16_t	pb_len;		/*!< posted buffer length	*/
	void		*buf_kva;	/*!< buffer virtual address	*/
	uint64_t	buf_pa;		/*!< buffer physical address	*/
	struct bfa_cb_qe_s hcb_qe;	/*!< comp: BFA comp qelem	*/
	struct bfa_sge_s sges[BFI_SGE_INLINE_MAX];
};
typedef struct bfa_uf_s bfa_uf_t;

/**
 * @brief
 *      Callback prototype for unsolicited frame receive handler.
 *
 * @param[in]           cbarg           callback arg for receive handler
 * @param[in]           uf              unsolicited frame descriptor
 *
 * @return None
 */
typedef void (*bfa_cb_uf_recv_t) (void *cbarg, struct bfa_uf_s *uf);

#define BFA_UF_BUFSZ	(2 * 1024 + 256)

/**
 * @todo private
 */
struct bfa_uf_buf_s {
	uint8_t		d[BFA_UF_BUFSZ];
};
typedef struct bfa_uf_buf_s bfa_uf_buf_t;

#define BFA_PER_UF_DMA_SZ	\
	(uint32_t)BFA_ROUNDUP(sizeof(struct bfa_uf_buf_s), BFA_DMA_ALIGN_SZ)

/* Max UF dma segs required */
#define BFA_UF_DMA_SEGS	BFI_MEM_DMA_NSEGS(BFA_UF_MAX, BFA_PER_UF_DMA_SZ)

struct bfa_uf_mod_s {
	struct bfa_s *bfa;		/*!< back pointer to BFA */
	struct bfa_uf_s *uf_list;	/*!< array of UFs */
	uint16_t	num_ufs;	/*!< num unsolicited rx frames */
	struct bfa_q_s 	uf_free_q;	/*!< free UFs */
	struct bfa_q_s 	uf_posted_q;	/*!< UFs posted to IOC */
	struct bfa_q_s	uf_unused_q;	/*!< unused UF's */
	struct bfi_uf_buf_post_s *uf_buf_posts;
	/*!< pre-built UF post msgs */
	bfa_cb_uf_recv_t ufrecv;	/*!< uf recv handler function */
	void		*cbarg;		/*!< uf receive handler arg */
	struct bfa_mem_dma_s	dma_seg[BFA_UF_DMA_SEGS];
	struct bfa_mem_kva_s	kva_seg;
};

#define BFA_UF_MOD(__bfa)	(&(__bfa)->modules.uf_mod)
#define BFA_MEM_UF_KVA(__bfa)	(&(BFA_UF_MOD(__bfa)->kva_seg))

#define ufm_pbs_pa(_ufmod, _uftag)			\
	bfa_mem_get_dmabuf_pa(_ufmod, _uftag, BFA_PER_UF_DMA_SZ)

void	bfa_uf_isr(struct bfa_s *bfa, struct bfi_msg_s *msg);
void bfa_uf_res_recfg(struct bfa_s *bfa, uint16_t num_uf_fw);

/**
 * LPS - bfa lport login/logout service interface
 */
struct bfa_lps_s {
	struct bfa_q_s	qe;		/*!< queue element		*/
	struct bfa_s	*bfa;		/*!< parent bfa instance	*/
	bfa_sm_t	sm;		/*!< finite state machine	*/
	uint8_t		bfa_tag;	/*!< lport bfa tag		*/
	uint8_t		fw_tag;		/*!< lport fw tag		*/
	uint8_t		reqq;		/*!< lport request queue	*/
	uint8_t		alpa;		/*!< ALPA for loop topologies	*/
	uint32_t	lp_pid;		/*!< lport port ID		*/
	bfa_boolean_t	fdisc;		/*!< snd FDISC instead of FLOGI	*/
	bfa_boolean_t	auth_en;	/*!< enable authentication	*/
	bfa_boolean_t	auth_req;	/*!< authentication required	*/
	bfa_boolean_t	npiv_en;	/*!< NPIV is allowed by peer	*/
	bfa_boolean_t	fport;		/*!< attached peer is F_PORT	*/
	bfa_boolean_t	brcd_switch;	/*!< attached peer is brcd sw	*/
	bfa_status_t	status;		/*!< login status		*/
	uint16_t	pdusz;		/*!< max receive PDU size 	*/
	uint16_t	pr_bbcred;	/*!< BB_CREDIT from peer 	*/
	uint8_t		lsrjt_rsn;	/*!< LSRJT reason		*/
	uint8_t		lsrjt_expl;	/*!< LSRJT explanation		*/
	uint8_t		lun_mask;	/*!< LUN mask flag 		*/
	wwn_t		pwwn;		/*!< port wwn of lport		*/
	wwn_t		nwwn;		/*!< node wwn of lport		*/
	wwn_t		pr_pwwn;	/*!< port wwn of lport peer	*/
	wwn_t		pr_nwwn;	/*!< node wwn of lport peer	*/
	mac_t		lp_mac;		/*!< fpma/spma MAC for lport	*/
	mac_t		fcf_mac;	/*!< FCF MAC of lport		*/
	struct bfa_reqq_wait_s	wqe;	/*!< request wait queue element	*/
	void		*uarg;		/*!< user callback arg		*/
	struct bfa_cb_qe_s hcb_qe;	/*!< comp: callback qelem	*/
	struct bfi_lps_login_rsp_s *loginrsp;
	bfa_eproto_status_t ext_status;
};
typedef struct bfa_lps_s bfa_lps_t;

struct bfa_lps_mod_s {
	struct bfa_q_s		lps_free_q;
	struct bfa_q_s		lps_active_q;
	struct bfa_q_s		lps_login_q;
	struct bfa_lps_s	*lps_arr;
	int			num_lps;
	struct bfa_mem_kva_s	kva_seg;
};

#define BFA_LPS_MOD(__bfa)		(&(__bfa)->modules.lps_mod)
#define BFA_LPS_FROM_TAG(__mod, __tag)	(&(__mod)->lps_arr[__tag])
#define BFA_MEM_LPS_KVA(__bfa) (&(BFA_LPS_MOD(__bfa)->kva_seg))

/*
 * external functions
 */
void	bfa_lps_isr(struct bfa_s *bfa, struct bfi_msg_s *msg);


/**
 * FCPORT related defines
 */

#define BFA_FCPORT(_bfa)	(&((_bfa)->modules.port))
#define	MAX_LUN_MASK_CFG	16

/**
 * Link notification data structure
 */
struct bfa_fcport_ln_s {
	struct bfa_fcport_s	*fcport;
	bfa_sm_t		sm;
	struct bfa_cb_qe_s	ln_qe;	/*!< BFA callback queue elem for ln */
	enum bfa_port_linkstate ln_event; /*!< ln event for callback */
};

struct bfa_fcport_trunk_s {
	struct bfa_trunk_attr_s	attr;
};

/**
 * BFA FC port data structure
 */
struct bfa_fcport_s {
	struct bfa_s 		*bfa;	/*!< parent BFA instance */
	bfa_sm_t		sm;	/*!< port state machine */
	wwn_t			nwwn;	/*!< node wwn of physical port */
	wwn_t			pwwn;	/*!< port wwn of physical oprt */
	enum bfa_port_speed speed_sup;
	/*!< supported speeds */
	enum bfa_port_speed speed;	/*!< current speed */
	enum bfa_port_topology topology;	/*!< current topology */
	uint8_t			rsvd[2];
	uint8_t			myalpa;	/*!< my ALPA in LOOP topology */
	uint8_t			alpabm_valid; /*!< alpa bitmap valid or not */
	fc_alpabm_t		alpabm; /*!< alpa bitmap */
	struct bfa_port_cfg_s	cfg;	/*!< current port configuration */
	bfa_boolean_t		use_flash_cfg; /*!< get port cfg from flash */
	struct bfa_qos_attr_s  qos_attr;   /* QoS Attributes */
	struct bfa_qos_vc_attr_s qos_vc_attr;  /*!< VC info from ELP */
	struct bfa_reqq_wait_s	reqq_wait;
	/*!< to wait for room in reqq */
	struct bfa_reqq_wait_s	svcreq_wait;
	/*!< to wait for room in reqq */
	struct bfa_reqq_wait_s	stats_reqq_wait;
	/*!< to wait for room in reqq (stats) */
	void			*event_cbarg;
	void			(*event_cbfn) (void *cbarg,
					       bfa_port_linkstate_t event);
	union {
		union bfi_fcport_i2h_msg_u i2hmsg;
	} event_arg;
	void			*bfad;	/*!< BFA driver handle */
	struct bfa_fcport_ln_s	ln; /*!< Link Notification */
	struct bfa_cb_qe_s	hcb_qe;	/*!< BFA callback queue elem */
	struct bfa_timer_s 	timer;	/*!< timer */
	uint32_t		msgtag;	/*!< fimrware msg tag for reply */
	uint8_t			*stats_kva;
	uint64_t		stats_pa;
	union bfa_fcport_stats_u *stats;
	bfa_status_t		stats_status; /*!< stats/statsclr status */
	struct bfa_q_s		stats_pending_q;
	struct bfa_q_s		statsclr_pending_q;
	bfa_boolean_t   	stats_qfull;
	uint32_t		stats_reset_time; /*!< stats reset time stamp */
	bfa_boolean_t   	diag_busy; /*!< diag busy status */
	bfa_boolean_t   	beacon; /*!< port beacon status */
	bfa_boolean_t   	link_e2e_beacon; /*!< link beacon status */
	struct bfa_fcport_trunk_s trunk;
	uint16_t		fcoe_vlan;
	struct bfa_mem_dma_s	fcport_dma;
	bfa_boolean_t		stats_dma_ready;
	struct bfa_bbcr_attr_s	bbcr_attr;
	bfa_fec_state_t		fec_state;
};

#define BFA_FCPORT_MOD(__bfa)	(&(__bfa)->modules.fcport)
#define BFA_MEM_FCPORT_DMA(__bfa) (&(BFA_FCPORT_MOD(__bfa)->fcport_dma))

/*
 * protected functions
 */
void bfa_fcport_init(struct bfa_s *bfa);
void bfa_fcport_isr(struct bfa_s *bfa, struct bfi_msg_s *msg);

/*
 * bfa fcport API functions
 */
bfa_status_t bfa_fcport_enable(struct bfa_s *bfa);
bfa_status_t bfa_fcport_disable(struct bfa_s *bfa);
bfa_status_t bfa_fcport_cfg_speed(struct bfa_s *bfa,
				  enum bfa_port_speed speed);
enum bfa_port_speed bfa_fcport_get_speed(struct bfa_s *bfa);
bfa_status_t bfa_fcport_cfg_topology(struct bfa_s *bfa,
				     enum bfa_port_topology topo);
enum bfa_port_topology bfa_fcport_get_topology(struct bfa_s *bfa);
enum bfa_port_topology bfa_fcport_get_cfg_topology(struct bfa_s *bfa);
bfa_status_t bfa_fcport_cfg_hardalpa(struct bfa_s *bfa, uint8_t alpa);
bfa_boolean_t bfa_fcport_get_hardalpa(struct bfa_s *bfa, uint8_t *alpa);
uint8_t bfa_fcport_get_myalpa(struct bfa_s *bfa);
bfa_status_t bfa_fcport_clr_hardalpa(struct bfa_s *bfa);
bfa_status_t bfa_fcport_cfg_maxfrsize(struct bfa_s *bfa, uint16_t maxsize);
uint16_t bfa_fcport_get_maxfrsize(struct bfa_s *bfa);
uint32_t bfa_fcport_mypid(struct bfa_s *bfa);
uint8_t bfa_fcport_get_rx_bbcredit(struct bfa_s *bfa);
void bfa_fcport_get_attr(struct bfa_s *bfa, struct bfa_port_attr_s *attr);
wwn_t bfa_fcport_get_wwn(struct bfa_s *bfa, bfa_boolean_t node);
void bfa_fcport_event_register(struct bfa_s *bfa,
			void (*event_cbfn) (void *cbarg,
			bfa_port_linkstate_t event), void *event_cbarg);
bfa_boolean_t bfa_fcport_is_disabled(struct bfa_s *bfa);
bfa_boolean_t bfa_fcport_is_dport(struct bfa_s *bfa);
bfa_boolean_t bfa_fcport_is_ddport(struct bfa_s *bfa);
void bfa_fcport_cfg_qos(struct bfa_s *bfa, bfa_boolean_t on_off);
bfa_status_t bfa_fcport_set_qos_bw(struct bfa_s *bfa,
			struct bfa_qos_bw_s *qos_bw);
void bfa_fcport_cfg_ratelim(struct bfa_s *bfa, bfa_boolean_t on_off);
bfa_status_t bfa_fcport_cfg_ratelim_speed(struct bfa_s *bfa,
					  enum bfa_port_speed speed);
enum bfa_port_speed bfa_fcport_get_ratelim_speed(struct bfa_s *bfa);
void bfa_fcport_cfg_faa(struct bfa_s *bfa, uint8_t state);

void bfa_fcport_set_tx_bbcredit(struct bfa_s *bfa, uint16_t tx_bbcredit);
void bfa_fcport_busy(struct bfa_s *bfa, bfa_boolean_t status);
void bfa_fcport_beacon(void *dev, bfa_boolean_t beacon,
		       bfa_boolean_t link_e2e_beacon);
void bfa_cb_port_event(void *cbarg, bfa_port_linkstate_t event);
void bfa_fcport_qos_get_attr(struct bfa_s *bfa,
			     struct bfa_qos_attr_s *qos_attr);
void bfa_fcport_qos_get_vc_attr(struct bfa_s *bfa,
				struct bfa_qos_vc_attr_s *qos_vc_attr);
bfa_status_t bfa_fcport_get_qos_stats(struct bfa_s *bfa,
					struct bfa_cb_pending_q_s *cb);
bfa_status_t bfa_fcport_clear_qos_stats(struct bfa_s *bfa,
					struct bfa_cb_pending_q_s *cb);
bfa_status_t bfa_fcport_get_fcoe_stats(struct bfa_s *bfa,
					struct bfa_cb_pending_q_s *cb);
bfa_status_t bfa_fcport_clear_fcoe_stats(struct bfa_s *bfa,
						struct bfa_cb_pending_q_s *cb);
bfa_boolean_t bfa_fcport_is_ratelim(struct bfa_s *bfa);
bfa_boolean_t bfa_fcport_is_linkup(bfa_t *bfa);
bfa_status_t bfa_fcport_get_stats(struct bfa_s *bfa,
					struct bfa_cb_pending_q_s *cb);
bfa_status_t bfa_fcport_clear_stats(struct bfa_s *bfa,
					struct bfa_cb_pending_q_s *cb);
bfa_boolean_t bfa_fcport_is_qos_enabled(struct bfa_s *bfa);
bfa_boolean_t bfa_fcport_is_trunk_enabled(struct bfa_s *bfa);
bfa_status_t bfa_fcport_is_pbcdisabled(struct bfa_s *bfa);
void bfa_fcport_dportenable(struct bfa_s *bfa);
void bfa_fcport_dportdisable(struct bfa_s *bfa);
void bfa_fcport_ddportenable(struct bfa_s *bfa);
void bfa_fcport_ddportdisable(struct bfa_s *bfa);
bfa_status_t bfa_fcport_cfg_bbcr(struct bfa_s *bfa,
			bfa_boolean_t on_off, uint8_t bb_scn);
bfa_status_t bfa_fcport_get_bbcr_attr(struct bfa_s *bfa,
			struct bfa_bbcr_attr_s *bbcr_attr);

/*
 * bfa rport API functions
 */
struct bfa_rport_s *bfa_rport_create(struct bfa_s *bfa, void *rport_drv);
void bfa_rport_delete(struct bfa_rport_s *rport);
void bfa_rport_online(struct bfa_rport_s *rport,
		      struct bfa_rport_info_s *rport_info);
void bfa_rport_offline(struct bfa_rport_s *rport);
void bfa_rport_speed(struct bfa_rport_s *rport, enum bfa_port_speed speed);
void bfa_rport_get_stats(struct bfa_rport_s *rport,
			 struct bfa_rport_hal_stats_s *stats);
void bfa_rport_clear_stats(struct bfa_rport_s *rport);
void bfa_cb_rport_online(void *rport);
void bfa_cb_rport_offline(void *rport);
void bfa_cb_rport_qos_scn_flowid(void *rport,
				 struct bfa_rport_qos_attr_s old_qos_attr,
				 struct bfa_rport_qos_attr_s new_qos_attr);
void bfa_cb_rport_scn_online(struct bfa_s *bfa);
void bfa_cb_rport_scn_offline(struct bfa_s *bfa);
void bfa_cb_rport_scn_no_dev(void *rp);
void bfa_cb_rport_qos_scn_prio(void *rport,
			       struct bfa_rport_qos_attr_s old_qos_attr,
			       struct bfa_rport_qos_attr_s new_qos_attr);
void bfa_rport_get_qos_attr(struct bfa_rport_s *rport,
			    struct bfa_rport_qos_attr_s *qos_attr);

bfa_boolean_t bfa_rport_iopf_status(struct bfa_rport_s *rport);
wwn_t bfa_cb_rport_get_pwwn(void *rport);

#define BFA_RPORT_TAG_INVALID 0xffff
#define BFA_LP_TAG_INVALID 0xff

/**
 * Rport LUN masking APIs
 */
void bfa_rport_set_lunmask(struct bfa_s *bfa, bfa_rport_t *rp);
void bfa_rport_unset_lunmask(struct bfa_s *bfa, bfa_rport_t *rp);
bfa_boolean_t bfa_rport_lunmask_active(bfa_rport_t *rp);
wwn_t bfa_rport_get_pwwn(struct bfa_s *bfa, bfa_rport_t *rp);
struct bfa_rport_s *bfa_rport_get_by_wwn(struct bfa_s *bfa, uint16_t vf_id,
	 wwn_t *lpwwn, wwn_t rpwwn);
void *bfa_cb_get_rp_by_wwn(void *arg, uint16_t vf_id,
	 wwn_t *lpwwn, wwn_t rpwwn);
void bfa_rport_iopf_add(bfa_rport_t *rp, bfa_fcpim_iopf_t *iopf);
void bfa_rport_iopf_remove(bfa_rport_t *rp, bfa_fcpim_iopf_t *iopf);
uint16_t bfa_rport_get_iopf_id(bfa_rport_t *rp, lun_t lun);

/*
 * bfa fcxp API functions
 */
struct bfa_fcxp_s *bfa_fcxp_req_rsp_alloc(void *bfad_fcxp, struct bfa_s *bfa,
				  int nreq_sgles, int nrsp_sgles,
				  bfa_fcxp_get_sgaddr_t get_req_sga,
				  bfa_fcxp_get_sglen_t get_req_sglen,
				  bfa_fcxp_get_sgaddr_t get_rsp_sga,
				  bfa_fcxp_get_sglen_t get_rsp_sglen,
				  bfa_boolean_t req);
void bfa_fcxp_req_rsp_alloc_wait(struct bfa_s *bfa, struct bfa_fcxp_wqe_s *wqe,
			          bfa_fcxp_alloc_cbfn_t alloc_cbfn,
				  void *cbarg, void *bfad_fcxp,
				  int nreq_sgles, int nrsp_sgles,
				  bfa_fcxp_get_sgaddr_t get_req_sga,
				  bfa_fcxp_get_sglen_t get_req_sglen,
				  bfa_fcxp_get_sgaddr_t get_rsp_sga,
				  bfa_fcxp_get_sglen_t get_rsp_sglen,
				  bfa_boolean_t req);
struct bfa_fcxp_s *bfa_fcxp_alloc(void *bfad_fcxp, struct bfa_s *bfa,
				  int nreq_sgles, int nrsp_sgles,
				  bfa_fcxp_get_sgaddr_t get_req_sga,
				  bfa_fcxp_get_sglen_t get_req_sglen,
				  bfa_fcxp_get_sgaddr_t get_rsp_sga,
				  bfa_fcxp_get_sglen_t get_rsp_sglen);
void bfa_fcxp_alloc_wait(struct bfa_s *bfa, struct bfa_fcxp_wqe_s *wqe,
			          bfa_fcxp_alloc_cbfn_t alloc_cbfn,
				  void *cbarg, void *bfad_fcxp,
				  int nreq_sgles, int nrsp_sgles,
				  bfa_fcxp_get_sgaddr_t get_req_sga,
				  bfa_fcxp_get_sglen_t get_req_sglen,
				  bfa_fcxp_get_sgaddr_t get_rsp_sga,
				  bfa_fcxp_get_sglen_t get_rsp_sglen);
void bfa_fcxp_walloc_cancel(struct bfa_s *bfa,
			    struct bfa_fcxp_wqe_s *wqe);
void bfa_fcxp_discard(struct bfa_fcxp_s *fcxp);

void *bfa_fcxp_get_reqbuf(struct bfa_fcxp_s *fcxp);
void *bfa_fcxp_get_rspbuf(struct bfa_fcxp_s *fcxp);

void bfa_fcxp_free(struct bfa_fcxp_s *fcxp);

void bfa_fcxp_send(struct bfa_fcxp_s *fcxp, struct bfa_rport_s *rport,
		   uint16_t vf_id, uint8_t lp_tag,
		   bfa_boolean_t cts, fc_cos_t cos,
		   uint32_t reqlen, fchs_t *fchs,
		   bfa_cb_fcxp_send_t cbfn,
		   void *cbarg,
		   uint32_t rsp_maxlen, uint8_t rsp_timeout);
bfa_status_t bfa_fcxp_abort(struct bfa_fcxp_s *fcxp);
uint32_t bfa_fcxp_get_reqbufsz(struct bfa_fcxp_s *fcxp);
uint32_t bfa_fcxp_get_maxrsp(struct bfa_s *bfa);
void bfa_fcxp_res_recfg(struct bfa_s *bfa, uint16_t num_fcxp_fw);

static inline void *
bfa_uf_get_frmbuf(struct bfa_uf_s *uf)
{
	return uf->data_ptr;
}

static inline   uint16_t
bfa_uf_get_frmlen(struct bfa_uf_s *uf)
{
	return uf->data_len;
}

/*
 * bfa uf API functions
 */
void bfa_uf_recv_register(struct bfa_s *bfa, bfa_cb_uf_recv_t ufrecv,
			  void *cbarg);
void bfa_uf_free(struct bfa_uf_s *uf);

/**
 * bfa lport service api
 */

uint32_t bfa_lps_get_max_vport(struct bfa_s *bfa);
struct bfa_lps_s *bfa_lps_alloc(struct bfa_s *bfa);
void bfa_lps_delete(struct bfa_lps_s *lps);
void bfa_lps_discard(struct bfa_lps_s *lps);
void bfa_lps_flogi(struct bfa_lps_s *lps, void *uarg, uint8_t alpa,
		   uint16_t pdusz, wwn_t pwwn, wwn_t nwwn,
		   bfa_boolean_t auth_en);
void bfa_lps_fdisc(struct bfa_lps_s *lps, void *uarg, uint16_t pdusz,
		   wwn_t pwwn, wwn_t nwwn);
void bfa_lps_flogo(struct bfa_lps_s *lps);
void bfa_lps_fdisclogo(struct bfa_lps_s *lps);
void bfa_lps_set_n2n_pid(struct bfa_lps_s *lps, uint32_t n2n_pid);
uint8_t bfa_lps_get_tag(struct bfa_lps_s *lps);
uint8_t bfa_lps_get_fwtag(struct bfa_s *bfa, uint8_t lp_tag);
bfa_boolean_t bfa_lps_is_npiv_en(struct bfa_lps_s *lps);
bfa_boolean_t bfa_lps_is_fport(struct bfa_lps_s *lps);
bfa_boolean_t bfa_lps_is_brcd_fabric(struct bfa_lps_s *lps);
bfa_boolean_t bfa_lps_is_authreq(struct bfa_lps_s *lps);
bfa_eproto_status_t bfa_lps_get_extstatus(struct bfa_lps_s *lps);
uint32_t bfa_lps_get_pid(struct bfa_lps_s *lps);
uint32_t bfa_lps_get_base_pid(struct bfa_s *bfa);
uint8_t bfa_lps_get_tag_from_pid(struct bfa_s *bfa, uint32_t pid);
uint16_t bfa_lps_get_peer_bbcredit(struct bfa_lps_s *lps);
wwn_t bfa_lps_get_peer_pwwn(struct bfa_lps_s *lps);
wwn_t bfa_lps_get_peer_nwwn(struct bfa_lps_s *lps);
uint8_t bfa_lps_get_lsrjt_rsn(struct bfa_lps_s *lps);
uint8_t bfa_lps_get_lsrjt_expl(struct bfa_lps_s *lps);
mac_t bfa_lps_get_lp_mac(struct bfa_lps_s *lps);
void bfa_cb_lps_flogi_comp(void *bfad, void *uarg, bfa_status_t status);
void bfa_cb_lps_flogo_comp(void *bfad, void *uarg);
void bfa_cb_lps_fdisc_comp(void *bfad, void *uarg, bfa_status_t status);
void bfa_cb_lps_fdisclogo_comp(void *bfad, void *uarg);
void bfa_cb_lps_cvl_event(void *bfad, void *uarg);

/**
 * LPS LUN Masking APIs
 */
bfa_lps_t *bfa_lps_get_from_pwwn(struct bfa_s *bfa, wwn_t lp_wwwn);

void bfa_trunk_enable_cfg(struct bfa_s *bfa);
bfa_status_t bfa_trunk_enable(struct bfa_s *bfa);
bfa_status_t bfa_trunk_disable(struct bfa_s *bfa);
bfa_status_t bfa_trunk_get_attr(struct bfa_s *bfa, bfa_trunk_attr_t *attr);
bfa_status_t bfa_vhba_query(struct bfa_s *bfa, struct bfa_vhba_attr_s *attr);
bfa_status_t bfa_faa_query(struct bfa_s *bfa, struct bfa_faa_attr_s *attr,
		bfa_cb_iocfc_t cbfn, void *cbarg);

#endif /* __BFA_SVC_H__ */
