/*
 * Copyright (c) 2005-2014 Brocade Communications Systems, Inc.
 * Copyright (c) 2014 QLogic Corporation.
 * All rights reserved
 * www.qlogic.com
 *
 * See LICENSE.bfa for copyright and licensing details.
 */
/*
 * fcbuild.h - FC link service frame building and parsing routines
 */

#ifndef __FCBUILD_H__
#define __FCBUILD_H__

#include <bfa_os_inc.h>
#include <protocol/bfa_fc.h>
#include <defs/bfa_defs_fcs.h>

/*
 * Utility Macros/functions
 */

#define wwn_is_equal(_wwn1, _wwn2)		\
	(memcmp(&(_wwn1), &(_wwn2), sizeof(wwn_t)) == 0)

#define fc_roundup(_l, _s) (((_l) + ((_s) - 1)) & ~((_s) - 1))

/*
 * Given the fc response length, this routine will return
 * the length of the actual payload bytes following the CT header.
 *
 * Assumes the input response length does not include the crc, eof, etc.
 */
static inline   uint32_t
fc_get_ctresp_pyld_len(uint32_t resp_len)
{
	return (resp_len - sizeof(ct_hdr_t));
}

/*
 * Convert bfa speed to rpsc speed value.
 */
static inline  enum bfa_port_speed
fc_rpsc_operspeed_to_bfa_speed(fc_rpsc_op_speed_t speed)
{
	switch (speed) {

	case RPSC_OP_SPEED_1G :
		return BFA_PORT_SPEED_1GBPS;

	case RPSC_OP_SPEED_2G :
		return BFA_PORT_SPEED_2GBPS;

	case RPSC_OP_SPEED_4G :
		return BFA_PORT_SPEED_4GBPS;

	case RPSC_OP_SPEED_8G :
		return BFA_PORT_SPEED_8GBPS;

	case RPSC_OP_SPEED_16G :
		return BFA_PORT_SPEED_16GBPS;

	case RPSC_OP_SPEED_10G :
		return BFA_PORT_SPEED_10GBPS;

	default:
		return BFA_PORT_SPEED_UNKNOWN;
	}
}

/*
 * Convert RPSC speed to bfa speed value.
 */
static inline   fc_rpsc_op_speed_t
fc_bfa_speed_to_rpsc_operspeed(enum bfa_port_speed op_speed)
{
	switch (op_speed) {

	case BFA_PORT_SPEED_1GBPS:
		return RPSC_OP_SPEED_1G;

	case BFA_PORT_SPEED_2GBPS:
		return RPSC_OP_SPEED_2G;

	case BFA_PORT_SPEED_4GBPS:
		return RPSC_OP_SPEED_4G;

	case BFA_PORT_SPEED_8GBPS:
		return RPSC_OP_SPEED_8G;

	case BFA_PORT_SPEED_16GBPS:
		return RPSC_OP_SPEED_16G;

	case BFA_PORT_SPEED_10GBPS:
		return RPSC_OP_SPEED_10G;

	default:
		return RPSC_OP_SPEED_NOT_EST;
	}
}
typedef enum {
	FC_PARSE_OK = 0,
	FC_PARSE_FAILURE = 1,
	FC_PARSE_BUSY = 2,
	FC_PARSE_LEN_INVAL,
	FC_PARSE_ACC_INVAL,
	FC_PARSE_PWWN_NOT_EQUAL,
	FC_PARSE_NWWN_NOT_EQUAL,
	FC_PARSE_RXSZ_INVAL,
	FC_PARSE_NOT_FCP,
	FC_PARSE_OPAFLAG_INVAL,
	FC_PARSE_RPAFLAG_INVAL,
	FC_PARSE_OPA_INVAL,
	FC_PARSE_RPA_INVAL,

} fc_parse_status_t;

typedef struct fc_templates {
	fchs_t		fc_els_req;
	fchs_t		fc_bls_req;
	fc_plogi_t      plogi;
	fc_rrq_t	rrq;
} fc_templates_t;

void	    fcbuild_init(void);

uint16_t	fc_flogi_build(fchs_t *fchs, fc_flogi_t *flogi, uint32_t s_id,
		       uint16_t ox_id, wwn_t port_name, wwn_t node_name,
		       uint16_t pdu_size, uint8_t set_npiv, uint8_t set_auth,
		       uint16_t local_bb_credits);

uint16_t	fc_fdisc_build(fchs_t *buf, fc_flogi_t *flogi, uint32_t s_id,
			       uint16_t ox_id, wwn_t port_name, wwn_t node_name,
			       uint16_t pdu_size);

uint16_t	fc_flogi_acc_build(fchs_t *fchs, fc_flogi_t *flogi,
				   uint32_t s_id, uint16_t ox_id,
				   wwn_t port_name, wwn_t node_name,
				   uint16_t pdu_size,
				   uint16_t local_bb_credits,
				   uint8_t bb_scn);

uint16_t	fc_plogi_build(fchs_t *fchs, void *pld, uint32_t d_id,
			       uint32_t s_id, uint16_t ox_id, wwn_t port_name,
			       wwn_t node_name, uint16_t pdu_size,
			       uint16_t bb_cr);

fc_parse_status_t fc_plogi_parse(fchs_t *fchs);

uint16_t	fc_abts_build(fchs_t *buf, uint32_t d_id, uint32_t s_id,
			      uint16_t ox_id);

fc_parse_status_t fc_abts_rsp_parse(fchs_t *buf, int len);

uint16_t	fc_rrq_build(fchs_t *buf, fc_rrq_t *rrq, uint32_t d_id,
		     uint32_t s_id, uint16_t ox_id, uint16_t rrq_oxid);
fc_parse_status_t fc_rrq_rsp_parse(fchs_t *buf, int len);

uint16_t	fc_rspnid_build(fchs_t *fchs, void *pld, uint32_t s_id,
				uint16_t ox_id, uint8_t *name);
uint16_t	fc_rsnn_nn_build(fchs_t *fchs, void *pld, uint32_t s_id,
				wwn_t node_name, uint8_t *name);

uint16_t	fc_rftid_build(fchs_t *fchs, void *pld, uint32_t s_id,
			       uint16_t ox_id, bfa_lport_role_t role);

uint16_t	fc_rftid_build_sol(fchs_t *fchs, void *pyld, uint32_t s_id,
				   uint16_t ox_id, uint8_t *fc4_bitmap,
				   uint32_t bitmap_size);

uint16_t	fc_rffid_build(fchs_t *fchs, void *pyld, uint32_t s_id,
			uint16_t ox_id, uint8_t fc4_type, uint8_t fc4_ftrs);

uint16_t	fc_gidpn_build(fchs_t *fchs, void *pyld, uint32_t s_id,
			       uint16_t ox_id, wwn_t port_name);

uint16_t	fc_gpnid_build(fchs_t *fchs, void *pld, uint32_t s_id,
			       uint16_t ox_id, uint32_t port_id);

uint16_t 	fc_gs_rjt_build(fchs_t *fchs, ct_hdr_t *cthdr,
				uint32_t d_id, uint32_t s_id, uint16_t ox_id,
				uint8_t reason_code, uint8_t reason_code_expl);

uint16_t	fc_scr_build(fchs_t *fchs, fc_scr_t *scr, uint8_t set_br_reg,
			uint32_t s_id, uint16_t ox_id);

uint16_t	fc_plogi_acc_build(fchs_t *fchs, void *pld, uint32_t d_id,
				   uint32_t s_id, uint16_t ox_id,
				   wwn_t port_name, wwn_t node_name,
				   uint16_t pdu_size, uint16_t bb_cr);

uint16_t	fc_adisc_build(fchs_t *fchs, fc_adisc_t *adisc, uint32_t d_id,
			       uint32_t s_id, uint16_t ox_id, wwn_t port_name,
			       wwn_t node_name);

fc_parse_status_t fc_adisc_parse(fchs_t *fchs, void *pld, uint32_t host_dap,
				 wwn_t node_name, wwn_t port_name);

fc_parse_status_t fc_adisc_rsp_parse(fc_adisc_t *adisc, int len,
				 wwn_t port_name, wwn_t node_name);

uint16_t	fc_adisc_acc_build(fchs_t *fchs, fc_adisc_t *adisc,
				   uint32_t d_id, uint32_t s_id, uint16_t ox_id,
				   wwn_t port_name, wwn_t node_name);
uint16_t	fc_ls_rjt_build(fchs_t *fchs, fc_ls_rjt_t *ls_rjt,
				uint32_t d_id, uint32_t s_id, uint16_t ox_id,
				uint8_t reason_code, uint8_t reason_code_expl);
uint16_t	fc_ls_acc_build(fchs_t *fchs, fc_els_cmd_t *els_cmd,
				uint32_t d_id, uint32_t s_id, uint16_t ox_id);
uint16_t	fc_prli_build(fchs_t *fchs, void *pld, uint32_t d_id,
			      uint32_t s_id, uint16_t ox_id);

fc_parse_status_t fc_prli_rsp_parse(fc_prli_t *prli, int len);

uint16_t	fc_prli_acc_build(fchs_t *fchs, void *pld, uint32_t d_id,
				  uint32_t s_id, uint16_t ox_id,
				  enum bfa_lport_role role);

uint16_t	fc_rnid_build(fchs_t *fchs, fc_rnid_cmd_t *rnid,
			      uint32_t d_id, uint32_t s_id, uint16_t ox_id,
			      uint32_t data_format);

uint16_t	fc_rnid_acc_build(fchs_t *fchs, fc_rnid_acc_t *rnid_acc,
				  uint32_t d_id, uint32_t s_id, uint16_t ox_id,
				  uint32_t data_format,
				  fc_rnid_common_id_data_t *common_id_data,
				  fc_rnid_general_topology_data_t *
				  gen_topo_data);

uint16_t	fc_rpsc2_build(fchs_t *fchs, fc_rpsc2_cmd_t *rps2c,
				uint32_t d_id, uint32_t s_id,
				uint32_t *pid_list, uint16_t npids);
uint16_t	fc_rpsc_build(fchs_t *fchs, fc_rpsc_cmd_t *rpsc,
			      uint32_t d_id, uint32_t s_id, uint16_t ox_id);
uint16_t	fc_rpsc_acc_build(fchs_t *fchs, fc_rpsc_acc_t *rpsc_acc,
				  uint32_t d_id, uint32_t s_id, uint16_t ox_id,
				  fc_rpsc_speed_info_t *oper_speed);
uint16_t	fc_gid_ft_build(fchs_t *fchs, void *pld, uint32_t s_id,
				uint8_t fc4_type);

uint16_t	fc_rpnid_build(fchs_t *fchs, void *pyld, uint32_t s_id,
			       uint32_t port_id, wwn_t port_name);

uint16_t	fc_rnnid_build(fchs_t *fchs, void *pyld, uint32_t s_id,
			       uint32_t port_id, wwn_t node_name);

uint16_t	fc_rcsid_build(fchs_t *fchs, void *pyld, uint32_t s_id,
			       uint32_t port_id, uint32_t cos);

uint16_t	fc_rptid_build(fchs_t *fchs, void *pyld, uint32_t s_id,
			       uint32_t port_id, uint8_t port_type);

uint16_t	fc_ganxt_build(fchs_t *fchs, void *pyld, uint32_t s_id,
			       uint32_t port_id);

uint16_t	fc_logo_build(fchs_t *fchs, fc_logo_t *logo, uint32_t d_id,
			      uint32_t s_id, uint16_t ox_id, wwn_t port_name);

uint16_t	fc_logo_acc_build(fchs_t *fchs, void *pld, uint32_t d_id,
				  uint32_t s_id, uint16_t ox_id);

uint16_t	fc_fdmi_reqhdr_build(fchs_t *fchs, void *pyld, uint32_t s_id,
				     uint16_t cmd_code);
uint16_t 	fc_gmal_req_build(fchs_t *fchs, void *pyld,
				uint32_t s_id, wwn_t wwn);
uint16_t 	fc_gfn_req_build(fchs_t *fchs, void *pyld,
				uint32_t s_id, wwn_t wwn);

void		fc_get_fc4type_bitmask(uint8_t fc4_type, uint8_t *bit_mask);

void		fc_els_req_build(fchs_t *fchs, uint32_t d_id, uint32_t s_id,
					 uint16_t ox_id);

fc_parse_status_t	fc_els_rsp_parse(fchs_t *fchs, int len);

fc_parse_status_t	fc_plogi_rsp_parse(fchs_t *fchs, int len,
					wwn_t port_name);

fc_parse_status_t	fc_prli_parse(fc_prli_t *prli);

fc_parse_status_t	fc_pdisc_parse(fchs_t *fchs, wwn_t node_name,
					wwn_t port_name);

uint16_t fc_ba_acc_build(fchs_t *fchs, fc_ba_acc_t *ba_acc, uint32_t d_id,
		uint32_t s_id, uint16_t ox_id, uint16_t rx_id);

int fc_logout_params_pages(fchs_t *fc_frame, uint8_t els_code);

uint16_t fc_tprlo_acc_build(fchs_t *fchs, fc_tprlo_acc_t *tprlo_acc,
		uint32_t d_id, uint32_t s_id, uint16_t ox_id, int num_pages);

uint16_t fc_prlo_acc_build(fchs_t *fchs, fc_prlo_acc_t *prlo_acc, uint32_t d_id,
		uint32_t s_id, uint16_t ox_id, int num_pages);

uint16_t fc_logo_rsp_parse(fchs_t *fchs, int len);

uint16_t fc_pdisc_build(fchs_t *fchs, uint32_t d_id, uint32_t s_id,
		uint16_t ox_id, wwn_t port_name, wwn_t node_name,
		uint16_t pdu_size);

uint16_t fc_pdisc_rsp_parse(fchs_t *fchs, int len, wwn_t port_name);

uint16_t fc_prlo_build(fchs_t *fchs, uint32_t d_id, uint32_t s_id,
		uint16_t ox_id, int num_pages);

uint16_t fc_prlo_rsp_parse(fchs_t *fchs, int len);

uint16_t fc_tprlo_build(fchs_t *fchs, uint32_t d_id, uint32_t s_id,
		uint16_t ox_id, int num_pages, fc_tprlo_type_t tprlo_type,
		uint32_t tpr_id);

uint16_t fc_tprlo_rsp_parse(fchs_t *fchs, int len);

uint16_t fc_ba_rjt_build(fchs_t *fchs, uint32_t d_id, uint32_t s_id,
		uint16_t ox_id, uint32_t reason_code, uint32_t reason_expl);

uint16_t fc_gnnid_build(fchs_t *fchs, void *pyld, uint32_t s_id, uint16_t ox_id,
		uint32_t port_id);

uint16_t fc_ct_rsp_parse(ct_hdr_t *cthdr);

uint16_t fc_rscn_build(fchs_t *fchs, fc_rscn_pl_t *rscn, uint32_t s_id,
		uint16_t ox_id);
void fcp_rsp_fchdr_build(fchs_t *fchs, uint32_t d_id, uint32_t s_id,
		uint16_t ox_id);
uint16_t fc_rec_acc_build(fchs_t *fchs, void * pyld, uint16_t ox_id,
		uint32_t s_id, uint32_t d_id, uint16_t io_oxid,
		uint16_t io_rxid, uint32_t count, uint32_t e_stat);
uint16_t fc_rrq_rsp_build(fchs_t *fchs, void *pyld, uint32_t s_id,
		uint32_t d_id, uint16_t ox_id);
uint16_t fc_srr_acc_build(fchs_t *fchs, void *pld, uint32_t d_id,
		uint32_t s_id, uint16_t ox_id);
#endif
