/*
 * Copyright (c) 2005-2014 Brocade Communications Systems, Inc.
 * Copyright (c) 2014 QLogic Corporation.
 * All rights reserved
 * www.qlogic.com
 *
 * See LICENSE.bfa for copyright and licensing details.
 */

#include <bfa/bfa_modules.h>
#include <bfa_cb_ioim_macros.h>
#include <protocol/scsi.h>

BFA_TRC_FILE(HAL, FCPIM);

#define bfa_fcpim_add_iostats(__l, __r, __stats)	\
	__l->__stats += __r->__stats

#define BFA_IOIM_BITS_IN_BYTE 8

/**
 *  BFA ITNIM Related definitions
 */
static void bfa_itnim_update_del_itn_stats(struct bfa_itnim_s *itnim);
static bfa_boolean_t bfa_ioim_lm_proc_rpl_data(struct bfa_ioim_s *ioim);
static bfa_boolean_t bfa_ioim_lm_proc_inq_data(struct bfa_ioim_s *ioim);
static void bfa_ioim_lm_init(struct bfa_s *bfa);
static void bfa_fcpim_iopf_start_io(struct bfa_ioim_s *ioim);
static void bfa_fcpim_iopf_comp_io(struct bfa_ioim_s *ioim,
		bfa_boolean_t fastpath);
static void bfa_fcpim_iopf_scsi_comp_stats(struct bfa_ioim_s *ioim,
		 scsi_sense_t *sense);
static void bfa_fcpim_iopf_reset_data(bfa_fcpim_iopf_t *iopf);
static void bfa_fcpim_iopf_sum_lun(struct bfa_s *bfa, bfa_fcpim_iopf_t *iopf,
		bfa_fcpim_iopf_t *iopf_in);
static void bfa_fcpim_iopf_sum_rp(struct bfa_s *bfa, bfa_fcpim_iopf_t *iopf,
		bfa_fcpim_iopf_t *iopf_in);
static bfa_status_t bfa_fcpim_iopf_show_rp(bfa_t *bfa, uint16_t vf_id,
	 wwn_t lpwwn, wwn_t rpwwn,  bfa_fcpim_iopf_t *iopf);
static void bfa_fcpim_iopf_show_process(bfa_t *bfa, bfa_fcpim_iopf_t *iopf,
		bfa_fcpim_iopf_t *iopf_in);
static void bfa_fcpim_iopf_unmark_showdone(bfa_t *bfa);

#define BFA_ITNIM_FROM_TAG(_fcpim, _tag)                                \
	((_fcpim)->itnim_arr + ((_tag) & ((_fcpim)->num_itnims - 1)))
#define bfa_lun_stats(__ioim, __stats)	\
	if ((__ioim)->iopf)		\
	(__ioim)->iopf->io_st.__stats++;

#define bfa_fcpim_additn(__itnim)					\
	bfa_q_enq(&(__itnim)->fcpim->itnim_q, __itnim)
#define bfa_fcpim_delitn(__itnim)	do {				\
	bfa_assert(bfa_q_is_on_q(&(__itnim)->fcpim->itnim_q, __itnim)); \
	bfa_itnim_update_del_itn_stats(__itnim);			\
	bfa_q_qe_deq(__itnim);						\
	bfa_assert(bfa_q_is_empty(&(__itnim)->io_q));			\
	bfa_assert(bfa_q_is_empty(&(__itnim)->io_cleanup_q));		\
	bfa_assert(bfa_q_is_empty(&(__itnim)->pending_q));		\
} while (0)

#define bfa_itnim_online_cb(__itnim) do {				\
	if ((__itnim)->bfa->fcs)					\
		bfa_cb_itnim_online((__itnim)->ditn);			\
	else {								\
		bfa_cb_queue((__itnim)->bfa, &(__itnim)->hcb_qe,	\
		__bfa_cb_itnim_online, (__itnim));			\
	}								\
} while (0)

#define bfa_itnim_offline_cb(__itnim) do {				\
	if ((__itnim)->bfa->fcs)					\
		bfa_cb_itnim_offline((__itnim)->ditn);			\
	else {								\
		bfa_cb_queue((__itnim)->bfa, &(__itnim)->hcb_qe,	\
		__bfa_cb_itnim_offline, (__itnim));			\
	}								\
} while (0)

#define bfa_ioim_rp_wwn(__ioim)						\
	bfa_rport_get_pwwn((__ioim)->bfa, (__ioim)->itnim->rport)

#define bfa_ioim_lp_wwn(__ioim)						\
	(BFA_LPS_FROM_TAG(BFA_LPS_MOD((__ioim)->bfa),			\
	(__ioim)->itnim->rport->rport_info.lp_tag))->pwwn		\

#define bfa_itnim_sler_cb(__itnim) do {					\
	if ((__itnim)->bfa->fcs)					\
		bfa_cb_itnim_sler((__itnim)->ditn);			\
	else {								\
		bfa_cb_queue((__itnim)->bfa, &(__itnim)->hcb_qe,	\
		__bfa_cb_itnim_sler, (__itnim));			\
	}								\
} while (0)

enum bfa_ioim_lm_status {
	BFA_IOIM_LM_PRESENT = 1,
	BFA_IOIM_LM_LUN_NOT_SUP = 2,
	BFA_IOIM_LM_RPL_DATA_CHANGED = 3,
	BFA_IOIM_LM_LUN_NOT_RDY = 4,
};

enum bfa_ioim_lm_ua_status {
	BFA_IOIM_LM_UA_RESET = 0,
	BFA_IOIM_LM_UA_SET = 1,
};
/**
 * @dg bfa_itnim_sm BFA itnim state machine
 * @img bfa_itnim_sm.jpg
 * @{
 */

/**
 * @brief BFA itnim state machine events
 */
enum bfa_itnim_event {
	BFA_ITNIM_SM_CREATE = 1,	/*!< itnim is created */
	BFA_ITNIM_SM_ONLINE = 2,	/*!< itnim is online */
	BFA_ITNIM_SM_OFFLINE = 3,	/*!< itnim is offline */
	BFA_ITNIM_SM_FWRSP = 4,		/*!< firmware response */
	BFA_ITNIM_SM_DELETE = 5,	/*!< deleting an existing itnim */
	BFA_ITNIM_SM_CLEANUP = 6,	/*!< IO cleanup completion */
	BFA_ITNIM_SM_SLER = 7,		/*!< second level error recovery */
	BFA_ITNIM_SM_HWFAIL = 8,	/*!< IOC h/w failure event */
	BFA_ITNIM_SM_QRESUME = 9,	/*!< queue space available */
};

/**
 *  BFA IOIM related definitions
 */
#define bfa_ioim_move_to_comp_q(__ioim) do {				\
	bfa_q_qe_mov(__ioim, &__ioim->fcpim->ioim_comp_q);		\
} while (0)

#define bfa_ioim_cb_profile_comp(__fcpim, __ioim, __fp) do {		\
	if ((__ioim->itnim->rport->rp_iopf_enabled) || (__ioim)->iopf)	\
		bfa_fcpim_iopf_comp_io(__ioim, __fp);			\
} while (0)


#define bfa_ioim_cb_iopf_start(__ioim)	do {				\
	if (__ioim->itnim->rport->rp_iopf_enabled) 			\
		(__ioim)->start_time = bfa_os_get_clock((__ioim)->bfa->bfad); \
	if (__ioim->itnim->rport->iopf_enabled)				\
		bfa_fcpim_iopf_start_io(__ioim);			\
} while (0)

/**
 * @ig hal_ioim_sm
 * @{
 */

/**
 * IO state machine events
 */
enum bfa_ioim_event {
	BFA_IOIM_SM_START	= 1,	/* !< io start request from host */
	BFA_IOIM_SM_COMP_GOOD	= 2,	/* !< io good comp, resource free */
	BFA_IOIM_SM_COMP	= 3,	/* !< io comp, resource is free */
	BFA_IOIM_SM_COMP_UTAG	= 4,	/* !< io comp, resource is free */
	BFA_IOIM_SM_DONE	= 5,	/* !< io comp, resource not free */
	BFA_IOIM_SM_FREE	= 6,	/* !< io resource is freed */
	BFA_IOIM_SM_ABORT	= 7,	/* !< abort request from scsi stack */
	BFA_IOIM_SM_ABORT_COMP	= 8,	/* !< abort from f/w */
	BFA_IOIM_SM_ABORT_DONE	= 9,	/* !< abort completion from f/w */
	BFA_IOIM_SM_QRESUME	= 10,	/* !< CQ space available to queue IO */
	BFA_IOIM_SM_SGALLOCED	= 11,	/* !< SG page allocation successful */
	BFA_IOIM_SM_SQRETRY	= 12,	/* !< sequence recovery retry */
	BFA_IOIM_SM_HCB		= 13,	/* !< bfa callback complete */
	BFA_IOIM_SM_CLEANUP	= 14,	/* !< IO cleanup from itnim */
	BFA_IOIM_SM_TMSTART	= 15,	/* !< IO cleanup from tskim */
	BFA_IOIM_SM_TMDONE	= 16,	/* !< IO cleanup from tskim */
	BFA_IOIM_SM_HWFAIL	= 17,	/* !< IOC h/w failure event */
	BFA_IOIM_SM_IOTOV	= 18,	/* !< ITN offline TOV */
	BFA_IOIM_SM_LM_LUN_NOT_SUP = 19, /* !< lunmask lun not supported */
	BFA_IOIM_SM_LM_RPL_DC = 20,	/* !< lunmask report-lun data changed */
	BFA_IOIM_SM_LM_LUN_NOT_RDY = 21, /* !< lunmask lun not ready */
};


/**
 *  BFA TSKIM related definitions
 */

/**
 * task management completion handling
 */
#define bfa_tskim_qcomp(__tskim, __cbfn) do {				\
	bfa_cb_queue((__tskim)->bfa, &(__tskim)->hcb_qe, __cbfn, (__tskim));\
	bfa_tskim_notify_comp(__tskim);					\
} while (0)

#define bfa_tskim_notify_comp(__tskim) do {				\
	if ((__tskim)->notify)						\
		bfa_itnim_tskdone((__tskim)->itnim);			\
} while (0)


enum bfa_tskim_event {
	BFA_TSKIM_SM_START	= 1,	/*!< TM command start		*/
	BFA_TSKIM_SM_DONE	= 2,	/*!< TM completion		*/
	BFA_TSKIM_SM_QRESUME	= 3,	/*!< resume after qfull		*/
	BFA_TSKIM_SM_HWFAIL	= 5,	/*!< IOC h/w failure event	*/
	BFA_TSKIM_SM_HCB	= 6,	/*!< BFA callback completion	*/
	BFA_TSKIM_SM_IOS_DONE	= 7,	/*!< IO and sub TM completions	*/
	BFA_TSKIM_SM_CLEANUP	= 8,	/*!< TM cleanup on ITN offline	*/
	BFA_TSKIM_SM_CLEANUP_DONE = 9,	/*!< TM abort completion	*/
	BFA_TSKIM_SM_UTAG	= 10,	/*!< TM completion unknown tag	*/
};

/**
 * forward declaration for BFA ITNIM functions
 */
static void     bfa_itnim_iocdisable_cleanup(struct bfa_itnim_s *itnim);
static bfa_boolean_t bfa_itnim_send_fwcreate(struct bfa_itnim_s *itnim);
static bfa_boolean_t bfa_itnim_send_fwdelete(struct bfa_itnim_s *itnim);
static void     bfa_itnim_cleanp_comp(void *itnim_cbarg);
static void     bfa_itnim_cleanup(struct bfa_itnim_s *itnim);
static void     __bfa_cb_itnim_online(void *cbarg, bfa_boolean_t complete);
static void     __bfa_cb_itnim_offline(void *cbarg, bfa_boolean_t complete);
static void     __bfa_cb_itnim_sler(void *cbarg, bfa_boolean_t complete);
static void     bfa_itnim_iotov_online(struct bfa_itnim_s *itnim);
static void     bfa_itnim_iotov_cleanup(struct bfa_itnim_s *itnim);
static void     bfa_itnim_iotov(void *itnim_arg);
static void     bfa_itnim_iotov_start(struct bfa_itnim_s *itnim);
static void     bfa_itnim_iotov_stop(struct bfa_itnim_s *itnim);
static void     bfa_itnim_iotov_delete(struct bfa_itnim_s *itnim);

/**
 * forward declaration of ITNIM state machine
 */
static void     bfa_itnim_sm_uninit(struct bfa_itnim_s *itnim,
					enum bfa_itnim_event event);
static void     bfa_itnim_sm_created(struct bfa_itnim_s *itnim,
					enum bfa_itnim_event event);
static void     bfa_itnim_sm_fwcreate(struct bfa_itnim_s *itnim,
					enum bfa_itnim_event event);
static void     bfa_itnim_sm_delete_pending(struct bfa_itnim_s *itnim,
					enum bfa_itnim_event event);
static void     bfa_itnim_sm_online(struct bfa_itnim_s *itnim,
					enum bfa_itnim_event event);
static void     bfa_itnim_sm_sler(struct bfa_itnim_s *itnim,
					enum bfa_itnim_event event);
static void     bfa_itnim_sm_cleanup_offline(struct bfa_itnim_s *itnim,
					enum bfa_itnim_event event);
static void     bfa_itnim_sm_cleanup_delete(struct bfa_itnim_s *itnim,
					enum bfa_itnim_event event);
static void     bfa_itnim_sm_fwdelete(struct bfa_itnim_s *itnim,
					enum bfa_itnim_event event);
static void     bfa_itnim_sm_offline(struct bfa_itnim_s *itnim,
					enum bfa_itnim_event event);
static void     bfa_itnim_sm_iocdisable(struct bfa_itnim_s *itnim,
					enum bfa_itnim_event event);
static void     bfa_itnim_sm_deleting(struct bfa_itnim_s *itnim,
					enum bfa_itnim_event event);
static void     bfa_itnim_sm_fwcreate_qfull(bfa_itnim_t *itnim,
					enum bfa_itnim_event event);
static void     bfa_itnim_sm_fwdelete_qfull(bfa_itnim_t *itnim,
					enum bfa_itnim_event event);
static void     bfa_itnim_sm_deleting_qfull(bfa_itnim_t *itnim,
					enum bfa_itnim_event event);

/** 
 * forward declaration for BFA IOIM functions
 */
static bfa_boolean_t	bfa_ioim_send_ioreq(struct bfa_ioim_s *ioim);
static bfa_boolean_t	bfa_ioim_sge_setup(struct bfa_ioim_s *ioim);
static void		bfa_ioim_sgpg_setup(struct bfa_ioim_s *ioim);
static bfa_boolean_t	bfa_ioim_send_abort(struct bfa_ioim_s *ioim);
static void		bfa_ioim_notify_cleanup(struct bfa_ioim_s *ioim);
static void __bfa_cb_ioim_good_comp(void *cbarg, bfa_boolean_t complete);
static void __bfa_cb_ioim_comp(void *cbarg, bfa_boolean_t complete);
static void __bfa_cb_ioim_abort(void *cbarg, bfa_boolean_t complete);
static void __bfa_cb_ioim_failed(void *cbarg, bfa_boolean_t complete);
static void __bfa_cb_ioim_lm_lun_not_sup(void *cbarg, bfa_boolean_t complete);
static void __bfa_cb_ioim_lm_rpl_dc(void *cbarg, bfa_boolean_t complete);
static void __bfa_cb_ioim_lm_lun_not_rdy(void *cbarg, bfa_boolean_t complete);
static void __bfa_cb_ioim_pathtov(void *cbarg, bfa_boolean_t complete);
static bfa_boolean_t    bfa_ioim_is_abortable(struct bfa_ioim_s *ioim);


/**
 * forward declaration of BFA IO state machine
 */
static void     bfa_ioim_sm_uninit(struct bfa_ioim_s *ioim,
					enum bfa_ioim_event event);
static void     bfa_ioim_sm_sgalloc(struct bfa_ioim_s *ioim,
					enum bfa_ioim_event event);
static void     bfa_ioim_sm_active(struct bfa_ioim_s *ioim,
					enum bfa_ioim_event event);
static void     bfa_ioim_sm_abort(struct bfa_ioim_s *ioim,
					enum bfa_ioim_event event);
static void     bfa_ioim_sm_cleanup(struct bfa_ioim_s *ioim,
					enum bfa_ioim_event event);
static void     bfa_ioim_sm_qfull(struct bfa_ioim_s *ioim,
					enum bfa_ioim_event event);
static void     bfa_ioim_sm_abort_qfull(struct bfa_ioim_s *ioim,
					enum bfa_ioim_event event);
static void     bfa_ioim_sm_cleanup_qfull(struct bfa_ioim_s *ioim,
					enum bfa_ioim_event event);
static void     bfa_ioim_sm_hcb(struct bfa_ioim_s *ioim,
					enum bfa_ioim_event event);
static void     bfa_ioim_sm_hcb_free(struct bfa_ioim_s *ioim,
					enum bfa_ioim_event event);
static void     bfa_ioim_sm_resfree(struct bfa_ioim_s *ioim,
					enum bfa_ioim_event event);
static void	bfa_ioim_sm_cmnd_retry(struct bfa_ioim_s *ioim,
					enum bfa_ioim_event event);

/**
 * forward declaration for BFA TSKIM functions
 */
static void     __bfa_cb_tskim_done(void *cbarg, bfa_boolean_t complete);
static void     __bfa_cb_tskim_failed(void *cbarg, bfa_boolean_t complete);
static bfa_boolean_t bfa_tskim_match_scope(struct bfa_tskim_s *tskim,
					lun_t lun);
static void     bfa_tskim_gather_ios(struct bfa_tskim_s *tskim);
static void     bfa_tskim_cleanp_comp(void *tskim_cbarg);
static void     bfa_tskim_cleanup_ios(struct bfa_tskim_s *tskim);
static bfa_boolean_t bfa_tskim_send(struct bfa_tskim_s *tskim);
static bfa_boolean_t bfa_tskim_send_abort(struct bfa_tskim_s *tskim);
static void     bfa_tskim_iocdisable_ios(struct bfa_tskim_s *tskim);


/**
 * forward declaration of BFA TSKIM state machine
 */
static void     bfa_tskim_sm_uninit(struct bfa_tskim_s *tskim,
					enum bfa_tskim_event event);
static void     bfa_tskim_sm_active(struct bfa_tskim_s *tskim,
					enum bfa_tskim_event event);
static void     bfa_tskim_sm_cleanup(struct bfa_tskim_s *tskim,
					enum bfa_tskim_event event);
static void     bfa_tskim_sm_iocleanup(struct bfa_tskim_s *tskim,
					enum bfa_tskim_event event);
static void     bfa_tskim_sm_qfull(struct bfa_tskim_s *tskim,
					enum bfa_tskim_event event);
static void     bfa_tskim_sm_cleanup_qfull(struct bfa_tskim_s *tskim,
					enum bfa_tskim_event event);
static void     bfa_tskim_sm_hcb(struct bfa_tskim_s *tskim,
					enum bfa_tskim_event event);

/**
 * @dg hal_fcpim_mod BFA FCP Initiator Mode module
 * @{
 */



/**
 * @brief
 * 		Compute and return memory needed by FCP(im) module.
*/
void
bfa_itnim_lun_latency_init(struct bfa_itnim_latency_s *io_lat)
{
	int j;

	/* bfal check for clock div to be non zero */
	io_lat->clock_res_mul = bfa_io_lat_clock_res_mul;
	io_lat->clock_res_div = bfa_io_lat_clock_res_div;

	for (j = 0; j < BFA_IOBUCKET_MAX; j++)
		io_lat->min[j] = ~0;
}

/**
 * @brief
 * 		Compute and return memory needed by FCP(im) module.
*/
void
bfa_iopf_attach(struct bfa_fcpim_s *fcpim)
{
	struct bfa_fcpim_iopf_qe_s *iopf;
	int i;

	bfa_q_init(&fcpim->iopf_free_q);
	bfa_q_init(&fcpim->iopf_active_q);

	iopf = (struct bfa_fcpim_iopf_qe_s *) bfa_mem_kva_curp(fcpim->fcp);
	bfa_mem_kva_curp(fcpim->fcp) += BFA_IOPF_MAX *
		 sizeof(struct bfa_fcpim_iopf_qe_s);

	fcpim->iopf_arr = iopf;
	for (i = 0; i < BFA_IOPF_MAX; i++) {
		bfa_os_memset(iopf, 0, sizeof(struct bfa_fcpim_iopf_qe_s));
		iopf->iopf.id = (uint16_t) i;
		bfa_q_qe_init(&(iopf->qe));
		bfa_q_enq(&fcpim->iopf_free_q, iopf);
		/* ITN and LUN  io latency init */
		bfa_itnim_lun_latency_init(&(iopf->iopf.io_lat));
		iopf++;
	}

}

/**
 * @brief
 * 		Compute and return memory needed by FCP(im) module.
 */
void
bfa_fcpim_meminfo(struct bfa_iocfc_cfg_s *cfg, uint32_t *km_len)
{
	bfa_itnim_meminfo(cfg, km_len);

	/**
	 * Advanced IO profile memory
	 */
	if (!cfg->drvcfg.min_cfg)
		*km_len += BFA_IOPF_MAX * sizeof(struct bfa_fcpim_iopf_qe_s);

	/**
	 * IO memory
	 */
	*km_len += cfg->fwcfg.num_ioim_reqs *
		(sizeof(struct bfa_ioim_s) + sizeof(struct bfa_ioim_sp_s));

	/**
	 * task management command memory
	 */
	if (cfg->fwcfg.num_tskim_reqs < BFA_TSKIM_MIN)
		cfg->fwcfg.num_tskim_reqs = BFA_TSKIM_MIN;
	*km_len += cfg->fwcfg.num_tskim_reqs * sizeof(struct bfa_tskim_s);
}

/**
 * @brief
 *
 */
void
bfa_fcpim_attach(struct bfa_fcp_mod_s *fcp, void *bfad,
		struct bfa_iocfc_cfg_s *cfg,
		struct bfa_pcidev_s *pcidev)
{
	struct bfa_fcpim_s *fcpim = &fcp->fcpim;
	struct bfa_s *bfa = fcp->bfa;

	bfa_trc(bfa, cfg->drvcfg.path_tov);
	bfa_trc(bfa, cfg->fwcfg.num_rports);
	bfa_trc(bfa, cfg->fwcfg.num_ioim_reqs);
	bfa_trc(bfa, cfg->fwcfg.num_tskim_reqs);

	fcpim->fcp		= fcp;
	fcpim->bfa		= bfa;
	fcpim->num_itnims	= cfg->fwcfg.num_rports;
	fcpim->num_tskim_reqs = cfg->fwcfg.num_tskim_reqs;
	fcpim->path_tov		= cfg->drvcfg.path_tov;
	fcpim->delay_comp	= cfg->drvcfg.delay_comp;
	fcpim->profile_comp = NULL;
	fcpim->profile_start = NULL;

	if (!cfg->drvcfg.min_cfg)
		bfa_iopf_attach(fcpim);
	bfa_itnim_attach(fcpim);
	bfa_ioim_attach(fcpim);
	bfa_tskim_attach(fcpim);
}

void
bfa_fcpim_detach(struct bfa_fcp_mod_s *fcp)
{
	struct bfa_fcpim_s *fcpim = &fcp->fcpim;

	bfa_ioim_detach(fcpim);
	bfa_tskim_detach(fcpim);
}

void
bfa_fcpim_start(struct bfa_fcp_mod_s *fcp)
{
	/*
	 bfa_init() with flash read is complete. now invalidate the stale
	 content of lun mask like unit attention, rp tag and lp tag.
	 */
	bfa_ioim_lm_init(fcp->bfa);
}

void
bfa_fcpim_stop(struct bfa_fcp_mod_s *fcp)
{
}

void
bfa_fcpim_iocdisable(struct bfa_fcp_mod_s *fcp)
{
	struct bfa_fcpim_s *fcpim = &fcp->fcpim;
	struct bfa_itnim_s *itnim;
	struct bfa_q_s *qe, *qen;

	/* Enqueue unused ioim resources to free_q */
	bfa_q_enq_q(&fcpim->tskim_free_q, &fcpim->tskim_unused_q);

	bfa_q_iter_safe(&fcpim->itnim_q, qe, qen) {
		itnim = (struct bfa_itnim_s *) qe;
		bfa_itnim_iocdisable(itnim);
	}
}

void
bfa_fcpim_add_stats(struct bfa_itnim_iostats_s *lstats,
		struct bfa_itnim_iostats_s *rstats)
{
	bfa_fcpim_add_iostats(lstats, rstats, total_ios);
	bfa_fcpim_add_iostats(lstats, rstats, qresumes);
	bfa_fcpim_add_iostats(lstats, rstats, no_iotags);
	bfa_fcpim_add_iostats(lstats, rstats, io_aborts);
	bfa_fcpim_add_iostats(lstats, rstats, no_tskims);
	bfa_fcpim_add_iostats(lstats, rstats, iocomp_ok);
	bfa_fcpim_add_iostats(lstats, rstats, iocomp_underrun);
	bfa_fcpim_add_iostats(lstats, rstats, iocomp_overrun);
	bfa_fcpim_add_iostats(lstats, rstats, iocomp_aborted);
	bfa_fcpim_add_iostats(lstats, rstats, iocomp_timedout);
	bfa_fcpim_add_iostats(lstats, rstats, iocom_nexus_abort);
	bfa_fcpim_add_iostats(lstats, rstats, iocom_proto_err);
	bfa_fcpim_add_iostats(lstats, rstats, iocom_dif_err);
	bfa_fcpim_add_iostats(lstats, rstats, iocom_sqer_needed);
	bfa_fcpim_add_iostats(lstats, rstats, iocom_res_free);
	bfa_fcpim_add_iostats(lstats, rstats, iocom_hostabrts);
	bfa_fcpim_add_iostats(lstats, rstats, iocom_utags);
	bfa_fcpim_add_iostats(lstats, rstats, io_cleanups);
	bfa_fcpim_add_iostats(lstats, rstats, io_tmaborts);
	bfa_fcpim_add_iostats(lstats, rstats, onlines);
	bfa_fcpim_add_iostats(lstats, rstats, offlines);
	bfa_fcpim_add_iostats(lstats, rstats, creates);
	bfa_fcpim_add_iostats(lstats, rstats, deletes);
	bfa_fcpim_add_iostats(lstats, rstats, create_comps);
	bfa_fcpim_add_iostats(lstats, rstats, delete_comps);
	bfa_fcpim_add_iostats(lstats, rstats, sler_events);
	bfa_fcpim_add_iostats(lstats, rstats, fw_create);
	bfa_fcpim_add_iostats(lstats, rstats, fw_delete);
	bfa_fcpim_add_iostats(lstats, rstats, ioc_disabled);
	bfa_fcpim_add_iostats(lstats, rstats, cleanup_comps);
	bfa_fcpim_add_iostats(lstats, rstats, tm_cmnds);
	bfa_fcpim_add_iostats(lstats, rstats, tm_fw_rsps);
	bfa_fcpim_add_iostats(lstats, rstats, tm_success);
	bfa_fcpim_add_iostats(lstats, rstats, tm_failures);
	bfa_fcpim_add_iostats(lstats, rstats, tm_io_comps);
	bfa_fcpim_add_iostats(lstats, rstats, tm_qresumes);
	bfa_fcpim_add_iostats(lstats, rstats, tm_iocdowns);
	bfa_fcpim_add_iostats(lstats, rstats, tm_cleanups);
	bfa_fcpim_add_iostats(lstats, rstats, tm_cleanup_comps);
	bfa_fcpim_add_iostats(lstats, rstats, io_comps);
	bfa_fcpim_add_iostats(lstats, rstats, input_reqs);
	bfa_fcpim_add_iostats(lstats, rstats, output_reqs);
	bfa_fcpim_add_iostats(lstats, rstats, rd_throughput);
	bfa_fcpim_add_iostats(lstats, rstats, wr_throughput);
	bfa_fcpim_add_iostats(lstats, rstats, lm_lun_across_sg);
	bfa_fcpim_add_iostats(lstats, rstats, lm_lun_not_sup);
	bfa_fcpim_add_iostats(lstats, rstats, lm_rpl_data_changed);
	bfa_fcpim_add_iostats(lstats, rstats, lm_wire_residue_changed);
	bfa_fcpim_add_iostats(lstats, rstats, lm_small_buf_addresidue);
	bfa_fcpim_add_iostats(lstats, rstats, lm_lun_not_rdy);
}

void
bfa_fcpim_path_tov_set(struct bfa_s *bfa, uint16_t path_tov)
{
	struct bfa_fcpim_s *fcpim = BFA_FCPIM(bfa);

	fcpim->path_tov = path_tov * 1000;
	if (fcpim->path_tov > BFA_FCPIM_PATHTOV_MAX)
		fcpim->path_tov = BFA_FCPIM_PATHTOV_MAX;
}

uint16_t
bfa_fcpim_path_tov_get(struct bfa_s *bfa)
{
	struct bfa_fcpim_s *fcpim = BFA_FCPIM(bfa);

	return (fcpim->path_tov / 1000);
}

bfa_status_t
bfa_fcpim_port_iostats(struct bfa_s *bfa, struct bfa_itnim_iostats_s *stats,
	uint8_t lp_tag)
{
	struct bfa_fcpim_s *fcpim = BFA_FCPIM(bfa);
	struct bfa_q_s *qe, *qen;
	struct bfa_itnim_s *itnim;

	/* accumulate IO stats from itnim */
	bfa_os_memset(stats, 0, sizeof(struct bfa_itnim_iostats_s));
	bfa_q_iter_safe(&fcpim->itnim_q, qe, qen) {
		itnim = (struct bfa_itnim_s *) qe;
		if (itnim->rport->rport_info.lp_tag != lp_tag)
			continue;
		bfa_fcpim_add_stats(stats, &(itnim->stats));
	}
	return BFA_STATUS_OK;
}
bfa_status_t
bfa_fcpim_get_modstats(struct bfa_s *bfa, struct bfa_itnim_iostats_s *modstats)
{
	struct bfa_fcpim_s *fcpim = BFA_FCPIM(bfa);
	struct bfa_q_s *qe, *qen;
	struct bfa_itnim_s *itnim;

	/* accumulate IO stats from itnim */
	bfa_os_memset(modstats, 0, sizeof(struct bfa_itnim_iostats_s));
	bfa_q_iter_safe(&fcpim->itnim_q, qe, qen) {
		itnim = (struct bfa_itnim_s *) qe;
		bfa_fcpim_add_stats(modstats, &(itnim->stats));
	}
	return BFA_STATUS_OK;
}

bfa_status_t
bfa_fcpim_get_del_itn_stats(struct bfa_s *bfa,
	 struct bfa_fcpim_del_itn_stats_s *modstats)
{
	struct bfa_fcpim_s *fcpim = BFA_FCPIM(bfa);

	*modstats = fcpim->del_itn_stats;

	return BFA_STATUS_OK;
}


bfa_status_t
bfa_fcpim_port_clear_iostats(struct bfa_s *bfa, uint8_t lp_tag)
{
	struct bfa_fcpim_s *fcpim = BFA_FCPIM(bfa);
	struct bfa_q_s *qe, *qen;
	struct bfa_itnim_s *itnim;

	/* clear IO stats from all active itnims */
	bfa_q_iter_safe(&fcpim->itnim_q, qe, qen) {
		itnim = (struct bfa_itnim_s *) qe;
		if (itnim->rport->rport_info.lp_tag != lp_tag)
			continue;
		bfa_itnim_clear_stats(itnim);
	}
	return BFA_STATUS_OK;

}

bfa_status_t
bfa_fcpim_clr_modstats(struct bfa_s *bfa)
{
	struct bfa_fcpim_s *fcpim = BFA_FCPIM(bfa);
	struct bfa_q_s *qe, *qen;
	struct bfa_itnim_s *itnim;

	/* clear IO stats from all active itnims */
	bfa_q_iter_safe(&fcpim->itnim_q, qe, qen) {
		itnim = (struct bfa_itnim_s *) qe;
		bfa_itnim_clear_stats(itnim);
	}
	bfa_os_memset(&fcpim->del_itn_stats, 0,
		sizeof(struct bfa_fcpim_del_itn_stats_s));

	return BFA_STATUS_OK;
}

void
bfa_fcpim_qdepth_set(struct bfa_s *bfa, uint16_t q_depth)
{
	struct bfa_fcpim_s *fcpim = BFA_FCPIM(bfa);

	bfa_assert(q_depth <= BFA_IOCFC_QDEPTH_MAX);

	fcpim->q_depth = q_depth;
}

uint16_t
bfa_fcpim_qdepth_get(struct bfa_s *bfa)
{
	struct bfa_fcpim_s *fcpim = BFA_FCPIM(bfa);

	return (fcpim->q_depth);
}

void
bfa_fcpim_update_ioredirect(struct bfa_s *bfa)
{
	bfa_boolean_t ioredirect;

	/*
	 * IO redirection is turned off when QoS is enabled and vice versa
	 */
	ioredirect = bfa_fcport_is_qos_enabled(bfa) ? BFA_FALSE : BFA_TRUE;

	/*
	 * Notify the bfad module of a possible state change in
	 * IO redirection capability, due to a QoS state change. bfad will
	 * check on the support for io redirection and update the
	 * fcpim's ioredirect state accordingly.
	 */
	bfa_cb_ioredirect_state_change((void *)(bfa->bfad), ioredirect);

}

void
bfa_fcpim_set_ioredirect(struct bfa_s *bfa, bfa_boolean_t state)
{
	struct bfa_fcpim_s *fcpim = BFA_FCPIM(bfa);
	fcpim->ioredirect = state;
}

bfa_itnim_t *
bfa_fcpim_get_itnim_by_wwn(struct bfa_fcpim_s *fcpim, wwn_t lpwwn, wwn_t rpwwn)
{
	bfa_itnim_t *itnim = NULL;
	wwn_t itnim_lpwwn = 0, itnim_rpwwn = 0;
	bfa_q_t *qe, *qen;

	bfa_q_iter_safe(&fcpim->itnim_q, qe, qen) {
		itnim = (bfa_itnim_t *)qe;
		itnim_lpwwn = BFA_ITNIM_GET_LPWWN(itnim);
		itnim_rpwwn = bfa_rport_get_pwwn(itnim->bfa, itnim->rport);
		if ((itnim_lpwwn == lpwwn) && (itnim_rpwwn == rpwwn))
			return itnim;
	}

	return NULL;
}

bfa_status_t
bfa_fcpim_trs_enable(struct bfa_s *bfa, wwn_t lpwwn, wwn_t rpwwn)
{
	bfa_fcpim_t *fcpim = BFA_FCPIM(bfa);
	bfa_itnim_t *itnim = NULL;

	itnim = bfa_fcpim_get_itnim_by_wwn(fcpim, lpwwn, rpwwn);
	if (!itnim)
		return BFA_STATUS_NO_FCPIM_NEXUS;

	/* Check if TRS is already enabled */
	if (!itnim->trs_blocked)
		return BFA_STATUS_TRS_ENABLED;

	/* Dequeue the TRS entry, free the memory & decrement the TRS count */
	bfa_cb_trs_dequeue_entry(bfa->bfad, lpwwn, rpwwn);

	/* Disable the `trs_blocked` flag in bfa itnim structure */
	itnim->trs_blocked = BFA_FALSE;
	return BFA_STATUS_OK;
}

bfa_status_t
bfa_fcpim_trs_disable(struct bfa_s *bfa, wwn_t lpwwn, wwn_t rpwwn)
{
	bfa_fcpim_t *fcpim = BFA_FCPIM(bfa);
	bfa_itnim_t *itnim = NULL;
	bfa_status_t status;

	itnim = bfa_fcpim_get_itnim_by_wwn(fcpim, lpwwn, rpwwn);
	if (!itnim)
		return BFA_STATUS_NO_FCPIM_NEXUS;

	/* Check if TRS is already disabled */
	if (itnim->trs_blocked)
		return BFA_STATUS_TRS_DISABLED;

	/* Alloc mem for TRS entry, queue it & increment the TRS count */
	status = bfa_cb_trs_enqueue_entry(bfa->bfad, lpwwn, rpwwn);
	if (status != BFA_STATUS_OK)
		return status;

	/* Enable the `trs_blocked` flag in bfa itnim structure */
	itnim->trs_blocked = BFA_TRUE;
	return BFA_STATUS_OK;
}

bfa_status_t
bfa_fcpim_trs_clear(struct bfa_s *bfa)
{
	bfa_trs_cfg_t *trs_cfg = bfa_cb_trs_get_cfg(bfa->bfad);
	bfa_fcpim_t *fcpim = BFA_FCPIM(bfa);
	bfa_itnim_t *itnim = NULL;
	bfa_trs_mask_t *trs_mask = NULL;
	bfa_q_t *qe, *qen;

	/* Disable the `trs_blocked` flag in bfa_itnim_t structure */
	bfa_q_iter_safe(&trs_cfg->trs_q, qe, qen) {
		trs_mask = (bfa_trs_mask_t *)qe;
		itnim = bfa_fcpim_get_itnim_by_wwn(fcpim,
			trs_mask->trs_list.lpwwn, trs_mask->trs_list.rpwwn);
		if (!itnim)
			continue;

		itnim->trs_blocked = BFA_FALSE;
	}

	/* Clear all entries in the TRS list & reset TRS count to zero */
	bfa_cb_trs_clear_list(bfa->bfad);
	return BFA_STATUS_OK;
}

static void
bfa_itnim_update_trs_state(struct bfa_itnim_s *itnim)
{
	bfa_trs_cfg_t *trs_cfg = bfa_cb_trs_get_cfg(itnim->bfa->bfad);
	wwn_t lpwwn = 0, rpwwn = 0;
	bfa_trs_mask_t *trs_mask = NULL;
	bfa_q_t *qe, *qen;

	/* TRS config will be NULL in EFI environment */
	if (!trs_cfg)
		return;

	lpwwn = BFA_ITNIM_GET_LPWWN(itnim);
	rpwwn = bfa_rport_get_pwwn(itnim->bfa, itnim->rport);
	bfa_q_iter_safe(&trs_cfg->trs_q, qe, qen) {
		trs_mask = (bfa_trs_mask_t *)qe;
		if (trs_mask->trs_list.lpwwn == lpwwn &&
				trs_mask->trs_list.rpwwn == rpwwn) {
			itnim->trs_blocked = BFA_TRUE;
			return;
		}
	}
}

/**
 *  BFA ITNIM module state machine functions
 */

/**
 * @brief
 * 		Beginning/unallocated state - no events expected.
 */
static void
bfa_itnim_sm_uninit(struct bfa_itnim_s *itnim, enum bfa_itnim_event event)
{
	bfa_trc(itnim->bfa, itnim->rport->rport_tag);
	bfa_trc(itnim->bfa, event);

	switch (event) {
	case BFA_ITNIM_SM_CREATE:
		bfa_sm_set_state(itnim, bfa_itnim_sm_created);
		itnim->is_online = BFA_FALSE;
		itnim->trs_blocked = BFA_FALSE;
		bfa_fcpim_additn(itnim);
		break;

	default:
		bfa_sm_fault(itnim->bfa, event);
	}
}

/**
 * @brief
 * 		Beginning state, only online event expected.
 */
static void
bfa_itnim_sm_created(struct bfa_itnim_s *itnim, enum bfa_itnim_event event)
{
	bfa_trc(itnim->bfa, itnim->rport->rport_tag);
	bfa_trc(itnim->bfa, event);

	switch (event) {
	case BFA_ITNIM_SM_ONLINE:
		if (bfa_itnim_send_fwcreate(itnim))
			bfa_sm_set_state(itnim, bfa_itnim_sm_fwcreate);
		else
			bfa_sm_set_state(itnim, bfa_itnim_sm_fwcreate_qfull);
		break;

	case BFA_ITNIM_SM_DELETE:
		bfa_sm_set_state(itnim, bfa_itnim_sm_uninit);
		bfa_fcpim_delitn(itnim);
		break;

	case BFA_ITNIM_SM_HWFAIL:
		bfa_sm_set_state(itnim, bfa_itnim_sm_iocdisable);
		break;

	default:
		bfa_sm_fault(itnim->bfa, event);
	}
}

/**
 * @brief
 * 		Waiting for itnim create response from firmware.
 */
static void
bfa_itnim_sm_fwcreate(struct bfa_itnim_s *itnim, enum bfa_itnim_event event)
{
	bfa_trc(itnim->bfa, itnim->rport->rport_tag);
	bfa_trc(itnim->bfa, event);

	switch (event) {
	case BFA_ITNIM_SM_FWRSP:
		bfa_sm_set_state(itnim, bfa_itnim_sm_online);
		itnim->is_online = BFA_TRUE;
		bfa_itnim_update_trs_state(itnim);
		bfa_itnim_iotov_online(itnim);
		bfa_itnim_online_cb(itnim);
		break;

	case BFA_ITNIM_SM_DELETE:
		bfa_sm_set_state(itnim, bfa_itnim_sm_delete_pending);
		break;

	case BFA_ITNIM_SM_OFFLINE:
		if (bfa_itnim_send_fwdelete(itnim))
			bfa_sm_set_state(itnim, bfa_itnim_sm_fwdelete);
		else
			bfa_sm_set_state(itnim, bfa_itnim_sm_fwdelete_qfull);
		break;

	case BFA_ITNIM_SM_HWFAIL:
		bfa_sm_set_state(itnim, bfa_itnim_sm_iocdisable);
		break;

	default:
		bfa_sm_fault(itnim->bfa, event);
	}
}

static void
bfa_itnim_sm_fwcreate_qfull(bfa_itnim_t *itnim, enum bfa_itnim_event event)
{
	bfa_trc(itnim->bfa, itnim->rport->rport_tag);
	bfa_trc(itnim->bfa, event);

	switch (event) {
	case BFA_ITNIM_SM_QRESUME:
		bfa_sm_set_state(itnim, bfa_itnim_sm_fwcreate);
		bfa_itnim_send_fwcreate(itnim);
		break;

	case BFA_ITNIM_SM_DELETE:
		bfa_sm_set_state(itnim, bfa_itnim_sm_uninit);
		bfa_reqq_wcancel(&itnim->reqq_wait);
		bfa_fcpim_delitn(itnim);
		break;

	case BFA_ITNIM_SM_OFFLINE:
		bfa_sm_set_state(itnim, bfa_itnim_sm_offline);
		bfa_reqq_wcancel(&itnim->reqq_wait);
		bfa_itnim_offline_cb(itnim);
		break;

	case BFA_ITNIM_SM_HWFAIL:
		bfa_sm_set_state(itnim, bfa_itnim_sm_iocdisable);
		bfa_reqq_wcancel(&itnim->reqq_wait);
		break;

	default:
		bfa_sm_fault(itnim->bfa, event);
	}
}

/**
 * @brief
 * 	Waiting for itnim create response from firmware, a delete is pending.
 */
static void
bfa_itnim_sm_delete_pending(struct bfa_itnim_s *itnim,
				enum bfa_itnim_event event)
{
	bfa_trc(itnim->bfa, itnim->rport->rport_tag);
	bfa_trc(itnim->bfa, event);

	switch (event) {
	case BFA_ITNIM_SM_FWRSP:
		if (bfa_itnim_send_fwdelete(itnim))
			bfa_sm_set_state(itnim, bfa_itnim_sm_deleting);
		else
			bfa_sm_set_state(itnim, bfa_itnim_sm_deleting_qfull);
		break;

	case BFA_ITNIM_SM_HWFAIL:
		bfa_sm_set_state(itnim, bfa_itnim_sm_uninit);
		bfa_fcpim_delitn(itnim);
		break;

	default:
		bfa_sm_fault(itnim->bfa, event);
	}
}

/**
 * @brief
 * 		Online state - normal parking state.
 */
static void
bfa_itnim_sm_online(struct bfa_itnim_s *itnim, enum bfa_itnim_event event)
{
	bfa_trc(itnim->bfa, itnim->rport->rport_tag);
	bfa_trc(itnim->bfa, event);

	switch (event) {
	case BFA_ITNIM_SM_OFFLINE:
		bfa_sm_set_state(itnim, bfa_itnim_sm_cleanup_offline);
		itnim->is_online = BFA_FALSE;
		bfa_itnim_iotov_start(itnim);
		bfa_itnim_cleanup(itnim);
		break;

	case BFA_ITNIM_SM_DELETE:
		bfa_sm_set_state(itnim, bfa_itnim_sm_cleanup_delete);
		itnim->is_online = BFA_FALSE;
		bfa_itnim_cleanup(itnim);
		break;

	case BFA_ITNIM_SM_SLER:
		bfa_sm_set_state(itnim, bfa_itnim_sm_sler);
		itnim->is_online = BFA_FALSE;
		bfa_itnim_iotov_start(itnim);
		bfa_itnim_sler_cb(itnim);
		break;

	case BFA_ITNIM_SM_HWFAIL:
		bfa_sm_set_state(itnim, bfa_itnim_sm_iocdisable);
		itnim->is_online = BFA_FALSE;
		bfa_itnim_iotov_start(itnim);
		bfa_itnim_iocdisable_cleanup(itnim);
		break;

	default:
		bfa_sm_fault(itnim->bfa, event);
	}
}

/**
 * @brief
 * 		Second level error recovery need.
 */
static void
bfa_itnim_sm_sler(struct bfa_itnim_s *itnim, enum bfa_itnim_event event)
{
	bfa_trc(itnim->bfa, itnim->rport->rport_tag);
	bfa_trc(itnim->bfa, event);

	switch (event) {
	case BFA_ITNIM_SM_OFFLINE:
		bfa_sm_set_state(itnim, bfa_itnim_sm_cleanup_offline);
		bfa_itnim_cleanup(itnim);
		break;

	case BFA_ITNIM_SM_DELETE:
		bfa_sm_set_state(itnim, bfa_itnim_sm_cleanup_delete);
		bfa_itnim_cleanup(itnim);
		bfa_itnim_iotov_delete(itnim);
		break;

	case BFA_ITNIM_SM_HWFAIL:
		bfa_sm_set_state(itnim, bfa_itnim_sm_iocdisable);
		bfa_itnim_iocdisable_cleanup(itnim);
		break;

	default:
		bfa_sm_fault(itnim->bfa, event);
	}
}

/**
 * @brief
 * 		Going offline. Waiting for active IO cleanup.
 */
static void
bfa_itnim_sm_cleanup_offline(struct bfa_itnim_s *itnim,
				 enum bfa_itnim_event event)
{
	bfa_trc(itnim->bfa, itnim->rport->rport_tag);
	bfa_trc(itnim->bfa, event);

	switch (event) {
	case BFA_ITNIM_SM_CLEANUP:
		if (bfa_itnim_send_fwdelete(itnim))
			bfa_sm_set_state(itnim, bfa_itnim_sm_fwdelete);
		else
			bfa_sm_set_state(itnim, bfa_itnim_sm_fwdelete_qfull);
		break;

	case BFA_ITNIM_SM_DELETE:
		bfa_sm_set_state(itnim, bfa_itnim_sm_cleanup_delete);
		bfa_itnim_iotov_delete(itnim);
		break;

	case BFA_ITNIM_SM_HWFAIL:
		bfa_sm_set_state(itnim, bfa_itnim_sm_iocdisable);
		bfa_itnim_iocdisable_cleanup(itnim);
		bfa_itnim_offline_cb(itnim);
		break;

	case BFA_ITNIM_SM_SLER:
		break;

	default:
		bfa_sm_fault(itnim->bfa, event);
	}
}

/**
 * @brief
 * 		Deleting itnim. Waiting for active IO cleanup.
 */
static void
bfa_itnim_sm_cleanup_delete(struct bfa_itnim_s *itnim,
				enum bfa_itnim_event event)
{
	bfa_trc(itnim->bfa, itnim->rport->rport_tag);
	bfa_trc(itnim->bfa, event);

	switch (event) {
	case BFA_ITNIM_SM_CLEANUP:
		if (bfa_itnim_send_fwdelete(itnim))
			bfa_sm_set_state(itnim, bfa_itnim_sm_deleting);
		else
			bfa_sm_set_state(itnim, bfa_itnim_sm_deleting_qfull);
		break;

	case BFA_ITNIM_SM_HWFAIL:
		bfa_sm_set_state(itnim, bfa_itnim_sm_iocdisable);
		bfa_itnim_iocdisable_cleanup(itnim);
		break;

	default:
		bfa_sm_fault(itnim->bfa, event);
	}
}

/**
 * @brief
 * Rport offline. Fimrware itnim is being deleted - awaiting f/w response.
 */
static void
bfa_itnim_sm_fwdelete(struct bfa_itnim_s *itnim, enum bfa_itnim_event event)
{
	bfa_trc(itnim->bfa, itnim->rport->rport_tag);
	bfa_trc(itnim->bfa, event);

	switch (event) {
	case BFA_ITNIM_SM_FWRSP:
		bfa_sm_set_state(itnim, bfa_itnim_sm_offline);
		bfa_itnim_offline_cb(itnim);
		break;

	case BFA_ITNIM_SM_DELETE:
		bfa_sm_set_state(itnim, bfa_itnim_sm_deleting);
		break;

	case BFA_ITNIM_SM_HWFAIL:
		bfa_sm_set_state(itnim, bfa_itnim_sm_iocdisable);
		bfa_itnim_offline_cb(itnim);
		break;

	default:
		bfa_sm_fault(itnim->bfa, event);
	}
}

static void
bfa_itnim_sm_fwdelete_qfull(bfa_itnim_t *itnim, enum bfa_itnim_event event)
{
	bfa_trc(itnim->bfa, itnim->rport->rport_tag);
	bfa_trc(itnim->bfa, event);

	switch (event) {
	case BFA_ITNIM_SM_QRESUME:
		bfa_sm_set_state(itnim, bfa_itnim_sm_fwdelete);
		bfa_itnim_send_fwdelete(itnim);
		break;

	case BFA_ITNIM_SM_DELETE:
		bfa_sm_set_state(itnim, bfa_itnim_sm_deleting_qfull);
		break;

	case BFA_ITNIM_SM_HWFAIL:
		bfa_sm_set_state(itnim, bfa_itnim_sm_iocdisable);
		bfa_reqq_wcancel(&itnim->reqq_wait);
		bfa_itnim_offline_cb(itnim);
		break;

	default:
		bfa_sm_fault(itnim->bfa, event);
	}
}

/**
 * @brief
 * 		Offline state.
 */
static void
bfa_itnim_sm_offline(struct bfa_itnim_s *itnim, enum bfa_itnim_event event)
{
	bfa_trc(itnim->bfa, itnim->rport->rport_tag);
	bfa_trc(itnim->bfa, event);

	switch (event) {
	case BFA_ITNIM_SM_DELETE:
		bfa_sm_set_state(itnim, bfa_itnim_sm_uninit);
		bfa_itnim_iotov_delete(itnim);
		bfa_fcpim_delitn(itnim);
		break;

	case BFA_ITNIM_SM_ONLINE:
		if (bfa_itnim_send_fwcreate(itnim))
			bfa_sm_set_state(itnim, bfa_itnim_sm_fwcreate);
		else
			bfa_sm_set_state(itnim, bfa_itnim_sm_fwcreate_qfull);
		break;

	case BFA_ITNIM_SM_HWFAIL:
		bfa_sm_set_state(itnim, bfa_itnim_sm_iocdisable);
		break;

	default:
		bfa_sm_fault(itnim->bfa, event);
	}
}

/**
 * @brief
 * 		IOC h/w failed state.
 */
static void
bfa_itnim_sm_iocdisable(struct bfa_itnim_s *itnim,
				enum bfa_itnim_event event)
{
	bfa_trc(itnim->bfa, itnim->rport->rport_tag);
	bfa_trc(itnim->bfa, event);

	switch (event) {
	case BFA_ITNIM_SM_DELETE:
		bfa_sm_set_state(itnim, bfa_itnim_sm_uninit);
		bfa_itnim_iotov_delete(itnim);
		bfa_fcpim_delitn(itnim);
		break;

	case BFA_ITNIM_SM_OFFLINE:
		bfa_itnim_offline_cb(itnim);
		break;

	case BFA_ITNIM_SM_ONLINE:
		if (bfa_itnim_send_fwcreate(itnim))
			bfa_sm_set_state(itnim, bfa_itnim_sm_fwcreate);
		else
			bfa_sm_set_state(itnim, bfa_itnim_sm_fwcreate_qfull);
		break;

	case BFA_ITNIM_SM_HWFAIL:
		break;

	default:
		bfa_sm_fault(itnim->bfa, event);
	}
}

/**
 * @brief
 * 		Itnim is deleted, waiting for firmware response to delete.
 */
static void
bfa_itnim_sm_deleting(struct bfa_itnim_s *itnim, enum bfa_itnim_event event)
{
	bfa_trc(itnim->bfa, itnim->rport->rport_tag);
	bfa_trc(itnim->bfa, event);

	switch (event) {
	case BFA_ITNIM_SM_FWRSP:
	case BFA_ITNIM_SM_HWFAIL:
		bfa_sm_set_state(itnim, bfa_itnim_sm_uninit);
		bfa_fcpim_delitn(itnim);
		break;

	default:
		bfa_sm_fault(itnim->bfa, event);
	}
}

static void
bfa_itnim_sm_deleting_qfull(bfa_itnim_t *itnim, enum bfa_itnim_event event)
{
	bfa_trc(itnim->bfa, itnim->rport->rport_tag);
	bfa_trc(itnim->bfa, event);

	switch (event) {
	case BFA_ITNIM_SM_QRESUME:
		bfa_sm_set_state(itnim, bfa_itnim_sm_deleting);
		bfa_itnim_send_fwdelete(itnim);
		break;

	case BFA_ITNIM_SM_HWFAIL:
		bfa_sm_set_state(itnim, bfa_itnim_sm_uninit);
		bfa_reqq_wcancel(&itnim->reqq_wait);
		bfa_fcpim_delitn(itnim);
		break;

	default:
		bfa_sm_fault(itnim->bfa, event);
	}
}

/**
 * @brief called from config lun profile start
 * @return NULL when no iopf if found
 */
static bfa_fcpim_iopf_t*
bfa_fcpim_iopf_alloc(struct bfa_fcpim_s *fcpim)
{
	bfa_fcpim_iopf_qe_t *iopf_qe;

	bfa_q_deq(&fcpim->iopf_free_q, &iopf_qe);
	if (iopf_qe) {
		bfa_q_enq(&fcpim->iopf_active_q, iopf_qe);
		bfa_trc(fcpim->bfa, iopf_qe->iopf.id);
		return &iopf_qe->iopf;
	}
	return NULL;
}

/**
 * @brief
 * 		Initiate cleanup of all IOs on an IOC failure.
 */
static void
bfa_itnim_iocdisable_cleanup(struct bfa_itnim_s *itnim)
{
	struct bfa_tskim_s *tskim;
	struct bfa_ioim_s *ioim;
	struct bfa_q_s	*qe, *qen;

	bfa_q_iter_safe(&itnim->tsk_q, qe, qen) {
		tskim = (struct bfa_tskim_s *) qe;
		bfa_tskim_iocdisable(tskim);
	}

	bfa_q_iter_safe(&itnim->io_q, qe, qen) {
		ioim = (struct bfa_ioim_s *) qe;
		bfa_ioim_iocdisable(ioim);
	}

	/**
	 * For IO request in pending queue, we pretend an early timeout.
	 */
	bfa_q_iter_safe(&itnim->pending_q, qe, qen) {
		ioim = (struct bfa_ioim_s *) qe;
		bfa_ioim_tov(ioim);
	}

	bfa_q_iter_safe(&itnim->io_cleanup_q, qe, qen) {
		ioim = (struct bfa_ioim_s *) qe;
		bfa_ioim_iocdisable(ioim);
	}
}

/**
 * @brief
 * 		IO cleanup completion
 */
static void
bfa_itnim_cleanp_comp(void *itnim_cbarg)
{
	struct bfa_itnim_s *itnim = itnim_cbarg;

	bfa_stats(itnim, cleanup_comps);
	bfa_sm_send_event(itnim, BFA_ITNIM_SM_CLEANUP);
}

/**
 * @brief
 * 		Initiate cleanup of all IOs.
 */
static void
bfa_itnim_cleanup(struct bfa_itnim_s *itnim)
{
	struct bfa_ioim_s  *ioim;
	struct bfa_tskim_s *tskim;
	struct bfa_q_s	*qe, *qen;

	bfa_wc_init(&itnim->wc, bfa_itnim_cleanp_comp, itnim);

	bfa_q_iter_safe(&itnim->io_q, qe, qen) {
		ioim = (struct bfa_ioim_s *) qe;

		/**
		 * Move IO to a cleanup queue from active queue so that a later
		 * TM will not pickup this IO.
		 */
		bfa_q_qe_deq(ioim);
		bfa_q_enq(&itnim->io_cleanup_q, ioim);

		bfa_wc_up(&itnim->wc);
		bfa_ioim_cleanup(ioim);
	}

	bfa_q_iter_safe(&itnim->tsk_q, qe, qen) {
		tskim = (struct bfa_tskim_s *) qe;
		bfa_wc_up(&itnim->wc);
		bfa_tskim_cleanup(tskim);
	}

	bfa_wc_wait(&itnim->wc);
}

static void
__bfa_cb_itnim_online(void *cbarg, bfa_boolean_t complete)
{
	struct bfa_itnim_s *itnim = cbarg;

	if (complete)
		bfa_cb_itnim_online(itnim->ditn);
}

static void
__bfa_cb_itnim_offline(void *cbarg, bfa_boolean_t complete)
{
	struct bfa_itnim_s *itnim = cbarg;

	if (complete)
		bfa_cb_itnim_offline(itnim->ditn);
}

static void
__bfa_cb_itnim_sler(void *cbarg, bfa_boolean_t complete)
{
	struct bfa_itnim_s *itnim = cbarg;

	if (complete)
		bfa_cb_itnim_sler(itnim->ditn);
}

/**
 * @brief
 * Call to resume any I/O requests waiting for room in request queue.
 */
static void
bfa_itnim_qresume(void *cbarg)
{
	struct bfa_itnim_s *itnim = cbarg;

	bfa_sm_send_event(itnim, BFA_ITNIM_SM_QRESUME);
}


/**
 * @}
 */

/**
 * @dg bfa_itnim_public
 * @{
 */

void
bfa_itnim_iodone(struct bfa_itnim_s *itnim)
{
	bfa_wc_down(&itnim->wc);
}

void
bfa_itnim_tskdone(struct bfa_itnim_s *itnim)
{
	bfa_wc_down(&itnim->wc);
}

void
bfa_itnim_meminfo(struct bfa_iocfc_cfg_s *cfg, uint32_t *km_len)
{
	/**
	 * ITN memory
	 */
	*km_len += cfg->fwcfg.num_rports * sizeof(struct bfa_itnim_s);
}

void
bfa_itnim_attach(struct bfa_fcpim_s *fcpim)
{
	struct bfa_s	*bfa = fcpim->bfa;
	struct bfa_fcp_mod_s	*fcp = fcpim->fcp;
	struct bfa_itnim_s *itnim;
	int	i, j;

	bfa_q_init(&fcpim->itnim_q);

	itnim = (struct bfa_itnim_s *) bfa_mem_kva_curp(fcp);
	fcpim->itnim_arr = itnim;

	for (i = 0; i < fcpim->num_itnims; i++, itnim++) {
		bfa_os_memset(itnim, 0, sizeof(struct bfa_itnim_s));
		itnim->bfa = bfa;
		itnim->fcpim = fcpim;
		itnim->reqq = BFA_REQQ_QOS_LO;
		itnim->rport = BFA_RPORT_FROM_TAG(bfa, i);
		itnim->iotov_active = BFA_FALSE;
		bfa_reqq_winit(&itnim->reqq_wait, bfa_itnim_qresume, itnim);

		bfa_q_init(&itnim->io_q);
		bfa_q_init(&itnim->io_cleanup_q);
		bfa_q_init(&itnim->pending_q);
		bfa_q_init(&itnim->tsk_q);
		bfa_q_init(&itnim->delay_comp_q);
		for (j = 0; j < BFA_IOBUCKET_MAX; j++)
			itnim->iolat.min[j] = ~0;
		bfa_sm_set_state(itnim, bfa_itnim_sm_uninit);
	}

	bfa_mem_kva_curp(fcp) = (uint8_t *) itnim;
}

void
bfa_itnim_iocdisable(struct bfa_itnim_s *itnim)
{
	bfa_stats(itnim, ioc_disabled);
	bfa_sm_send_event(itnim, BFA_ITNIM_SM_HWFAIL);
}

static bfa_boolean_t
bfa_itnim_send_fwcreate(struct bfa_itnim_s *itnim)
{
	struct bfi_itn_create_req_s *m;

	itnim->msg_no++;

	/**
	 * check for room in queue to send request now
	 */
	m = bfa_reqq_next(itnim->bfa, itnim->reqq);
	if (!m) {
		bfa_reqq_wait(itnim->bfa, itnim->reqq, &itnim->reqq_wait);
		return BFA_FALSE;
	}

	bfi_h2i_set(m->mh, BFI_MC_ITN, BFI_ITN_H2I_CREATE_REQ,
			bfa_fn_lpu(itnim->bfa));
	m->fw_handle = itnim->rport->fw_handle;
	m->class = FC_CLASS_3;
	m->seq_rec = itnim->seq_rec;
	m->msg_no = itnim->msg_no;
	bfa_stats(itnim, fw_create);

	/**
	 * queue I/O message to firmware
	 */
	bfa_reqq_produce(itnim->bfa, itnim->reqq, m->mh);
	return BFA_TRUE;
}

static bfa_boolean_t
bfa_itnim_send_fwdelete(struct bfa_itnim_s *itnim)
{
	struct bfi_itn_delete_req_s *m;

	/**
	 * check for room in queue to send request now
	 */
	m = bfa_reqq_next(itnim->bfa, itnim->reqq);
	if (!m) {
		bfa_reqq_wait(itnim->bfa, itnim->reqq, &itnim->reqq_wait);
		return BFA_FALSE;
	}

	bfi_h2i_set(m->mh, BFI_MC_ITN, BFI_ITN_H2I_DELETE_REQ,
			bfa_fn_lpu(itnim->bfa));
	m->fw_handle = itnim->rport->fw_handle;
	bfa_stats(itnim, fw_delete);

	/**
	 * queue I/O message to firmware
	 */
	bfa_reqq_produce(itnim->bfa, itnim->reqq, m->mh);
	return BFA_TRUE;
}

/**
 * Cleanup all pending failed inflight requests.
 */
static void
bfa_itnim_delayed_comp(struct bfa_itnim_s *itnim, bfa_boolean_t iotov)
{
	struct bfa_ioim_s *ioim;
	struct bfa_q_s *qe, *qen;

	bfa_q_iter_safe(&itnim->delay_comp_q, qe, qen) {
		ioim = (struct bfa_ioim_s *)qe;
		bfa_ioim_delayed_comp(ioim, iotov);
	}
}

/**
 * Start all pending IO requests.
 */
static void
bfa_itnim_iotov_online(struct bfa_itnim_s *itnim)
{
	struct bfa_ioim_s *ioim;

	bfa_itnim_iotov_stop(itnim);

	/**
	 * Abort all inflight IO requests in the queue
	 */
	bfa_itnim_delayed_comp(itnim, BFA_FALSE);

	/**
	 * Start all pending IO requests.
	 */
	while (!bfa_q_is_empty(&itnim->pending_q)) {
		bfa_q_deq(&itnim->pending_q, &ioim);
		bfa_q_enq(&itnim->io_q, ioim);
		/**
		 it it failed due to lun mask ioim is in free q and uninit
		 state after calling.
		 */
		bfa_ioim_start(ioim, ioim->mode);
	}
}

/**
 * Fail all pending IO requests
 */
static void
bfa_itnim_iotov_cleanup(struct bfa_itnim_s *itnim)
{
	struct bfa_ioim_s *ioim;

	/**
	 * Fail all inflight IO requests in the queue
	 */
	bfa_itnim_delayed_comp(itnim, BFA_TRUE);

	/**
	 * Fail any pending IO requests.
	 */
	while (!bfa_q_is_empty(&itnim->pending_q)) {
		bfa_q_deq(&itnim->pending_q, &ioim);
		bfa_q_enq(&ioim->fcpim->ioim_comp_q, ioim);
		bfa_ioim_tov(ioim);
	}
}

/**
 * IO TOV timer callback. Fail any pending IO requests.
 */
static void
bfa_itnim_iotov(void *itnim_arg)
{
	struct bfa_itnim_s *itnim = itnim_arg;

	itnim->iotov_active = BFA_FALSE;

	bfa_cb_itnim_tov_begin(itnim->ditn);
	bfa_itnim_iotov_cleanup(itnim);
	bfa_cb_itnim_tov(itnim->ditn);
}

/**
 * Start IO TOV timer for failing back pending IO requests in offline state.
 */
static void
bfa_itnim_iotov_start(struct bfa_itnim_s *itnim)
{
	if (itnim->fcpim->path_tov > 0)  {

		itnim->iotov_active = BFA_TRUE;
		bfa_assert(bfa_itnim_hold_io(itnim));
		bfa_timer_start(itnim->bfa, &itnim->timer,
			bfa_itnim_iotov, itnim, itnim->fcpim->path_tov);
	}
}

/**
 * Stop IO TOV timer.
 */
static void
bfa_itnim_iotov_stop(struct bfa_itnim_s *itnim)
{
	if (itnim->iotov_active) {
		itnim->iotov_active = BFA_FALSE;
		bfa_timer_stop(&itnim->timer);
	}
}

/**
 * Stop IO TOV timer.
 */
static void
bfa_itnim_iotov_delete(struct bfa_itnim_s *itnim)
{
	bfa_boolean_t pathtov_active = BFA_FALSE;

	if (itnim->iotov_active)
		pathtov_active = BFA_TRUE;

	bfa_itnim_iotov_stop(itnim);
	if (pathtov_active)
		bfa_cb_itnim_tov_begin(itnim->ditn);
	bfa_itnim_iotov_cleanup(itnim);
	if (pathtov_active)
		bfa_cb_itnim_tov(itnim->ditn);
}

static void
bfa_itnim_update_del_itn_stats(struct bfa_itnim_s *itnim)
{
	struct bfa_fcpim_s *fcpim = BFA_FCPIM(itnim->bfa);
	fcpim->del_itn_stats.del_itn_iocomp_aborted +=
		itnim->stats.iocomp_aborted;
	fcpim->del_itn_stats.del_itn_iocomp_timedout +=
		itnim->stats.iocomp_timedout;
	fcpim->del_itn_stats.del_itn_iocom_sqer_needed +=
		itnim->stats.iocom_sqer_needed;
	fcpim->del_itn_stats.del_itn_iocom_res_free +=
		itnim->stats.iocom_res_free;
	fcpim->del_itn_stats.del_itn_iocom_hostabrts +=
		itnim->stats.iocom_hostabrts;
	fcpim->del_itn_stats.del_itn_total_ios += itnim->stats.total_ios;
	fcpim->del_itn_stats.del_io_iocdowns += itnim->stats.io_iocdowns;
	fcpim->del_itn_stats.del_tm_iocdowns += itnim->stats.tm_iocdowns;
}

/**
 * @}
 */

/**
 * @ig bfa_itnim_public
 * @{
 */

/**
 * @brief
 * 		Itnim interrupt processing.
 */
void
bfa_itnim_isr(struct bfa_s *bfa, struct bfi_msg_s *m)
{
	struct bfa_fcpim_s *fcpim = BFA_FCPIM(bfa);
	union bfi_itn_i2h_msg_u msg;
	struct bfa_itnim_s *itnim;

	bfa_trc(bfa, m->mhdr.msg_id);

	msg.msg = m;

	switch (m->mhdr.msg_id) {
	case BFI_ITN_I2H_CREATE_RSP:
		itnim = BFA_ITNIM_FROM_TAG(fcpim,
						msg.create_rsp->bfa_handle);
		bfa_assert(msg.create_rsp->status == BFA_STATUS_OK);
		bfa_stats(itnim, create_comps);
		bfa_sm_send_event(itnim, BFA_ITNIM_SM_FWRSP);
		break;

	case BFI_ITN_I2H_DELETE_RSP:
		itnim = BFA_ITNIM_FROM_TAG(fcpim,
						msg.delete_rsp->bfa_handle);
		bfa_assert(msg.delete_rsp->status == BFA_STATUS_OK);
		bfa_stats(itnim, delete_comps);
		bfa_sm_send_event(itnim, BFA_ITNIM_SM_FWRSP);
		break;

	case BFI_ITN_I2H_SLER_EVENT:
		itnim = BFA_ITNIM_FROM_TAG(fcpim,
						msg.sler_event->bfa_handle);
		bfa_stats(itnim, sler_events);
		bfa_sm_send_event(itnim, BFA_ITNIM_SM_SLER);
		break;

	default:
		bfa_trc(bfa, m->mhdr.msg_id);
		bfa_assert(0);
	}
}

/**
 * @}
 */

/**
 * @ig bfa_itnim_api
 * @{
 */

struct bfa_itnim_s *
bfa_itnim_create(struct bfa_s *bfa, struct bfa_rport_s *rport, void *ditn)
{
	struct bfa_fcpim_s *fcpim = BFA_FCPIM(bfa);
	struct bfa_itnim_s *itnim;

	bfa_itn_create(bfa, rport, bfa_itnim_isr);

	itnim = BFA_ITNIM_FROM_TAG(fcpim, rport->rport_tag);
	bfa_assert(itnim->rport == rport);

	itnim->ditn = ditn;

	bfa_stats(itnim, creates);
	bfa_sm_send_event(itnim, BFA_ITNIM_SM_CREATE);

	return (itnim);
}

void
bfa_itnim_delete(struct bfa_itnim_s *itnim)
{
	bfa_stats(itnim, deletes);
	bfa_sm_send_event(itnim, BFA_ITNIM_SM_DELETE);
}

void
bfa_itnim_online(struct bfa_itnim_s *itnim, bfa_boolean_t seq_rec)
{
	itnim->seq_rec = seq_rec;
	bfa_stats(itnim, onlines);
	bfa_sm_send_event(itnim, BFA_ITNIM_SM_ONLINE);
}

void
bfa_itnim_offline(struct bfa_itnim_s *itnim)
{
	bfa_stats(itnim, offlines);
	bfa_sm_send_event(itnim, BFA_ITNIM_SM_OFFLINE);
}

/**
 * Return true if itnim is considered offline for holding off IO request.
 * IO is not held if itnim is being deleted.
 */
bfa_boolean_t
bfa_itnim_hold_io(struct bfa_itnim_s *itnim)
{
	return (itnim->fcpim->path_tov && itnim->iotov_active &&
		(bfa_sm_cmp_state(itnim, bfa_itnim_sm_fwcreate) ||
		 bfa_sm_cmp_state(itnim, bfa_itnim_sm_sler) ||
		 bfa_sm_cmp_state(itnim, bfa_itnim_sm_cleanup_offline) ||
		 bfa_sm_cmp_state(itnim, bfa_itnim_sm_fwdelete) ||
		 bfa_sm_cmp_state(itnim, bfa_itnim_sm_offline) ||
		 bfa_sm_cmp_state(itnim, bfa_itnim_sm_iocdisable)));
}

void
bfa_itnim_get_stats(struct bfa_itnim_s *itnim,
	struct bfa_itnim_iostats_s *stats)
{
	if (itnim)
		*stats = itnim->stats;
}

void
bfa_itnim_clear_stats(struct bfa_itnim_s *itnim)
{
	int j;

	if (!itnim)
		return;

	bfa_os_memset(&itnim->stats, 0, sizeof(itnim->stats));
	bfa_os_memset(&itnim->iolat, 0, sizeof(itnim->iolat));
	for (j = 0; j < BFA_IOBUCKET_MAX; j++)
		itnim->iolat.min[j] = ~0;
}

/**
 *  BFA IO module state machine functions
 */

/**
 * @brief
 * 		IO is not started (unallocated).
 */
static void
bfa_ioim_sm_uninit(struct bfa_ioim_s *ioim, enum bfa_ioim_event event)
{
	bfa_trc_fp(ioim->bfa, ioim->iotag);
	bfa_trc_fp(ioim->bfa, event);

	switch (event) {
	case BFA_IOIM_SM_START:
		if (!bfa_itnim_is_online(ioim->itnim)) {
			if (!bfa_itnim_hold_io(ioim->itnim)) {
				bfa_sm_set_state(ioim, bfa_ioim_sm_hcb);
				bfa_q_qe_deq(ioim);
				bfa_q_enq(&ioim->fcpim->ioim_comp_q, ioim);
				bfa_cb_queue(ioim->bfa, &ioim->hcb_qe,
						__bfa_cb_ioim_pathtov, ioim);
			} else {
				bfa_q_qe_deq(ioim);
				bfa_q_enq(&ioim->itnim->pending_q, ioim);
			}
			break;
		}

		if (ioim->nsges > BFI_SGE_INLINE) {
			if (!bfa_ioim_sge_setup(ioim)) {
				bfa_sm_set_state(ioim, bfa_ioim_sm_sgalloc);
				return;
			}
		}

		if (!bfa_ioim_send_ioreq(ioim)) {
			bfa_sm_set_state(ioim, bfa_ioim_sm_qfull);
			break;
		}

		bfa_sm_set_state(ioim, bfa_ioim_sm_active);
		break;

	case BFA_IOIM_SM_IOTOV:
		bfa_sm_set_state(ioim, bfa_ioim_sm_hcb);
		bfa_ioim_move_to_comp_q(ioim);
		bfa_cb_queue(ioim->bfa, &ioim->hcb_qe,
				__bfa_cb_ioim_pathtov, ioim);
		break;

	case BFA_IOIM_SM_ABORT:
		/**
		 * IO in pending queue can get abort requests. Complete abort
		 * requests immediately.
		 */
		bfa_sm_set_state(ioim, bfa_ioim_sm_hcb);
		bfa_assert(bfa_q_is_on_q(&ioim->itnim->pending_q, ioim));
		bfa_cb_queue(ioim->bfa, &ioim->hcb_qe,
				__bfa_cb_ioim_abort, ioim);
		break;

	case	BFA_IOIM_SM_LM_LUN_NOT_SUP:
		bfa_sm_set_state(ioim, bfa_ioim_sm_hcb);
		bfa_ioim_move_to_comp_q(ioim);
		bfa_cb_queue(ioim->bfa, &ioim->hcb_qe,
				__bfa_cb_ioim_lm_lun_not_sup, ioim);
		break;

	case	BFA_IOIM_SM_LM_RPL_DC:
		bfa_sm_set_state(ioim, bfa_ioim_sm_hcb);
		bfa_ioim_move_to_comp_q(ioim);
		bfa_cb_queue(ioim->bfa, &ioim->hcb_qe,
				__bfa_cb_ioim_lm_rpl_dc, ioim);
		break;


	case	BFA_IOIM_SM_LM_LUN_NOT_RDY:
		bfa_sm_set_state(ioim, bfa_ioim_sm_hcb);
		bfa_ioim_move_to_comp_q(ioim);
		bfa_cb_queue(ioim->bfa, &ioim->hcb_qe,
				__bfa_cb_ioim_lm_lun_not_rdy, ioim);
		break;

	default:
		bfa_sm_fault(ioim->bfa, event);
	}
}

/**
 * @brief
 * 		IO is waiting for SG pages.
 */
static void
bfa_ioim_sm_sgalloc(struct bfa_ioim_s *ioim, enum bfa_ioim_event event)
{
	bfa_trc(ioim->bfa, ioim->iotag);
	bfa_trc(ioim->bfa, event);

	switch (event) {
	case BFA_IOIM_SM_SGALLOCED:
		if (!bfa_ioim_send_ioreq(ioim)) {
			bfa_sm_set_state(ioim, bfa_ioim_sm_qfull);
			break;
		}
		bfa_sm_set_state(ioim, bfa_ioim_sm_active);
		break;

	case BFA_IOIM_SM_CLEANUP:
		bfa_sm_set_state(ioim, bfa_ioim_sm_hcb);
		bfa_sgpg_wcancel(ioim->bfa, &ioim->iosp->sgpg_wqe);
		bfa_cb_queue(ioim->bfa, &ioim->hcb_qe, __bfa_cb_ioim_failed,
			      ioim);
		bfa_ioim_notify_cleanup(ioim);
		break;

	case BFA_IOIM_SM_ABORT:
		bfa_sm_set_state(ioim, bfa_ioim_sm_hcb);
		bfa_sgpg_wcancel(ioim->bfa, &ioim->iosp->sgpg_wqe);
		bfa_ioim_move_to_comp_q(ioim);
		bfa_cb_queue(ioim->bfa, &ioim->hcb_qe, __bfa_cb_ioim_abort,
			      ioim);
		break;

	case BFA_IOIM_SM_HWFAIL:
		bfa_sm_set_state(ioim, bfa_ioim_sm_hcb);
		bfa_sgpg_wcancel(ioim->bfa, &ioim->iosp->sgpg_wqe);
		bfa_ioim_move_to_comp_q(ioim);
		bfa_cb_queue(ioim->bfa, &ioim->hcb_qe, __bfa_cb_ioim_failed,
			      ioim);
		break;

	default:
		bfa_sm_fault(ioim->bfa, event);
	}
}

/**
 * @brief
 * 		IO is active.
 */
static void
bfa_ioim_sm_active(struct bfa_ioim_s *ioim, enum bfa_ioim_event event)
{
	bfa_trc_fp(ioim->bfa, ioim->iotag);
	bfa_trc_fp(ioim->bfa, event);

	switch (event) {
	case BFA_IOIM_SM_COMP_GOOD:
		bfa_sm_set_state(ioim, bfa_ioim_sm_hcb);
		bfa_ioim_move_to_comp_q(ioim);
		bfa_cb_queue(ioim->bfa, &ioim->hcb_qe,
			      __bfa_cb_ioim_good_comp, ioim);
		break;

	case BFA_IOIM_SM_COMP:
		bfa_sm_set_state(ioim, bfa_ioim_sm_hcb);
		bfa_ioim_move_to_comp_q(ioim);
		bfa_cb_queue(ioim->bfa, &ioim->hcb_qe, __bfa_cb_ioim_comp,
			      ioim);
		break;

	case BFA_IOIM_SM_DONE:
		bfa_sm_set_state(ioim, bfa_ioim_sm_hcb_free);
		bfa_ioim_move_to_comp_q(ioim);
		bfa_cb_queue(ioim->bfa, &ioim->hcb_qe, __bfa_cb_ioim_comp,
			      ioim);
		break;

	case BFA_IOIM_SM_ABORT:
		ioim->iosp->abort_explicit = BFA_TRUE;
		ioim->io_cbfn = __bfa_cb_ioim_abort;

		if (bfa_ioim_send_abort(ioim))
			bfa_sm_set_state(ioim, bfa_ioim_sm_abort);
		else {
			bfa_sm_set_state(ioim, bfa_ioim_sm_abort_qfull);
			bfa_stats(ioim->itnim, qwait);
			bfa_reqq_wait(ioim->bfa, ioim->reqq,
					  &ioim->iosp->reqq_wait);
		}
		break;

	case BFA_IOIM_SM_CLEANUP:
		ioim->iosp->abort_explicit = BFA_FALSE;
		ioim->io_cbfn = __bfa_cb_ioim_failed;

		if (bfa_ioim_send_abort(ioim))
			bfa_sm_set_state(ioim, bfa_ioim_sm_cleanup);
		else {
			bfa_sm_set_state(ioim, bfa_ioim_sm_cleanup_qfull);
			bfa_stats(ioim->itnim, qwait);
			bfa_reqq_wait(ioim->bfa, ioim->reqq,
					  &ioim->iosp->reqq_wait);
		}
		break;

	case BFA_IOIM_SM_HWFAIL:
		bfa_sm_set_state(ioim, bfa_ioim_sm_hcb);
		bfa_ioim_move_to_comp_q(ioim);
		bfa_cb_queue(ioim->bfa, &ioim->hcb_qe, __bfa_cb_ioim_failed,
			      ioim);
		break;

	case BFA_IOIM_SM_SQRETRY:
		if (bfa_ioim_maxretry_reached(ioim)) {
			/* max retry reached, free IO */
			bfa_sm_set_state(ioim, bfa_ioim_sm_hcb_free);
			bfa_ioim_move_to_comp_q(ioim);
			bfa_cb_queue(ioim->bfa, &ioim->hcb_qe,
					__bfa_cb_ioim_failed, ioim);
			break;
		}
		/* waiting for IO tag resource free */
		bfa_sm_set_state(ioim, bfa_ioim_sm_cmnd_retry);
		break;

	default:
		bfa_sm_fault(ioim->bfa, event);
	}
}

/**
* @brief
*	IO is retried with new tag.
*/
static void
bfa_ioim_sm_cmnd_retry(struct bfa_ioim_s *ioim, enum bfa_ioim_event event)
{
	bfa_trc_fp(ioim->bfa, ioim->iotag);
	bfa_trc_fp(ioim->bfa, event);

	switch (event) {
	case BFA_IOIM_SM_FREE:
		/* abts and rrq done. Now retry the IO with new tag */
		bfa_ioim_update_iotag(ioim);
		if (!bfa_ioim_send_ioreq(ioim)) {
			bfa_sm_set_state(ioim, bfa_ioim_sm_qfull);
			break;
		}
		bfa_sm_set_state(ioim, bfa_ioim_sm_active);
	break;

	case BFA_IOIM_SM_CLEANUP:
		ioim->iosp->abort_explicit = BFA_FALSE;
		ioim->io_cbfn = __bfa_cb_ioim_failed;

		if (bfa_ioim_send_abort(ioim))
			bfa_sm_set_state(ioim, bfa_ioim_sm_cleanup);
		else {
			bfa_sm_set_state(ioim, bfa_ioim_sm_cleanup_qfull);
			bfa_stats(ioim->itnim, qwait);
			bfa_reqq_wait(ioim->bfa, ioim->reqq,
					  &ioim->iosp->reqq_wait);
		}
	break;

	case BFA_IOIM_SM_HWFAIL:
		bfa_sm_set_state(ioim, bfa_ioim_sm_hcb);
		bfa_ioim_move_to_comp_q(ioim);
		bfa_cb_queue(ioim->bfa, &ioim->hcb_qe,
			 __bfa_cb_ioim_failed, ioim);
		break;

	case BFA_IOIM_SM_ABORT:
		/** in this state IO abort is done.
		 * Waiting for IO tag resource free.
		 */
		bfa_sm_set_state(ioim, bfa_ioim_sm_hcb_free);
		bfa_cb_queue(ioim->bfa, &ioim->hcb_qe, __bfa_cb_ioim_abort,
			      ioim);
		break;

	default:
		bfa_sm_fault(ioim->bfa, event);
	}
}

/**
 * @brief
 * 		IO is being aborted, waiting for completion from firmware.
 */
static void
bfa_ioim_sm_abort(struct bfa_ioim_s *ioim, enum bfa_ioim_event event)
{
	bfa_trc(ioim->bfa, ioim->iotag);
	bfa_trc(ioim->bfa, event);

	switch (event) {
	case BFA_IOIM_SM_COMP_GOOD:
	case BFA_IOIM_SM_COMP:
	case BFA_IOIM_SM_DONE:
	case BFA_IOIM_SM_FREE:
		break;

	case BFA_IOIM_SM_ABORT_DONE:
		bfa_sm_set_state(ioim, bfa_ioim_sm_hcb_free);
		bfa_cb_queue(ioim->bfa, &ioim->hcb_qe, __bfa_cb_ioim_abort,
			      ioim);
		break;

	case BFA_IOIM_SM_ABORT_COMP:
		bfa_sm_set_state(ioim, bfa_ioim_sm_hcb);
		bfa_ioim_move_to_comp_q(ioim);
		bfa_cb_queue(ioim->bfa, &ioim->hcb_qe, __bfa_cb_ioim_abort,
			      ioim);
		break;

	case BFA_IOIM_SM_COMP_UTAG:
		bfa_sm_set_state(ioim, bfa_ioim_sm_hcb);
		bfa_ioim_move_to_comp_q(ioim);
		bfa_cb_queue(ioim->bfa, &ioim->hcb_qe, __bfa_cb_ioim_abort,
			      ioim);
		break;

	case BFA_IOIM_SM_CLEANUP:
		bfa_assert(ioim->iosp->abort_explicit == BFA_TRUE);
		ioim->iosp->abort_explicit = BFA_FALSE;

		if (bfa_ioim_send_abort(ioim))
			bfa_sm_set_state(ioim, bfa_ioim_sm_cleanup);
		else {
			bfa_sm_set_state(ioim, bfa_ioim_sm_cleanup_qfull);
			bfa_stats(ioim->itnim, qwait);
			bfa_reqq_wait(ioim->bfa, ioim->reqq,
					  &ioim->iosp->reqq_wait);
		}
		break;

	case BFA_IOIM_SM_HWFAIL:
		bfa_sm_set_state(ioim, bfa_ioim_sm_hcb);
		bfa_ioim_move_to_comp_q(ioim);
		bfa_cb_queue(ioim->bfa, &ioim->hcb_qe, __bfa_cb_ioim_failed,
			      ioim);
		break;

	default:
		bfa_sm_fault(ioim->bfa, event);
	}
}

/**
 * @brief
 * IO is being cleaned up (implicit abort), waiting for completion from
 * firmware.
 */
static void
bfa_ioim_sm_cleanup(struct bfa_ioim_s *ioim, enum bfa_ioim_event event)
{
	bfa_trc(ioim->bfa, ioim->iotag);
	bfa_trc(ioim->bfa, event);

	switch (event) {
	case BFA_IOIM_SM_COMP_GOOD:
	case BFA_IOIM_SM_COMP:
	case BFA_IOIM_SM_DONE:
	case BFA_IOIM_SM_FREE:
		break;

	case BFA_IOIM_SM_ABORT:
		/**
		 * IO is already being aborted implicitly
		 */
		ioim->io_cbfn = __bfa_cb_ioim_abort;
		break;

	case BFA_IOIM_SM_ABORT_DONE:
		bfa_sm_set_state(ioim, bfa_ioim_sm_hcb_free);
		bfa_cb_queue(ioim->bfa, &ioim->hcb_qe, ioim->io_cbfn, ioim);
		bfa_ioim_notify_cleanup(ioim);
		break;

	case BFA_IOIM_SM_ABORT_COMP:
		bfa_sm_set_state(ioim, bfa_ioim_sm_hcb);
		bfa_cb_queue(ioim->bfa, &ioim->hcb_qe, ioim->io_cbfn, ioim);
		bfa_ioim_notify_cleanup(ioim);
		break;

	case BFA_IOIM_SM_COMP_UTAG:
		bfa_sm_set_state(ioim, bfa_ioim_sm_hcb);
		bfa_cb_queue(ioim->bfa, &ioim->hcb_qe, ioim->io_cbfn, ioim);
		bfa_ioim_notify_cleanup(ioim);
		break;

	case BFA_IOIM_SM_HWFAIL:
		bfa_sm_set_state(ioim, bfa_ioim_sm_hcb);
		bfa_ioim_move_to_comp_q(ioim);
		bfa_cb_queue(ioim->bfa, &ioim->hcb_qe, __bfa_cb_ioim_failed,
			      ioim);
		break;

	case BFA_IOIM_SM_CLEANUP:
		/**
		 * IO can be in cleanup state already due to TM command.
		 * 2nd cleanup request comes from ITN offline event.
		 */
		break;

	default:
		bfa_sm_fault(ioim->bfa, event);
	}
}

/**
 * @brief
 * 		IO is waiting for room in request CQ
 */
static void
bfa_ioim_sm_qfull(struct bfa_ioim_s *ioim, enum bfa_ioim_event event)
{
	bfa_trc(ioim->bfa, ioim->iotag);
	bfa_trc(ioim->bfa, event);

	switch (event) {
	case BFA_IOIM_SM_QRESUME:
		bfa_sm_set_state(ioim, bfa_ioim_sm_active);
		bfa_ioim_send_ioreq(ioim);
		break;

	case BFA_IOIM_SM_ABORT:
		bfa_sm_set_state(ioim, bfa_ioim_sm_hcb);
		bfa_reqq_wcancel(&ioim->iosp->reqq_wait);
		bfa_ioim_move_to_comp_q(ioim);
		bfa_cb_queue(ioim->bfa, &ioim->hcb_qe, __bfa_cb_ioim_abort,
			      ioim);
		break;

	case BFA_IOIM_SM_CLEANUP:
		bfa_sm_set_state(ioim, bfa_ioim_sm_hcb);
		bfa_reqq_wcancel(&ioim->iosp->reqq_wait);
		bfa_cb_queue(ioim->bfa, &ioim->hcb_qe, __bfa_cb_ioim_failed,
			      ioim);
		bfa_ioim_notify_cleanup(ioim);
		break;

	case BFA_IOIM_SM_HWFAIL:
		bfa_sm_set_state(ioim, bfa_ioim_sm_hcb);
		bfa_reqq_wcancel(&ioim->iosp->reqq_wait);
		bfa_ioim_move_to_comp_q(ioim);
		bfa_cb_queue(ioim->bfa, &ioim->hcb_qe, __bfa_cb_ioim_failed,
			      ioim);
		break;

	default:
		bfa_sm_fault(ioim->bfa, event);
	}
}

/**
 * @brief
 * 		Active IO is being aborted, waiting for room in request CQ.
 */
static void
bfa_ioim_sm_abort_qfull(struct bfa_ioim_s *ioim, enum bfa_ioim_event event)
{
	bfa_trc(ioim->bfa, ioim->iotag);
	bfa_trc(ioim->bfa, event);

	switch (event) {
	case BFA_IOIM_SM_QRESUME:
		bfa_sm_set_state(ioim, bfa_ioim_sm_abort);
		bfa_ioim_send_abort(ioim);
		break;

	case BFA_IOIM_SM_CLEANUP:
		bfa_assert(ioim->iosp->abort_explicit == BFA_TRUE);
		ioim->iosp->abort_explicit = BFA_FALSE;
		bfa_sm_set_state(ioim, bfa_ioim_sm_cleanup_qfull);
		break;

	case BFA_IOIM_SM_COMP_GOOD:
	case BFA_IOIM_SM_COMP:
		bfa_sm_set_state(ioim, bfa_ioim_sm_hcb);
		bfa_reqq_wcancel(&ioim->iosp->reqq_wait);
		bfa_ioim_move_to_comp_q(ioim);
		bfa_cb_queue(ioim->bfa, &ioim->hcb_qe, __bfa_cb_ioim_abort,
			      ioim);
		break;

	case BFA_IOIM_SM_DONE:
		bfa_sm_set_state(ioim, bfa_ioim_sm_hcb_free);
		bfa_reqq_wcancel(&ioim->iosp->reqq_wait);
		bfa_ioim_move_to_comp_q(ioim);
		bfa_cb_queue(ioim->bfa, &ioim->hcb_qe, __bfa_cb_ioim_abort,
			      ioim);
		break;

	case BFA_IOIM_SM_HWFAIL:
		bfa_sm_set_state(ioim, bfa_ioim_sm_hcb);
		bfa_reqq_wcancel(&ioim->iosp->reqq_wait);
		bfa_ioim_move_to_comp_q(ioim);
		bfa_cb_queue(ioim->bfa, &ioim->hcb_qe, __bfa_cb_ioim_failed,
			      ioim);
		break;

	default:
		bfa_sm_fault(ioim->bfa, event);
	}
}

/**
 * @brief
 * 		Active IO is being cleaned up, waiting for room in request CQ.
 */
static void
bfa_ioim_sm_cleanup_qfull(struct bfa_ioim_s *ioim, enum bfa_ioim_event event)
{
	bfa_trc(ioim->bfa, ioim->iotag);
	bfa_trc(ioim->bfa, event);

	switch (event) {
	case BFA_IOIM_SM_QRESUME:
		bfa_sm_set_state(ioim, bfa_ioim_sm_cleanup);
		bfa_ioim_send_abort(ioim);
		break;

	case BFA_IOIM_SM_ABORT:
		/**
		 * IO is alraedy being cleaned up implicitly
		 */
		ioim->io_cbfn = __bfa_cb_ioim_abort;
		break;

	case BFA_IOIM_SM_COMP_GOOD:
	case BFA_IOIM_SM_COMP:
		bfa_sm_set_state(ioim, bfa_ioim_sm_hcb);
		bfa_reqq_wcancel(&ioim->iosp->reqq_wait);
		bfa_cb_queue(ioim->bfa, &ioim->hcb_qe, ioim->io_cbfn, ioim);
		bfa_ioim_notify_cleanup(ioim);
		break;

	case BFA_IOIM_SM_DONE:
		bfa_sm_set_state(ioim, bfa_ioim_sm_hcb_free);
		bfa_reqq_wcancel(&ioim->iosp->reqq_wait);
		bfa_cb_queue(ioim->bfa, &ioim->hcb_qe, ioim->io_cbfn, ioim);
		bfa_ioim_notify_cleanup(ioim);
		break;

	case BFA_IOIM_SM_HWFAIL:
		bfa_sm_set_state(ioim, bfa_ioim_sm_hcb);
		bfa_reqq_wcancel(&ioim->iosp->reqq_wait);
		bfa_ioim_move_to_comp_q(ioim);
		bfa_cb_queue(ioim->bfa, &ioim->hcb_qe, __bfa_cb_ioim_failed,
			      ioim);
		break;

	default:
		bfa_sm_fault(ioim->bfa, event);
	}
}

/**
 * @brief
 * IO bfa callback is pending.
 */
static void
bfa_ioim_sm_hcb(struct bfa_ioim_s *ioim, enum bfa_ioim_event event)
{
	bfa_trc_fp(ioim->bfa, ioim->iotag);
	bfa_trc_fp(ioim->bfa, event);

	switch (event) {
	case BFA_IOIM_SM_HCB:
		bfa_sm_set_state(ioim, bfa_ioim_sm_uninit);
		bfa_ioim_free(ioim);
		bfa_cb_ioim_resfree(ioim->bfa->bfad);
		break;

	case BFA_IOIM_SM_CLEANUP:
		bfa_ioim_notify_cleanup(ioim);
		break;

	case BFA_IOIM_SM_HWFAIL:
		break;

	default:
		bfa_sm_fault(ioim->bfa, event);
	}
}

/**
 * @brief
 * IO bfa callback is pending. IO resource cannot be freed.
 */
static void
bfa_ioim_sm_hcb_free(struct bfa_ioim_s *ioim, enum bfa_ioim_event event)
{
	bfa_trc(ioim->bfa, ioim->iotag);
	bfa_trc(ioim->bfa, event);

	switch (event) {
	case BFA_IOIM_SM_HCB:
		bfa_sm_set_state(ioim, bfa_ioim_sm_resfree);
		bfa_q_qe_deq(ioim);
		bfa_q_enq(&ioim->fcpim->ioim_resfree_q, ioim);
		break;

	case BFA_IOIM_SM_FREE:
		bfa_sm_set_state(ioim, bfa_ioim_sm_hcb);
		break;

	case BFA_IOIM_SM_CLEANUP:
		bfa_ioim_notify_cleanup(ioim);
		break;

	case BFA_IOIM_SM_HWFAIL:
		bfa_sm_set_state(ioim, bfa_ioim_sm_hcb);
		break;

	default:
		bfa_sm_fault(ioim->bfa, event);
	}
}

/**
 * @brief
 * IO is completed, waiting resource free from firmware.
 */
static void
bfa_ioim_sm_resfree(struct bfa_ioim_s *ioim, enum bfa_ioim_event event)
{
	bfa_trc(ioim->bfa, ioim->iotag);
	bfa_trc(ioim->bfa, event);

	switch (event) {
	case BFA_IOIM_SM_FREE:
		bfa_sm_set_state(ioim, bfa_ioim_sm_uninit);
		bfa_ioim_free(ioim);
		bfa_cb_ioim_resfree(ioim->bfa->bfad);
		break;

	case BFA_IOIM_SM_CLEANUP:
		bfa_ioim_notify_cleanup(ioim);
		break;

	case BFA_IOIM_SM_HWFAIL:
		break;

	default:
		bfa_sm_fault(ioim->bfa, event);
	}
}

/**
 * @}
 */

/**
 * @ig hal_ioim_private
 * @{
 */

/*
  This is called from bfa_fcpim_start after the bfa_init() with flash read
  is complete by driver. now invalidate the stale content of lun mask
  like unit attention, rp tag and lp tag.
 */
static void
bfa_ioim_lm_init(struct bfa_s *bfa)
{
	bfa_lun_mask_t *lunm_list;
	int	i;

	if (bfa_get_lun_mask_status(bfa) == BFA_LUNMASK_MINCFG)
		 return;

	lunm_list = bfa_get_lun_mask_list(bfa);
	for (i = 0; i < MAX_LUN_MASK_CFG; i++) {
		lunm_list[i].ua = BFA_IOIM_LM_UA_RESET;
		lunm_list[i].lp_tag = BFA_LP_TAG_INVALID;
		lunm_list[i].rp_tag = BFA_RPORT_TAG_INVALID;
	}
}
/**
 * set UA for all active luns in LM DB
 */
static void
bfa_ioim_lm_set_ua(struct bfa_s *bfa)
{
	bfa_lun_mask_t *lunm_list;
	int	i;

	lunm_list = bfa_get_lun_mask_list(bfa);
	if (!lunm_list)
		return;

	for (i = 0; i < MAX_LUN_MASK_CFG; i++) {
		if (lunm_list[i].state != BFA_IOIM_LUN_MASK_ACTIVE) continue;
		lunm_list[i].ua = BFA_IOIM_LM_UA_SET;
	}
}

/**
 * Validate LUN for LUN masking
 */
static enum bfa_ioim_lm_status
bfa_ioim_lm_check(struct bfa_ioim_s *ioim, bfa_lps_t *lps,
					bfa_rport_t *rp, lun_t lun)
{
	uint8_t	i;
	uint8_t	lp_tag = (uint8_t)ioim->itnim->rport->rport_info.lp_tag;
	uint16_t rp_tag = rp->rport_tag;
	struct bfa_s *bfa = ioim->bfa;
	bfa_lun_mask_t *lun_list = bfa_get_lun_mask_list(bfa);
	scsi_cdb_t	*cdb;

	cdb = (scsi_cdb_t *)bfa_cb_ioim_get_cdb(ioim->dio);

	if (cdb->scsi_cdb[0] == SCSI_OP_REPORT_LUNS) {
		ioim->proc_rsp_data = bfa_ioim_lm_proc_rpl_data;
		return BFA_IOIM_LM_PRESENT;
	}

	if (!lun_list)
		return BFA_IOIM_LM_LUN_NOT_SUP;

	for (i = 0; i < MAX_LUN_MASK_CFG; i++) {
		if (lun_list[i].state != BFA_IOIM_LUN_MASK_ACTIVE)
			continue;
		if ((lun_list[i].lun == lun) &&
				(rp_tag == lun_list[i].rp_tag) &&
				(lp_tag == lun_list[i].lp_tag)) {

			bfa_trc(bfa, lun_list[i].rp_tag);
			bfa_trc(bfa, lun_list[i].lp_tag);
			bfa_trc(bfa, lun_list[i].lun);

			if (bfa_ioc_is_uefi(&bfa->ioc))
				return BFA_IOIM_LM_PRESENT;

			if ((lun_list[i].ua == BFA_IOIM_LM_UA_SET) &&
				(cdb->scsi_cdb[0] != SCSI_OP_INQUIRY)) {
				lun_list[i].ua = BFA_IOIM_LM_UA_RESET;
				return BFA_IOIM_LM_RPL_DATA_CHANGED;
			}

			if (cdb->scsi_cdb[0] == SCSI_OP_REPORT_LUNS)
				ioim->proc_rsp_data =
					bfa_ioim_lm_proc_rpl_data;
			return BFA_IOIM_LM_PRESENT;
		}
	}

	if (bfa_ioc_is_uefi(&bfa->ioc))
		return BFA_IOIM_LM_LUN_NOT_SUP;

	/**
	 * for luns not in lun mask DB we need to change inquiry response
	 * peripheral device qualifier  to  SCSI_DEVQUAL_NOT_CONNECTED
	 */
	if (cdb->scsi_cdb[0] == SCSI_OP_INQUIRY) {
		ioim->proc_rsp_data = bfa_ioim_lm_proc_inq_data;
		return BFA_IOIM_LM_PRESENT;
	}
	if (cdb->scsi_cdb[0] == SCSI_OP_TEST_UNIT_READY)
		return BFA_IOIM_LM_LUN_NOT_RDY;

	return BFA_IOIM_LM_LUN_NOT_SUP;
}

static bfa_boolean_t
bfa_ioim_lm_proc_rsp_data_dummy(struct bfa_ioim_s *ioim)
{
	return BFA_TRUE;
}

static void
bfa_ioim_lm_fetch_lun(struct bfa_ioim_s *ioim, uint8_t *rl_data, int offset,
	 int buf_lun_cnt)
{
	bfa_lun_mask_t *lun_list = bfa_get_lun_mask_list(ioim->bfa);
	lun_t *lun_data = (lun_t *)(rl_data + offset);
	lun_t lun;
	int i, j;

	bfa_trc(ioim->bfa, buf_lun_cnt);
	if (!lun_list)
		return;

	for (j = 0; j < buf_lun_cnt; j++) {
		lun = *((lun_t *)(lun_data + j));
		for (i = 0; i < MAX_LUN_MASK_CFG; i++) {
			if (lun_list[i].state != BFA_IOIM_LUN_MASK_ACTIVE)
				continue;
			if ((lun_list[i].rp_wwn == bfa_ioim_rp_wwn(ioim)) &&
				(lun_list[i].lp_wwn == bfa_ioim_lp_wwn(ioim)) &&
				(lun_list[i].lun == lun)) {
				lun_list[i].state = BFA_IOIM_LUN_MASK_FETCHED;
				break;
			}
		} /* next lun in mask DB */
	} /* next lun in buf */
}

static int
bfa_ioim_lm_update_lun_sg(struct bfa_ioim_s *ioim, uint32_t *pgdlen,
	 scsi_report_luns_data_t *rl)
{
	int sge_count = bfa_cb_ioim_get_sg_count(ioim->dio);
	void *sg = BFA_SG_FIRST(ioim->dio);
	void *prev_sg = sg;
	uint8_t  *rl_data;
	lun_t *prev_rl_data;
	lun_t *base_rl_data;
	bfa_lun_mask_t *lun_list = bfa_get_lun_mask_list(ioim->bfa);
	uint32_t lun_count = 0;
	int lun_fetched_cnt = 0;
	int cur_sg_len = 0;
	int prev_sg_len = 0;
	int i, j, buf_lun_cnt, sgeid = 0, base_count = 0;
	int lun_across_sg_bytes = 0, bytes_from_next_buf = 0;
	lun_t last_lun;
	lun_t temp_last_lun = 0;

	lun_count = bfa_os_htonl(rl->lun_list_length) / sizeof(lun_t);
	rl_data = (uint8_t  *)(rl->lun);
	base_rl_data = (lun_t *)rl_data;
	cur_sg_len = bfa_cb_ioim_get_sglen(ioim->dio, sg);

	/* 1st sg has buf_lun_cnt in initial 8 bytes */
	buf_lun_cnt = (cur_sg_len / sizeof(lun_t)) - 1;
	base_count = buf_lun_cnt;

	bfa_ioim_lm_fetch_lun(ioim, rl_data, 0,  buf_lun_cnt);

	for (sgeid = 1; sgeid < sge_count; sgeid++) {
		bytes_from_next_buf = 0;
		lun_across_sg_bytes = 0;
		prev_sg = sg;
		prev_sg_len = cur_sg_len;
		BFA_SG_NEXT(sg, ioim->dio, sgeid);
		cur_sg_len = bfa_cb_ioim_get_sglen(ioim->dio, sg);
		buf_lun_cnt = cur_sg_len / sizeof(lun_t);
		*pgdlen += cur_sg_len;
		prev_rl_data = (lun_t *)rl_data;
		rl_data = (uint8_t *)bfa_cb_ioim_phys_to_virt(
				ioim->dio, sg);

		/* if the buf is having more data */
		lun_across_sg_bytes = prev_sg_len % sizeof(lun_t);
		if (lun_across_sg_bytes) {
			bfa_trc(ioim->bfa, lun_across_sg_bytes);
			bfa_stats(ioim->itnim, lm_lun_across_sg);
			bytes_from_next_buf = sizeof(lun_t) -
				lun_across_sg_bytes;
			last_lun = 0;
			/* from next buf take higher bytes */
			temp_last_lun = *((lun_t *)rl_data);
			last_lun |= temp_last_lun >> (lun_across_sg_bytes *
					BFA_IOIM_BITS_IN_BYTE);

			/* from prev buf take higher bytes */
			temp_last_lun = *((lun_t *)(prev_rl_data + (
					prev_sg_len - lun_across_sg_bytes)));
			temp_last_lun >>= bytes_from_next_buf *
				BFA_IOIM_BITS_IN_BYTE;
			last_lun = last_lun | (temp_last_lun <<
				(bytes_from_next_buf * BFA_IOIM_BITS_IN_BYTE));

			bfa_ioim_lm_fetch_lun(ioim, (uint8_t*)&last_lun, 0, 1);
		}

		bfa_ioim_lm_fetch_lun(ioim, rl_data, bytes_from_next_buf,
			 buf_lun_cnt);
	} /* next sgeid */

	j = 0;
	sgeid = 0;
	if (!lun_list)
		return lun_fetched_cnt;

	/* for linux reset the sg next loop */
	sg = BFA_SG_FIRST(ioim->dio);
	for (i = 0; i < MAX_LUN_MASK_CFG; i++) {
		if (lun_list[i].state == BFA_IOIM_LUN_MASK_FETCHED) {
			base_rl_data[j] = lun_list[i].lun;
			lun_list[i].state = BFA_IOIM_LUN_MASK_ACTIVE;
			j++; lun_fetched_cnt++;

		}
		if (j > base_count) {
			sgeid ++;
			j = 0;
			BFA_SG_NEXT(sg, ioim->dio, sgeid);
			base_rl_data = (lun_t *)bfa_cb_ioim_phys_to_virt(
					ioim->dio, sg);
			base_count = bfa_cb_ioim_get_sglen(
					ioim->dio, sg) / sizeof(lun_t);
		}
	}
	bfa_trc(ioim->bfa, lun_fetched_cnt);
	return lun_fetched_cnt;

}

/* returns luns masked for each buf */
static int
bfa_ioim_lm_update_lun_nosg(struct bfa_ioim_s *ioim, uint8_t *lun_data,
				int lun_count)
{
	int i, j = 0;
	bfa_lun_mask_t *lun_list = bfa_get_lun_mask_list(ioim->bfa);
	lun_t *rl_data = (lun_t *)lun_data;

	if (!lun_list)
		return j;

	bfa_ioim_lm_fetch_lun(ioim, lun_data, 0, lun_count);
	for (i = 0; i < MAX_LUN_MASK_CFG; i++) {
		if (lun_list[i].state == BFA_IOIM_LUN_MASK_FETCHED) {
			rl_data[j] = lun_list[i].lun;
			lun_list[i].state = BFA_IOIM_LUN_MASK_ACTIVE;
			j++;
			bfa_assert(j <= lun_count);
		}
	}
	return j;
}

static bfa_boolean_t
bfa_ioim_lm_proc_inq_data(struct bfa_ioim_s *ioim)
{
	scsi_inquiry_data_t *inq = NULL;
	void *sg;

	ioim->proc_rsp_data = bfa_ioim_lm_proc_rsp_data_dummy;
	if (ioim->mode == BFA_IOIM_NO_SG) {
		inq = (scsi_inquiry_data_t *)bfa_cb_ioim_get_kva(ioim->dio);
	} else {
		sg = BFA_SG_FIRST(ioim->dio);
		inq = (scsi_inquiry_data_t *)
			bfa_cb_ioim_phys_to_virt(ioim->dio, sg);
	}
	bfa_trc(ioim->bfa, inq->device_type);
	inq->peripheral_qual = SCSI_DEVQUAL_NOT_CONNECTED;
	return 0;
}

static bfa_boolean_t
bfa_ioim_lm_proc_rpl_data(struct bfa_ioim_s *ioim)
{
	int lun_count = 0;
	void *sg;
	int lun_fetched_cnt = 0;
	uint32_t	residue;
	scsi_report_luns_data_t *rl = NULL;
	uint8_t *rl_data;
	struct bfa_s *bfa = ioim->bfa;
	struct bfi_ioim_rsp_s *m;
	uint32_t	pgdlen = 0;
	uint32_t	lun_list_len = 0;
	uint32_t	loop_lun_cnt = 0;

	ioim->proc_rsp_data = bfa_ioim_lm_proc_rsp_data_dummy;
	if (bfa_get_lun_mask_status(bfa) != BFA_LUNMASK_ENABLED) {
		return BFA_TRUE;
	}

	m = (struct bfi_ioim_rsp_s *) &ioim->iosp->comp_rspmsg;
	if (m->scsi_status == SCSI_STATUS_CHECK_CONDITION) {
		bfa_trc(ioim->bfa, m->scsi_status);
		return BFA_TRUE;
	}

	sg = BFA_SG_FIRST(ioim->dio);
	pgdlen = bfa_cb_ioim_get_sglen(ioim->dio, sg);
	bfa_trc(ioim->bfa, pgdlen);

	if (ioim->mode == BFA_IOIM_NO_SG) {
		rl = (scsi_report_luns_data_t *)bfa_cb_ioim_get_kva(ioim->dio);
		rl_data = (uint8_t *)rl->lun;
		lun_list_len = bfa_os_htonl(rl->lun_list_length);
		pgdlen = pgdlen - 8; /* remove for lun count bytes */
		/* for windows if rpl fcp_dl is small it retries the rpl cmd */
		if (pgdlen < lun_list_len) {
			bfa_trc(ioim->bfa, m->resid_flags);
			bfa_trc(ioim->bfa, lun_list_len);
			bfa_trc(ioim->bfa, pgdlen);
			m->resid_flags = FCP_RESID_OVER;
			m->residue = pgdlen - lun_list_len;
			bfa_trc(ioim->bfa, m->residue);
			return BFA_TRUE;
		}
		lun_count = lun_list_len / sizeof(lun_t);
		loop_lun_cnt = (lun_list_len <= pgdlen) ? lun_count: (pgdlen /
				sizeof(lun_t));
		lun_fetched_cnt = bfa_ioim_lm_update_lun_nosg(ioim,
				rl_data, loop_lun_cnt);
	} else {
		rl = (scsi_report_luns_data_t *)
			bfa_cb_ioim_phys_to_virt(ioim->dio, sg);

		lun_count = bfa_os_htonl(rl->lun_list_length) /
			 sizeof(lun_t);

		lun_fetched_cnt = bfa_ioim_lm_update_lun_sg(
				ioim, &pgdlen, rl);
	}
	if (lun_count == lun_fetched_cnt)
		return BFA_TRUE;

	bfa_trc(ioim->bfa, lun_count);
	bfa_trc(ioim->bfa, lun_fetched_cnt);
	bfa_trc(ioim->bfa, bfa_os_ntohl(rl->lun_list_length));

	if (bfa_os_ntohl(rl->lun_list_length) <= pgdlen)
		rl->lun_list_length = bfa_os_ntohl(lun_fetched_cnt) *
				 sizeof(lun_t);
	else
		bfa_stats(ioim->itnim, lm_small_buf_addresidue);

	bfa_trc(ioim->bfa, bfa_os_ntohl(rl->lun_list_length));

	bfa_trc(ioim->bfa, bfa_os_ntohl(m->residue));
	residue = bfa_os_ntohl(m->residue);
	residue += (lun_count - lun_fetched_cnt) * sizeof(lun_t);
	bfa_stats(ioim->itnim, lm_wire_residue_changed);
	m->residue = bfa_os_ntohl(residue);
	bfa_trc(ioim->bfa, ioim->nsges);
	return BFA_FALSE;
}

static void
__bfa_cb_ioim_good_comp(void *cbarg, bfa_boolean_t complete)
{
	struct bfa_ioim_s *ioim = cbarg;

	if (!complete) {
		bfa_sm_send_event(ioim, BFA_IOIM_SM_HCB);
		return;
	}

	bfa_cb_ioim_good_comp(ioim->bfa->bfad, ioim->dio);
}

static void
__bfa_cb_ioim_comp(void *cbarg, bfa_boolean_t complete)
{
	struct bfa_ioim_s	*ioim = cbarg;
	struct bfi_ioim_rsp_s *m;
	uint8_t	*snsinfo = NULL;
	uint8_t	sns_len = 0;
	int32_t	residue = 0;

	if (!complete) {
		bfa_sm_send_event(ioim, BFA_IOIM_SM_HCB);
		return;
	}

	m = (struct bfi_ioim_rsp_s *) &ioim->iosp->comp_rspmsg;
	if (m->io_status == BFI_IOIM_STS_OK) {
		/**
		 * setup sense information, if present
		 */
		if ((m->scsi_status == SCSI_STATUS_CHECK_CONDITION) &&
					m->sns_len) {

			sns_len = m->sns_len;
			snsinfo = BFA_SNSINFO_FROM_TAG(
					ioim->fcpim->fcp, ioim->iotag);
			bfa_stats(ioim->itnim, lun_scsi_check_cond);
			bfa_lun_stats(ioim, lun_scsi_check_cond);

			/* add lun/itnim scsi err stats  based on sense */
			bfa_fcpim_iopf_scsi_comp_stats(ioim,
					(scsi_sense_t *)snsinfo);
		}
		switch (m->scsi_status) {

		case SCSI_STATUS_BUSY:
			bfa_stats(ioim->itnim, lun_scsi_busy);
			bfa_lun_stats(ioim, lun_scsi_busy);
			break;

		case  SCSI_STATUS_RESERVATION_CONFLICT:
			bfa_stats(ioim->itnim, lun_scsi_res_conflict);
			bfa_lun_stats(ioim, lun_scsi_res_conflict);
			break;

		case SCSI_STATUS_COMMAND_TERMINATED:
			bfa_stats(ioim->itnim, lun_scsi_cmd_term);
			bfa_lun_stats(ioim, lun_scsi_cmd_term);
			break;

		case SCSI_STATUS_QUEUE_FULL:
			bfa_stats(ioim->itnim, lun_scsi_q_full);
			bfa_lun_stats(ioim, lun_scsi_q_full);
			break;

		case SCSI_STATUS_ACA_ACTIVE:
			bfa_stats(ioim->itnim, lun_scsi_aca_act);
			bfa_lun_stats(ioim, lun_scsi_aca_act);
			break;

		default :
			break;
		}

		/**
		 * setup residue value correctly for normal completions
		 */
		if (m->resid_flags == FCP_RESID_UNDER) {
			residue = bfa_os_ntohl(m->residue);
			bfa_stats(ioim->itnim, iocomp_underrun);
			bfa_lun_stats(ioim, iocomp_underrun);
		}
		if (m->resid_flags == FCP_RESID_OVER) {
			residue = bfa_os_ntohl(m->residue);
			residue = -residue;
			bfa_stats(ioim->itnim, iocomp_overrun);
			bfa_lun_stats(ioim, iocomp_overrun);
		}
	}

	bfa_cb_ioim_done(ioim->bfa->bfad, ioim->dio, m->io_status,
			 m->scsi_status, sns_len, snsinfo, residue);
}

static void
__bfa_cb_ioim_lm_lun_not_sup(void *cbarg, bfa_boolean_t complete)
{
	struct bfa_ioim_s *ioim = cbarg;
	int sns_len = 0xD;
	uint32_t	residue = bfa_cb_ioim_get_size(ioim->dio);
	scsi_sense_t *snsinfo;

	if (!complete) {
		bfa_sm_send_event(ioim, BFA_IOIM_SM_HCB);
		return;
	}
	snsinfo = (scsi_sense_t *)BFA_SNSINFO_FROM_TAG(
			ioim->fcpim->fcp, ioim->iotag);
	snsinfo->rsp_code = SCSI_SENSE_CUR_ERR;
	snsinfo->add_sense_length = 0xa;
	snsinfo->asc = SCSI_ASC_LUN_NOT_SUPPORTED;
	snsinfo->sense_key = SCSI_SK_ILLEGAL_REQ;
	bfa_trc(ioim->bfa, residue);
	bfa_cb_ioim_done(ioim->bfa->bfad, ioim->dio, BFI_IOIM_STS_OK,
			SCSI_STATUS_CHECK_CONDITION, sns_len,
			(uint8_t*)snsinfo, residue);

}

static void
__bfa_cb_ioim_lm_rpl_dc(void *cbarg, bfa_boolean_t complete)
{
	struct bfa_ioim_s *ioim = cbarg;
	int sns_len = 0xD;
	uint32_t	residue = bfa_cb_ioim_get_size(ioim->dio);
	scsi_sense_t *snsinfo;

	if (!complete) {
		bfa_sm_send_event(ioim, BFA_IOIM_SM_HCB);
		return;
	}
	snsinfo = (scsi_sense_t *)BFA_SNSINFO_FROM_TAG(ioim->fcpim->fcp,
		 ioim->iotag);
	snsinfo->rsp_code = SCSI_SENSE_CUR_ERR;
	snsinfo->sense_key = SCSI_MP_IEC_UNIT_ATTN;
	snsinfo->asc = SCSI_ASC_TOCC;
	snsinfo->add_sense_length = 0x6;
	snsinfo->ascq = SCSI_ASCQ_RL_DATA_CHANGED;
	bfa_trc(ioim->bfa, residue);
	bfa_cb_ioim_done(ioim->bfa->bfad, ioim->dio, BFI_IOIM_STS_OK,
			SCSI_STATUS_CHECK_CONDITION, sns_len,
			(uint8_t*)snsinfo, residue);

}

static void
__bfa_cb_ioim_lm_lun_not_rdy(void *cbarg, bfa_boolean_t complete)
{
	struct bfa_ioim_s *ioim = cbarg;
	int sns_len = 0xD;
	uint32_t	residue = bfa_cb_ioim_get_size(ioim->dio);
	scsi_sense_t *snsinfo;

	if (!complete) {
		bfa_sm_send_event(ioim, BFA_IOIM_SM_HCB);
		return;
	}
	snsinfo = (scsi_sense_t *)BFA_SNSINFO_FROM_TAG(
			ioim->fcpim->fcp, ioim->iotag);
	snsinfo->rsp_code = SCSI_SENSE_CUR_ERR;
	snsinfo->add_sense_length = 0xa;
	snsinfo->sense_key = SCSI_SK_NOT_READY;
	snsinfo->asc = SCSI_ASC_LUN_NOT_READY;
	snsinfo->ascq = SCSI_ASCQ_MAN_INTR_REQ;
	bfa_trc(ioim->bfa, residue);
	bfa_cb_ioim_done(ioim->bfa->bfad, ioim->dio, BFI_IOIM_STS_OK,
			SCSI_STATUS_CHECK_CONDITION, sns_len,
			(uint8_t*)snsinfo, residue);
}

static void
__bfa_cb_ioim_failed(void *cbarg, bfa_boolean_t complete)
{
	struct bfa_ioim_s *ioim = cbarg;

	if (!complete) {
		bfa_sm_send_event(ioim, BFA_IOIM_SM_HCB);
		return;
	}

	ioim->proc_rsp_data = bfa_ioim_lm_proc_rsp_data_dummy;
	bfa_cb_ioim_done(ioim->bfa->bfad, ioim->dio, BFI_IOIM_STS_ABORTED,
			  0, 0, NULL, 0);
}

static void
__bfa_cb_ioim_pathtov(void *cbarg, bfa_boolean_t complete)
{
	struct bfa_ioim_s *ioim = cbarg;

	bfa_stats(ioim->itnim, path_tov_expired);
	if (!complete) {
		bfa_sm_send_event(ioim, BFA_IOIM_SM_HCB);
		return;
	}

	ioim->proc_rsp_data = bfa_ioim_lm_proc_rsp_data_dummy;
	bfa_cb_ioim_done(ioim->bfa->bfad, ioim->dio, BFI_IOIM_STS_PATHTOV,
			  0, 0, NULL, 0);
}

static void
__bfa_cb_ioim_abort(void *cbarg, bfa_boolean_t complete)
{
	struct bfa_ioim_s *ioim = cbarg;

	if (!complete) {
		bfa_sm_send_event(ioim, BFA_IOIM_SM_HCB);
		return;
	}

	ioim->proc_rsp_data = bfa_ioim_lm_proc_rsp_data_dummy;
	bfa_cb_ioim_abort(ioim->bfa->bfad, ioim->dio);
}

static void
bfa_ioim_sgpg_alloced(void *cbarg)
{
	struct bfa_ioim_s *ioim = cbarg;

	ioim->nsgpgs = BFA_SGPG_NPAGE(ioim->nsges);
	bfa_q_mv(&ioim->iosp->sgpg_wqe.sgpg_q, &ioim->sgpg_q);
	bfa_ioim_sgpg_setup(ioim);
	bfa_sm_send_event(ioim, BFA_IOIM_SM_SGALLOCED);
}

/**
 * @brief
 * Send I/O request to firmware.
 */
static	bfa_boolean_t
bfa_ioim_send_ioreq(struct bfa_ioim_s *ioim)
{
	struct bfa_itnim_s *itnim = ioim->itnim;
	struct bfi_ioim_req_s *m;
	static fcp_cmnd_t cmnd_z0 = { 0 };
	struct bfi_sge_s      *sge;
	uint32_t	pgdlen = 0;
	uint32_t	fcp_dl;
	void *sg;
	/**
	 * check for room in queue to send request now
	 */
	m = bfa_reqq_next(ioim->bfa, ioim->reqq);
	if (!m) {
		bfa_stats(ioim->itnim, qwait);
		bfa_reqq_wait(ioim->bfa, ioim->reqq,
				  &ioim->iosp->reqq_wait);
		return BFA_FALSE;
	}

	/**
	 * build i/o request message next
	 */
	m->io_tag = bfa_os_htons(ioim->iotag);
	m->rport_hdl = ioim->itnim->rport->fw_handle;
	m->io_timeout = bfa_cb_ioim_get_timeout(ioim->dio);

	/**
	 * build inline IO SG element here
	 */
	sge = &m->sges[0];
	if (ioim->nsges) {
		sg = BFA_SG_FIRST(ioim->dio);
		sge->sga = bfa_cb_ioim_get_sgaddr(ioim->dio, sg);
		pgdlen = bfa_cb_ioim_get_sglen(ioim->dio, sg);
		sge->sg_len = pgdlen;
		sge->flags = (ioim->nsges > BFI_SGE_INLINE) ?
					BFI_SGE_DATA_CPL : BFI_SGE_DATA_LAST;
		bfa_sge_to_be(sge);
		sge++;
		BFA_SG_RESET(sg);
	}

	if (ioim->nsges > BFI_SGE_INLINE) {
		sge->sga = ioim->sgpg->sgpg_pa;
	} else {
		sge->sga.a32.addr_lo = 0;
		sge->sga.a32.addr_hi = 0;
	}
	sge->sg_len = pgdlen;
	sge->flags = BFI_SGE_PGDLEN;
	bfa_sge_to_be(sge);

	/**
	 * set up I/O command parameters
	 */
	bfa_os_assign(m->cmnd, cmnd_z0);
	m->cmnd.lun = bfa_cb_ioim_get_lun(ioim->dio);
	m->cmnd.iodir = bfa_cb_ioim_get_iodir(ioim->dio);
	bfa_os_assign(m->cmnd.cdb,
			*(scsi_cdb_t *)bfa_cb_ioim_get_cdb(ioim->dio));
	fcp_dl = bfa_cb_ioim_get_size(ioim->dio);
	m->cmnd.fcp_dl = bfa_os_htonl(fcp_dl);

	/**
	 * set up I/O message header
	 */
	switch (m->cmnd.iodir) {
	case FCP_IODIR_READ:
		bfi_h2i_set(m->mh, BFI_MC_IOIM_READ, 0, bfa_fn_lpu(ioim->bfa));
		bfa_stats(itnim, input_reqs);
		ioim->itnim->stats.rd_throughput += fcp_dl;

		break;
	case FCP_IODIR_WRITE:
		bfi_h2i_set(m->mh, BFI_MC_IOIM_WRITE, 0, bfa_fn_lpu(ioim->bfa));
		bfa_stats(itnim, output_reqs);
		ioim->itnim->stats.wr_throughput += fcp_dl;
		break;
	case FCP_IODIR_RW:
		bfa_stats(itnim, input_reqs);
		bfa_stats(itnim, output_reqs);
	default:
		bfi_h2i_set(m->mh, BFI_MC_IOIM_IO, 0, bfa_fn_lpu(ioim->bfa));
	}
	if (itnim->seq_rec ||
	    (bfa_cb_ioim_get_size(ioim->dio) & (sizeof(uint32_t) - 1)))
		bfi_h2i_set(m->mh, BFI_MC_IOIM_IO, 0, bfa_fn_lpu(ioim->bfa));

#ifdef IOIM_ADVANCED
	m->cmnd.crn = bfa_cb_ioim_get_crn(ioim->dio);
	m->cmnd.priority = bfa_cb_ioim_get_priority(ioim->dio);
	m->cmnd.taskattr = bfa_cb_ioim_get_taskattr(ioim->dio);

	/**
	 * Handle large CDB (>16 bytes).
	 */
	m->cmnd.addl_cdb_len = (bfa_cb_ioim_get_cdblen(ioim->dio) -
					FCP_CMND_CDB_LEN) / sizeof(uint32_t);
	if (m->cmnd.addl_cdb_len) {
		bfa_os_memcpy(&m->cmnd.cdb + 1, (scsi_cdb_t *)
				bfa_cb_ioim_get_cdb(ioim->dio) + 1,
				m->cmnd.addl_cdb_len * sizeof(uint32_t));
		fcp_cmnd_fcpdl(&m->cmnd) =
				bfa_os_htonl(bfa_cb_ioim_get_size(ioim->dio));
	}
#endif

	/**
	 * queue I/O message to firmware
	 */
	bfa_reqq_produce(ioim->bfa, ioim->reqq, m->mh);
	return BFA_TRUE;
}

/**
 * @brief
 * Setup any additional SG pages needed.Inline SG element is setup
 * at queuing time.
 */
static bfa_boolean_t
bfa_ioim_sge_setup(struct bfa_ioim_s *ioim)
{
	uint16_t	nsgpgs;

	bfa_assert(ioim->nsges > BFI_SGE_INLINE);

	/**
	 * allocate SG pages needed
	 */
	nsgpgs = BFA_SGPG_NPAGE(ioim->nsges);
	if (!nsgpgs)
		return BFA_TRUE;

	if (bfa_sgpg_malloc(ioim->bfa, &ioim->sgpg_q, nsgpgs)
	    != BFA_STATUS_OK) {
		bfa_sgpg_wait(ioim->bfa, &ioim->iosp->sgpg_wqe, nsgpgs);
		return BFA_FALSE;
	}

	ioim->nsgpgs = nsgpgs;
	bfa_ioim_sgpg_setup(ioim);

	return BFA_TRUE;
}

static void
bfa_ioim_sgpg_setup(struct bfa_ioim_s *ioim)
{
	int		sgeid, nsges, i;
	struct bfi_sge_s      *sge;
	struct bfa_sgpg_s *sgpg;
	uint32_t	pgcumsz;
	void *sg;

	sgeid = BFI_SGE_INLINE;
	ioim->sgpg = sgpg = bfa_q_first(&ioim->sgpg_q);
	sg = BFA_SG_FIRST(ioim->dio);

	do {
		sge = sgpg->sgpg->sges;
		nsges = ioim->nsges - sgeid;
		if (nsges > BFI_SGPG_DATA_SGES)
			nsges = BFI_SGPG_DATA_SGES;

		pgcumsz = 0;
		for (i = 0; i < nsges; i++, sge++, sgeid++) {
			BFA_SG_NEXT(sg, ioim->dio, sgeid);

			sge->sga = bfa_cb_ioim_get_sgaddr(ioim->dio, sg);
			sge->sg_len = bfa_cb_ioim_get_sglen(ioim->dio, sg);
			pgcumsz += sge->sg_len;

			/**
			 * set flags
			 */
			if (i < (nsges - 1))
				sge->flags = BFI_SGE_DATA;
			else if (sgeid < (ioim->nsges - 1))
				sge->flags = BFI_SGE_DATA_CPL;
			else
				sge->flags = BFI_SGE_DATA_LAST;

			bfa_sge_to_le(sge);

		}

		sgpg = (struct bfa_sgpg_s *) bfa_q_next(sgpg);

		/**
		 * set the link element of each page
		 */
		if (sgeid == ioim->nsges) {
			sge->flags = BFI_SGE_PGDLEN;
			sge->sga.a32.addr_lo = 0;
			sge->sga.a32.addr_hi = 0;
		} else {
			sge->flags = BFI_SGE_LINK;
			sge->sga = sgpg->sgpg_pa;
		}
		sge->sg_len = pgcumsz;

		bfa_sge_to_le(sge);
	} while (sgeid < ioim->nsges);

	BFA_SG_RESET(sg);
}

/**
 * @brief
 * Send I/O abort request to firmware.
 */
static	bfa_boolean_t
bfa_ioim_send_abort(struct bfa_ioim_s *ioim)
{
	struct bfi_ioim_abort_req_s *m;
	enum bfi_ioim_h2i	msgop;

	/**
	 * check for room in queue to send request now
	 */
	m = bfa_reqq_next(ioim->bfa, ioim->reqq);
	if (!m)
		return BFA_FALSE;

	/**
	 * build i/o request message next
	 */
	if (ioim->iosp->abort_explicit)
		msgop = BFI_IOIM_H2I_IOABORT_REQ;
	else
		msgop = BFI_IOIM_H2I_IOCLEANUP_REQ;

	bfi_h2i_set(m->mh, BFI_MC_IOIM, msgop, bfa_fn_lpu(ioim->bfa));
	m->io_tag    = bfa_os_htons(ioim->iotag);
	m->abort_tag = ++ioim->abort_tag;

	/**
	 * queue I/O message to firmware
	 */
	bfa_reqq_produce(ioim->bfa, ioim->reqq, m->mh);
	return BFA_TRUE;
}

/**
 * @brief
 * Call to resume any I/O requests waiting for room in request queue.
 */
static void
bfa_ioim_qresume(void *cbarg)
{
	struct bfa_ioim_s *ioim = cbarg;

	bfa_stats(ioim->itnim, qresumes);
	bfa_sm_send_event(ioim, BFA_IOIM_SM_QRESUME);
}

/**
 * @brief Used to notify itn of IO completion as part of itn cleanup.
 */
static void
bfa_ioim_notify_cleanup(struct bfa_ioim_s *ioim)
{
	/**
	 * Move IO from itnim queue to fcpim global queue since itnim will be
	 * freed.
	 */
	bfa_q_qe_deq(ioim);
	bfa_q_enq(&ioim->fcpim->ioim_comp_q, ioim);

	if (!ioim->iosp->tskim) {
		if (ioim->fcpim->delay_comp && ioim->itnim->iotov_active) {
			bfa_cb_dequeue(&ioim->hcb_qe);
			bfa_q_qe_deq(ioim);
			bfa_q_enq(&ioim->itnim->delay_comp_q, ioim);
		}
		bfa_itnim_iodone(ioim->itnim);
	} else
		bfa_tskim_iodone(ioim->iosp->tskim);
}

static bfa_boolean_t
bfa_ioim_is_abortable(struct bfa_ioim_s *ioim)
{
	if ((bfa_sm_cmp_state(ioim, bfa_ioim_sm_uninit) &&
		(!bfa_q_is_on_q(&ioim->itnim->pending_q, ioim)))||
	    (bfa_sm_cmp_state(ioim, bfa_ioim_sm_abort))		||
	    (bfa_sm_cmp_state(ioim, bfa_ioim_sm_abort_qfull))	||
	    (bfa_sm_cmp_state(ioim, bfa_ioim_sm_hcb))		||
	    (bfa_sm_cmp_state(ioim, bfa_ioim_sm_hcb_free))	||
	    (bfa_sm_cmp_state(ioim, bfa_ioim_sm_resfree)))
		return BFA_FALSE;

	return BFA_TRUE;
}

/**
 * @brief Used to notify itn of IO completion after path_tov timeout
 * 		or after the link comes back.
 */
void
bfa_ioim_delayed_comp(struct bfa_ioim_s *ioim, bfa_boolean_t iotov)
{
	/**
	 * If path tov timer expired, failback with PATHTOV status - these
	 * IO requests are not normally retried by IO stack.
	 *
	 * Otherwise device cameback online and fail it with normal failed
	 * status so that IO stack retries these failed IO requests.
	 */
	if (iotov)
		ioim->io_cbfn = __bfa_cb_ioim_pathtov;
	else {
		ioim->io_cbfn = __bfa_cb_ioim_failed;
		bfa_stats(ioim->itnim, iocom_nexus_abort);
		bfa_lun_stats(ioim, iocom_nexus_abort);
	}
	bfa_cb_queue(ioim->bfa, &ioim->hcb_qe, ioim->io_cbfn, ioim);

	/**
	 * Move IO to fcpim global queue since itnim will be
	 * freed.
	 */
	bfa_q_qe_deq(ioim);
	bfa_q_enq(&ioim->fcpim->ioim_comp_q, ioim);
}

/**
 * @}
 */

/**
 * @ig hal_ioim_friend
 * @{
 */


/**
 * @brief
 * Memory allocation and initialization.
 */
void
bfa_ioim_attach(struct bfa_fcpim_s *fcpim)
{
	struct bfa_ioim_s		*ioim;
	struct bfa_fcp_mod_s	*fcp = fcpim->fcp;
	struct bfa_ioim_sp_s	*iosp;
	uint16_t		i;

	/**
	 * claim memory first
	 */
	ioim = (struct bfa_ioim_s *) bfa_mem_kva_curp(fcp);
	fcpim->ioim_arr = ioim;
	bfa_mem_kva_curp(fcp) = (uint8_t *) (ioim + fcpim->fcp->num_ioim_reqs);

	iosp = (struct bfa_ioim_sp_s *) bfa_mem_kva_curp(fcp);
	fcpim->ioim_sp_arr = iosp;
	bfa_mem_kva_curp(fcp) = (uint8_t *) (iosp + fcpim->fcp->num_ioim_reqs);

	/**
	 * Initialize ioim free queues
	 */
	bfa_q_init(&fcpim->ioim_resfree_q);
	bfa_q_init(&fcpim->ioim_comp_q);

	for (i = 0; i < fcpim->fcp->num_ioim_reqs;
	     i++, ioim++, iosp++) {
		/*
		 * initialize IOIM
		 */
		bfa_os_memset(ioim, 0, sizeof(struct bfa_ioim_s));
		ioim->iotag   = i;
		ioim->bfa     = fcpim->bfa;
		ioim->fcpim   = fcpim;
		ioim->iosp    = iosp;
		ioim->proc_rsp_data = bfa_ioim_lm_proc_rsp_data_dummy;
		bfa_q_init(&ioim->sgpg_q);
		bfa_reqq_winit(&ioim->iosp->reqq_wait,
				   bfa_ioim_qresume, ioim);
		bfa_sgpg_winit(&ioim->iosp->sgpg_wqe,
				   bfa_ioim_sgpg_alloced, ioim);
		bfa_sm_set_state(ioim, bfa_ioim_sm_uninit);
	}
}

/**
 * @brief
 * Driver detach time call.
 */
void
bfa_ioim_detach(struct bfa_fcpim_s *fcpim)
{
}

void
bfa_ioim_isr(struct bfa_s *bfa, struct bfi_msg_s *m)
{
	struct bfa_fcpim_s *fcpim = BFA_FCPIM(bfa);
	struct bfi_ioim_rsp_s *rsp = (struct bfi_ioim_rsp_s *) m;
	struct bfa_ioim_s *ioim;
	uint16_t	iotag;
	enum bfa_ioim_event evt = BFA_IOIM_SM_COMP;

	iotag = bfa_os_ntohs(rsp->io_tag);

	ioim = BFA_IOIM_FROM_TAG(fcpim, iotag);
	bfa_assert(ioim->iotag == iotag);

	bfa_trc(ioim->bfa, ioim->iotag);
	bfa_trc(ioim->bfa, rsp->io_status);
	bfa_trc(ioim->bfa, rsp->reuse_io_tag);

	if (bfa_sm_cmp_state(ioim, bfa_ioim_sm_active))
		bfa_os_assign(ioim->iosp->comp_rspmsg, *m);

	switch (rsp->io_status) {
	case BFI_IOIM_STS_OK:
		bfa_stats(ioim->itnim, iocomp_ok);
		if (rsp->reuse_io_tag == 0) {
			bfa_ioim_cb_profile_comp(fcpim, ioim, BFA_FALSE);
			evt = BFA_IOIM_SM_DONE;
		} else
			evt = BFA_IOIM_SM_COMP;
		ioim->proc_rsp_data(ioim);
		break;

	case BFI_IOIM_STS_TIMEDOUT:
		bfa_stats(ioim->itnim, iocomp_timedout);
		bfa_lun_stats(ioim, iocomp_timedout);
	case BFI_IOIM_STS_ABORTED:
		bfa_stats(ioim->itnim, iocomp_aborted);
		bfa_lun_stats(ioim, iocomp_aborted);
		if (rsp->reuse_io_tag == 0)
			evt = BFA_IOIM_SM_DONE;
		else
			evt = BFA_IOIM_SM_COMP;
		break;

	case BFI_IOIM_STS_PROTO_ERR:
		bfa_stats(ioim->itnim, iocom_proto_err);
		bfa_lun_stats(ioim, iocom_proto_err);
		bfa_assert(rsp->reuse_io_tag);
		evt = BFA_IOIM_SM_COMP;
		break;

	case BFI_IOIM_STS_SQER_NEEDED:
		bfa_stats(ioim->itnim, iocom_sqer_needed);
		bfa_lun_stats(ioim, iocom_sqer_needed);
		bfa_assert(rsp->reuse_io_tag == 0);
		evt = BFA_IOIM_SM_SQRETRY;
		break;

	case BFI_IOIM_STS_RES_FREE:
		bfa_stats(ioim->itnim, iocom_res_free);
		bfa_lun_stats(ioim, iocom_res_free);
		evt = BFA_IOIM_SM_FREE;
		break;

	case BFI_IOIM_STS_HOST_ABORTED:
		bfa_stats(ioim->itnim, iocom_hostabrts);
		bfa_lun_stats(ioim, iocom_hostabrts);
		if (rsp->abort_tag != ioim->abort_tag) {
			bfa_trc(ioim->bfa, rsp->abort_tag);
			bfa_trc(ioim->bfa, ioim->abort_tag);
			ioim->proc_rsp_data = bfa_ioim_lm_proc_rsp_data_dummy;
			return;
		}

		if (rsp->reuse_io_tag)
			evt = BFA_IOIM_SM_ABORT_COMP;
		else
			evt = BFA_IOIM_SM_ABORT_DONE;
		break;

	case BFI_IOIM_STS_UTAG:
		bfa_stats(ioim->itnim, iocom_utags);
		bfa_lun_stats(ioim, iocom_utags);
		evt = BFA_IOIM_SM_COMP_UTAG;
		break;

	default:
		bfa_assert(0);
	}
	ioim->proc_rsp_data = bfa_ioim_lm_proc_rsp_data_dummy;

	bfa_sm_send_event(ioim, evt);
}

void
bfa_ioim_good_comp_isr(struct bfa_s *bfa, struct bfi_msg_s *m)
{
	struct bfa_fcpim_s *fcpim = BFA_FCPIM(bfa);
	struct bfi_ioim_rsp_s *rsp = (struct bfi_ioim_rsp_s *) m;
	struct bfa_ioim_s *ioim;
	uint16_t	iotag;

	iotag = bfa_os_ntohs(rsp->io_tag);

	ioim = BFA_IOIM_FROM_TAG(fcpim, iotag);
	bfa_assert(ioim->iotag == iotag);

	bfa_trc_fp(ioim->bfa, ioim->iotag);
	bfa_ioim_cb_profile_comp(fcpim, ioim, BFA_TRUE);

	if (bfa_get_lun_mask_status(bfa) != BFA_LUNMASK_ENABLED)  {
		bfa_sm_send_event(ioim, BFA_IOIM_SM_COMP_GOOD);
		return;
	}
	if (ioim->proc_rsp_data(ioim) == BFA_TRUE)
		bfa_sm_send_event(ioim, BFA_IOIM_SM_COMP_GOOD);
	else
		bfa_sm_send_event(ioim, BFA_IOIM_SM_COMP);
}

/**
 * @brief
 * Called by itnim to clean up IO while going offline.
 */
void
bfa_ioim_cleanup(struct bfa_ioim_s *ioim)
{
	bfa_trc(ioim->bfa, ioim->iotag);
	bfa_stats(ioim->itnim, io_cleanups);
	bfa_lun_stats(ioim, io_cleanups);

	ioim->iosp->tskim = NULL;
	bfa_sm_send_event(ioim, BFA_IOIM_SM_CLEANUP);
}

void
bfa_ioim_cleanup_tm(struct bfa_ioim_s *ioim, struct bfa_tskim_s *tskim)
{
	bfa_trc(ioim->bfa, ioim->iotag);
	bfa_stats(ioim->itnim, io_tmaborts);
	bfa_lun_stats(ioim, io_tmaborts);

	ioim->iosp->tskim = tskim;
	bfa_sm_send_event(ioim, BFA_IOIM_SM_CLEANUP);
}

/**
 * @brief
 * IOC failure handling.
 */
void
bfa_ioim_iocdisable(struct bfa_ioim_s *ioim)
{
	bfa_trc(ioim->bfa, ioim->iotag);
	bfa_stats(ioim->itnim, io_iocdowns);
	bfa_lun_stats(ioim, io_iocdowns);
	bfa_sm_send_event(ioim, BFA_IOIM_SM_HWFAIL);
}

/**
 * @brief
 * IO offline TOV popped. Fail the pending IO.
 */
void
bfa_ioim_tov(struct bfa_ioim_s *ioim)
{
	bfa_trc(ioim->bfa, ioim->iotag);
	bfa_sm_send_event(ioim, BFA_IOIM_SM_IOTOV);
}

/**
 * @}
 */

/**
 * @ig hal_ioim_api
 * @{
 */

/**
 * @brief
 * Allocate IOIM resource for initiator mode I/O request.
 */
struct bfa_ioim_s *
bfa_ioim_alloc(struct bfa_s *bfa, struct bfad_ioim_s *dio,
		struct bfa_itnim_s *itnim, uint16_t nsges)
{
	struct bfa_fcpim_s *fcpim = BFA_FCPIM(bfa);
	struct bfa_ioim_s *ioim;
	uint16_t iotag;
	bfa_status_t status;

	/**
	 * alocate IOIM resource
	 */
	status = bfa_fcp_alloc_iotag(bfa, BFI_ITN_ROLE_IM, &iotag);
	if (status != BFA_STATUS_OK) {
		bfa_stats(itnim, no_iotags);
		return NULL;
	}

	ioim = BFA_IOIM_FROM_TAG(fcpim, iotag);

	ioim->dio = dio;
	ioim->itnim = itnim;
	ioim->nsges = nsges;
	ioim->nsgpgs = 0;
	ioim->latency = 0;

	bfa_stats(itnim, total_ios);
	fcpim->ios_active++;

	bfa_q_enq(&itnim->io_q, ioim);
	bfa_trc_fp(ioim->bfa, ioim->iotag);

	return ioim;
}

void
bfa_ioim_free(struct bfa_ioim_s *ioim)
{
	struct bfa_fcpim_s *fcpim = ioim->fcpim;
	struct bfa_s *bfa = fcpim->bfa;

	bfa_trc_fp(ioim->bfa, ioim->iotag);
	bfa_assert_fp(bfa_sm_cmp_state(ioim, bfa_ioim_sm_uninit));

	bfa_assert_fp(bfa_q_is_empty(&ioim->sgpg_q) ||
			(ioim->nsges > BFI_SGE_INLINE));

	if (ioim->nsgpgs > 0)
		bfa_sgpg_mfree(ioim->bfa, &ioim->sgpg_q, ioim->nsgpgs);

	bfa_stats(ioim->itnim, io_comps);
	fcpim->ios_active--;

	ioim->iotag &= BFA_IOIM_IOTAG_MASK;
	ioim->iopf = NULL;
	bfa_fcp_release_iotag(bfa, ioim->iotag);
	bfa_q_qe_deq(ioim);
}

void
bfa_ioim_start(struct bfa_ioim_s *ioim, uint8_t mode)
{
	lun_t 			lun;
	bfa_lps_t		*lps;
	enum bfa_ioim_lm_status	status;

	bfa_trc_fp(ioim->bfa, ioim->iotag);

	if (BFAD_DEF_LUNMASK_SUPPORT == BFA_FALSE &&
	    bfa_get_lun_mask_status(ioim->bfa) == BFA_LUNMASK_ENABLED) {
		lps = BFA_IOIM_TO_LPS(ioim);
		lun = bfa_cb_ioim_get_lun(ioim->dio);
		status = bfa_ioim_lm_check(ioim, lps, ioim->itnim->rport, lun);
		if (status == BFA_IOIM_LM_LUN_NOT_RDY) {
			bfa_sm_send_event(ioim, BFA_IOIM_SM_LM_LUN_NOT_RDY);
			bfa_stats(ioim->itnim, lm_lun_not_rdy);
			return;
		}

		if (status == BFA_IOIM_LM_LUN_NOT_SUP) {
			bfa_sm_send_event(ioim, BFA_IOIM_SM_LM_LUN_NOT_SUP);
			bfa_stats(ioim->itnim, lm_lun_not_sup);
			return;
		}
		if (status == BFA_IOIM_LM_RPL_DATA_CHANGED) {
			bfa_sm_send_event(ioim, BFA_IOIM_SM_LM_RPL_DC);
			bfa_stats(ioim->itnim, lm_rpl_data_changed);
			return;
		}
	}

	bfa_ioim_cb_iopf_start(ioim);

	ioim->mode = mode;
	/**
	 * Obtain the queue over which this request has to be issued
	 */
	ioim->reqq = bfa_fcpim_ioredirect_enabled(ioim->bfa) ?
			bfa_cb_ioim_get_reqq(ioim->dio) :
			bfa_itnim_get_reqq(ioim);

	bfa_sm_send_event(ioim, BFA_IOIM_SM_START);
}

/**
 * @brief
 * Driver I/O abort request.
 */
bfa_status_t
bfa_ioim_abort(struct bfa_ioim_s *ioim)
{

	bfa_trc(ioim->bfa, ioim->iotag);

	if (!bfa_ioim_is_abortable(ioim))
		return BFA_STATUS_FAILED;

	bfa_stats(ioim->itnim, io_aborts);
	bfa_lun_stats(ioim, io_aborts);
	bfa_sm_send_event(ioim, BFA_IOIM_SM_ABORT);

	return BFA_STATUS_OK;
}


/**
 *  BFA TSKIM state machine functions
 */

/**
 * @brief
 *	Task management command beginning state.
 */
static void
bfa_tskim_sm_uninit(struct bfa_tskim_s *tskim, enum bfa_tskim_event event)
{
	bfa_trc(tskim->bfa, tskim->tsk_tag << 16 | event);

	switch (event) {
	case BFA_TSKIM_SM_START:
		bfa_sm_set_state(tskim, bfa_tskim_sm_active);
		bfa_tskim_gather_ios(tskim);

		/**
		 * If device is offline, do not send TM on wire. Just cleanup
		 * any pending IO requests and complete TM request.
		 */
		if (!bfa_itnim_is_online(tskim->itnim)) {
			bfa_sm_set_state(tskim, bfa_tskim_sm_iocleanup);
			tskim->tsk_status = BFI_TSKIM_STS_OK;
			bfa_tskim_cleanup_ios(tskim);
			return;
		}

		if (!bfa_tskim_send(tskim)) {
			bfa_sm_set_state(tskim, bfa_tskim_sm_qfull);
			bfa_stats(tskim->itnim, tm_qwait);
			bfa_reqq_wait(tskim->bfa, tskim->itnim->reqq,
					  &tskim->reqq_wait);
		}
		break;

	default:
		bfa_sm_fault(tskim->bfa, event);
	}
}

/**
 * brief
 *	TM command is active, awaiting completion from firmware to
 *	cleanup IO requests in TM scope.
 */
static void
bfa_tskim_sm_active(struct bfa_tskim_s *tskim, enum bfa_tskim_event event)
{
	bfa_trc(tskim->bfa, tskim->tsk_tag << 16 | event);

	switch (event) {
	case BFA_TSKIM_SM_DONE:
		bfa_sm_set_state(tskim, bfa_tskim_sm_iocleanup);
		bfa_tskim_cleanup_ios(tskim);
		break;

	case BFA_TSKIM_SM_CLEANUP:
		bfa_sm_set_state(tskim, bfa_tskim_sm_cleanup);
		if (!bfa_tskim_send_abort(tskim)) {
			bfa_sm_set_state(tskim, bfa_tskim_sm_cleanup_qfull);
			bfa_stats(tskim->itnim, tm_qwait);
			bfa_reqq_wait(tskim->bfa, tskim->itnim->reqq,
				&tskim->reqq_wait);
		}
		break;

	case BFA_TSKIM_SM_HWFAIL:
		bfa_sm_set_state(tskim, bfa_tskim_sm_hcb);
		bfa_tskim_iocdisable_ios(tskim);
		bfa_tskim_qcomp(tskim, __bfa_cb_tskim_failed);
		break;

	default:
		bfa_sm_fault(tskim->bfa, event);
	}
}

/**
 * @brief
 *	An active TM is being cleaned up since ITN is offline. Awaiting cleanup
 *	completion event from firmware.
 */
static void
bfa_tskim_sm_cleanup(struct bfa_tskim_s *tskim, enum bfa_tskim_event event)
{
	bfa_trc(tskim->bfa, tskim->tsk_tag << 16 | event);

	switch (event) {
	case BFA_TSKIM_SM_DONE:
		/**
		 * Ignore and wait for ABORT completion from firmware.
		 */
		break;

	case BFA_TSKIM_SM_UTAG:
	case BFA_TSKIM_SM_CLEANUP_DONE:
		bfa_sm_set_state(tskim, bfa_tskim_sm_iocleanup);
		bfa_tskim_cleanup_ios(tskim);
		break;

	case BFA_TSKIM_SM_HWFAIL:
		bfa_sm_set_state(tskim, bfa_tskim_sm_hcb);
		bfa_tskim_iocdisable_ios(tskim);
		bfa_tskim_qcomp(tskim, __bfa_cb_tskim_failed);
		break;

	default:
		bfa_sm_fault(tskim->bfa, event);
	}
}

static void
bfa_tskim_sm_iocleanup(struct bfa_tskim_s *tskim, enum bfa_tskim_event event)
{
	bfa_trc(tskim->bfa, tskim->tsk_tag << 16 | event);

	switch (event) {
	case BFA_TSKIM_SM_IOS_DONE:
		bfa_sm_set_state(tskim, bfa_tskim_sm_hcb);
		bfa_tskim_qcomp(tskim, __bfa_cb_tskim_done);
		break;

	case BFA_TSKIM_SM_CLEANUP:
		/**
		 * Ignore, TM command completed on wire.
		 * Notify TM conmpletion on IO cleanup completion.
		 */
		break;

	case BFA_TSKIM_SM_HWFAIL:
		bfa_sm_set_state(tskim, bfa_tskim_sm_hcb);
		bfa_tskim_iocdisable_ios(tskim);
		bfa_tskim_qcomp(tskim, __bfa_cb_tskim_failed);
		break;

	default:
		bfa_sm_fault(tskim->bfa, event);
	}
}

/**
 * @brief
 *	Task management command is waiting for room in request CQ
 */
static void
bfa_tskim_sm_qfull(struct bfa_tskim_s *tskim, enum bfa_tskim_event event)
{
	bfa_trc(tskim->bfa, tskim->tsk_tag << 16 | event);

	switch (event) {
	case BFA_TSKIM_SM_QRESUME:
		bfa_sm_set_state(tskim, bfa_tskim_sm_active);
		bfa_tskim_send(tskim);
		break;

	case BFA_TSKIM_SM_CLEANUP:
		/**
		 * No need to send TM on wire since ITN is offline.
		 */
		bfa_sm_set_state(tskim, bfa_tskim_sm_iocleanup);
		bfa_reqq_wcancel(&tskim->reqq_wait);
		bfa_tskim_cleanup_ios(tskim);
		break;

	case BFA_TSKIM_SM_HWFAIL:
		bfa_sm_set_state(tskim, bfa_tskim_sm_hcb);
		bfa_reqq_wcancel(&tskim->reqq_wait);
		bfa_tskim_iocdisable_ios(tskim);
		bfa_tskim_qcomp(tskim, __bfa_cb_tskim_failed);
		break;

	default:
		bfa_sm_fault(tskim->bfa, event);
	}
}

/**
 * @brief
 *	Task management command is active, awaiting for room in request CQ
 *	to send clean up request.
 */
static void
bfa_tskim_sm_cleanup_qfull(struct bfa_tskim_s *tskim,
		enum bfa_tskim_event event)
{
	bfa_trc(tskim->bfa, tskim->tsk_tag << 16 | event);

	switch (event) {
	case BFA_TSKIM_SM_DONE:
		bfa_reqq_wcancel(&tskim->reqq_wait);
		/**
		 *
		 * Fall through !!!
		 */

	case BFA_TSKIM_SM_QRESUME:
		bfa_sm_set_state(tskim, bfa_tskim_sm_cleanup);
		bfa_tskim_send_abort(tskim);
		break;

	case BFA_TSKIM_SM_HWFAIL:
		bfa_sm_set_state(tskim, bfa_tskim_sm_hcb);
		bfa_reqq_wcancel(&tskim->reqq_wait);
		bfa_tskim_iocdisable_ios(tskim);
		bfa_tskim_qcomp(tskim, __bfa_cb_tskim_failed);
		break;

	default:
		bfa_sm_fault(tskim->bfa, event);
	}
}

/**
 * @brief
 *	BFA callback is pending
 */
static void
bfa_tskim_sm_hcb(struct bfa_tskim_s *tskim, enum bfa_tskim_event event)
{
	bfa_trc(tskim->bfa, tskim->tsk_tag << 16 | event);

	switch (event) {
	case BFA_TSKIM_SM_HCB:
		bfa_sm_set_state(tskim, bfa_tskim_sm_uninit);
		bfa_tskim_free(tskim);
		break;

	case BFA_TSKIM_SM_CLEANUP:
		bfa_tskim_notify_comp(tskim);
		break;

	case BFA_TSKIM_SM_HWFAIL:
		break;

	default:
		bfa_sm_fault(tskim->bfa, event);
	}
}

/**
 * @}
 */

/**
 * @ig hal_tskim_private
 * @{
 */

static void
__bfa_cb_tskim_done(void *cbarg, bfa_boolean_t complete)
{
	struct bfa_tskim_s *tskim = cbarg;

	if (!complete) {
		bfa_sm_send_event(tskim, BFA_TSKIM_SM_HCB);
		return;
	}

	bfa_stats(tskim->itnim, tm_success);
	bfa_cb_tskim_done(tskim->bfa->bfad, tskim->dtsk, tskim->tsk_status);
}

static void
__bfa_cb_tskim_failed(void *cbarg, bfa_boolean_t complete)
{
	struct bfa_tskim_s *tskim = cbarg;

	if (!complete) {
		bfa_sm_send_event(tskim, BFA_TSKIM_SM_HCB);
		return;
	}

	bfa_stats(tskim->itnim, tm_failures);
	bfa_cb_tskim_done(tskim->bfa->bfad, tskim->dtsk,
				BFI_TSKIM_STS_FAILED);
}

static	bfa_boolean_t
bfa_tskim_match_scope(struct bfa_tskim_s *tskim, lun_t lun)
{
	switch (tskim->tm_cmnd) {
	case FCP_TM_TARGET_RESET:
		return BFA_TRUE;

	case FCP_TM_ABORT_TASK_SET:
	case FCP_TM_CLEAR_TASK_SET:
	case FCP_TM_LUN_RESET:
	case FCP_TM_CLEAR_ACA:
		return (tskim->lun == lun);

	default:
		bfa_assert(0);
	}

	return BFA_FALSE;
}

/**
 * @brief
 *	Gather affected IO requests and task management commands.
 */
static void
bfa_tskim_gather_ios(struct bfa_tskim_s *tskim)
{
	struct bfa_itnim_s *itnim = tskim->itnim;
	struct bfa_ioim_s *ioim;
	struct bfa_q_s	*qe, *qen;

	bfa_q_init(&tskim->io_q);

	/**
	 * Gather any active IO requests first.
	 */
	bfa_q_iter_safe(&itnim->io_q, qe, qen) {
		ioim = (struct bfa_ioim_s *) qe;
		if (bfa_tskim_match_scope
			(tskim, bfa_cb_ioim_get_lun(ioim->dio))) {
			bfa_q_qe_deq(ioim);
			bfa_q_enq(&tskim->io_q, ioim);
		}
	}

	/**
	 * Failback any pending IO requests immediately.
	 */
	bfa_q_iter_safe(&itnim->pending_q, qe, qen) {
		ioim = (struct bfa_ioim_s *) qe;
		if (bfa_tskim_match_scope
			(tskim, bfa_cb_ioim_get_lun(ioim->dio))) {
			bfa_q_qe_deq(ioim);
			bfa_q_enq(&ioim->fcpim->ioim_comp_q, ioim);
			bfa_ioim_tov(ioim);
		}
	}
}

/**
 * @brief
 * 		IO cleanup completion
 */
static void
bfa_tskim_cleanp_comp(void *tskim_cbarg)
{
	struct bfa_tskim_s *tskim = tskim_cbarg;

	bfa_stats(tskim->itnim, tm_io_comps);
	bfa_sm_send_event(tskim, BFA_TSKIM_SM_IOS_DONE);
}

/**
 * @brief
 *	Gather affected IO requests and task management commands.
 */
static void
bfa_tskim_cleanup_ios(struct bfa_tskim_s *tskim)
{
	struct bfa_ioim_s *ioim;
	struct bfa_q_s	*qe, *qen;

	bfa_wc_init(&tskim->wc, bfa_tskim_cleanp_comp, tskim);

	bfa_q_iter_safe(&tskim->io_q, qe, qen) {
		ioim = (struct bfa_ioim_s *) qe;
		bfa_wc_up(&tskim->wc);
		bfa_ioim_cleanup_tm(ioim, tskim);
	}

	bfa_wc_wait(&tskim->wc);
}

/**
 * @brief
 *	Send task management request to firmware.
 */
static bfa_boolean_t
bfa_tskim_send(struct bfa_tskim_s *tskim)
{
	struct bfa_itnim_s *itnim = tskim->itnim;
	struct bfi_tskim_req_s *m;

	/**
	 * check for room in queue to send request now
	 */
	m = bfa_reqq_next(tskim->bfa, itnim->reqq);
	if (!m)
		return BFA_FALSE;

	/**
	 * build i/o request message next
	 */
	bfi_h2i_set(m->mh, BFI_MC_TSKIM, BFI_TSKIM_H2I_TM_REQ,
			bfa_fn_lpu(tskim->bfa));

	m->tsk_tag = bfa_os_htons(tskim->tsk_tag);
	m->itn_fhdl = tskim->itnim->rport->fw_handle;
	m->t_secs = tskim->tsecs;
	m->lun = tskim->lun;
	m->tm_flags = tskim->tm_cmnd;

	/**
	 * queue I/O message to firmware
	 */
	bfa_reqq_produce(tskim->bfa, itnim->reqq, m->mh);
	return BFA_TRUE;
}

/**
 * @brief
 *	Send abort request to cleanup an active TM to firmware.
 */
static bfa_boolean_t
bfa_tskim_send_abort(struct bfa_tskim_s *tskim)
{
	struct bfa_itnim_s	*itnim = tskim->itnim;
	struct bfi_tskim_abortreq_s	*m;

	/**
	 * check for room in queue to send request now
	 */
	m = bfa_reqq_next(tskim->bfa, itnim->reqq);
	if (!m)
		return BFA_FALSE;

	/**
	 * build i/o request message next
	 */
	bfi_h2i_set(m->mh, BFI_MC_TSKIM, BFI_TSKIM_H2I_ABORT_REQ,
			bfa_fn_lpu(tskim->bfa));

	m->tsk_tag  = bfa_os_htons(tskim->tsk_tag);

	/**
	 * queue I/O message to firmware
	 */
	bfa_reqq_produce(tskim->bfa, itnim->reqq, m->mh);
	return BFA_TRUE;
}

/**
 * @brief
 *	Call to resume task management cmnd waiting for room in request queue.
 */
static void
bfa_tskim_qresume(void *cbarg)
{
	struct bfa_tskim_s *tskim = cbarg;

	bfa_stats(tskim->itnim, tm_qresumes);
	bfa_sm_send_event(tskim, BFA_TSKIM_SM_QRESUME);
}

/**
 * Cleanup IOs associated with a task mangement command on IOC failures.
 */
static void
bfa_tskim_iocdisable_ios(struct bfa_tskim_s *tskim)
{
	struct bfa_ioim_s *ioim;
	struct bfa_q_s	*qe, *qen;

	bfa_q_iter_safe(&tskim->io_q, qe, qen) {
		ioim = (struct bfa_ioim_s *) qe;
		bfa_ioim_iocdisable(ioim);
	}
}

/**
 * @}
 */

/**
 * @ig hal_tskim_friend
 * @{
 */

/**
 * Notification on completions from related ioim.
 */
void
bfa_tskim_iodone(struct bfa_tskim_s *tskim)
{
	bfa_wc_down(&tskim->wc);
}

/**
 * Handle IOC h/w failure notification from itnim.
 */
void
bfa_tskim_iocdisable(struct bfa_tskim_s *tskim)
{
	tskim->notify = BFA_FALSE;
	bfa_stats(tskim->itnim, tm_iocdowns);
	bfa_sm_send_event(tskim, BFA_TSKIM_SM_HWFAIL);
}

/**
 * Cleanup TM command and associated IOs as part of ITNIM offline.
 */
void
bfa_tskim_cleanup(struct bfa_tskim_s *tskim)
{
	tskim->notify = BFA_TRUE;
	bfa_stats(tskim->itnim, tm_cleanups);
	bfa_sm_send_event(tskim, BFA_TSKIM_SM_CLEANUP);
}

/**
 * @brief
 *	Memory allocation and initialization.
 */
void
bfa_tskim_attach(struct bfa_fcpim_s *fcpim)
{
	struct bfa_tskim_s *tskim;
	struct bfa_fcp_mod_s	*fcp = fcpim->fcp;
	uint16_t	i;

	bfa_q_init(&fcpim->tskim_free_q);
	bfa_q_init(&fcpim->tskim_unused_q);

	tskim = (struct bfa_tskim_s *) bfa_mem_kva_curp(fcp);
	fcpim->tskim_arr = tskim;

	for (i = 0; i < fcpim->num_tskim_reqs; i++, tskim++) {
		/*
		 * initialize TSKIM
		 */
		bfa_os_memset(tskim, 0, sizeof(struct bfa_tskim_s));
		tskim->tsk_tag = i;
		tskim->bfa	= fcpim->bfa;
		tskim->fcpim	= fcpim;
		tskim->notify  = BFA_FALSE;
		bfa_reqq_winit(&tskim->reqq_wait, bfa_tskim_qresume,
					tskim);
		bfa_sm_set_state(tskim, bfa_tskim_sm_uninit);

		bfa_q_enq(&fcpim->tskim_free_q, tskim);
	}

	bfa_mem_kva_curp(fcp) = (uint8_t *) tskim;
}

void
bfa_tskim_detach(struct bfa_fcpim_s *fcpim)
{
	/**
	* @todo
	*/
}

void
bfa_tskim_isr(struct bfa_s *bfa, struct bfi_msg_s *m)
{
	struct bfa_fcpim_s *fcpim = BFA_FCPIM(bfa);
	struct bfi_tskim_rsp_s *rsp = (struct bfi_tskim_rsp_s *) m;
	struct bfa_tskim_s *tskim;
	uint16_t	tsk_tag = bfa_os_ntohs(rsp->tsk_tag);

	tskim = BFA_TSKIM_FROM_TAG(fcpim, tsk_tag);
	bfa_assert(tskim->tsk_tag == tsk_tag);

	tskim->tsk_status = rsp->tsk_status;

	/**
	 * Firmware sends BFI_TSKIM_STS_ABORTED status for abort
	 * requests. All other statuses are for normal completions.
	 */
	if (rsp->tsk_status == BFI_TSKIM_STS_ABORTED) {
		bfa_stats(tskim->itnim, tm_cleanup_comps);
		bfa_sm_send_event(tskim, BFA_TSKIM_SM_CLEANUP_DONE);
	} else if (rsp->tsk_status == BFI_TSKIM_STS_UTAG) {
		bfa_sm_send_event(tskim, BFA_TSKIM_SM_UTAG);
	} else {
		bfa_stats(tskim->itnim, tm_fw_rsps);
		bfa_sm_send_event(tskim, BFA_TSKIM_SM_DONE);
	}
}

/**
 * @}
 */

/**
 * @ig hal_tskim_api
 * @{
 */

/**
 * @brief
 *
 */
struct bfa_tskim_s *
bfa_tskim_alloc(struct bfa_s *bfa, struct bfad_tskim_s *dtsk)
{
	struct bfa_fcpim_s *fcpim = BFA_FCPIM(bfa);
	struct bfa_tskim_s *tskim;

	bfa_q_deq(&fcpim->tskim_free_q, &tskim);

	if (tskim)
		tskim->dtsk = dtsk;

	return tskim;
}

void
bfa_tskim_free(struct bfa_tskim_s *tskim)
{
	bfa_assert(bfa_q_is_on_q_func(&tskim->itnim->tsk_q, &tskim->qe));
	bfa_q_qe_deq(tskim);
	bfa_q_enq(&tskim->fcpim->tskim_free_q, tskim);
}

/**
 * @brief
 *	Start a task management command.
 *
 * @param[in]	tskim	BFA task management command instance
 * @param[in]	itnim	i-t nexus for the task management command
 * @param[in]	lun	lun, if applicable
 * @param[in]	tm_cmnd	Task management command code.
 * @param[in]	t_secs	Timeout in seconds
 *
 * @return Return status.
 */
bfa_status_t
bfa_tskim_start(struct bfa_tskim_s *tskim, struct bfa_itnim_s *itnim, lun_t lun,
			fcp_tm_cmnd_t tm_cmnd, uint8_t tsecs)
{
	tskim->itnim	= itnim;
	tskim->lun	= lun;
	tskim->tm_cmnd	= tm_cmnd;
	tskim->tsecs	= tsecs;
	tskim->notify  = BFA_FALSE;
	bfa_stats(itnim, tm_cmnds);

	bfa_q_enq(&itnim->tsk_q, tskim);

	/* Block TRS cmnd from sending out if `trs_blocked` flag is set */
	if (BFA_TM_CMD_IS_TRS(tskim->tm_cmnd) &&
			BFA_ITNIM_IS_TRS_DISABLED(tskim->itnim)) {
		tskim->tsk_status = BFI_TSKIM_STS_NOT_SUPP;
		return BFA_STATUS_TRS_DISABLED;
	} else {
		bfa_sm_send_event(tskim, BFA_TSKIM_SM_START);
	}

	return BFA_STATUS_OK;
}

void
bfa_tskim_res_recfg(struct bfa_s *bfa, uint16_t num_tskim_fw)
{
	struct bfa_fcpim_s *fcpim = BFA_FCPIM(bfa);
	struct bfa_q_s      *qe;
	int	i;

	for (i = 0; i < (fcpim->num_tskim_reqs - num_tskim_fw); i++) {
		bfa_q_deq_tail(&fcpim->tskim_free_q, &qe);
		bfa_q_enq(&fcpim->tskim_unused_q, qe);
	}
}

bfa_status_t
bfa_fcpim_lunmask_update(struct bfa_s *bfa, uint32_t update)
{
	bfa_lunmask_cfg_t *lun_mask;

	bfa_trc(bfa, bfa_get_lun_mask_status(bfa));
	if (bfa_get_lun_mask_status(bfa) == BFA_LUNMASK_MINCFG)
		 return BFA_STATUS_FAILED;

	if (bfa_get_lun_mask_status(bfa) == (uint8_t) update)
		return BFA_STATUS_NO_CHANGE;
	lun_mask = bfa_get_lun_mask(bfa);
	lun_mask->status  = (uint8_t) update;
	if (bfa_get_lun_mask_status(bfa) == BFA_LUNMASK_ENABLED)
		bfa_ioim_lm_set_ua(bfa);

	return  bfa_dconf_update(bfa);
}

bfa_status_t
bfa_fcpim_lunmask_add(struct bfa_s *bfa, uint16_t vf_id, wwn_t *pwwn,
		 wwn_t rpwwn, lun_t lun)
{
	bfa_lun_mask_t *lunm_list;
	bfa_rport_t *rp = NULL;
	int i;
	int free_index = MAX_LUN_MASK_CFG + 1;

	bfa_trc(bfa, bfa_get_lun_mask_status(bfa));
	if (bfa_get_lun_mask_status(bfa) == BFA_LUNMASK_MINCFG)
		 return BFA_STATUS_FAILED;

	rp =  bfa_rport_get_by_wwn(bfa, vf_id, pwwn, rpwwn);
	lunm_list = bfa_get_lun_mask_list(bfa);
	/* if entry exists */
	for (i = 0; i < MAX_LUN_MASK_CFG; i++) {
		if (lunm_list[i].state != BFA_IOIM_LUN_MASK_ACTIVE)
			free_index = i;
		if ((lunm_list[i].lp_wwn == *pwwn) &&
				(lunm_list[i].rp_wwn == rpwwn) &&
				(lunm_list[i].lun == lun)) {
			return  BFA_STATUS_ENTRY_EXISTS;
		}
	}

	if (free_index > MAX_LUN_MASK_CFG) return BFA_STATUS_MAX_ENTRY_REACHED;

	if (rp) {
		lunm_list[free_index].lp_tag = bfa_lps_get_tag_from_pid(bfa,
				rp->rport_info.local_pid);
		lunm_list[free_index].rp_tag = rp->rport_tag;
	} else {
		lunm_list[free_index].lp_tag = BFA_LP_TAG_INVALID;
		lunm_list[free_index].rp_tag = BFA_RPORT_TAG_INVALID;
	}
	lunm_list[free_index].lp_wwn = *pwwn;
	lunm_list[free_index].rp_wwn = rpwwn;
	lunm_list[free_index].lun = lun;
	lunm_list[free_index].state = BFA_IOIM_LUN_MASK_ACTIVE;
	/* set for all luns in this rp */
	for (i = 0; i < MAX_LUN_MASK_CFG; i++) {
		if ((lunm_list[i].lp_wwn == *pwwn) &&
				(lunm_list[i].rp_wwn == rpwwn))
			lunm_list[i].ua = BFA_IOIM_LM_UA_SET;
	}


	return bfa_dconf_update(bfa);
}

bfa_status_t
bfa_fcpim_lunmask_delete(struct bfa_s *bfa, uint16_t vf_id, wwn_t *pwwn,
	 wwn_t rpwwn, lun_t lun)
{
	bfa_lun_mask_t *lunm_list;
	bfa_rport_t *rp = NULL;
	int i;

	/* in min cfg lunm_list could be NULL but  no commands should run. */
	if (bfa_get_lun_mask_status(bfa) == BFA_LUNMASK_MINCFG)
		 return BFA_STATUS_FAILED;

	bfa_trc(bfa, bfa_get_lun_mask_status(bfa));
	bfa_trc(bfa, *pwwn);
	bfa_trc(bfa, rpwwn);
	bfa_trc(bfa, lun);

	if (*pwwn == 0) {
		rp =  bfa_rport_get_by_wwn(bfa, vf_id, pwwn, rpwwn);
	}
	lunm_list = bfa_get_lun_mask_list(bfa);
	for (i = 0; i < MAX_LUN_MASK_CFG; i++) {
		if ((lunm_list[i].lp_wwn == *pwwn) &&
				(lunm_list[i].rp_wwn == rpwwn) &&
				(lunm_list[i].lun == lun)) {
			lunm_list[i].lp_wwn = 0;
			lunm_list[i].rp_wwn = 0;
			lunm_list[i].lun = 0;
			lunm_list[i].state = BFA_IOIM_LUN_MASK_INACTIVE;
			if (lunm_list[i].rp_tag != BFA_RPORT_TAG_INVALID) {
				lunm_list[i].rp_tag = BFA_RPORT_TAG_INVALID;
				lunm_list[i].lp_tag = BFA_LP_TAG_INVALID;
			}
			return bfa_dconf_update(bfa);
		}
	}
	/* set for all luns in this rp */
	for (i = 0; i < MAX_LUN_MASK_CFG; i++) {
		if ((lunm_list[i].lp_wwn == *pwwn) &&
				(lunm_list[i].rp_wwn == rpwwn))
			lunm_list[i].ua = BFA_IOIM_LM_UA_SET;
	}

	return BFA_STATUS_ENTRY_NOT_EXISTS;
}

bfa_status_t
bfa_fcpim_lunmask_query(struct bfa_s *bfa, void *buf)
{
	bfa_lunmask_cfg_t *lun_mask;

	bfa_trc(bfa, bfa_get_lun_mask_status(bfa));
	if (bfa_get_lun_mask_status(bfa) == BFA_LUNMASK_MINCFG)
		 return BFA_STATUS_FAILED;

	lun_mask = bfa_get_lun_mask(bfa);
	bfa_os_memcpy(buf, lun_mask, sizeof(bfa_lunmask_cfg_t));
	return BFA_STATUS_OK;
}

bfa_status_t
bfa_fcpim_lunmask_clear(struct bfa_s *bfa)
{
	int i;
	bfa_lun_mask_t *lunm_list;

	bfa_trc(bfa, bfa_get_lun_mask_status(bfa));
	if (bfa_get_lun_mask_status(bfa) == BFA_LUNMASK_MINCFG)
		 return BFA_STATUS_FAILED;

	lunm_list = bfa_get_lun_mask_list(bfa);
	for (i = 0; i < MAX_LUN_MASK_CFG; i++) {
		if (lunm_list[i].state == BFA_IOIM_LUN_MASK_ACTIVE) {
			if (lunm_list[i].rp_tag != BFA_RPORT_TAG_INVALID) {
				bfa_rport_unset_lunmask(bfa,
				BFA_RPORT_FROM_TAG(bfa, lunm_list[i].rp_tag));
			}
		}
	}

	bfa_os_memset(lunm_list, 0, sizeof(bfa_lun_mask_t) * MAX_LUN_MASK_CFG);
	return bfa_dconf_update(bfa);
}

void
bfa_fcpim_lunmask_rp_update(bfa_t *bfa, wwn_t lp_wwn, wwn_t rp_wwn,
	uint16_t rp_tag, uint8_t lp_tag)
{
	bfa_lun_mask_t *lun_list;
	uint8_t			i;

	if (bfa_get_lun_mask_status(bfa) == BFA_LUNMASK_MINCFG)
		 return;

	lun_list = bfa_get_lun_mask_list(bfa);
	for (i = 0; i < MAX_LUN_MASK_CFG; i++) {
		if (lun_list[i].state == BFA_IOIM_LUN_MASK_ACTIVE) {
			if ((lun_list[i].lp_wwn == lp_wwn) &&
				(lun_list[i].rp_wwn == rp_wwn)) {
				lun_list[i].rp_tag = rp_tag;
				lun_list[i].lp_tag = lp_tag;
			}
		}
	}

}

/**
 *  @brief free iopf and move from active q to free queue.
 */
void
bfa_fcpim_iopf_free(struct bfa_fcpim_s *fcpim, bfa_fcpim_iopf_t *iopf)
{
	bfa_fcpim_iopf_qe_t *iopf_qe =
		(bfa_fcpim_iopf_qe_t *) BFA_IOPF_QE_FROM_ID(fcpim, iopf->id);

	bfa_trc(fcpim->bfa, iopf->id);

	iopf->lun = 0;
	iopf->lpwwn = 0;
	iopf->rpwwn = 0;
	iopf->iopf_type = 0;
	iopf->lun_status = BFA_FALSE;
	bfa_q_qe_deq(iopf_qe);
	bfa_q_enq(&fcpim->iopf_free_q, iopf_qe);
}

/**
 *  @brief find iopf with same lpwwn rpwwn and lun
 * @return TRUE when it finds same iopf
 */
bfa_fcpim_iopf_t *
bfa_fcpim_get_iopf(struct bfa_fcpim_s *fcpim, wwn_t lpwwn, wwn_t rpwwn,
	 lun_t lun)
{
	bfa_fcpim_iopf_t *iopf;
	bfa_fcpim_iopf_qe_t *iopf_qe;
	struct bfa_q_s *qe, *qen;

	bfa_q_iter_safe(&fcpim->iopf_active_q, qe, qen) {
		iopf_qe = (bfa_fcpim_iopf_qe_t *)qe;
		iopf = &iopf_qe->iopf;
		if ((iopf->lpwwn == lpwwn) && (iopf->rpwwn == rpwwn) &&
				(iopf->lun == lun)) {
			bfa_trc(fcpim->bfa, iopf->id);
			return iopf;
		}
	}
	return NULL;
}

/**
 * @brief
 *	Start advanced IO profile on given lun/rp/lp
 *
 * @param[in]	bfa BFA
 * @param[in]	vf_id not used
 * @param[in]	lpwwn lport wwn  valid value
 * @param[in]	rpwwn rport wwn valid value
 * @param[in]	flow_count	0 if IT level stats
 * @param[in]	lun array of LUNS of lun count
 *
 * @return success/failure on MAX reached
 */

bfa_status_t bfa_fcpim_iopf_start(struct bfa_s *bfa, uint16_t vf_id,
	wwn_t lpwwn, wwn_t rpwwn, int flow_count,
	lun_t lun[BFA_IOPF_MAX], uint8_t lun_status[BFA_IOPF_MAX],
	uint32_t time)
{
	struct bfa_fcpim_s *fcpim = BFA_FCPIM(bfa);
	bfa_fcpim_iopf_t *iopf;
	bfa_rport_t *rp = NULL;
	struct bfa_itnim_s *itnim;
	int i;


	bfa_trc(bfa, lpwwn);
	bfa_trc(bfa, rpwwn);
	bfa_trc(bfa, flow_count);


	if (flow_count) {
		/* for all in count look at lun array */
		/* valid LUN  its lp and rp case */
		for (i = 0; i < flow_count; i++) {
			iopf = bfa_fcpim_get_iopf(fcpim, lpwwn, rpwwn, lun[i]);
			if (iopf) {
				if (flow_count == 1)
					return BFA_STATUS_ENTRY_EXISTS;
				else
					continue;
			}
			iopf = bfa_fcpim_iopf_alloc(fcpim);
			if (!iopf)
				return BFA_STATUS_MAX_ENTRY_REACHED;


			bfa_trc(bfa, iopf->id);
			bfa_trc(bfa, lun[i]);

			iopf->lun = lun[i];
			iopf->lpwwn = lpwwn;
			iopf->rpwwn = rpwwn;
			iopf->iopf_type = BFA_FCPIM_IOPF_ONE_LUN;
			/* todo get LUN status from user space */
			iopf->lun_status = lun_status[i];
			iopf->io_lat.iopf_start_time = time;
			iopf->io_lat.clock_res_mul = bfa_io_lat_clock_res_mul;
			iopf->io_lat.clock_res_div = bfa_io_lat_clock_res_div;
			rp =  bfa_rport_get_by_wwn(bfa, 0, &lpwwn, rpwwn);
			/* no rport found  so no need to update lun id */
			if (rp) {
				/* inactive rp and lun can be added */
				bfa_rport_iopf_add(rp, iopf);
				itnim = BFA_ITNIM_FROM_TAG(fcpim,
						rp->rport_tag);
				bfa_itnim_clear_stats(itnim);
			}
		}
	} else {
		iopf = bfa_fcpim_get_iopf(fcpim, lpwwn, rpwwn,
					BFA_FCPIM_LUN_INVALID);

		if (iopf)
			return BFA_STATUS_ENTRY_EXISTS;
		/* rport levelk state */
		iopf = bfa_fcpim_iopf_alloc(fcpim);
		if (!iopf)
			return BFA_STATUS_MAX_ENTRY_REACHED;

		bfa_trc(bfa, iopf->id);

		iopf->lun = BFA_FCPIM_LUN_INVALID;
		iopf->lpwwn = lpwwn;
		iopf->rpwwn = rpwwn;
		iopf->lun_status = BFA_FALSE;
		iopf->iopf_type = BFA_FCPIM_IOPF_NONE;
		iopf->io_lat.iopf_start_time = time;
		iopf->io_lat.clock_res_mul = bfa_io_lat_clock_res_mul;
		iopf->io_lat.clock_res_div = bfa_io_lat_clock_res_div;
		rp =  bfa_rport_get_by_wwn(bfa, 0, &lpwwn, rpwwn);
		if (!rp) {
			bfa_fcpim_iopf_free(fcpim, iopf);
			return BFA_STATUS_NO_RPORTS;
			/* no rport found */
		}

		// todo add this to IT level only
		bfa_rport_iopf_add(rp, iopf);
		itnim = BFA_ITNIM_FROM_TAG(fcpim, rp->rport_tag);
		bfa_itnim_clear_stats(itnim);

	}

	return BFA_STATUS_OK;
}
/**
 * @brief
 *	Show profile data for specified flow.
 *
 * @param[in]	bfa BFA
 * @param[in]	pwwn lport wwn , 0 if all
 * @param[in]	rpwwn rport wwn 0 if all
 * @param[in]	itn_lun	 flag 0 means itn 1 means LUNa level profile data
 * @param[in]	iopf
 *
 * @return None.
 */
bfa_status_t
bfa_fcpim_iopf_show(struct bfa_s *bfa, uint16_t vf_id,
	wwn_t lpwwn, wwn_t rpwwn, lun_t lun,
	enum bfa_fcpim_iopf_data_type itn_lun, bfa_fcpim_iopf_t *iopf)
{
	bfa_fcpim_iopf_qe_t * iopf_in;
	bfa_fcpim_iopf_t * iopf_tmp;
	struct bfa_fcpim_s *fcpim = BFA_FCPIM(bfa);
	bfa_rport_t *rp = NULL;
	struct bfa_q_s *qe, *qen;
	bfa_boolean_t rp_found = BFA_FALSE;

	bfa_trc(bfa, lpwwn);
	bfa_trc(bfa, rpwwn);

	bfa_os_memset(iopf, 0, sizeof(*iopf));
	bfa_itnim_lun_latency_init(&iopf->io_lat);

	/* show is for a flow */
	if (itn_lun == BFA_FCPIM_IOPF_DATA_LUN) {
		/* valid rp and lun case */
		/* to fill lpwwn if userspace did not give */
		rp =  bfa_rport_get_by_wwn(bfa, vf_id,
				&lpwwn, rpwwn);
		iopf_tmp = bfa_fcpim_get_iopf(fcpim, lpwwn, rpwwn, lun);
		if (!iopf_tmp)
			return BFA_STATUS_IOPF_NOT_EXISTS;

		bfa_os_memcpy(iopf, iopf_tmp,
				sizeof(bfa_fcpim_iopf_t));
		return BFA_STATUS_OK;

	}
	if (lpwwn == 0) {
		if (rpwwn == 0) {
			bfa_q_iter_safe(&fcpim->iopf_active_q, qe, qen) {
				iopf_in = (bfa_fcpim_iopf_qe_t *)qe;
				/* sum all  profile rports */
				rp_found  = BFA_TRUE;
				bfa_fcpim_iopf_show_process(bfa, iopf,
						&iopf_in->iopf);
			}
		} else { /* lpwwn == 0; rpwwn valid */
			bfa_q_iter_safe(&fcpim->iopf_active_q, qe, qen) {
				iopf_in = (bfa_fcpim_iopf_qe_t *)qe;
				if (iopf_in->iopf.rpwwn == rpwwn) {
					rp_found  = BFA_TRUE;
					bfa_fcpim_iopf_show_process(bfa,
							iopf, &iopf_in->iopf);
				}
			}
		}
	} else {
		if (rpwwn == 0) {
			/* rpwwn == 0; lpwwn valid */
			bfa_q_iter_safe(&fcpim->iopf_active_q, qe, qen) {
				iopf_in = (bfa_fcpim_iopf_qe_t *)qe;
				/* sum all  profile rports for lpwwn */
				if (iopf_in->iopf.lpwwn == lpwwn) {
					rp_found  = BFA_TRUE;
					bfa_fcpim_iopf_show_process(bfa,
							iopf, &iopf_in->iopf);
				}
			}
		} else {
			/* lpwwn and rpwwn is valid find a rp */
			return bfa_fcpim_iopf_show_rp(bfa, vf_id, lpwwn,
					rpwwn, iopf);
		}
	}

	/* mark show done FALSE */
	bfa_fcpim_iopf_unmark_showdone(bfa);

	if (rp_found)
		return BFA_STATUS_OK;
	else
		return BFA_STATUS_IOPF_NOT_EXISTS;
}

/**
 * @brief
 *	Stop lun profile on lun
 * for clr Bfal has to call stop for for rport it will clear LUNS too
 *
 * @param[in]	bfa BFA
 * @param[in]	pwwn lport wwn , 0 if all
 * @param[in]	rpwwn rport wwn 0 if all
 * @param[in]	lun	lun,  -1 if all
 * @param[in]	flow_count	lun,  -1 if all
 * @param[in]	lun_array
 *
 * @return success/failure.
 */
bfa_status_t
bfa_fcpim_iopf_stop(struct bfa_s *bfa, uint16_t vf_id,
	wwn_t lpwwn, wwn_t rpwwn, int flow_count, lun_t lunarray[BFA_IOPF_MAX])
{
	struct bfa_fcpim_s *fcpim = BFA_FCPIM(bfa);
	bfa_fcpim_iopf_t *iopf = NULL;
	bfa_fcpim_iopf_qe_t *iopf_qe = NULL;
	bfa_rport_t *rp = NULL;
	struct bfa_q_s *qe, *qen;
	struct bfa_itnim_s *itn;
	int i;

	bfa_trc(bfa, lpwwn);
	bfa_trc(bfa, rpwwn);
	bfa_trc(bfa, flow_count);

	/* if rp not found then also clear cfg data */
	rp =  bfa_rport_get_by_wwn(bfa, vf_id, &lpwwn, rpwwn);
	/**
	 * clr call may have rpwwn and lpwwn 0
	 * but all flows are in fcpim->iopf_active_q
	 */
	if (!flow_count) {

		/* for clear BFAL will give us LUNS */
		bfa_q_iter_safe(&fcpim->iopf_active_q, qe, qen) {
			iopf_qe = (bfa_fcpim_iopf_qe_t *)qe;
			iopf = &iopf_qe->iopf;
			/* clear for all LUNS in that rport */
			if ((iopf->lpwwn == lpwwn) && (iopf->rpwwn == rpwwn)) {
				if (rp) {
					bfa_rport_iopf_remove(rp, iopf);
					itn = BFA_ITNIM_FROM_TAG(fcpim,
						 rp->rport_tag);
					bfa_itnim_clear_stats(itn);
				}
				bfa_fcpim_iopf_reset_data(iopf);
				bfa_fcpim_iopf_free(fcpim, iopf);
			}
		}
		return BFA_STATUS_OK;
	} else {
		/* for all in count look at lun array */
		/* valid LUN  its lp and rp case */

		for (i = 0; i < flow_count; i++) {

			bfa_q_iter_safe(&fcpim->iopf_active_q, qe, qen) {
				iopf_qe = (bfa_fcpim_iopf_qe_t *)qe;
				iopf = &iopf_qe->iopf;
				if ((iopf->lpwwn == lpwwn) && (iopf->rpwwn ==
					rpwwn) && (iopf->lun == lunarray[i])) {
					if (rp) {
						bfa_rport_iopf_remove(rp, iopf);
					}
					bfa_fcpim_iopf_reset_data(iopf);
					bfa_fcpim_iopf_free(fcpim, iopf);
				}
			}
		}
	}

	return BFA_STATUS_OK;
}

/**
 * @brief
 *	query a lun or all luns
 *
 * @param[in]	bfa BFA
 * @param[in]	bfa_fcpim_iopf_cfg_t
 *
 * @return None.
 */
bfa_status_t
bfa_fcpim_iopf_query(struct bfa_s *bfa, uint32_t *flow_count,
	bfa_fcpim_iopf_cfg_t *iopf_cfg)
{
	struct bfa_fcpim_s *fcpim = BFA_FCPIM(bfa);
	bfa_fcpim_iopf_t *iopf;
	bfa_fcpim_iopf_qe_t *iopf_qe;
	struct bfa_q_s *qe, *qen;
	uint32_t i = 0;

	bfa_q_iter_safe(&fcpim->iopf_active_q, qe, qen) {
		iopf_qe = (bfa_fcpim_iopf_qe_t *)qe;
		iopf = &iopf_qe->iopf;

		bfa_trc(bfa, iopf->id);
		bfa_trc(bfa, iopf->lpwwn);
		bfa_trc(bfa, iopf->rpwwn);
		bfa_trc(bfa, iopf->lun);
		bfa_trc(bfa, ((iopf->lun_status << 8) | iopf->iopf_type));

		iopf_cfg[i].lun_id = iopf->lun;
		iopf_cfg[i].lpwwn = iopf->lpwwn;
		iopf_cfg[i].rpwwn = iopf->rpwwn;
		iopf_cfg[i].lun_status = iopf->lun_status;
		iopf_cfg[i].iopf_type = iopf->iopf_type;
		i++;
	}
	bfa_trc(bfa, i);
	*flow_count = i;
	return BFA_STATUS_OK;
}

/**
 * @brief
 *	Reset data for a flow.
 * BFAL has to call for one flow. user can just give port id
 * or lpwwn. bfal has to find rports in that call reset ioctl for each flow.
 * @param[in]	bfa BFA
 * @param[in]	pwwn lport wwn , 0 if all
 * @param[in]	rpwwn rport wwn 0 if all
 * @param[in]	flow_count	flow lun count 0 means at IT level
 * @param[in]	lun data array
 *
 * @return None.
 */
bfa_status_t
bfa_fcpim_iopf_reset(struct bfa_s *bfa, uint16_t vf_id,
	wwn_t lpwwn, wwn_t rpwwn, uint32_t time, int flow_count,
	lun_t lunarray[BFA_IOPF_MAX])
{
	struct bfa_fcpim_s *fcpim = BFA_FCPIM(bfa);
	bfa_rport_t *rp = NULL;
	struct bfa_itnim_s *itnim;
	bfa_fcpim_iopf_qe_t *iopf_qe;
	bfa_fcpim_iopf_t *iopf;
	wwn_t temp_wwn;
	struct bfa_q_s *qe, *qen;
	int i;
	bfa_boolean_t rp_found = BFA_FALSE;

	bfa_trc(bfa, lpwwn);
	bfa_trc(bfa, rpwwn);
	bfa_trc(bfa, flow_count);

	for (i = 0; i < flow_count; i++) {
		/* this means LP and RP wwn were valid */
		bfa_q_iter_safe(&fcpim->iopf_active_q, qe, qen) {
			iopf_qe = (bfa_fcpim_iopf_qe_t *)qe;
			iopf = &iopf_qe->iopf;

			bfa_trc(bfa, iopf->id);
			bfa_trc(bfa, lunarray[i]);
			bfa_trc(bfa, iopf->lun);

			if ((iopf->lpwwn == lpwwn) && (iopf->rpwwn ==
					rpwwn) && (iopf->lun == lunarray[i])) {
				bfa_fcpim_iopf_reset_data(iopf);
				iopf->io_lat.iopf_start_time = time;
				bfa_trc(bfa, iopf->id);
				rp_found = BFA_TRUE;
			}
		}
	}
	if (flow_count) {
		if (rp_found)
			return BFA_STATUS_OK;
		else
			return BFA_STATUS_IOPF_NOT_EXISTS;
	}

	if (!lpwwn) {
		if (!rpwwn) {
			/* Reset all */
			bfa_q_iter_safe(&fcpim->iopf_active_q, qe, qen) {
				iopf_qe = (bfa_fcpim_iopf_qe_t *)qe;
				iopf = &iopf_qe->iopf;

				bfa_trc(bfa, iopf->id);
				temp_wwn = iopf->lpwwn;

				rp_found = BFA_TRUE;
				bfa_fcpim_iopf_reset_data(iopf);
				iopf->io_lat.iopf_start_time = time;
				bfa_trc(bfa, iopf->id);
				rp =  bfa_rport_get_by_wwn(bfa, vf_id,
						&temp_wwn, iopf->rpwwn);
				if (!rp) {
					continue;
					/* no rport found */
				}

				bfa_trc(bfa, rp->rport_tag);
				itnim = BFA_ITNIM_FROM_TAG(fcpim,
						rp->rport_tag);
				bfa_itnim_clear_stats(itnim);

			}
		} else { /* valid rpwwn no lpwwn */
			bfa_q_iter_safe(&fcpim->iopf_active_q, qe, qen) {
				iopf_qe = (bfa_fcpim_iopf_qe_t *)qe;
				iopf = &iopf_qe->iopf;
				bfa_trc(bfa, iopf->id);
				temp_wwn = iopf->lpwwn;
				if (iopf->rpwwn == rpwwn) {
					rp_found = BFA_TRUE;
					bfa_fcpim_iopf_reset_data(iopf);
					iopf->io_lat.iopf_start_time = time;
					bfa_trc(bfa, iopf->id);
					rp =  bfa_rport_get_by_wwn(bfa, vf_id,
							&temp_wwn, rpwwn);
					if (!rp) {
						continue;
						/* no rport found */
					}

					bfa_trc(bfa, rp->rport_tag);
					itnim = BFA_ITNIM_FROM_TAG(fcpim,
							rp->rport_tag);
					bfa_itnim_clear_stats(itnim);
					/* if LUN find that from fcpim lunarr */
				}
			}
		}
	} else { /* valid lpwwn */
		if (!rpwwn) {
			bfa_q_iter_safe(&fcpim->iopf_active_q, qe,
					qen) {
				iopf_qe = (bfa_fcpim_iopf_qe_t *)qe;
				iopf = &iopf_qe->iopf;

				if (iopf->lpwwn == lpwwn) {
					rp_found = BFA_TRUE;
					bfa_fcpim_iopf_reset_data(iopf);
					iopf->io_lat.iopf_start_time = time;
					bfa_trc(bfa, iopf->id);
					rp =  bfa_rport_get_by_wwn(bfa, vf_id,
							&lpwwn, iopf->rpwwn);
					if (!rp) {
						continue;
						/* no rport found */
					}

					bfa_trc(bfa, rp->rport_tag);
					itnim = BFA_ITNIM_FROM_TAG(fcpim,
							rp->rport_tag);
					bfa_itnim_clear_stats(itnim);
					/* if LUN find that from fcpim lunarr */
				}
			}
		} else { /* valid rpwwn  and  lpwwn  and flow count 0 */
			bfa_q_iter_safe(&fcpim->iopf_active_q, qe,
					qen) {
				iopf_qe = (bfa_fcpim_iopf_qe_t *)qe;
				iopf = &iopf_qe->iopf;

				if ((iopf->lpwwn == lpwwn) &&
						(iopf->rpwwn == rpwwn)) {
					rp_found = BFA_TRUE;
					bfa_fcpim_iopf_reset_data(iopf);
					iopf->io_lat.iopf_start_time = time;
					bfa_trc(bfa, iopf->id);
					rp =  bfa_rport_get_by_wwn(bfa, vf_id,
							&lpwwn, rpwwn);
					if (!rp) {
						continue;
						/* no rport found */
					}

					bfa_trc(bfa, rp->rport_tag);
					itnim = BFA_ITNIM_FROM_TAG(fcpim,
							rp->rport_tag);
					bfa_itnim_clear_stats(itnim);
				}
			}
		}
	}

	if (rp_found)
		return BFA_STATUS_OK;
	else
		return BFA_STATUS_IOPF_NOT_EXISTS;
}

static void
bfa_fcpim_iopf_mark_showdone(bfa_t *bfa, bfa_fcpim_iopf_t *iopf)
{
	struct bfa_fcpim_s *fcpim = BFA_FCPIM(bfa);
	struct bfa_q_s *qe, *qen;
	bfa_fcpim_iopf_qe_t * iopf_in;

	bfa_q_iter_safe(&fcpim->iopf_active_q, qe, qen) {
		iopf_in = (bfa_fcpim_iopf_qe_t *)qe;
		if ((iopf_in->iopf.iopf_type != BFA_FCPIM_IOPF_NONE) &&
			(iopf_in->iopf.lpwwn == iopf->lpwwn) &&
			(iopf_in->iopf.rpwwn == iopf->rpwwn))
		iopf_in->iopf.show_done = BFA_TRUE;
	}
}

static void
bfa_fcpim_iopf_unmark_showdone(bfa_t *bfa)
{
	struct bfa_fcpim_s *fcpim = BFA_FCPIM(bfa);
	struct bfa_q_s *qe, *qen;
	bfa_fcpim_iopf_qe_t * iopf_in;

	bfa_q_iter_safe(&fcpim->iopf_active_q, qe, qen) {
		iopf_in = (bfa_fcpim_iopf_qe_t *)qe;
		iopf_in->iopf.show_done = BFA_FALSE;
	}
}

static void
bfa_fcpim_iopf_show_process(bfa_t *bfa, bfa_fcpim_iopf_t *iopf,
		bfa_fcpim_iopf_t *iopf_in)
{
	bfa_rport_t *rp = NULL;


	if (iopf_in->iopf_type == BFA_FCPIM_IOPF_NONE) {
		bfa_fcpim_iopf_mark_showdone(bfa, iopf_in);
		bfa_fcpim_iopf_sum_rp(bfa, iopf, iopf_in);
	} else {
		/**
		 * at this rport the IO profile is enabled
		 * dont worry about LUNS.
		 */
		rp =  bfa_rport_get_by_wwn(bfa, 0, &iopf_in->lpwwn,
				iopf_in->rpwwn);
		if (rp && rp->rp_iopf_enabled)
			return;
		/* if not marked show done */
		bfa_fcpim_iopf_sum_lun(bfa, iopf,
				iopf_in);
	}
}

static bfa_status_t
bfa_fcpim_iopf_show_rp(bfa_t *bfa, uint16_t vf_id, wwn_t lpwwn,
		 wwn_t rpwwn,  bfa_fcpim_iopf_t *iopf)
{
	struct bfa_fcpim_s *fcpim = BFA_FCPIM(bfa);
	bfa_rport_t *rp = NULL;
	struct bfa_itnim_s *itn;
	bfa_fcpim_iopf_t *iopf_tmp;

	rp =  bfa_rport_get_by_wwn(bfa, vf_id, &lpwwn, rpwwn);
	if (!rp) {
		return BFA_STATUS_NO_RPORTS;
		/* no rport found */
	}
	if (bfa_rport_iopf_status(rp) == BFA_FALSE) {
		return BFA_STATUS_IOPF_NOT_EXISTS;
	}
	itn = BFA_ITNIM_FROM_TAG(fcpim, rp->rport_tag);
	/* now copy from itn io_lat and stats to iopf */
	bfa_os_memcpy(&(iopf->io_lat), &itn->iolat,
			sizeof(iopf->io_lat));

	bfa_os_memcpy(&(iopf->io_st), &itn->stats,
			sizeof(iopf->io_st));
	/* bfal check for clock div to be non zero */
	iopf->io_lat.clock_res_mul = bfa_io_lat_clock_res_mul;
	iopf->io_lat.clock_res_div = bfa_io_lat_clock_res_div;
	iopf_tmp = bfa_fcpim_get_iopf(fcpim, lpwwn, rpwwn,
			BFA_FCPIM_LUN_INVALID);
	if (iopf_tmp) {
		iopf->io_lat.iopf_start_time =
			iopf_tmp->io_lat.iopf_start_time;
	}
	return BFA_STATUS_OK;
}

static void
bfa_fcpim_iopf_sum_lun(struct bfa_s *bfa, bfa_fcpim_iopf_t *iopf,
		bfa_fcpim_iopf_t *iopf_in)
{
	bfa_rport_t *rp = NULL;
	uint32_t *tmp, *tmp_in;
	int i, j;

	if (iopf_in->show_done)
		return;

	rp =  bfa_rport_get_by_wwn(bfa, 0, &iopf_in->lpwwn, iopf_in->rpwwn);
	if (!rp) {
		return;
		/* no rport found */
	}

	if (!iopf->io_lat.iopf_start_time)
		iopf->io_lat.iopf_start_time =
			iopf_in->io_lat.iopf_start_time;

	if (iopf->io_lat.iopf_start_time >
			iopf_in->io_lat.iopf_start_time)
		iopf->io_lat.iopf_start_time =
			iopf_in->io_lat.iopf_start_time;

	/* even if one lun in rp is added to iopf. gather the lat values */
	if (iopf_in->iopf_type == BFA_FCPIM_IOPF_ONE_LUN) {
		for (j = 0; j < BFA_IOBUCKET_MAX; j++) {
			if (iopf->io_lat.min[j] > iopf_in->io_lat.min[j])
				iopf->io_lat.min[j] = iopf_in->io_lat.min[j];
			if (iopf->io_lat.max[j] < iopf_in->io_lat.max[j])
				iopf->io_lat.max[j] = iopf_in->io_lat.max[j];
			iopf->io_lat.count[j] += iopf_in->io_lat.count[j];
			iopf->io_lat.avg[j] += iopf_in->io_lat.avg[j];
			iopf->io_lat.iocomps[j] += iopf_in->io_lat.iocomps[j];
		}
	}

	tmp = (uint32_t *)&(iopf->io_st);
	tmp_in = (uint32_t *)&(iopf_in->io_st);
	for (i = 0; i < sizeof(iopf_in->io_st)/sizeof(uint32_t); i++) {
		*tmp += *tmp_in;
		tmp++;
		tmp_in++;
	}
}

static void
bfa_fcpim_iopf_sum_rp(struct bfa_s *bfa, bfa_fcpim_iopf_t *iopf,
		bfa_fcpim_iopf_t *iopf_in)
{
	bfa_rport_t *rp = NULL;
	struct bfa_fcpim_s *fcpim = BFA_FCPIM(bfa);
	struct bfa_itnim_s *itnim = NULL;
	uint32_t *tmp, *tmp_in;
	int i, j;

	bfa_assert(iopf_in->iopf_type != BFA_FCPIM_IOPF_ONE_LUN);

	rp =  bfa_rport_get_by_wwn(bfa, 0, &iopf_in->lpwwn, iopf_in->rpwwn);
	if (!rp) {
		return;
		/* no rport found */
	}

	if (!iopf->io_lat.iopf_start_time)
		iopf->io_lat.iopf_start_time =
			iopf_in->io_lat.iopf_start_time;

	if (iopf->io_lat.iopf_start_time >
			iopf_in->io_lat.iopf_start_time)
		iopf->io_lat.iopf_start_time =
			iopf_in->io_lat.iopf_start_time;

	itnim = BFA_ITNIM_FROM_TAG(fcpim, rp->rport_tag);

	tmp = (uint32_t *)&(iopf->io_st);
	tmp_in = (uint32_t *)&(itnim->stats);
	for (i = 0; i < sizeof(itnim->stats)/sizeof(uint32_t); i++) {
		*tmp += *tmp_in;
		tmp++;
		tmp_in++;
	}

	for (j = 0; j < BFA_IOBUCKET_MAX; j++) {
		if (iopf->io_lat.min[j] > itnim->iolat.min[j])
			iopf->io_lat.min[j] = itnim->iolat.min[j];
		if (iopf->io_lat.max[j] < itnim->iolat.max[j])
			iopf->io_lat.max[j] = itnim->iolat.max[j];
		iopf->io_lat.count[j] += itnim->iolat.count[j];
		iopf->io_lat.avg[j] += itnim->iolat.avg[j];
		iopf->io_lat.iocomps[j] += itnim->iolat.iocomps[j];
	}
	/* bfal check for clock div to be non zero */
	iopf->io_lat.clock_res_mul = bfa_io_lat_clock_res_mul;
	iopf->io_lat.clock_res_div = bfa_io_lat_clock_res_div;
}

static void
bfa_fcpim_iopf_reset_data(bfa_fcpim_iopf_t *iopf)
{
	int j;

	bfa_os_memset(&iopf->io_st, 0, sizeof(iopf->io_st));
	bfa_os_memset(&iopf->io_lat, 0, sizeof(iopf->io_lat));
	for (j = 0; j < BFA_IOBUCKET_MAX; j++)
		iopf->io_lat.min[j] = ~0;

	iopf->io_lat.clock_res_mul = bfa_io_lat_clock_res_mul;
	iopf->io_lat.clock_res_div = bfa_io_lat_clock_res_div;
}

static void
bfa_fcpim_iopf_start_io(struct bfa_ioim_s *ioim)
{
	uint32_t fcp_dl;
	bfa_fcpim_iopf_qe_t *iopf_qe;
	struct bfa_fcpim_s *fcpim = BFA_FCPIM(ioim->bfa);
	uint16_t iopf_id = bfa_rport_get_iopf_id(ioim->itnim->rport,
			bfa_cb_ioim_get_lun(ioim->dio));

	ioim->start_time = bfa_os_get_clock(ioim->bfa->bfad);
	if (iopf_id < BFA_IOPF_MAX) {
		iopf_qe = (bfa_fcpim_iopf_qe_t *)BFA_IOPF_QE_FROM_ID(fcpim,
				iopf_id);
		ioim->iopf = &iopf_qe->iopf;
		bfa_lun_stats(ioim, total_ios);

		fcp_dl = bfa_cb_ioim_get_size(ioim->dio);
		switch (bfa_cb_ioim_get_iodir(ioim->dio)) {
			case FCP_IODIR_READ:
				ioim->iopf->io_st.rd_throughput += fcp_dl;
				bfa_lun_stats(ioim, input_reqs);
				break;
			case FCP_IODIR_WRITE:
				ioim->iopf->io_st.wr_throughput += fcp_dl;
				bfa_lun_stats(ioim, output_reqs);
				break;
			default:
				bfa_lun_stats(ioim, input_reqs);
				bfa_lun_stats(ioim, output_reqs);
				break;
		}
	}
}

static void
bfa_fcpim_ioim_lat(struct bfa_ioim_s *ioim, bfa_itnim_latency_t *io_lat,
		 uint64_t val)
{
	uint32_t fcp_dl = bfa_cb_ioim_get_size(ioim->dio);
	uint32_t index = bfa_ioim_get_index(fcp_dl);

	io_lat->count[index]++;
	io_lat->iocomps[index]++;
	io_lat->min[index] = (io_lat->min[index] < val) ?
		io_lat->min[index] : val;
	io_lat->max[index] = (io_lat->max[index] > val) ?
		io_lat->max[index] : val;
	io_lat->avg[index] += val;

}

static void
bfa_fcpim_iopf_comp_io(struct bfa_ioim_s *ioim, bfa_boolean_t fastpath)
{
	uint64_t end_time = bfa_os_get_clock(ioim->bfa->bfad);
	uint64_t val;
	bfa_itnim_latency_t *io_lat;

	if (end_time < ioim->start_time)
		return;

	val = end_time - ioim->start_time;
	ioim->latency = val;
	if (ioim->itnim->rport->rp_iopf_enabled) {
		io_lat = &(ioim->itnim->iolat);
		bfa_fcpim_ioim_lat(ioim, io_lat, val);
	}

	if (ioim->iopf) {

		if (fastpath)
			ioim->iopf->io_st.io_comps++;
		else
			ioim->iopf->io_st.iocomp_ok++;

		io_lat = &(ioim->iopf->io_lat);
		bfa_fcpim_ioim_lat(ioim, io_lat, val);
	}

}

static void
bfa_fcpim_iopf_scsi_comp_stats(struct bfa_ioim_s *ioim, scsi_sense_t *sense)
{
	if (sense->asc == SCSI_ASC_BAD_LBA) {
		bfa_stats(ioim->itnim, lun_scsi_bad_lba);
		bfa_lun_stats(ioim, lun_scsi_bad_lba);
	} else if (sense->asc == SCSI_ASC_LUN_NOT_SUPPORTED) {
		bfa_stats(ioim->itnim, lun_scsi_lun_n_sup);
		bfa_lun_stats(ioim, lun_scsi_lun_n_sup);
	} else if (sense->asc == SCSI_ASC_LUN_NOT_READY) {
		bfa_stats(ioim->itnim, lun_scsi_lun_n_rdy);
		bfa_lun_stats(ioim, lun_scsi_lun_n_rdy);
	} else if ((sense->asc == SCSI_ASC_POWERON_BDR) &&
			(sense->ascq == SCSI_ASCQ_POR)) {
		bfa_stats(ioim->itnim, lun_scsi_power_on_rst);
		bfa_lun_stats(ioim, lun_scsi_power_on_rst);
	} else if ((sense->asc == SCSI_ASC_TOCC) &&
			(sense->ascq == SCSI_ASCQ_INQ_CHANGED)) {
		bfa_stats(ioim->itnim, lun_scsi_inq_ch);
		bfa_lun_stats(ioim, lun_scsi_inq_ch);
	} else if ((sense->asc == SCSI_ASC_TOCC) &&
			(sense->ascq == SCSI_ASCQ_RL_DATA_CHANGED)) {
		bfa_stats(ioim->itnim, lun_scsi_rpl_ch);
		bfa_lun_stats(ioim, lun_scsi_rpl_ch);
	} else if (sense->asc == SCSI_ASC_LUN_WRITE_PROTECT) {
		bfa_stats(ioim->itnim, lun_scsi_wr_prct);
		bfa_lun_stats(ioim, lun_scsi_wr_prct);
	} else if ((sense->asc == SCSI_ASC_PARAMS_CHANGED) &&
			(sense->ascq == SCSI_ASCQ_RESERVATIONS_PREEMPTED)) {
		bfa_stats(ioim->itnim, lun_scsi_res_preempt);
		bfa_lun_stats(ioim, lun_scsi_res_preempt);
	}
}

/**
 * To send config req, first try to use throttle value from flash
 * If 0, then use driver parameter
 * We need to use min(flash_val, drv_val) because
 * memory allocation was done based on this cfg'd value
 */
uint16_t
bfa_fcpim_get_throttle_cfg(struct bfa_s *bfa, uint16_t drv_cfg_param)
{
	uint16_t tmp;
	struct bfa_fcp_mod_s *fcp = BFA_FCP_MOD(bfa);

	/*
	 * If throttle value from flash is already in effect after driver is
	 * loaded then until next load, always return current value instead
	 * of actual flash value
	 */
	if (!fcp->throttle_update_required)
		return (uint16_t)(fcp->num_ioim_reqs);

	tmp = (bfa_dconf_read_data_valid(bfa)) ?
			bfa_fcpim_read_throttle(bfa) : 0;
	if ((!tmp) || (tmp > drv_cfg_param))
		tmp = drv_cfg_param;

	return (tmp);
}

bfa_status_t
bfa_fcpim_write_throttle(struct bfa_s *bfa, uint16_t value)
{
	if (!bfa_dconf_get_min_cfg(bfa)) {
		BFA_DCONF_MOD(bfa)->dconf->throttle_cfg.value = value;
		BFA_DCONF_MOD(bfa)->dconf->throttle_cfg.is_valid = 1;
		return BFA_STATUS_OK;
	}

	return BFA_STATUS_FAILED;
}

uint16_t
bfa_fcpim_read_throttle(struct bfa_s *bfa)
{
	struct bfa_throttle_cfg_s *throttle_cfg =
			&(BFA_DCONF_MOD(bfa)->dconf->throttle_cfg);

	return ((!bfa_dconf_get_min_cfg(bfa)) ?
	((throttle_cfg->is_valid == 1) ? (throttle_cfg->value) : 0) : 0);
}

bfa_status_t
bfa_fcpim_throttle_set(struct bfa_s *bfa, uint16_t value)
{
	/* in min cfg no commands should run. */
	if ((bfa_dconf_get_min_cfg(bfa) == BFA_TRUE) ||
		(!bfa_dconf_read_data_valid(bfa)))
		 return BFA_STATUS_FAILED;

	bfa_fcpim_write_throttle(bfa, value);

	return bfa_dconf_update(bfa);
}

bfa_status_t
bfa_fcpim_throttle_get(struct bfa_s *bfa, void *buf)
{
	struct bfa_fcpim_s *fcpim = BFA_FCPIM(bfa);
	bfa_defs_fcpim_throttle_t throttle;

	if ((bfa_dconf_get_min_cfg(bfa) == BFA_TRUE) ||
		(!bfa_dconf_read_data_valid(bfa)))
		 return BFA_STATUS_FAILED;

	bfa_os_memset(&throttle, 0, sizeof(bfa_defs_fcpim_throttle_t));

	throttle.cur_value = (uint16_t)(fcpim->fcp->num_ioim_reqs);
	throttle.cfg_value = bfa_fcpim_read_throttle(bfa);
	if (!throttle.cfg_value)
		throttle.cfg_value = throttle.cur_value;
	throttle.max_value = (uint16_t)(fcpim->fcp->max_ioim_reqs);
	bfa_os_memcpy(buf, &throttle, sizeof(bfa_defs_fcpim_throttle_t));

	return BFA_STATUS_OK;
}
